#!/usr/bin/perl
#
# $Id: lynx-dev.pl 519 2016-08-26 05:40:19Z gouzy $
#

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 519 $ =~ /\d+/g); $r[0]};
    use FindBin;
    my $dirlib = ( $0 =~ /-dev.pl/ ) ? "lib-dev" : "lib";
    unshift(@INC,"$FindBin::Bin/../$dirlib");
	print STDERR "$0 version: $VERSION\n";
	print STDERR "CMD: ",join(' ',@ARGV),"\n";
}

use strict;
use warnings;

use General;
use GeneralBioinfo;
use ParamParser;
use HSPSet::M8;
use HSPSet::RMAP;
use SequenceSet::Hash;
use SequenceSet::RMAP;
use Data::Dumper;
use Scaffolder;
use LinkSet;

use constant ILLUMINA_MAXINSERTLEN =>  25000;
use constant DEFOVL => 200;
use constant DEFOVH => 200;
use constant DEFOVH_PCLEN => 5;

my $MINCTGLEN        = 200;
my $MINSPAN          = 400;
my $RATIOMINSPAN     = 5;
my $GAPLEN_TOLERANCE = 50;
my $NB_CPUS = 8;

my $VALID_STEPS = "Scaffolding CleanUpLinks CleanUpMoreLinks TagWeakLinks ValidatePaths RestoreNotAmbiguousWeakLinks LinkScaffolds AddRepeats ValidateScaffoldsDirect ValidateScaffoldsIndirect ValidateScaffoldsSpan MicroScaffolding ScaffoldingValidatedPaths TrackNextScaffolds ReassembleScaffolds SeedScaffolding FillGapsWithScaffolds TrimScaffolds ValidateUnambiguousLinks TrimRepeats BuildPseudomolecules Finishing ContigAnchoring KeepOnlyOverlappingContigs NoSpe DEBUGScaffoldingRepeats";

my $DEBUG = 1;

MAIN:
{

	my $o_param = New ParamParser('GETOPTLONG',\&Usage,'help','cfg=s','outdir=s','minctglen=i','outprefix=s','minspan=i','ratiominspan=i','savehsps','savelinks=s','loadlinks=s','loadscaffolds=s','onlyspecifichsps','scaffolder_list=s','scaffolder_steps=s','debug_ctg=s','gaplen_tolerance=i','loadvpaths=s','nohspfiltering','no_seed_consistency','scaffolder_anchors=s','nb_cpus=i','linkvslength=i','checklinksbyalignment','removingredundantcontigs','initfromfasta');

	$o_param->SetUnlessDefined('minspan',$MINSPAN);
	$o_param->SetUnlessDefined('ratiominspan',$RATIOMINSPAN);	# controle la tolerance a la difference de span (2=>le span sur la plus courte ne doit pas etre 2 x plus eleve que sur la plus longue)
	$o_param->SetUnlessDefined('minctglen',$MINCTGLEN);
	$o_param->SetUnlessDefined('gaplen_tolerance',$GAPLEN_TOLERANCE);

	$o_param->SetUnlessDefined('nb_cpus',$NB_CPUS);

	$o_param->SetUnlessDefined('mingaplen',10);
	$o_param->SetUnlessDefined('debug_ctg','NODEBUG');
	
	$o_param->SetUnlessDefined('outdir','.');
	$o_param->SetUnlessDefined('outprefix',"lynx_".time.".".$$);

	$o_param->AssertFileExists('cfg');
	$o_param->Update($o_param->Get('cfg'),'A');

	$o_param->AssertDefined('hsp_list','sequence_list');
	$o_param->AssertDirExists('outdir');

	foreach my $param ('minspan','ratiominspan','minctglen')
	{
		print STDERR "$param\t".$o_param->Get($param)."\n";
	}
	
	if ( $o_param->IsDefined('checklinksbyalignment') )
	{
		&GetExePath("aaln5");  # je l'utile aussi pour calculer le chev dans GetSequence
	}

	foreach my $infile_key ($o_param->GetKeys("infile"))
	{
		$o_param->AssertFileExists($infile_key) if ( $o_param->Get($infile_key) ne "" );
	}


	my @a_seqlist=split(' ',$o_param->Get('sequence_list'));
	my %h_param=();

	my $flag_seed_consistency =  ( $o_param->IsDefined('no_seed_consistency') ) ? &FALSE : &TRUE;
	# the seed IDs must ^Chr(\d+)

	#
	# Load sequence datasets
	#
	my @a_insert_size=();
	my %h_o_seqds=();
	my $min_insert_size = &MAXINT;
	my $max_insert_size = 0;
	foreach my $seqset (@a_seqlist)
	{
		$o_param->SelectNameSpace("sequence_${seqset}_");
		my @a_opt = $o_param->GetKeys();
		my %h_param = ();
		foreach my $opt (@a_opt)
		{
			$h_param{$opt} = $o_param->Get($opt);
		}
		if ( $h_param{format} eq 'Fasta' )
		{
			$h_o_seqds{$seqset} = New SequenceSet::Hash(%h_param);
		}
		elsif ( $h_param{format} eq 'RMAP' )
		{
			$h_o_seqds{$seqset} = New SequenceSet::RMAP(%h_param);
		}
		else
		{
			Carp::croak("Unknown format >$h_param{format}<\n");
		}
		$h_o_seqds{$seqset}->Load(%h_param);
		$h_o_seqds{$seqset}->PrintInfo();
		if ( defined($h_param{insert_size}) )
		{
			$min_insert_size = &Min($min_insert_size,$h_param{insert_size});
			$max_insert_size = &Max($max_insert_size,$h_param{insert_size});
		}
		if ( defined($h_param{source}) && $h_param{source} =~ /illumina/ )
		{
			push(@a_insert_size,$h_param{insert_size});
		}
	}
	$o_param->SelectNameSpace('');

	my $linkvslen = ( $o_param->IsDefined('linkvslength') ) ? $o_param->Get('linkvslength') : &FALSE;

	$o_param->Set('max_insert_size',$max_insert_size);
	$o_param->Set('maxvalidationlength',$max_insert_size);

	my $o_scaffolder = New Scaffolder(	'minspan'  =>$o_param->Get('minspan'),
										'ratiominspan' => $o_param->Get('ratiominspan'),
										'mingaplen'=>$o_param->Get('mingaplen'),
										'gaplen_tolerance'=>$o_param->Get('gaplen_tolerance'),
										'seed_consistency'=>$flag_seed_consistency,
										'debug_ctg'=>$o_param->Get('debug_ctg'),
										'max_insert_size'=>$o_param->Get('max_insert_size'),
										'maxvalidationlength'=>$o_param->Get('maxvalidationlength'),
										'linkvslength'=>$linkvslen,
										'nb_cpus'=>$o_param->Get('nb_cpus')
									 );

	@a_insert_size = sort { $a <=> $b } @a_insert_size;
	my @a_querylist=();
	&LoadQueryList($o_param,\@a_querylist);

	$o_scaffolder->_Set('_rh_o_seqds',\%h_o_seqds);
	$o_scaffolder->_Set('_ra_seqlist',\@a_querylist);

	my $dumpfile = $o_param->GetIfDefined('loadlinks');
	my $dumpfilewithoverlaps = undef;
	if ( $o_param->IsDefined('initfromfasta') ) # en fait init contigs+links from fasta, pas les scaffolds
	{
		# initialise scaffolds and links from fasta files
		$o_scaffolder->InitFromFasta();
		# pre compute overlaps
		$dumpfilewithoverlaps = $o_scaffolder->AlignLinks(&GetParams($o_param,'AlignLinks'));
	}
	elsif (! defined($dumpfile) )
	{
		&LoadHSPs($o_param,$o_scaffolder,\%h_o_seqds);
	}
	else
	{
		$o_param->AssertFileExists('loadlinks');
		$o_scaffolder->LoadLinks('dumpfile'=>$dumpfile);
	}

	if ( $o_param->IsDefined('checklinksbyalignment') )
	{	
	    # selection des candidats attention il y a les filtres classiques sur les liens et des filtres sur les hsp (minovlpci,minovllen,maxovlovh)
		$dumpfilewithoverlaps = $o_scaffolder->AlignLinks(&GetParams($o_param,'AlignLinks'));
	}
	if ( $o_param->IsDefined('removingredundantcontigs') )
	{
		$o_scaffolder->RemovingRedundantContigs(&GetParams($o_param,'RemovingRedundantContigs'));
	}
	if ( my $vpaths = $o_param->GetIfDefined('loadvpaths') )
	{
		$o_param->AssertFileExists('loadvpaths');
		$o_scaffolder->LoadValidatedPaths($vpaths);
	}
	if ( my $scaffolds = $o_param->GetIfDefined('loadscaffolds') )
	{
		$o_param->AssertFileExists('loadscaffolds');
		$o_scaffolder->LoadScaffolds($scaffolds);
	}

	foreach my $cfg_opt ('scaffolder_anchors','scaffolder_opticalmap')
	{
		if ( $o_param->IsDefined($cfg_opt) && $o_param->Get($cfg_opt) ne '' )
		{
			print STDERR "Lynx: $cfg_opt activated using ".$o_param->Get($cfg_opt)."\n";
		}
		else
		{
			print STDERR "Lynx: $cfg_opt not found in cfg file\n";
		}
	}
	if ( $o_param->IsDefined('scaffolder_anchors') )	# il faut que je charge les ancres avant les scaffolds pour pouvoir verifier la coherence
														# je l'ai mis apres car j'ai besoin de savoir si une ancre est utilisee plusieurs fois
														# est ce que c'est toujours vrai depuis le refactoring ?
	{
		$o_scaffolder->LoadAnchors(&GetParams($o_param,'anchors'));
	}
	if ( $o_param->IsDefined('scaffolder_opticalmap') )
	{
		$o_scaffolder->LoadOpticalMap(&GetParams($o_param,'opticalmap'));
	}

	if ( $o_param->IsDefined('scaffolder_statistics') )
	{
		$o_scaffolder->LinkDatasetStatistics(('tag'=>'AV'));
	}

	# filtrage preliminaire des liens
	if ( $o_param->IsDefined('scaffolder_preliminary_filter') )
	{
		$o_scaffolder->LinkDatasetFilter(&GetParams($o_param,'preliminary_filter'));
	}

	if ( $o_param->IsDefined('scaffolder_statistics') )
	{
		$o_scaffolder->LinkDatasetStatistics(('tag'=>'AP'));
	}

	$o_scaffolder->InitBestLinks(); # needed for LinkDatasetFilterNeighboursOfNeighboursConsistencyCheck

	if ( $o_param->IsDefined('scaffolder_anchors') )
	{
		$o_scaffolder->AnalyseBestAnchoring();
	}

	# je ne peux pas le faire avant le CleanUp car dans ces methodes je suppose qu'il n'y a qu'un seul lien entre 2 contigs
	if ( $o_param->IsDefined('scaffolder_remove_unreliable_neighbours') )
	{
		my %h_param = &GetParams($o_param,'remove_unreliable_neighbours');
		# supprime les liens longues distances avec contigs courts et inserts longs lorsque plusieurs banques donnent le lien (pe contamination dans banques mp)
		$o_scaffolder->LinkDatasetFilterNeighboursNlibConsistencyCheck(\@a_insert_size,%h_param) if ( exists($h_param{NeighboursNlibConsistencyCheck}) && $h_param{NeighboursNlibConsistencyCheck} );
		# ne garde que les voisins qui sont lies a des voisins
		$o_scaffolder->LinkDatasetFilterNeighboursOfNeighbours(%h_param) if ( exists($h_param{NeighboursOfNeighbours}) && $h_param{NeighboursOfNeighbours} );
		# lorsque les distances entres liens englobants et chemin plus long, ne garde que le meilleur chemin
		$o_scaffolder->LinkDatasetFilterNeighboursOfNeighboursConsistencyCheck(%h_param) if (exists($h_param{NeighboursOfNeighboursConsistencyCheck}) && $h_param{NeighboursOfNeighboursConsistencyCheck} );
	}
	if ( defined($dumpfile) )
	{
		$o_scaffolder->DumpLinks('dumpfile'=>"$dumpfile.post_consistencychecks.links.gz");
	}

	&Scaffolding($o_param,$o_scaffolder,\@a_insert_size);

}

sub LoadQueryList
{
my($o_param,$ra_querylist)=@_;

	$o_param->SelectNameSpace('');
	my %h_querylist=();
	my @a_hsplist=split(' ',$o_param->Get('hsp_list'));
	foreach my $hspset (@a_hsplist)
	{
		$o_param->SelectNameSpace("hsp_${hspset}_");
		my @a_opt = $o_param->GetKeys();
		my %h_param = ();
		foreach my $opt (@a_opt)
		{
			$h_param{$opt} = $o_param->Get($opt);
			if ( $opt eq 'query' )
			{
				$h_querylist{$h_param{$opt}}=1;
			}
		}
	}
	@$ra_querylist = keys(%h_querylist);
	$o_param->SelectNameSpace('');
}

#
# Load HSP datasets
#

sub LoadHSPs
{
my($o_param,$o_scaffolder,$rh_o_seqds)=@_;

	my @a_hsplist=split(' ',$o_param->Get('hsp_list'));
	my $minlen = $o_param->Get('minctglen');
	$o_param->SelectNameSpace('');
	$o_param->Set('noseq');
	my $onlyspecifichsps = $o_param->IsDefined('onlyspecifichsps');
	my $nohspfiltering   = $o_param->IsDefined('nohspfiltering');

	my $savehsps = ( $o_param->IsDefined('savehsps') ) ? &TRUE : &FALSE;
	# l'ordre defini l'ordre de priorite
	foreach my $hspset (@a_hsplist)
	{
		my %h_o_hspds=(); # 20150105 je le mets dans la boucle car sinon cela consomme beaucoup trop de RAM
		$o_param->SelectNameSpace("hsp_${hspset}_");
		my @a_opt = $o_param->GetKeys();
		my %h_param = ();
		foreach my $opt (@a_opt)
		{
			$h_param{$opt} = $o_param->Get($opt);
			if ( $opt eq 'query' || $opt eq 'target' )
			{
				if ( $o_param->IsDefined('noseq') )
				{
					if ( !defined($$rh_o_seqds{$h_param{$opt}}) )
					{
						Carp::croak("Sequence dataset >$h_param{$opt}< must be loaded before loading any related HSP dataset\n");
					}
				}
				$h_param{$opt} = $$rh_o_seqds{$h_param{$opt}};
			}
		}

		if ( $h_param{format} eq 'M8' )
		{
			$h_o_hspds{$hspset} = New HSPSet::M8(%h_param);
		}
		elsif ( $h_param{format} eq 'RMAP' )
		{
			$h_o_hspds{$hspset} = New HSPSet::RMAP(%h_param);
		}
		else
		{
			Carp::croak("Unknown format >$h_param{format}<\n");
		}

		my $pattern  = $h_o_hspds{$hspset}->_Get('_target')->_Get('_pattern');
		my $linktype = $h_o_hspds{$hspset}->_Get('_linktype');

		my $mode = $h_o_hspds{$hspset}->_Get('_query')->_Get('_oriented');
		$mode = (defined($mode) && ($mode eq 'true' || $mode eq 'yes' || $mode eq 'oriented')) ? 'oriented' : undef;
		if ( ! defined($mode) )
		{
			# LinkBySeq
			$mode = $h_o_hspds{$hspset}->_Get('_target')->_Get('_spliced');
			$mode = (defined($mode) && ($mode eq 'true' || $mode eq 'yes' || $mode eq '1' )) ? 'mrna' : undef;
		}

		my $type='';
		if  ( !defined($linktype) || $linktype eq '' || $linktype =~ /^pe$/i )
		{
			$type = 'LinkByPE';
		}
		elsif ( $linktype =~ /^seq$/i || $linktype =~ /^sequence$/i )
		{
			$type = 'LinkBySeq';
		}
		elsif ( $linktype =~ /^ovl$/i ||  $linktype =~ /^overlap$/i )
		{
			$type = 'LinkByOverlap';
		}
		elsif ( $linktype =~ /^ref$/i ||  $linktype =~ /^reference$/i )
		{
			$type = 'LinkByReference';
		}
		else
		{
			Carp::croak("Error: unknown value for linktype: >$linktype< not in pe,seq[quence],ovl,overlap,ref,reference\n");
		}

		print STDERR "linktype=$type\n";

		# 20141027, je ne fais plus ca, on le fais via le fichier de conf
#		if ( $type =~ /(LinkByPE|LinkByOverlap)/ )
#		{
#			$o_param->Set("hsp_${hspset}_maxcount",1);
#			$o_param->Set("hsp_${hspset}_mincount",1);
#		}

		#print STDERR "type=$type/$linktype\n";

		if ( $type eq 'LinkBySeq' )
		{
			$h_o_hspds{$hspset}->_Get('_target')->_Set('_maxcount',&MAXINT);
			$h_o_hspds{$hspset}->_Get('_target')->_Set('_mincount',2);
			my $ovl = $h_o_hspds{$hspset}->_Get('_overlap');
			my $ovh = $h_o_hspds{$hspset}->_Get('_overhang');
			my $ovh_pclen = $h_o_hspds{$hspset}->_Get('_overhang_pclen');
			if ( !defined($ovl) || $ovl eq '' )
			{
				print STDERR "Warning: overlap cutoff not defined for hsp set #$hspset ==> set to ".&DEFOVL."\n";
				$h_o_hspds{$hspset}->_Set('_overlap',&DEFOVL);
			}
			if ( !defined($ovh) || $ovh eq '' )
			{
				print STDERR "Warning: overhang cutoff not defined for hsp set #$hspset ==> set to ".&DEFOVH."\n";
				$h_o_hspds{$hspset}->_Set('_overhang',&DEFOVH);
			}
			if ( !defined($ovh_pclen) || $ovh_pclen eq '' )
			{
				print STDERR "Warning: overhang_pclen cutoff not defined for hsp set #$hspset ==> set to ".&DEFOVH_PCLEN."\n";
				$h_o_hspds{$hspset}->_Set('_overhang_pclen',&DEFOVH_PCLEN);
			}
		}

		$h_o_hspds{$hspset}->Load(%h_param);
	
		# identifier les codes pour les deux extremites
	#	if ( $h_param{source} eq 'illumina' )
	#	{
		$h_o_hspds{$hspset}->MinQueryLen($minlen,$h_o_hspds{$hspset}->_Get('_query'));
		$h_o_hspds{$hspset}->PrintInfo();

		if ( defined($pattern) && $pattern !~ /^\s*$/ ) # assuming LinkByPE
		{
			if ( $pattern =~ /\// )
			{
				$pattern=~s,\(\.\+\),,;
				my ($end1,$end2)=$pattern=~/\((.+)\|(.+)\)/;
				$h_o_hspds{$hspset}->SetEndLabels($end1,$end2);
			}
			else
			{
				$h_o_hspds{$hspset}->InitEndLabels();	# c'est trop long sur du solexa
			}
			#$h_o_hspds{$hspset}->PrintInfo();
			# supprimer les HSPs dont les extremites ne sont pas cannoniques
			# $h_o_hspds{$hspset}->CheckEndLabels(); # c'est trop long sur du solexa
			# suppression des HSPs impliquant des sequences trop courtes
			# suppression des HSPs ne mappant qu'une seule query
			if ( $onlyspecifichsps )
			{
				# le mincount et le maxcount sont mis a 1 plus haut
				$h_o_hspds{$hspset}->QueryCount(); 
			}
			if ( ! $nohspfiltering )
			{
				#$h_o_hspds{$hspset}->LinkByPE();
				$h_o_hspds{$hspset}->OneSpecificEnd(); 	# ca fait comme LinkByPE mais j'aurais les liens avec les repeats
				# suppression des HSPs impliquant des targets mappees plusieurs fois
				# suppression de la redondance dans les liens (=inserts mappant au memes endroits des deux cotes)
				my ($srctype) = $h_o_hspds{$hspset}->_Get('_target')->_Get('_source');
				$h_o_hspds{$hspset}->NonRedundantPE() if ( defined($srctype) && $srctype eq 'illumina' );	# pour les BES je veux la garder
				# inutile, c'est fait aussi par LinearizeScaff $h_o_hspds{$hspset}->CompatScaff();
				$h_o_hspds{$hspset}->CompatScaff();
				# lorsque je linearize (j'ai A -> C et A -> B -> C et je ne garde que ABC)
				# au cas ou ce soit un repeat je pourrais : 
				#    1 creer un nouvelle sequence B' qui ne serait lie qu'a A et C
				#    2 supprimer les liens AB et BC
				# comme ca, B pourrait etre utilise par ailleurs si possible
				#$h_o_hspds{$hspset}->LinearizeScaff();
			}
		}
		elsif ($type eq 'LinkBySeq' )
		{
			if ( ! $nohspfiltering ) 
			{
				my ($srctype) = $h_o_hspds{$hspset}->_Get('_target')->_Get('_source');
				if (  defined($srctype) && $srctype eq 'illumina' )
				{
					Carp::croak("Configuration error, an illumina target database cannot be used without a pattern (LinkByPE mode)\n");
				}
				# je le fais d'abord pour simplifier 
				$h_o_hspds{$hspset}->QueryCount(); 
				$h_o_hspds{$hspset}->LinkBySeq(); 
				$h_o_hspds{$hspset}->Compatible($h_o_hspds{$hspset}->_Get('_target'),$h_o_hspds{$hspset}->_Get('_query'));
				# puis pour nettoyer
				$h_o_hspds{$hspset}->QueryCount(); 
			}
		}
		elsif ($type eq 'LinkByOverlap')
		{
			$h_o_hspds{$hspset}->NoSelf();
			$h_o_hspds{$hspset}->Compatible($h_o_hspds{$hspset}->_Get('_target'),$h_o_hspds{$hspset}->_Get('_query'));
		}
		elsif ($type eq 'LinkByReference')
		{
			# le Best doit etre fait apres avoir switche target et query switch=&TRUE
			$h_o_hspds{$hspset}->Best(&TRUE); # je place la target sur plusieurs query
		}
		print STDERR "========***\n";
		$h_o_hspds{$hspset}->PrintInfo();
		$h_o_hspds{$hspset}->DumpM8(&GetStreamOut($h_param{outfile})) if ( $savehsps && defined($h_param{outfile}) );
		$o_scaffolder->AddLinks($h_o_hspds{$hspset},$type,$mode);
		$o_param->SelectNameSpace('');
	}
	$o_param->SelectNameSpace('');
	my $dumpfile = $o_param->GetIfDefined('savelinks');
	if ( defined($dumpfile) )
	{
		$o_scaffolder->DumpLinks('dumpfile'=>"$dumpfile.raw");
	}
	$o_scaffolder->CleanUpLinks(&GetParams($o_param,'CleanUpLinks'));  # calcul des span et suppression des ambiguites
	if ( defined($dumpfile) )
	{
		$o_scaffolder->DumpLinks('dumpfile'=>$dumpfile);
	}
}


sub Scaffolding
{
my($o_param,$o_scaffolder,$ra_insert_size)=@_;

	#$o_scaffolder->InitGraph();
	#print STDERR $o_scaffolder->FindLongestPath("#seq04499m20321 -2705,7294 #seq00250m56113",0,undef);
	#exit;

	$o_param->SelectNameSpace('');
	my %h_steps = ();
	foreach my $k (split(' ',$o_param->Get('scaffolder_steps')))
	{
		$h_steps{$k} = 1;
		if ( $VALID_STEPS !~ /\b$k\b/ )
		{
			print STDERR "Warning: unknown Scaffolding step >$k<\n";
		}
	}

	my $outp = $o_param->Get('outprefix');
	my $outdir = $o_param->Get('outdir');

	if ( defined($h_steps{ContigAnchoring}) )
	{
		my $outroot = "$outdir/$outp.contiganchoring";
		$o_scaffolder->ContigAnchoring(&GetParams($o_param,'ContigAnchoring'));
		$o_scaffolder->DumpAnchors("$outroot.anchors");
		print STDERR "Warning: after the ContigAnchoring step the consistency of the scaffold anchoring is no more guaranty, the program must stop\n";
		exit;
	}

	$o_scaffolder->CleanUpLinks(&GetParams($o_param,'CleanUpLinks')) if ( defined($h_steps{CleanUpLinks}) );  # calcul des span et suppression des ambiguites

	$o_scaffolder->TagWeakLinks($ra_insert_size,&GetParams($o_param,'TagWeakLinks')) if ( defined($h_steps{TagWeakLinks}) );

	$o_scaffolder->RestoreNotAmbiguousWeakLinks()             if ( defined($h_steps{RestoreNotAmbiguousWeakLinks}) );

	$o_scaffolder->ValidateUnambiguousLinks()                 if ( defined($h_steps{ValidateUnambiguousLinks}) );

	# Saving links before any filtering steps. Filtered links can be dump by providing dumpfile=filename to $o_scaffolder->Scaffoling (via cfg file)


	if ( defined($h_steps{ValidatePaths}) )
	{
		$o_scaffolder->ValidatePaths("$outdir/$outp.vpaths.$$.1.".time,&GetParams($o_param,'ValidatePaths'));
		# je le fais 2 fois car lorsque je peux, dans la 1ere j'affine la taille des gap
		#$o_scaffolder->ValidatePaths("$outdir/$outp.vpaths.$$.2.".time,&GetParams($o_param,'ValidatePaths'));
	}

	if ( defined($h_steps{CleanUpMoreLinks}) )	# seulement si pas hardlink, plus maintenant on peut choisir
	{
		$o_scaffolder->CleanUpMoreLinks(&GetParams($o_param,'CleanUpMoreLinks'));  
											# pour une gamme de gaplen courts (def <2000), je ne garde que le successeurs 
											# qui a le meilleur span pour chaque gamme d'inserts et pour chaque extremitee
	}

	my %h_param=('dumpfile'=>"$outdir/$outp.links");
	if ( defined($h_steps{ScaffoldingValidatedPaths}) )
	{
		$o_scaffolder->ScaffoldingValidatedPaths(&GetParams($o_param,'ScaffoldingValidatedPaths'));
		$o_scaffolder->DumpFasta("$outdir/$outp.scaffolds.vpath.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$h_param{dumpfile}.=".vpath";
		$o_scaffolder->DumpLinks(%h_param);
	}

 	if ( defined($h_steps{ScaffoldingRepeats}) )
	{
		$o_scaffolder->ScaffoldingRepeats(&GetParams($o_param,'ScaffoldingRepeats'));
		$o_scaffolder->DumpFasta("$outdir/$outp.scaffolds.scaffrepeats.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$h_param{dumpfile}.=".scaffrepeats";
		$o_scaffolder->DumpLinks(%h_param);
	}

 	if ( defined($h_steps{SeedScaffolding}) )
	{
		$o_scaffolder->SeedScaffolding(&GetParams($o_param,'SeedScaffolding'));
		$o_scaffolder->DumpFasta("$outdir/$outp.scaffolds.seedscaff.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$h_param{dumpfile}.=".seedscaff";
		$o_scaffolder->DumpLinks(%h_param);
	}
		
 	if ( defined($h_steps{MicroScaffolding}) )
	{
		$o_scaffolder->MicroScaffolding(&GetParams($o_param,'MicroScaffolding'));
		$o_scaffolder->DumpFasta("$outdir/$outp.scaffolds.microscaff.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$h_param{dumpfile}.=".microscaff";
		$o_scaffolder->DumpLinks(%h_param);
	}

	

#	$o_scaffolder->DumpFasta("$outdir/$outp.pre-scaffolds.validatedpath.fa",\%h_o_seqds,\@a_querylist);
#	$o_scaffolder->InitScaffolds();

	my @a_scaff=split(' ',$o_param->Get('scaffolder_list'));
	my $param_general = $o_param->Get("scaffolder_general_param");
	if ( $param_general ne '' )
	{
		$param_general .= " ";
	}
	foreach my $scaff (@a_scaff)
	{
		print STDERR "Scaffolding: iteration $scaff of ".join(' ',@a_scaff)."\n";
		$o_param->SetUnlessDefined("scaffolder_${scaff}_param",'');
		my $param = $o_param->Get("scaffolder_${scaff}_param");
		my %h_param=split(/[\s=]+/,"$param_general$param");
#		%h_param=split(/[\s=]+/,$param); # surchage des param par defaut
		if ( defined($h_param{dumpfile}) && $h_param{dumpfile} !~ /^\// )
		{
			$h_param{dumpfile} = "$outdir/$outp.$scaff.$h_param{dumpfile}";
		}
		if ( defined($h_param{usescafflinks}) && $h_param{usescafflinks} && defined($h_steps{LinkScaffolds}) )
		{
			# je lance le calcul des liens si ce n'est deja fait
			$o_scaffolder->LinkScaffolds("$outdir/$outp.scaffolds.$scaff-pre.fa.links","update",&GetParams($o_param,'LinkScaffolds'));
		}

		if ( $scaff =~ /^cl/ ) # contig
		{
			$o_scaffolder->ScaffoldingContigCentered(%h_param) if ( defined($h_steps{Scaffolding}) );
		}
		else	# liens
		{
			$o_scaffolder->Scaffolding(%h_param) if ( defined($h_steps{Scaffolding}) );
		}

		# refait tout a partir de tous les liens d'un scaffold
		$o_scaffolder->ReassembleScaffolds(&GetParams($o_param,'ReassembleScaffolds')) if ( defined($h_steps{ReassembleScaffolds}) );
		
		# cherche les liens interscaffolds
		$o_scaffolder->TrackNextScaffolds(&GetParams($o_param,'TrackNextScaffolds')) if ( defined($h_steps{TrackNextScaffolds}) );

		if ( defined($h_steps{AddRepeats}) )
		{
			my %h_param = &GetParams($o_param,'AddRepeats');
			#print STDERR Dumper(%h_param);
			$o_scaffolder->AddRepeats(%h_param) if ( defined($h_param{allsteps}) );
		}

		# je supprime les liens non direct aux extremites
		$o_scaffolder->TrimScaffolds(&GetParams($o_param,'TrimScaffolds')) if ( defined($h_steps{TrimScaffolds}) );

		# je supprime les repeats aux extremites
		$o_scaffolder->TrimRepeats("Lynx") if ( defined($h_steps{TrimRepeats}) );

		# les contigs adjacents sont lies en direct
		$o_scaffolder->ValidateScaffoldsDirect(&GetParams($o_param,'ValidateScaffoldsDirect')) if ( defined($h_steps{ValidateScaffoldsDirect}) );

		# les contigs (<maxlen) sont lies a au moins 2 contigs 
		$o_scaffolder->ValidateScaffoldsIndirect(&GetParams($o_param,'ValidateScaffoldsIndirect')) if ( defined($h_steps{ValidateScaffoldsIndirect}) );

		# je rajoute des contigs internes lies en direct d'un cote et indirect de l'autre
		if ( defined($h_steps{Finishing}) )
		{
			my %h_param = &GetParams($o_param,'Finishing');
			$o_scaffolder->Finishing(%h_param) if ( defined($h_param{allsteps}) );
		}

		if ( defined($h_steps{KeepOnlyOverlappingContigs}) )
		{
			# Attention dans la fonction je modifie les regles pour le GetSequence du DumpFasta
			$o_scaffolder->KeepOnlyOverlappingContigs(&GetParams($o_param,'KeepOnlyOverlappingContigs'));
			$o_scaffolder->DumpFasta("$outdir/$outp.scaffolds.$scaff.fa.gz");
			$o_scaffolder->DumpLinks('dumpfile'=>"$outdir/$outp.scaffolds.$scaff.fa.links");
			$o_scaffolder->ResetGetSequenceRules();
		}
		else
		{
			$o_scaffolder->DumpFasta("$outdir/$outp.scaffolds.$scaff.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		}

		# je le fais apres pour avoir les memes id que dans le fasta


		# utilise les longs inserts pour calculer la couverture des contigs
		# et pour trouver des points de ruptures/branchement avec les autres scaffolds
		# EN COURS DE DEV, fait juste le diag
		$o_scaffolder->ValidateScaffoldsSpan("$outdir/$outp.scaffolds.$scaff.span",&GetParams($o_param,'ValidateScaffoldsSpan')) if ( defined($h_steps{ValidateScaffoldsSpan}) );

		# lui il marche
		$o_scaffolder->LinkScaffolds("$outdir/$outp.scaffolds.$scaff.fa.links",undef,&GetParams($o_param,'LinkScaffolds')) if ( defined($h_steps{LinkScaffolds}) );

		$o_scaffolder->DumpAnchors("$outdir/$outp.scaffolds.$scaff.fa.anchors");
	}

	my $id = $a_scaff[$#a_scaff];
	$id = '000' if ( ! defined($id) );
	my $outroot = "$outdir/$outp.scaffolds.$id";

	$o_scaffolder->DumpAnchors("$outroot.anchors");

	my $modified = &FALSE;

	if ( defined($h_steps{AddRepeats}) )
	{
		$o_scaffolder->AddRepeats(&GetParams($o_param,'AddRepeats'));
		$outroot .= ".repeats";
		$o_scaffolder->DumpFasta("$outroot.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$o_scaffolder->LinkScaffolds("$outroot.fa.links",undef,&GetParams($o_param,'LinkScaffolds')) if ( defined($h_steps{LinkScaffolds}) );
		$modified=&TRUE;
	}

	if ( defined($h_steps{TrimScaffolds}) )
	{
		$o_scaffolder->TrimScaffolds(&GetParams($o_param,'TrimScaffolds'));
		$outroot .= ".trimmed";
		$o_scaffolder->DumpFasta("$outroot.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$modified=&TRUE;
	}

	if ( defined($h_steps{Finishing}) )
	{
		$o_scaffolder->Finishing(&GetParams($o_param,'Finishing'));
		$outroot .= ".finishing";
		$o_scaffolder->DumpFasta("$outroot.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$modified=&TRUE;
	}

	if ( defined($h_steps{FillGapsWithScaffolds}) )
	{
		$o_scaffolder->FillGapsWithScaffolds(&GetParams($o_param,'FillGapsWithScaffolds'));
		$outroot .= ".filledscaff";
		$o_scaffolder->DumpFasta("$outroot.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$modified=&TRUE;
	}

	if ( !defined($h_steps{KeepOnlyOverlappingContigs}) && ! defined($h_steps{BuildPseudomolecules}) && ! defined($h_steps{DEBUG}) )
	{
		$o_scaffolder->DumpAnchors("$outroot.anchors");
		$o_scaffolder->DumpRepeats("$outroot.repeatdb.gz");
		$outroot .= ".finalized";
		$o_scaffolder->DumpFasta("$outroot.fa.gz",&TRUE,&FALSE,&GetParams($o_param,'DumpFasta'));
	}

	if ( defined($h_steps{DEBUG}) )
	{
		$o_scaffolder->DumpFasta("$outroot.test.fa.gz",&FALSE,&TRUE,&GetParams($o_param,'DumpFasta'));
	}

	if ( defined($h_steps{BuildPseudomolecules}) )
	{
		$o_scaffolder->BuildPseudomolecules(&GetParams($o_param,'BuildPseudomolecules'));
		$outroot .= ".pseudomolecules";
		$o_scaffolder->DumpFasta("$outroot.fa.gz",&FALSE,&FALSE,&GetParams($o_param,'DumpFasta'));
		$o_scaffolder->LinkScaffolds("$outroot.fa.links",undef,&GetParams($o_param,'LinkScaffolds')) if ( defined($h_steps{LinkScaffolds}) );
		$modified=&TRUE;
	}

	$o_scaffolder->ConsistencyChecks("$outroot.consistency_checks.tsv");

	$o_param->SelectNameSpace('');	# sinon le Get sur les param cli ne va pas marcher


	#$o_scaffolder->DumpAGP();
}


sub GetParams
{
my($o_param,$prg)=@_;

	return split(/[\s=]+/,$o_param->Get("scaffolder_$prg"));
}


sub Usage
{
print STDERR<<END
$0
	--help

Mandatory
	--cfg    filename       configuration file
	--outdir dirname        where processed files will be saved

Optional
	--minctglen        integer  [$MINCTGLEN]    mininum length of contigs to be considered in scaffolding
	--minspan          integer  [$MINSPAN]      minimum span on each end for considering the link as strong
	--ratiominspan     float    [$RATIOMINSPAN] maximum ratio of span comparing both ends 
	--gaplen_tolerance integer  [$GAPLEN_TOLERANCE] tolerance in percentage of gap length
	--savehsps                                  HSPs are dumped if outfile is provided
	--loadlinks    filename links are loaded from the file
	--savelinks    filename links are dumped in the file
	--loadscaffolds filename 
	--loadvpaths   filename 
	--onlyspecifichsps
	--nohspfiltering               desactivate HSP filtering (Warning: pre-filtrered HSPs are required)
	--no_seed_consistency          desactivate the exclusion of seeds (contigs named Chr\\d+)

	--scaffolder_list  string 
	--scaffolder_steps string      Scaffolding CleanUpLinks CleanUpMoreLinks TagWeakLinks ValidatePaths 
	                               RestoreNotAmbiguousWeakLinks ValidateScaffoldsDirect ValidateScaffoldsIndirect
	                               ValidateScaffoldsSpan LinkScaffolds MicroScaffolding ScaffoldingValidatedPaths 
	                               TrackNextScaffolds ReassembleScaffolds SeedScaffolding FillGapsWithScaffolds 
	                               TrimScaffolds AddRepeats TrimRepeats LoadOpticalMap Finishing ContigAnchoring
	                               BuildPseudomolecules ScaffoldingRepeats

	--scaffolder_general_param string           quoted list of parameters (key=value)

	--debug_ctg  string     --debug "(seq15473|seq10017)"
	--checklinksbyalignment
	--initfromfasta
	--nb_cpus     integer       [$NB_CPUS]
END
}
