#
# $Id: Scaffolder.pm 519 2016-08-26 05:40:19Z gouzy $
#

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

package Scaffolder;

use strict;
use Contig;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use Graph::Undirected;
use strict;
use warnings;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use Contig;
use ContigRegistry;
use Scaffold;
use ScaffGraph;
use OpticalMap;
use SequenceSet;
use threads;
use threads::shared;
use File::Temp qw/ tempfile tempdir /;
use HSPSet::AALN;

use constant MIN_GAPSIZE         => 30;
use constant MIN_UNKNOWN_GAPSIZE => 10000;

$Data::Dumper::Maxdepth = 10;

use constant MAXVALIDATION_LENGTH          => 200000;
use constant MAXVALIDATION_STEPS           => 5;
use constant LARGE_INSERT_LENGTH           => 30000;
use constant MAX_DISTANCE                  => 200000;    # MicroScaffoling ValidateScaffoldsSpan
use constant MAXGAPLENSTEPS                => 5;
use constant OM_ANCHORING_MIN_SCAFF_LENGTH => 150000;
use constant NB_CPUS                       => 8;
use constant REPEATDEGREECUTOFF            => 15;

my $OVL_MARGIN                     = 0.95;               # % de la longueur
my $MINCTGLINKED_BETWEEN_SCAFFOLDS = 3;                  # LinkScaffolds

my $DEBUG_LINK_PATTERN = undef;                          # cf constructeur
my $GAPLEN_MASKABLE    = 2000;                           # pour le tagging weak avec gaplen vs insert_size

my %H_SCAFF2EXT    = ("p p" => "3 5", "p m" => "3 3", "m p" => "5 5", "m m" => "5 3");
my %H_CTG2SCAFFEXT = ("p 5" => 5,     "p 3" => 3,     "m 5" => 3,     "m 3" => 5);
my %H_SCAFFEXT2CTG = %H_CTG2SCAFFEXT;
my %H_EXT2STRAND   = ("3 5" => "p p", "3 3" => "p m", "5 5" => "m p", "5 3" => "m m")
  ;                                                      # n'existait pas avant 01/12/2012 lorsque j'ai cree AlignLinks
my $SEP = '###';

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 519 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
    my ($self, %h_param) = @_;

    my @a_expected = ('gaplen_tolerance');
    $self->AssertParams(\%h_param, \@a_expected);

    $self->SUPER::_Init();

    $self->_Set('_om_anchoring_minlen',   &OM_ANCHORING_MIN_SCAFF_LENGTH);
    $self->_Set('_anchoring_consistency', &TRUE);
    $self->_Set('_ontheflyanchoring',     &TRUE);
    $self->_Set('_mingaplen',             &MIN_GAPSIZE);
    $self->_Set('_linkvslength',          &FALSE);
    $self->_Set('_repeat_degree_cutoff',  &REPEATDEGREECUTOFF);

    $self->_Set('_nb_cpus', &NB_CPUS);

    foreach my $param (keys(%h_param))
    {
        $self->_Set("_$param", $h_param{$param});
    }
    $DEBUG_LINK_PATTERN = $self->_FastOneGet("_debug_ctg");
    $DEBUG_LINK_PATTERN = 'DEBUG' if (!defined($DEBUG_LINK_PATTERN));

    my $o_graph = New ScaffGraph('maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
    $self->_Set('_o_graph', $o_graph);
    my $o_linkds = New LinkSet();
    $self->_Set('_o_linkds', $o_linkds);
    $self->_Set('_rh_scaffolds',      {});
    $self->_Set('_rh_longestpaths',   {});
    $self->_Set('_rh_validatedpaths', {});
    $self->_Set('_rh_contigs',        {});
    $self->_Set('_rh_bests',          {});
    $self->_Set('_rh_repeats',        undef);
    my $o_omap = New OpticalMap();
    $self->_Set('_o_omap',     $o_omap);
    $self->_Set('_rh_o_seqds', undef);
    $self->_Set('_ra_seqlist', undef);

    my $o_linksscaff = New LinkSet();
    $self->_Set('_o_linksscaff', $o_linksscaff);

    $self->_Set('_nextid',1);
}


sub AssertParams
{
    my ($self, $rh_param, $ra_expected) = @_;

    foreach my $param (@{$ra_expected})
    {
        if (!exists($$rh_param{$param}))
        {
            print STDERR Dumper(%$rh_param);
            Carp::confess(
                          "Fatal error in constructor: $param must be provided\n");
        }
    }
}

sub InitScaffolds
{
    my ($self) = @_;

    $self->_Set('_rh_scaffolds', {});
    $self->_Set('_nextid',1);
}

sub Reset
{
    my ($self) = @_;

    $self->_Set('_rh_scaffolds',      {});
    $self->_Set('_rh_longestpaths',   {});
    $self->_Set('_rh_validatedpaths', {});
    $self->_Set('_rh_contigs',        {});
    $self->_Set('_rh_bests',          {});
    $self->_Set('_rh_repeats',        undef);
    $self->_Set('_nextid',1);
}

sub SetLinks
{
    my ($self, $o_linkds) = @_;

    $self->_Set('_o_linkds', $o_linkds);
}

sub AddLinks
{
    my ($self, $o_hspds, $type, $mode) = @_;

    if ($type eq 'LinkByPE')
    {
        $self->_FastOneGet('_o_linkds')->LoadHSPSetByPE($o_hspds, undef);
    }
    elsif ($type eq 'LinkBySeq')
    {
        $self->_FastOneGet('_o_linkds')->LoadHSPSetBySeq($o_hspds, $mode);
    }
    elsif ($type eq 'LinkByOverlap')
    {
        $self->_FastOneGet('_o_linkds')->LoadHSPSetByOverlap($o_hspds, $mode);
    }
    elsif ($type eq 'LinkByReference')
    {
        $self->_FastOneGet('_o_linkds')->LoadHSPSetByReference($o_hspds, $mode);
    }
}

sub DumpLinks
{
    my ($self, %h_param) = @_;

    $self->_FastOneGet('_o_linkds')->Dump(%h_param);
}

sub InitGraph
{
    my ($self, $use_shared, @a_param) = @_;

    my $o_linkds = $self->_FastOneGet('_o_linkds');
    my $o_graph  = undef;
    if ($use_shared)
    {
        $o_graph = $self->_FastOneGet(
                               '_o_graph');
    }
    else
    {
        $o_graph =
          New ScaffGraph('maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'))
          ;    # trop long de reinit tout a chaque fois
    }
    $o_graph->InitGraph($o_linkds, @a_param);
    if ($use_shared)
    {
        $self->_Set('_o_graph', $o_graph);
    }
    return ($o_graph->GetGraph(), $o_graph);    # pour rester compatible avec l'existant
}

# pour filtrer un fichier de depart tres bruite
sub LinkDatasetFilter
{
my($self,%h_param)=@_;

    my $o_linkds = $self->_FastOneGet('_o_linkds');
    my $o_linkds_filt = New LinkSet();
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
    $self->_Set('_o_linkds', $o_linkds_filt);
}

# $h_best{ctg}{ext}{best} = "ctg list"
sub InitBestLinks
{
my($self,%h_param)=@_;

    my $o_linkds = $self->_FastOneGet('_o_linkds');
	my %h_best=();
    $o_linkds->Best(\%h_best);
    $self->_Set('_rh_bests', \%h_best);
}

sub AnalyseBestAnchoring
{
my($self,%h_param)=@_;

	my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $rh_best      = $self->_FastOneGet('_rh_bests');
	my %h_val=();
	foreach my $ctg (keys(%$rh_contigs))
	{
		foreach my $ext ('5','3')
		{
			if ( &Exists($rh_best,$ctg,$ext,'best') )
			{
				my $best = $$rh_best{$ctg}{$ext}{best};
				next if( ! defined($$rh_contigs{$ctg}) || ! defined($$rh_contigs{$best}) );
				my $srcanchor  = $$rh_contigs{$ctg}->GetAnchors();
				my $bestanchor = $$rh_contigs{$best}->GetAnchors();
				if ( defined($srcanchor) && defined($bestanchor) && $srcanchor ne $bestanchor )
				{
					my %h_val = ();
					if ( $o_linkds->GetKV2($ctg,$best,\%h_val) && $h_val{srclen} > 500 && $h_val{tgtlen} > 500 )
					{
						print STDERR "$ctg\t".$$rh_contigs{$ctg}->GetLength()."\t$ext\t$best\t".$$rh_contigs{$best}->GetLength()." srclen=$h_val{srclen} tgtlen=$h_val{tgtlen} count=$h_val{count} weak=$h_val{weak}\n";
					}
				}
			}
		}
	}

}

sub ConsistencyChecks
{
my ($self,$outfile) = @_;

	$self->_Print("ConsistencyChecks started\n");
	my $o_linkds = $self->_FastOneGet('_o_linkds');
	my $rh_best  = $self->_FastOneGet('_rh_bests');
    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $rh_scaff     = $self->_FastOneGet('_rh_scaffolds');
    $self->_Print("# Contigs: " . scalar(keys(%$rh_contigs)) . "\n");
    $self->_Print("# Scaffolds: " . scalar(keys(%$rh_scaff)) . "\n");
	my @a_contigs = sort keys(%{$rh_contigs});
	my %h_scaff2ctgstatus=();
	my %h_ctgstatus=();
	my $f_out = &GetStreamOut("$outfile");
	my $ctg_both_ok = 0;
	my $ctg_both_ok_len = 0;
	my $ctg_one_ok = 0;
	my $ctg_one_ok_len = 0;
	my $ctg_nok = 0;
	my $ctg_nok_len = 0;
	my $totlen=0;
	foreach my $ctg (@a_contigs)
	{
		my $o_ctgsrc = $self->GetRegisteredContig($ctg,  'p', 0);
    	my $scaffid  = $o_ctgsrc->GetScaffoldsIds();
		%h_ctgstatus=();
		$h_ctgstatus{$ctg}{5} = -1;
		$h_ctgstatus{$ctg}{3} = -1;
		foreach my $ext ('5','3')
		{
			#print STDERR "$rh_best,$ctg,$ext,best\n";
			next if ( ! &Exists($rh_best,$ctg,$ext,'best') );

			my $found = &FALSE;
			my @a_bests = split(' ',$$rh_best{$ctg}{$ext}{best});
			#print STDERR "\t$$rh_best{$ctg}{$ext}{best}\n";
			foreach my $best (@a_bests)
			{
				next if ( !defined($$rh_contigs{$best}) );
				my $o_ctgbest    = $self->GetRegisteredContig($best,  'p', 0);
				my $bestscaffid  = $o_ctgbest->GetScaffoldsIds();
				#print STDERR "$scaffid $bestscaffid\n";
				if ( defined($bestscaffid) && defined($scaffid) && $bestscaffid eq $scaffid )
				{
					$h_scaff2ctgstatus{$scaffid}{$ctg}{$ext}=1;
					$h_ctgstatus{$ctg}{$ext}=1;
					$found = &TRUE;
				}
			}
			if ( ! $found )
			{
				$h_scaff2ctgstatus{$scaffid}{$ctg}{$ext}=0;
				$h_ctgstatus{$ctg}{$ext}=0;
			}
		}
		my $ctglen = $o_ctgsrc->GetLength();
		my $code = 'UNK';
		if ( $h_ctgstatus{$ctg}{5} > 0 && $h_ctgstatus{$ctg}{3} > 0 )
		{
			$ctg_both_ok++;
			$ctg_both_ok_len+=$ctglen;
			$code = 'BOTH';
		}
		elsif ( ($h_ctgstatus{$ctg}{5} > 0)  || ($h_ctgstatus{$ctg}{3} > 0) )
		{
			$ctg_one_ok++;
			$ctg_one_ok_len+=$ctglen;
			$code = 'ONE';
		}
		elsif ( ($h_ctgstatus{$ctg}{5} == 0)  || ($h_ctgstatus{$ctg}{3} == 0) )
		{
			$ctg_nok++;
			$ctg_nok_len+=$ctglen;
			$code = 'NOK';
		}
		print $f_out "$ctg\t$ctglen\t$code\t$h_ctgstatus{$ctg}{5}\t$h_ctgstatus{$ctg}{3}\n";
		$totlen+=$ctglen;
	}
	printf $f_out "ctg_both_ok=$ctg_both_ok len=$ctg_both_ok_len %.2f\n",$ctg_both_ok_len*100/$totlen;
	printf $f_out "ctg_one_ok=$ctg_one_ok len=$ctg_one_ok_len %.2f\n",$ctg_one_ok_len*100/$totlen;
	printf $f_out "ctg_nok=$ctg_nok len=$ctg_nok_len %.2f\n",$ctg_nok_len*100/$totlen;
	$f_out->close;
	$self->_Print("ConsistencyChecks done\n");
}


# 2 contigs courts ne peuvent pas etre lies par plusieurs lib (contaminant pe dans lib mp solexa)
# il faut que je fasse ce filtre avant de choisir l'orientation des contigs (mais je n'ai pas le nlib)

sub LinkDatasetFilterNeighboursNlibConsistencyCheck
{
my($self,$ra_insert_size,%h_param)=@_;

	$self->_Print("LinkDatasetFilterNeighboursNlibConsistencyCheck started\n");
	my %h_limit = ();
	foreach my $insert_size (@$ra_insert_size)
	{
		$h_limit{$insert_size}=$insert_size; # je fais simple pour commencer
	}
	print STDERR Dumper(%h_limit);
	my $o_linkds = $self->_FastOneGet('_o_linkds');
    my @a_links    = $o_linkds->GetLinks();
	my %h_val =();
	my $nb_removed=0;
	for(my $i=0;$i<scalar(@a_links);$i++)
	{
		if ( $o_linkds->GetKV($a_links[$i],\%h_val) )
		{
			my $sumlen = $h_val{srclen}+$h_val{tgtlen};
			next if (! exists($h_limit{$h_val{insert_size}}) );
			if ( $sumlen < $h_limit{$h_val{insert_size}} && $h_val{nlib} > 1 )
			{
				printf STDERR "DEL-NLIB $a_links[$i] sum=$sumlen insert_size=$h_val{insert_size} : %s\n", join(' ',%h_val);
				$o_linkds->Del($a_links[$i]);
				$nb_removed++;
			}
		}
	}
	$self->_Print("LinkDatasetFilterNeighboursNlibConsistencyCheck done; nb_removed=$nb_removed\n");
}

# lorsqu'il y a un chemin direct englobant un chemin avec un 3e contig je verifie que le 3e contig peut contenir dans le chemin englobant
# si ce n'est pas le cas soit je casse le chemin englobant soit les deux liens internes en fonction des comptages

sub LinkDatasetFilterNeighboursOfNeighboursConsistencyCheck
{
my($self,%h_param)=@_;

	$self->_Print("LinkDatasetFilterNeighboursOfNeighboursConsistencyCheck started\n");
	my $o_linkds = $self->_FastOneGet('_o_linkds');
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance})) ? $h_param{gaplen_tolerance} : $gaplen_tolerance_ori;
    my ($rh_graph) = $self->InitGraph(&FALSE, undef, &TRUE);    # je fais un graphe avec les weak
    my @a_links    = $o_linkds->GetLinks();
    my %h_key      = ();
    my %h_val1      = ();
    my %h_val2      = ();
    my %h_val3      = ();
    my $nb_removed = 0;
	my @a_contigs = sort keys(%{$rh_graph});
    my $nb_tot = scalar(@a_links);
	my $rh_best = $self->_FastOneGet('_rh_bests'); # pas utilise
	for(my $i=0;$i<scalar(@a_contigs);$i++)
	{
		foreach my $ext ('5','3')
		{
			my @a_neighbours = sort keys(%{$$rh_graph{$a_contigs[$i]}{$ext}});
			for(my $j=0;$j<scalar(@a_neighbours);$j++)
			{
				next if ( $a_neighbours[$j] lt $a_contigs[$i] ); # => src=$a_contigs[$i]
				if ( $o_linkds->GetKV2($a_contigs[$i],$a_neighbours[$j],\%h_val1) )
				{
					if ( $h_val1{srcext} ne $ext )
					{
						print STDERR "LinkDatasetFilterNeighboursOfNeighboursConsistencyCheck WARNING: $a_contigs[$i],$a_neighbours[$j] inconsistent link\n";
						next;
					}
					my $newtgtext = ( $h_val1{tgtext} eq '5' ) ? 3 : 5;
					my @a_neighbours_neighbours = sort keys(%{$$rh_graph{$a_neighbours[$j]}{$newtgtext}});
					for(my $k=0;$k<scalar(@a_neighbours_neighbours);$k++)
					{
						if ( $o_linkds->GetKV2($a_contigs[$i],$a_neighbours_neighbours[$k],\%h_val2) )
						{
							# je dois verifier que l'orientation du nextnext avec l'englobant est la meme que pour le contig interne
							my $nextnextext = ( $a_contigs[$i] lt $a_neighbours_neighbours[$k] ) ? $h_val2{tgtext} : $h_val2{srcext};

							my $included_len = $$rh_graph{$a_neighbours[$j]}{len};
							my $maxinsert = $h_val2{gaplen_max} + $h_val2{gaplen_max}*($gaplen_tolerance/100);
							if ( $included_len > $maxinsert )
							{
								if ( $o_linkds->GetKV2Oriented($a_neighbours[$j],$newtgtext,$a_neighbours_neighbours[$k],$nextnextext,\%h_val3) )
								{
									print STDERR "INCONSISTENT $a_contigs[$i] - $a_neighbours_neighbours[$k] cannot include $a_neighbours[$j] (len=$included_len) details=".join(' ',%h_val2)."\n";
									if ( $h_val2{count} > $h_val1{count} && $h_val2{count} > $h_val3{count} )
									{
										print STDERR "\tEnglogbant kept: $h_val2{count} > $h_val1{count} && $h_val2{count} > $h_val3{count}\n";
										$nb_removed += &_DelLink($o_linkds,$a_contigs[$i],$a_neighbours[$j]);
										$nb_removed += &_DelLink($o_linkds,$a_neighbours[$j],$a_neighbours_neighbours[$k]);
									}
									else
									{
									    print STDERR "\tInternal kept: $h_val2{count} > $h_val1{count} && $h_val2{count} > $h_val3{count}\n";
										$nb_removed += &_DelLink($o_linkds,$a_contigs[$i],$a_neighbours_neighbours[$k]);
									}
								}
							}
						}
						else
						{
							; # pas de pb cela peut arriver
						}
					}
				}
			}
		}
	}
	$self->_Print("LinkDatasetFilterNeighboursOfNeighboursConsistencyCheck done, nb_removed=$nb_removed\n");
}

# supprime les voisins qui ne sont pas apparentes aux voisins (ou qui n'est pas le meilleur)

sub LinkDatasetFilterNeighboursOfNeighbours
{
my($self,%h_param)=@_;

	$self->_Print("LinkDatasetFilterNeighboursOfNeighbours started\n");
	my $o_linkds = $self->_FastOneGet('_o_linkds');
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my ($rh_graph) = $self->InitGraph(&TRUE, undef, &TRUE);    # je fais un graphe avec les weak
    my @a_links    = $o_linkds->GetLinks();
    my %h_key      = ();
    my %h_val      = ();
    my $nb_removed = 0;
	my @a_contigs = sort keys(%{$rh_graph});
    my $nb_tot = scalar(@a_links);
	my $rh_best = $self->_FastOneGet('_rh_bests');
	for(my $i=0;$i<scalar(@a_contigs);$i++)
	{
		my @a_src_neighbours = sort (keys(%{$$rh_graph{$a_contigs[$i]}{5}}),keys(%{$$rh_graph{$a_contigs[$i]}{3}}));
		for(my $j=0;$j<scalar(@a_src_neighbours);$j++)
		{
			next if ( $a_src_neighbours[$j] lt $a_contigs[$i] );
			my $found = undef;
			my @a_neighbours_neighbours = sort (keys(%{$$rh_graph{$a_src_neighbours[$j]}{5}}),keys(%{$$rh_graph{$a_src_neighbours[$j]}{3}}));
			for (my $k=0;$k<scalar(@a_neighbours_neighbours);$k++)
			{
				next if ( $a_neighbours_neighbours[$k] eq $a_contigs[$i] );
				if ( &Exists($rh_graph,$a_contigs[$i],5,$a_neighbours_neighbours[$k]) || &Exists($rh_graph,$a_contigs[$i],3,$a_neighbours_neighbours[$k]) )
				{
					if ( $o_linkds->GetKV2($a_contigs[$i],$a_neighbours_neighbours[$k],\%h_val) )
					{	
						my $gaplen = undef;
						if ($a_contigs[$i] lt $a_neighbours_neighbours[$k] && &Exists($rh_graph,$a_contigs[$i],$h_val{srcext},$a_neighbours_neighbours[$k],$h_val{tgtext}) )
						{
							$gaplen = $$rh_graph{$a_contigs[$i]}{$h_val{srcext}}{$a_neighbours_neighbours[$k]}{$h_val{tgtext}};
						}
						elsif ( $a_contigs[$i] gt $a_neighbours_neighbours[$k] && &Exists($rh_graph,$a_contigs[$i],$h_val{tgtext},$a_neighbours_neighbours[$k],$h_val{srcext}) )
						{
							$gaplen = $$rh_graph{$a_contigs[$i]}{$h_val{tgtext}}{$a_neighbours_neighbours[$k]}{$h_val{srcext}};
						}
						if ( defined($gaplen) ) # je recupere gaplen car je voulais mettre une verif que l'englobant peut contenir l'intermediaire mais je ne peux pas car je ne sais pas qui englobe qui
						{
							$found = $a_neighbours_neighbours[$k];
							last;
						}
						else
						{
							print STDERR "WARN Neighbours of neighbours $a_contigs[$i],$a_src_neighbours[$j] with inconsistent orientation\n";
						}
					}
				}
			}
			if ( ! defined($found) )
			{
				if ( (&Exists($rh_best,$a_contigs[$i],5,'best') && $$rh_best{$a_contigs[$i]}{5}{best} =~ /\b$a_src_neighbours[$j]\b/) ||
				     (&Exists($rh_best,$a_contigs[$i],3,'best') && $$rh_best{$a_contigs[$i]}{3}{best} =~ /\b$a_src_neighbours[$j]\b/) ||
				     (&Exists($rh_best,$a_src_neighbours[$j],5,'best') && $$rh_best{$a_src_neighbours[$j]}{5}{best} =~ /\b$a_contigs[$i]\b/) ||
				     (&Exists($rh_best,$a_src_neighbours[$j],3,'best') && $$rh_best{$a_src_neighbours[$j]}{3}{best} =~ /\b$a_contigs[$i]\b/) 
				   )
				{
					#print STDERR "RESCUE $a_contigs[$i],$a_src_neighbours[$j] : ".$o_linkds->Get(sprintf("src=%s tgt=%s",sort($a_contigs[$i],$a_src_neighbours[$j])))."\n";
				}
				else
				{
					$nb_removed += &_DelLink($o_linkds,$a_contigs[$i],$a_src_neighbours[$j]);
				}
			}
			else
			{
				#print STDERR "OK $a_contigs[$i],$a_src_neighbours[$j] by $found : ".$o_linkds->Get(sprintf("src=%s tgt=%s",sort($a_contigs[$i],$a_src_neighbours[$j])))."\n";
			}
		}
	}
	$self->_Print("LinkDatasetFilterNeighboursOfNeighbours #removed=$nb_removed / $nb_tot\n");
}

sub _DelLink
{
my($o_linkds,$c1,$c2) = @_;

	my $key = sprintf("src=%s tgt=%s",sort($c1,$c2));
	my $details = $o_linkds->Get($key);
	if ( defined($details) )
	{
		printf STDERR "DEL $c1,$c2 : %s\n", (defined($details)) ? $details : "";
		$o_linkds->Del2($c1,$c2);
		return 1;
	}
	return 0;
}

#
# Edition des liens (weak)
#

sub ValidateUnambiguousLinks
{
    my ($self) = @_;

    $self->_Print("ValidateUnambiguousLinks starting\n");
    my $o_linkds = $self->_FastOneGet('_o_linkds');
    my ($rh_graph) = $self->InitGraph(&TRUE, undef, &TRUE);    # je fais un graphe avec les weak
    my @a_links    = $o_linkds->GetLinks();
    my %h_key      = ();
    my $nb_restored = 0;
    foreach my $link (@a_links)
    {
        my %h_val = ();
        $o_linkds->GetKV($link, \%h_val);
        %h_key = split(/[=\s]+/, $link);
        next if ($h_val{hardlink});
        my @a_srclinks = keys(%{$$rh_graph{$h_key{src}}{$h_val{srcext}}});
        my @a_tgtlinks = keys(%{$$rh_graph{$h_key{tgt}}{$h_val{tgtext}}});
        if (scalar(@a_srclinks) == 1 && scalar(@a_tgtlinks) == 1)
        {
            $self->_Print("ValidateUnambiguousLinks: $link " . join(' ', %h_val) . "\n");
            $h_val{hardlink} = 3;
            $o_linkds->SetKV($link, \%h_val);
        }
    }
    $self->_Print("ValidateUnambiguousLinks done\n");
}

sub RestoreNotAmbiguousWeakLinks
{
    my ($self) = @_;

    $self->_Print("RestoreNotAmbiguousWeakLinks starting\n");
    my $o_linkds = $self->_FastOneGet('_o_linkds');
    my ($rh_graph) = $self->InitGraph(&TRUE, undef, &TRUE);    # je fais un graphe avec les weak
    my @a_links    = $o_linkds->GetLinks();
    my %h_val      = ();
    my %h_key      = ();
    my $nb_restored = 0;
    foreach my $link (@a_links)
    {
        $o_linkds->GetKV($link, \%h_val);
        %h_key = split(/[=\s]+/, $link);
        next if (!$h_val{weak});
        my @a_srclinks = keys(%{$$rh_graph{$h_key{src}}{$h_val{srcext}}});
        my @a_tgtlinks = keys(%{$$rh_graph{$h_key{tgt}}{$h_val{tgtext}}});
        if (scalar(@a_srclinks) == 1 && scalar(@a_tgtlinks) == 1)
        {
            $self->_Print("RestoreNotAmbiguousWeakLinks: $link " . join(' ', %h_val) . "\n");
            $h_val{weak} = 0;
            $o_linkds->SetKV($link, \%h_val);
        }
    }
    $self->_Print("RestoreNotAmbiguousWeakLinks done\n");
}

# je ne fais plus le seuil sur le count a ce niveau, je prefere le faire a la volee et ne pas modifier les liens
#
# le span doit depasser un certain niveau
# le span sur la sequence la plus longue doit etre ~ superieur a celui de la sequence la plus courte

# ce qui est suspect c'est un gaplen faible donne par un lien avec un grand insert_size car lorsque l'on a des inser_size de taille infrieure
# ce sont eux qui auraient du donner le lien (on ne garde que la taille d'insert la plus petite)
# le probleme c'est que si je sous estime la taille on rentre dans la deviation standard qui peut etre grande pour les grands inserts
# ==> le tagging des inserts marche lorsque je prends gaplen=max mais pas si je prends gaplen=insert_size

sub TagWeakLinks
{
    my ($self, $ra_insert_size, %h_param) = @_;

    my $minspan      = $self->_FastOneGet('_minspan');
    my $ratiominspan = $self->_FastOneGet('_ratiominspan');
    $ratiominspan = &MAXINT if (!defined($ratiominspan) || $ratiominspan eq '');
    $self->_Print("TagWeakLinks minspan=$minspan rationminspan=$ratiominspan\n");
    my $o_linkds   = $self->_FastOneGet('_o_linkds');
    my @a_links    = $o_linkds->GetLinks();
    my %h_res      = ();
    my $nb_removed = 0;
    foreach my $link (@a_links)
    {
        my $linkdetails = $o_linkds->Get($link);
        $linkdetails =~ s/##$//;
        my @a_ = split("##", $linkdetails);
        Carp::confess("TagWeakLinks: at that stage, ambiguities must be solved\n") if ($#a_);
        my $c       = 0;
        my $spansrc = 0;
        my $spantgt = 0;
        my $srclen  = 0;
        my $tgtlen  = 0;
        %h_res = split(/[=\s]+/, $a_[0]);
        next if (defined($h_res{ovllen}) && $h_res{ovllen});    # 20130202, un lien avec chev ne peut pas etre weak
        $c += $h_res{count};
        $spansrc = &Max($spansrc, $h_res{srcend} - $h_res{srcstart} + 1);
        $spantgt = &Max($spantgt, $h_res{tgtend} - $h_res{tgtstart} + 1);
        $srclen  = $h_res{srclen};
        $tgtlen  = $h_res{tgtlen};

        if (exists($h_param{gaplen_vs_insert}) && $h_param{gaplen_vs_insert})
        {
# je cherche les gap courts qui sont predits par des insert longs et qui auraient du etre predits par des inserts plus courts
# par contre je ne veux pas masquer des insert plus grand (e.g BES)
            foreach my $insert_size (sort {$a <=> $b} @$ra_insert_size)
            {
#if ( $h_res{gaplen} < $insert_size && $h_res{insert_size} > $insert_size && $h_res{insert_size} <= $$ra_insert_size[$#$ra_insert_size] )
                if (   $h_res{gaplen_max} < $insert_size
                    && $h_res{insert_size} > $insert_size
                    && $h_res{insert_size} <= $$ra_insert_size[$#$ra_insert_size])
                {
                    $self->_Print( "TagWeakLinks INSERT tagging $link:  $h_res{gaplen_max} < $insert_size   $h_res{insert_size}\n");
                    $h_res{weak} = 1;
                    $o_linkds->SetKV($link, \%h_res);
                    $nb_removed++;
                    next;
                }
            }
        }

        # pour les petits inserts la contrainte sur le minspan est trop dure
        if (exists($h_param{minspan}) && $h_param{minspan} && $h_res{insert_size} > (2 * $h_param{minspan}))
        {
            if (($spansrc < $minspan || $spantgt < $minspan))
            {
				$self->_Print( "TagWeakLinks MINSPAN tagging $link:  (srcspan=$spansrc < $minspan || tgtspan=$spantgt < $minspan)\n");
                $h_res{weak} = 1;
                $o_linkds->SetKV($link, \%h_res);
                $nb_removed++;
                next;
            }
        }

		# le span sur la plus petite ne peut pas etre plus important que sur la plus longue
        if (exists($h_param{ratiominspan}) && $h_param{ratiominspan} && $ratiominspan != &MAXINT)
        {
            my $to_remove = &FALSE;
            if ($srclen < $tgtlen)
            {
                $to_remove = &TRUE if ($spansrc > ($ratiominspan * $spantgt));
            }
            else
            {
                $to_remove = &TRUE if ($spantgt > ($ratiominspan * $spansrc));
            }
            if ($to_remove)
            {
                $self->_Print("TagWeakLinks RATIOMINSPAN tagging $link:  count=$c spansrc=$spansrc/$srclen spantgt=$spantgt/$tgtlen  span coverage too different\n");
                $h_res{weak} = 1;
                $o_linkds->SetKV($link, \%h_res);
                $nb_removed++;
            }
        }
    }
    $self->_Print(sprintf("TagWeakLinks done: $nb_removed links tagged / %d\n", $#a_links + 1));
    return $nb_removed;
}

# suppression des ambiguites d'orientation

sub CleanUpLinks
{
    my ($self, %h_param) = @_;

    my $o_linkds = $self->_FastOneGet('_o_linkds');
    $o_linkds->CleanUp(%h_param);

}

# pour une gamme de gaplen courts (e.g <2000), je ne garde que le successeurs
# qui a le meilleur span pour chaque gamme d'inserts et pour chaque extremitee

sub CleanUpMoreLinks
{
    my ($self, %h_param) = @_;

    $self->_PrintParams("CleanUpMoreLinks", %h_param);
    my $o_linkds = $self->_FastOneGet('_o_linkds');

    # suppression des ambiguites, pour une extremite je ne garde que le meilleur span
    my ($rh_graph) = $self->InitGraph(&TRUE, undef, &FALSE);    # le probleme c'est que je degage des pas weak et que je garde les weak

    my $gaplen_maskable = (exists($h_param{gaplen_maskable})) ? $h_param{gaplen_maskable} : $GAPLEN_MASKABLE;
    my $taghardlinks    = (exists($h_param{taghardlinks}))    ? $h_param{taghardlinks}    : 0;

    my @a_seqid = keys(%$rh_graph);
    foreach my $src (@a_seqid)
    {
        foreach my $srcext ('5', '3')
        {
            my @a_next = keys(%{$$rh_graph{$src}{$srcext}});
            next if (scalar(@a_next) == 1);
            my %h_max = ();

            # identification of the maximum span
            foreach my $next (@a_next)
            {
                foreach my $nextext ('5', '3')
                {
                    # c'est lourdingue mais je manque d'inspiration
                    my ($lsrc, $lsrcext, $lnext, $lnextext, $which) =
                        ($src lt $next)
                      ? ($src, $srcext, $next, $nextext, 'src')
                      : ($next, $nextext, $src, $srcext, 'tgt');
                    my %h_val = ();
                    next if (!$o_linkds->GetKV2Oriented($lsrc, $lsrcext, $lnext, $lnextext, \%h_val));
                    my ($srcspan) = $h_val{$which . "end"} - $h_val{$which . "start"};
                    my $insert_size = $h_val{insert_size};
                    if (!exists($h_max{$insert_size}{NT}) || $srcspan > $h_max{$insert_size}{NT})
                    {
                        $h_max{$insert_size}{NT} = $srcspan;
                        $h_max{$insert_size}{ID} = $next;
                    }
                }
            }

            my $best = '';
            foreach my $insert_size (keys(%h_max))
            {
                $best .= "##$h_max{$insert_size}{ID}##";
            }
            foreach my $next (@a_next)
            {
                if ($best !~ /##$next##/)
                {
                    my %h_res = ();
                    $o_linkds->GetKV2($src, $next, \%h_res);
                    next if (!%h_res);
                    next if (!$taghardlinks && $h_res{hardlink});
                    if ($h_res{gaplen_max} < $gaplen_maskable)    # pour des gap plus gros je ne peux rien dire
                           # en fait ca depend du nb de lib et de la distrib de la lib
                    {
                        $self->_Print("CleanUpMoreLinks: masking $src,$next ($best)\n");
                        $h_res{weak} = 1;
                        $o_linkds->SetKV2($src, $next, \%h_res);

                        #$o_linkds->Del($link);
                    }
                }
            }
        }
    }

    #$self -> _Set ('_rh_graph', undef);
}

# Migr
sub LoadLinks
{
    my ($self, %h_param) = @_;

    $self->_FastOneGet('_o_linkds')->Load(%h_param);

    my $o_linkds = $self->_FastOneGet('_o_linkds');

    # 20121216 je veux gerer les gap negatifs (=chev) je desactive
    if (0)
    {
        my @a_links = $o_linkds->GetLinks();

        foreach my $link (@a_links)
        {
            my %h_val = ();
            $o_linkds->GetKV($link, \%h_val);
            $h_val{gaplen_min} = 1 if ($h_val{gaplen_min} < 1);
            $o_linkds->SetKV($link, \%h_val);
        }
    }

}

sub RemoveUnexpectedOverlappingLinks
{
my ($self, %h_param) = @_;

    $self->_FastOneGet('_o_linkds')->RemoveUnexpectedOverlappingLinks(%h_param);
}

sub LinkDatasetStatistics
{
my ($self, %h_param) = @_;

    $self->_FastOneGet('_o_linkds')->InitStatistics(%h_param);
}

# Migr
sub GetNextScaffoldID
{
my ($self) = @_;

    my $max      = 0;
    my $nextid   = $self->_FastOneGet('_nextid');
    my $rh_scaff = $self->_FastOneGet('_rh_scaffolds');
	my $newnext  = $nextid+1;
	if ( defined($$rh_scaff{$newnext}) ) # en principe plus necessaire mais je le garde au cas ou
	{
		foreach my $id (keys(%$rh_scaff))
		{
			$max = &Max($max, $id);
		}
		$newnext = $max+1;
	}
	$self->_Set('_nextid',$newnext);
    return $nextid;
}

sub InitFromFasta
{
    my ($self) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    my $rh_contigs = $self->_FastOneGet('_rh_contigs');
    my $o_linkds   = $self->_FastOneGet('_o_linkds');
    $self->_Print("InitFromFasta started\n");

    my $cpt = 1;
    foreach my $seqds (@$ra_seqlist)
    {
        my @a_ids =
          sort {$$rh_o_seqds{$seqds}->GetLen($b) <=> $$rh_o_seqds{$seqds}->GetLen($a)}
          $$rh_o_seqds{$seqds}->GetSeqIds($seqds);
        foreach my $id (@a_ids)
        {
            my $seq = $$rh_o_seqds{$seqds}->GetSeq($id);
            if (!defined($seq) || $seq eq '')
            {
                Carp::confess("Error: unable to find the sequence of >$id<\n");
            }
            my @a_contigs = split(/N/, $seq);
            next if (scalar(@a_contigs) == 1);
            $$rh_o_seqds{$seqds}->Del($id);
            my $n       = length($#a_contigs);
            my $previd  = '';
            my $prevlen = 0;
            my $delta   = 0;
            my $path    = '';

            for (my $i = 0; $i < scalar(@a_contigs); $i++)
            {
                $delta++;
                if ($a_contigs[$i] eq '')
                {
                    next;
                }
                my $ctgid = sprintf("$id.%0${n}d", $i + 1);    # c'est important il faut que src lt tgt
                $$rh_o_seqds{$seqds}->Add($ctgid, \$a_contigs[$i]);
                my $ctglen = $$rh_o_seqds{$seqds}->GetLen($ctgid);
                if ($previd ne '')
                {
                    $delta--;
                    my $key = "src=$previd tgt=$ctgid";
                    my $val =
                      "srcext=3 srclen=$prevlen tgtext=5 tgtlen=$ctglen gaplen_max=$delta gaplen_min=$delta srcstart=$prevlen srcend=$prevlen tgtstart=1 tgtend=1 count=1 insert_size=$delta hardlink=0 weak=0 used=0 tgtspan=1 srcspan=1 nlib=1 ambiguous_orient=0";
                    $o_linkds->Add($key, $val);
                    $path .= " $delta,$delta #${ctgid}p$ctglen";
                    $delta = 0;
                }
                else
                {
                    $path = "#${ctgid}p$ctglen";
                }
                $previd  = $ctgid;
                $prevlen = $ctglen;
            }
            my $o_scaff = New Scaffold('name' => $cpt++, 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
            $o_scaff->InitPathFromString(\$path, $self, &FALSE);
            my $idscaff = $self->AddScaffold($o_scaff, &TRUE);
        }
    }
    $self->_Print("InitFromFasta done, " . ($cpt - 1) . " scaffolds loaded\n");
}

# Migr
sub LoadAnchors
{
    my ($self, %h_param) = @_;

    if (!exists($h_param{file}))
    {
        $self->_Print("LoadAnchors warning: no file found, skipped\n");
        return;
    }
    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    my $rh_contigs = $self->_FastOneGet('_rh_contigs');

    $self->_Print("LoadAnchors >$h_param{file}< started\n");
    $self->_Print("expected format: ctg len anchor1 anchor2 with anchor chr:pos or chr\n");
    my $f_in      = &GetStreamIn($h_param{file});
    my $cpt       = 0;
    my $markerlen = 0;

    #seq00939 OM_Chr2.mapsc:36516030:36621950:-
    while (my $l = <$f_in>)
    {
        chomp($l);
        my @a_ctg  = split(' ', $l);
        my $ctg    = shift(@a_ctg);
        my $ctglen = shift(@a_ctg);
        my $o_ctg  = $self->GetRegisteredContig($ctg, 'p', $ctglen);
        foreach my $anchors (@a_ctg)
        {
            $anchors =~ s/:$//;
            if ($anchors =~ /:/)
            {
                my @a_ = split(':', $anchors, 2);
                $o_ctg->SetAnchor($a_[0], $a_[1]);
            }
            else
            {
                $o_ctg->SetAnchor($anchors, '');
            }
        }

        #print STDERR Dumper($$rh_anchors{$ctg});
        $cpt++;
    }
    $f_in->close;
    $self->ScaffoldsAnchoring();
    $self->_Print("LoadAnchors: $cpt markers loaded, done\n");

    #print STDERR Dumper($rh_contigs);
}

# Migre
# Loaded after anchors
sub LoadValidatedPaths
{
    my ($self, $filename) = @_;

    my $rh_valpaths = $self->_FastOneGet('_rh_validatedpaths');
    my $rh_contigs  = $self->_FastOneGet('_rh_contigs');

    #seq01123 3 seq03366 3   1,1943  1       1082    #seq01123p24162 1,309 #seq38850m529 1,244 #seq03366m14430
    my $f_in = &GetStreamIn($filename);
    my @a_keys = sort {$b <=> $a} keys(%$rh_valpaths);
    my ($cpt) = (scalar(@a_keys)) ? ($a_keys[0] + 1) : 1;
    while (my $l = <$f_in>)
    {
        chomp($l);
        my @a_      = split(' ', $l, 8);
        my $newpath = $a_[7];
        my $o_scaff = New Scaffold('name' => $cpt, 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
        next if (!$o_scaff->InitPathFromString(\$newpath, $self, &TRUE));

#$self->UpdateContigRegistryAndAnchors($o_scaff); il ne faut pas le faire ici mais a l'utilisation, lorsqu'on les valide
        $$rh_valpaths{$cpt++} = $o_scaff;
    }
    $f_in->close;
    $self->_Print("LoadValidatedPaths: $cpt validated paths loaded\n");

    # 20120717 je ne suis pas sur que ce soit une bonne idee, je commente
    #print STDERR "Verifier que le TrimRepeats Dans le LoadValidatedPaths est une bonne idee\n";
    #$self->TrimRepeats("ValidatePaths",$rh_valpaths);
}

#
# supprime les contigs des extremites qui ne sont pas relies au suivant
# cela peut etre dangeureux si le suivant est petit
# mettre une protection lorsque contig est grand ? verifier s'il y a un lien indirect (mais il doit y en avoir un) ?
#

# Migr
sub TrimScaffolds
{
    my ($self) = @_;

    $self->_Print("TrimScaffolds started\n");
    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my @a_ids        = sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength();} keys(%$rh_scaffolds);
    my $cpt          = 0;
    my $cpt_edit     = 0;
    foreach my $idscaff (@a_ids)
    {
        my $nb_contigs_before_trimming = $$rh_scaffolds{$idscaff}->GetNumberOfContigs();
        $$rh_scaffolds{$idscaff}->Trim($o_linkds);
        my $nb_contigs_after_trimming = $$rh_scaffolds{$idscaff}->GetNumberOfContigs();
        if ($nb_contigs_after_trimming == 0)
        {
            $self->_Print("Trim " . $$rh_scaffolds{$idscaff}->GetName() . ": Deleting $idscaff\n");
            delete($$rh_scaffolds{$idscaff});
        }
        $cpt++;
        $cpt_edit += ($nb_contigs_before_trimming - $nb_contigs_after_trimming);
    }
    $self->_Print(sprintf("TrimScaffolds: post $cpt paths, $cpt_edit contigs removed\n"));
    $self->_Print("TrimScaffolds done\n");
}

# Migr
sub InitContig2Scaffold
{
    my ($self, $rh_contig2scaff, $r_max, $idscaff, $rh_paths) = @_;

    my $rh_scaffolds = (defined($rh_paths)) ? $rh_paths : $self->_FastOneGet('_rh_scaffolds');
    return if (!defined($rh_scaffolds));

    my $rh_contigs = $self->_FastOneGet('_rh_contigs');

    my @a_ids = ();
    if (defined($idscaff))
    {
        #		Carp::croak("InitContig2Scaffold NE PLUS UTILISER avec un seul scaffid\n"); plus maintenant
        @a_ids = ($idscaff);    # tres dangeureux de l'utiliser sur un seul scaff
    }
    else
    {
        %$rh_contig2scaff = ();
        @a_ids = sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength();} keys(%$rh_scaffolds);
    }

    # je veux synchroniser les contigs du repository avec les scaffolds
    # c'est important par exemple apres un dumpfasta ou il y a eu renumerotation des scaffods

    # le soucis c'est lorsque j'arrive avec un seul scaffold, je supprime le lien avec tous les scaffolds

#	foreach my $ctg (keys(%$rh_contigs))
#	{
#		if (! defined($$rh_contigs{$ctg}->GetScaffoldsIds()) )
#		{
#			my $anchors = $$rh_contigs{$ctg}->GetAnchors();
#			next if ( defined($anchors) && $anchors ne '' );
#			print STDERR "Warning: InitContig2Scaffold: ".$$rh_contigs{$ctg}->GetName()." associe a aucun scaffold\n";
#			print STDERR $$rh_contigs{$ctg}->GetName()."\n";
#			print STDERR Dumper($$rh_contigs{$ctg});
#			print STDERR "Cela arrive dans Scaffolding car on ne fait pas le Merge si l'un des Contigs est associe a plus de 1 scaff\n";
#			print STDERR "Je le sors du Repository\n";
#			delete($$rh_contigs{$ctg});
#		}
#else
#{
#	my @a_scafflist=split(' ',$$rh_contigs{$ctg}->GetScaffoldsIds());
#	foreach my $scaff (@a_scafflist)
#	{
#		$$rh_contigs{$ctg}->DelScaffold($scaff);
#	}
#}
#	}

    foreach my $idscaff (@a_ids)
    {
        my $ra_ctg = $$rh_scaffolds{$idscaff}->GetArrayContigs();
        for (my $i = 0; $i < scalar(@$ra_ctg); $i++)
        {
            my $id = $$ra_ctg[$i]->GetName();
            if (exists($$rh_contig2scaff{$id}) && $$rh_contig2scaff{$id} ne $idscaff)
            {
                # a contig (e.g hopefully a repeat) can appear several times
                # the scaffold code is the longest one
            }
            else
            {
                $$rh_contig2scaff{$id} = $idscaff;
                $$r_max = &Max($$r_max, $idscaff) if (defined($r_max));
            }

#			next if ( defined($rh_paths) ); # lorsque j'arrive avec $rh_paths c'est souvent avant d'avoir initialise le repository

            #if ( ref($$rh_contigs{$id}) ne 'Contig' )
            #
            #print STDERR Dumper($rh_contigs);
            #Carp::confess("InitContig2Scaffold ERROR: $id not found in the central Contig repository\n");
            #
            #lse
            #
            #$$rh_contigs{$id}->SetScaffold($idscaff,'');
            #
        }

        # je recupere les ancres
        #$self->UpdateContigRegistryAndAnchors($$rh_scaffolds{$idscaff});
    }

    #print STDERR Dumper($rh_contig2scaff);
}

sub GetRegisteredContig
{
    my ($self, $ctg, $strand, $len) = @_;

    my $rh_contigs = $self->_FastOneGet('_rh_contigs');
    if (!defined($$rh_contigs{$ctg}))
    {
        $$rh_contigs{$ctg} = New ContigRegistry('name' => $ctg, 'len' => $len);
    }
    return $$rh_contigs{$ctg}->GetMainContig($ctg, $strand, $len);
}

sub GetNewRegisteredContig
{
    my ($self, $ctg, $strand, $len) = @_;

    my $rh_contigs = $self->_FastOneGet('_rh_contigs');
    if (!defined($$rh_contigs{$ctg}))
    {
        $$rh_contigs{$ctg} = New ContigRegistry('name' => $ctg, 'len' => $len);
    }
    return $$rh_contigs{$ctg}->GetNewContigInstance($ctg, $strand, $len);
}

# supprime tous les contigs qui sont lies a des scaffolds qui n'existent plus

sub SynchronizeContigRegistry
{
    my ($self) = @_;

    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    return if (!defined($rh_contigs));

    foreach my $ctg (sort keys(%{$rh_contigs}))
    {
        #print STDERR "$ctg ";
        my $ra_ctg = $$rh_contigs{$ctg}->GetContigArray();
        if (!defined($ra_ctg)
            || ($$rh_contigs{$ctg}->GetNumberOfScaffolds() == 0 && $$rh_contigs{$ctg}->GetNumberOfAnchors() == 0))
        {
            delete($$rh_contigs{$ctg});
            next;
        }
        $$rh_contigs{$ctg}->SetRepeat();

        # premier passage je recupere les ancres chargees c'est a dire non affectees a des contigs)
        my %h_anch = ();
        for (my $i = 0; $i < scalar(@$ra_ctg); $i++)
        {
            next if (!defined($$ra_ctg[$i]));
            $h_anch{$$ra_ctg[$i]->GetAnchors()} = 1
              if (   $$ra_ctg[$i]->GetNumberOfScaffolds() == 0
                  && defined($$ra_ctg[$i]->GetAnchors())
                  && $$ra_ctg[$i]->GetAnchors() ne '');
        }
        my @a_ = keys(%h_anch);
        my $anchor =
          (scalar(@a_) == 1 && !$$rh_contigs{$ctg}->IsRepeat())
          ? shift(@a_)
          : undef;    # j'ignore les ancres sur des repeats
                      #print STDERR " anchor=$anchor\n";
        for (my $i = 0; $i < scalar(@$ra_ctg); $i++)
        {
            next if (!defined($$ra_ctg[$i]));

            #$$ra_ctg[$i]->RemoveTemporaryScaffolds();
            my $nbscaff = $$ra_ctg[$i]->GetNumberOfScaffolds();
            if ($nbscaff == 0)
            {
                if (!defined($$ra_ctg[$i]->GetAnchors()))
                {
                    delete($$ra_ctg[$i]);
                }
                next;
            }
            elsif ($nbscaff > 1)
            {
                print STDERR Dumper($$ra_ctg[$i]);
                Carp::confess(
                             "SynchronizeContigRegistry error: the number of scaffold per Contig instance must be 1\n");
            }
            $$ra_ctg[$i]->SetAnchor($anchor, "") if (defined($anchor) && !defined($$ra_ctg[$i]->GetAnchor($anchor)));
            my $scaff = $$ra_ctg[$i]->GetScaffoldsIds();
            if (!defined($scaff))
            {
                #print STDERR Dumper($$ra_ctg[$i]);
            }
            elsif (!exists($$rh_scaffolds{$scaff}))
            {
                #print STDERR "$scaff does not exist anymore, removing ".$$ra_ctg[$i]->Get()."\n";
                delete($$ra_ctg[$i]);
            }
            elsif (!$$rh_scaffolds{$scaff}->IsContainingContig($$ra_ctg[$i]))
            {
                print STDERR "$scaff does not contain " . $$ra_ctg[$i]->Get() . " anymore\n";
                delete($$ra_ctg[$i]);
            }

            #else
            #{
            #	print STDERR "$scaff found, keeping ".$$ra_ctg[$i]->Get()."\n";
            #}
        }
        delete($$rh_contigs{$ctg}) if (@$ra_ctg == 0);    # modifie le 20120816, je perds des ancres
    }

    #print STDERR Dumper($rh_contigs);
    $self->_Print("SynchronizeContigRegistry done\n");
}

# Migr
sub ScaffoldsAnchoring
{
    my ($self, $idscaff) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    my $rh_contigs = $self->_FastOneGet('_rh_contigs');

    my $rh_scaffolds          = $self->_FastOneGet('_rh_scaffolds');
    my $o_omap                = $self->_FastOneGet('_o_omap');
    my $o_linkds              = $self->_FastOneGet('_o_linkds');
    my $anchoring_consistency = $self->_FastOneGet('_anchoring_consistency');
    $anchoring_consistency = (!defined($anchoring_consistency) || $anchoring_consistency) ? &TRUE : &FALSE;

    my @a_ids = ();
    if (defined($idscaff))
    {
        @a_ids = ($idscaff);
    }
    else
    {
        @a_ids = sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength();} keys(%$rh_scaffolds);
    }

    foreach my $idscaff (@a_ids)
    {
        $$rh_scaffolds{$idscaff}->InitAnchor(&TRUE);    # je le fais avant au cas ou je puisse deja le faire
        next if (!$self->_FastOneGet('_ontheflyanchoring'));

        next if (!defined($rh_o_seqds) || $o_omap->GetNumberOfMaps() < 1);
        next if (defined($$rh_scaffolds{$idscaff}->GetAnchor()));
        next if ($$rh_scaffolds{$idscaff}->GetNumberOfContigs() < 1);

        my ($r_scaffseq, $rh_ctg_pos) =
          $$rh_scaffolds{$idscaff}->GetSequence($o_linkds, $rh_o_seqds, $ra_seqlist, &MIN_GAPSIZE);
        my @a_results = $o_omap->MapSequence($idscaff, $r_scaffseq, 'short', 1, &FALSE);
        next if (scalar(@a_results) == 0);

        my ($anchor, $start_seq, $end_seq, $start_omap, $end_omap, $strand) = split(/\t/, $a_results[0]);

        next if (!defined($anchor));

        $self->_Print("ScaffoldsAnchoring $idscaff: $anchor,$start_seq,$end_seq,$start_omap,$end_omap,$strand\n");
        my $ra_ctg = $$rh_scaffolds{$idscaff}->GetArrayContigs();
        for (my $i = 0; $i < scalar(@$ra_ctg); $i++)
        {
            my @a_pos = split(/\t/, $$rh_ctg_pos{$$ra_ctg[$i]->GetName()});
            if (scalar(@$ra_ctg) == 1
                || ($a_pos[0] >= $start_seq && $a_pos[1] <= $end_seq)
              )    # je ne veux ancrer que les contigs qui sont dans le hit avec l'om
                   # avec un cas particulier lorsqu'il n'y a qu'un contig qui ne serait jamais inclus
            {
                $self->_Print(  "ScaffoldsAnchoring $idscaff: contig="
                              . $$ra_ctg[$i]->GetName()
                              . " anchor=$anchor $start_omap:$end_omap:$strand\n");
                $$ra_ctg[$i]->SetAnchor($anchor, "$start_omap:$end_omap:$strand");
            }
        }
        $$rh_scaffolds{$idscaff}->InitAnchor(&TRUE);    # et je le fais apres au cas ou j'ai rajoute des infos
        $self->_Print("ScaffoldsAnchoring $idscaff: anchor=" . $$rh_scaffolds{$idscaff}->GetAnchor() . "\n");
    }
}

# Migr
sub AddScaffold
{
    my ($self, $o_scaff, $noname) = @_;

    my $prevname = $o_scaff->GetName();
    $o_scaff->Check();
    my $rh_scaff = $self->_FastOneGet('_rh_scaffolds');
    my $idscaff  = $self->GetNextScaffoldID();

    #$o_scaff->SetName($idscaff) if  ( !defined($noname) || $noname != &TRUE);
    $o_scaff->SetName($idscaff);
    $o_scaff->SetContigs($prevname);
    $$rh_scaff{$idscaff} = $o_scaff;
    $self->ScaffoldsAnchoring($idscaff);
    return $idscaff;
}

# Migr
sub GetAllScaffolds
{
    my ($self) = @_;

    my $rh_scaff = $self->_FastOneGet('_rh_scaffolds');
    return () if (!defined($rh_scaff));
    return %{$rh_scaff};
}

# Migr
#sub GetScaffolds
#{
#my($self,$idscaff)=@_;
#
#	my $rh_scaff = $self -> _FastOneGet ('_rh_scaffolds');
#	return undef  if ( !defined($rh_scaff) );
#	return $$rh_scaff{$idscaff};
#}

sub GetValidatedPaths
{
    my ($self) = @_;

    return $self->_FastOneGet('_rh_validatedpaths');
}

# Migr
sub TrimRepeats
{
    my ($self, $tag, $rh_paths) = @_;

    my $rh_scaffolds = (defined($rh_paths)) ? $rh_paths : $self->_FastOneGet('_rh_scaffolds');
    return if (!defined($rh_scaffolds));

    $self->_Print("TrimRepeats started\n");

    my %h_contig2scaff = ();
    $self->InitContig2Scaffold(\%h_contig2scaff, undef, undef, $rh_scaffolds);

# deuxieme passe, je supprime les sequences des extremites des scaffolds qui sont deja dans des scaffolds plus grands
# par contre je garde en double des sequences qui sont internes a deux scaffolds
# je veux aussi supprimer lorsque le scaffold est pratiquement englobe dans un autre, pour eviter les petits spe-gros pas spe-petits spe
    my $cpt   = 0;
    my $totnt = 0;
    $self->_Print("TrimRepeats/$tag: pre $cpt " . scalar(keys(%$rh_scaffolds)) . " paths\n");
    my $cpt_del = 0;
    foreach
      my $idscaff (sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength()} keys(%$rh_scaffolds))
    {
        $$rh_scaffolds{$idscaff}->TrimRepeats($tag, \%h_contig2scaff);
        if ($$rh_scaffolds{$idscaff}->GetNumberOfContigs() == 0)
        {
            delete($$rh_scaffolds{$idscaff});
            $cpt_del++;
            next;
        }
        $$rh_scaffolds{$idscaff}->InitAnchor(&TRUE);
        $totnt += $$rh_scaffolds{$idscaff}->GetLength();
        $cpt++;
    }
    $self->_Print(
                  sprintf("TrimRepeats/$tag: post $cpt paths remaining ($cpt_del deleted), totnt=$totnt, mean=%.2f\n",
                          $totnt / $cpt)
      )
      if ($cpt);
}

sub KeepOnlyOverlappingContigs
{
    my ($self, %h_param) = @_;

    $self->_Print("KeepOnlyOverlappingContigs started\n");

    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    my $rh_o_seqds   = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist   = $self->_FastOneGet('_ra_seqlist');
    my $o_linkds     = $self->_FastOneGet('_o_linkds');

    my %h_contig2scaff = ();
    $self->InitContig2Scaffold(\%h_contig2scaff, undef, undef, $rh_scaffolds);
    foreach
      my $idscaff (sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength()} keys(%$rh_scaffolds))
    {
        print STDERR "$idscaff idscaff ori: " . $$rh_scaffolds{$idscaff}->Get() . "\n";
        my $rh_new_scaff =
          $$rh_scaffolds{$idscaff}->KeepOnlyOverlappingContigs($o_linkds, $rh_o_seqds, $ra_seqlist, \%h_param);
        if (defined($rh_new_scaff))
        {
            foreach my $scaff (sort {$a <=> $b} keys(%$rh_new_scaff))
            {
                my $id = $self->AddScaffold($$rh_new_scaff{$scaff});
                $$rh_scaffolds{$id}->SetOverlapParams(1, 1, &MAXINT)
                  ;    # cette init c'est pour le GetSequence du Fasta post Keep
                print STDERR "added from $idscaff ==> scaff = $id\n";
                print STDERR "idscaff new: " . $$rh_scaffolds{$id}->Get() . "\n";
            }
        }

# ATTENTION, ATTENTION apres avoir calcule si overlap je supprime les seuils car dans le GetSequence je ne suis pas forcement sur la meme sequence
# et je dois etre plus tolerant (mais en princip je n'ai conserve que les chevauchements pertinents et je dois les retrouver)
        $$rh_scaffolds{$idscaff}->SetOverlapParams(1, 1, &MAXINT)
          ;    # cette init c'est pour le GetSequence du Fasta post Keep
    }
    $self->SynchronizeContigRegistry();
    $self->_Print(sprintf("KeepOnlyOverlappingContigs done\n"));
}

sub ResetGetSequenceRules
{
    my ($self) = @_;

    $self->_Print(sprintf("ResetGetSequenceRules started\n"));
    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    foreach
      my $idscaff (sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength()} keys(%$rh_scaffolds))
    {
        $$rh_scaffolds{$idscaff}->ResetGetSequenceRules();
    }
    $self->_Print(sprintf("ResetGetSequenceRules done\n"));
}

# Migre
# ca degrade les resultats en amont/pendant le scaffolding

sub CheckGaps
{
    my ($self, %h_param) = @_;

    my $src = (!defined($h_param{source})) ? 'unknown' : $h_param{source};
    $self->_Print("CheckGaps $src started - recompute gap lengths when possible\n");
    my $rh_scaff = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds = $self->_FastOneGet('_o_linkds');
    foreach my $idscaff (sort {$$rh_scaff{$b}->GetLength() <=> $$rh_scaff{$a}->GetLength()} keys(%$rh_scaff))
    {
        $$rh_scaff{$idscaff}->CheckGapsOnePath($o_linkds, &TRUE);
    }
    $self->_Print("CheckGaps $src done\n");
}

# Migr
sub AddRepeats
{
    my ($self, %h_param) = @_;

    $self->_Print("AddRepeats starting\n");
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance})) ? $h_param{gaplen_tolerance} : $gaplen_tolerance_ori;
    my $only_best = (exists($h_param{onlybest})) ? $h_param{onlybest} : &FALSE;
    my $rh_scaff         = $self->_FastOneGet('_rh_scaffolds');
    my $rh_contigs       = $self->_FastOneGet('_rh_contigs');
    my $nb_cpus          = $self->_FastOneGet('_nb_cpus');

    # premiere passe j'allonge
    my $o_linkds = $self->_FastOneGet('_o_linkds');
    my @a_weak = grep (/weak/, keys(%h_param));
    if (scalar(@a_weak) == 0)
    {
        $self->_Print("AddRepeats warning: minweak/maxweak not initialized => the default is to use weak links (can take a long time)\n");
    }
    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux lien (et plus au graphe) je suis oblige de travailler sur un temporaire filtre
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
	if ( $only_best )
	{
		my $rh_best = $self->_FastOneGet('_rh_bests');
    	my $ra_links_filt = $o_linkds_filt->GetLinksRef();
		printf STDERR "Pre Best Filter = %d\n",scalar(@$ra_links_filt);
		my %h_val=();
		my %h_key=();
		for(my $i=0;$i<scalar(@$ra_links_filt);$i++)
		{
			if ( $o_linkds_filt->GetKV($$ra_links_filt[$i],\%h_val) )
			{
				%h_key=split(/[=\s]/,$$ra_links_filt[$i]);
				if ( (&Exists($rh_best,$h_key{src},$h_val{srcext},'best') && $$rh_best{$h_key{src}}{$h_val{srcext}}{best} =~ /\b$h_key{tgt}\b/) || 
				     (&Exists($rh_best,$h_key{tgt},$h_val{tgtext},'best') && $$rh_best{$h_key{tgt}}{$h_val{tgtext}}{best} =~ /\b$h_key{src}\b/ ) )
				{
					;
					#print STDERR "$$ra_links_filt[$i] = BEST\n";
				}
				else
				{
					#print STDERR "$$ra_links_filt[$i] = NOTBEST, removed\n";
					$o_linkds_filt->Del($$ra_links_filt[$i]);
				}
			}
		}
	}
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();
	printf STDERR "Post Best Filter = %d\n",scalar(@$ra_links_filt) if ( $only_best );
    my ($rh_graph,$o_graph) = $self->InitGraph(&FALSE, $ra_links_filt, &TRUE);    # ici je ne veux dans le graphe que ce qui passe le seuil

# modifie le 20130331
#    my @a_filt_links = ();
#    $o_linkds->Filter(\@a_filt_links, %h_param);
# pour pouvoir faire le AddRepeat a chaque etape tout en beneficiant du cache InitGraph je travaille sur un graphe temporaire
    #my ($rh_graph, $o_graph) = $self->InitGraph(&FALSE, \@a_filt_links, &TRUE)
      ;    # je confirme que je veux utiliser les liens eventuellement filtres
           #my %h_contig2scaff=();
           #$self->InitContig2Scaffold(\%h_contig2scaff);


    $self->ScaffoldsAnchoring();

# je ne peux pas faire de controle d'ancrage a l'interieur car les problemes sont du a des repeats et pendant le processus
# d'longation je ne sais pas encore que ce sont des repeats

    my %sh_newpaths : shared = ();
    &Multithreading(\&SliceAddRepeats, $nb_cpus, $rh_scaff, \%sh_newpaths, $o_graph, $o_linkds_filt, $self, $gaplen_tolerance);

    #	foreach my $idscaff (sort { $$rh_scaff{$b}->GetLength() <=> $$rh_scaff{$a}->GetLength() } keys(%$rh_scaff))
    #	{
    #		$self->_Print("AddRepeats: $idscaff time=".time."\n");
    #		# attention sur Mt, gapplen_tolerance=200 c'est moins bien que 100
    #		my $o_longestnewpath = $o_graph->FindLongestPath($$rh_scaff{$idscaff},$o_linkds,$self,$gaplen_tolerance,undef);
    #		if (defined($o_longestnewpath))
    #		{
    #			$$rh_scaff{$idscaff}=$o_longestnewpath;
    #			$$rh_scaff{$idscaff}->Check();
    #		}
    #	}

	my $error = 0;
    foreach my $idscaff (keys(%sh_newpaths))
    {
        next if (!defined($sh_newpaths{$idscaff}));
		my $prevlen = $$rh_scaff{$idscaff}->GetLength(&TRUE); # longueur avec maxgaplen
        $$rh_scaff{$idscaff}->Reset();
        $$rh_scaff{$idscaff}->InitPathFromString(\$sh_newpaths{$idscaff}, $self, &TRUE);
		my $newlen  = $$rh_scaff{$idscaff}->GetLength(&FALSE); # longueur sans gaplen
		if ( $newlen > $prevlen )
		{
			$self->_Print("AddRepeat: error in length: $idscaff: prevlen (with gaps) = $prevlen; new length (without gaps) = $newlen\n");
			$error++;
		}
    }

	if ( $error )
	{
		#Carp::croak("ERROR in AddRepeats: $error sequences are too long compared to expected length\n");
		$self->_Print("ERROR in AddRepeats: $error sequences are too long compared to expected length\n");
	}

    $self->ScaffoldsAnchoring();
    $self->TrimRepeats("AddRepeats");
    $self->SynchronizeContigRegistry();
    $self->_Print("AddRepeats done\n");
}

sub SliceAddRepeats
{
    my ($i, $ra_tasks, $rh_scaff, $sh_results, @a_param) = @_;

    my ($o_graph, $o_linkds, $o_scaffolder, $gaplen_tolerance) = @a_param;
    my $task = $$ra_tasks[$i];
    my $av   = time;
    my $o_longestnewpath = $o_graph->FindLongestPathString($$rh_scaff{$task}, $o_linkds, $o_scaffolder, $gaplen_tolerance, undef);
    if (defined($o_longestnewpath))
    {
        $$sh_results{$task} = $o_longestnewpath;    # c'est une string
    }
    printf STDERR "SliceAddRepeats: $task %d s\n", time - $av;
}

# Migr non teste
# desormais renvoie un Scaffold
sub GetLongestValidatedPath
{
    my ($self, $ctg) = @_;

    my $vpath    = undef;
    my $rh_vpath = $self->GetValidatedPaths();
    foreach my $path (sort {$$rh_vpath{$b}->GetLength() <=> $$rh_vpath{$a}->GetLength()} keys(%$rh_vpath))
    {
        return $$rh_vpath{$path} if ($$rh_vpath{$path}->IsContainingContig($ctg));
    }
    return $vpath;
}

#
# Chargement des scaffolds/liens
# Migr
sub LoadScaffolds
{
    my ($self, $filename) = @_;

    my $rh_contigs            = $self->_FastOneGet('_rh_contigs');
    my $anchoring_consistency = $self->_FastOneGet('_anchoring_consistency');
    $self->_Print("LoadScaffolds starting\n");
    my %h_newscaff  = ();
    my $f_in        = &GetStreamIn($filename);
    my $cpt         = 0;
    my $cpttmp      = 0;
    my $lastprevctg = '';
    while (my $l = <$f_in>)
    {
        chomp($l);
        $l =~ s/\s*$//;

        #Scaff05902 previd=23972 nb_ctg=3 nt_len=2873 lst=#seq34875p1471 1,2245 #seq65314m331 1,1 #seq40064p1033
        if ($l =~ /^ScaffNEW.+lst=(.+)\s*/)
        {
            $cpttmp++;
            $h_newscaff{"999999999$cpttmp"} = $1;
        }
        elsif ($l =~ /Scaff0*(\d+).+nb_ctg=(\d+).+lst=(.+)\s*/)
        {
            my ($id, $nbctg, $lst) = ($1, $2, $3);
            next if ($nbctg < 2);
            $h_newscaff{$id} = $lst;
        }
        elsif ($l !~ /^\s*$/)
        {
            Carp::confess("LoadScaffolds: unexpected format >$l<\n");
        }
    }
    $self->InitScaffolds();

    foreach my $id (sort {$a <=> $b} keys(%h_newscaff))
    {
        my $o_scaff = New Scaffold('name' => $id, 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
        next if (!$o_scaff->InitPathFromString(\$h_newscaff{$id}, $self, $anchoring_consistency));
        my $setname = ($id =~ /^999999999/) ? &TRUE : &FALSE;
        my $idscaff = $self->AddScaffold($o_scaff, $setname);
        $cpt++;
        if ($id =~ /^999999999/)
        {
            $self->_Print("LoadScaffolds adding NEW scaffold $idscaff = " . $o_scaff->Get() . "\n");
        }
    }
    $f_in->close;
    my %h_param = ('source' => 'LoadScaffolds');

    #$self->CheckGaps(%h_param); ca degrade
    $self->SynchronizeContigRegistry();
    $self->_Print("LoadScaffolds done. $cpt scaffolds loaded\n");

    #my %h_contig2scaff=();
    #$self->InitContig2Scaffold(\%h_contig2scaff);
    #my $rh_scaffolds = $self->_FastOneGet ('_rh_scaffolds');
    #print STDERR Dumper($rh_scaffolds);
}

sub DumpAnchors
{
    my ($self, $outfile) = @_;

    my $rh_contigs = $self->_FastOneGet('_rh_contigs');
    $self->_Print("DumpAnchors started, prefix=$outfile\n");
    my $f_anchors = &GetStreamOut("$outfile");
    foreach my $ctg (sort keys(%{$rh_contigs}))
    {
        my $all_anchors = $$rh_contigs{$ctg}->GetAnchors();
        next if (!defined($all_anchors));
        my @a_anchors = split(' ', $all_anchors);
        next if (!scalar(@a_anchors));
        print $f_anchors $$rh_contigs{$ctg}->GetName() . "\t" . $$rh_contigs{$ctg}->GetLength();
        foreach my $anchor (@a_anchors)
        {
            print $f_anchors " $anchor:" . $$rh_contigs{$ctg}->GetAnchor($anchor);
        }
        print $f_anchors "\n";
    }
    $f_anchors->close;
    if (-e $outfile && -z $outfile)
    {
        $self->_Print("DumpAnchors: removing empty anchor file\n");
        unlink($outfile);
    }
    $self->_Print("DumpAnchors done\n");
}

sub DumpRepeats
{
    my ($self, $outfile) = @_;

    $self->_Print("DumpRepeats started, prefix=$outfile\n");
    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    my $rh_contigs = $self->_FastOneGet('_rh_contigs');
    $self->SynchronizeContigRegistry();
    my $f_fa = &GetStreamOut("$outfile");
    foreach my $creg (sort {$$rh_contigs{$b}->GetLength() <=> $$rh_contigs{$a}->GetLength()} keys(%$rh_contigs))
    {
        next if ($$rh_contigs{$creg}->GetNumberOfScaffolds() < 2);
        my $id = $$rh_contigs{$creg}->GetName();
        print $f_fa ">$id len="
          . $$rh_contigs{$creg}->GetLength()
          . " numscaff="
          . $$rh_contigs{$creg}->GetNumberOfScaffolds()
          . " scaff="
          . $$rh_contigs{$creg}->GetScaffoldsIds() . "\n";
        my $seq = '';
        foreach my $seqds (@$ra_seqlist)
        {
            last if ($seq = $$rh_o_seqds{$seqds}->GetSeq($id));
        }
        print $f_fa &FormatSeq(\$seq) . "\n";

    }
    $f_fa->close;
}

#	Fasta dump
#
# il y a des overlap inferieurs a la longueur de kmer utilisee pour contiguer,
# je rajoute une heuristique pour fermer ces trous
# pour le moment je ne mets pas de protection sur la taille des gaps car je ne sais pas encore bien la calculer
# et des mots exacts de 30nt ca ne doit pas etre si frequent
# Migr

sub DumpFasta
{
    my ($self, $outfile, $finalize, $dumpgff3, %h_param) = @_;

    $self->_Print("DumpFasta started, prefix=$outfile\n");
    my $rh_o_seqds       = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist       = $self->_FastOneGet('_ra_seqlist');
    my $rh_scaff         = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds         = $self->_FastOneGet('_o_linkds');
    my $rh_contigs       = $self->_FastOneGet('_rh_contigs');
    my $gaplen_tolerance = $self->_FastOneGet('_gaplen_tolerance');
    my $minmingaplen     = $self->_FastOneGet('_mingaplen');
    my %h_seqiddeja      = ();
    my %h_seq            = ();                                 # je charge en memoire pour pouvoir trier sur la longueur
    my $lenid            = length(scalar(keys(%$rh_scaff)));
    my $cpt              = 0;
    my $f_info           = &GetStreamOut("$outfile.scaff");

    my $f_gff3 = undef;
    if (defined($dumpgff3) && $dumpgff3)
    {
        $f_gff3 = &GetStreamOut("$outfile.scaff.gff3");
    }

    $self->SynchronizeContigRegistry();

    #print STDERR Dumper($rh_scaff);
    if (defined($finalize) && $finalize)
    {
       #		$self->CheckGaps(source=>"DumpFasta - finalize"); # ca degrade pendant le scaffolding mais a la fin je le fais
    }

    my %h_newscaff = ();

    # comme j'ai pris la taille max d'insert, il faudrait que je diminue la taille des gaps

    $self->_Set('_ontheflyanchoring', &FALSE);
    $self->ScaffoldsAnchoring();
    $self->_Set('_ontheflyanchoring', &TRUE);

    my $nb_overlaps  = 0;
    my $nb_gaps      = 0;
    my %h_anchor     = ();
    my %h_anchor_sum = ();

    my $nb_cpus = $self->_FastOneGet('_nb_cpus');

    my %sh_sequence : shared = ();
    if ($finalize || defined($h_param{minovlpci}) || defined($h_param{minovllen}) || defined($h_param{maxovlovh}))
    {
        foreach my $idscaff (sort {$$rh_scaff{$b}->GetLength() <=> $$rh_scaff{$a}->GetLength()} keys(%$rh_scaff))
        {
            $$rh_scaff{$idscaff}->_Set('_fast_getsequence', &FALSE);
            $$rh_scaff{$idscaff}->SetOverlapParams($h_param{minovlpci}, $h_param{minovllen}, $h_param{maxovlovh});
        }
    }
    my $avtime = time;
    &Multithreading(
                    \&SliceGetSequenceScaffold, $nb_cpus,    \%$rh_scaff, \%sh_sequence,
                    $o_linkds,                  $rh_o_seqds, $ra_seqlist, $minmingaplen,
                    $f_gff3,                    $gaplen_tolerance
                    );
    printf STDERR "Multithreading / SliceGetSequenceScaffold: time=%d\n", time - $avtime;

    foreach my $idscaff (sort {$$rh_scaff{$b}->GetLength() <=> $$rh_scaff{$a}->GetLength()} keys(%$rh_scaff))
    {
        my $ra_ctg = $$rh_scaff{$idscaff}->GetArrayContigs();
        my $nb_ctg = $$rh_scaff{$idscaff}->GetNumberOfContigs();
        my $newid  = '';
        if ($nb_ctg < 1)
        {
            delete($$rh_scaff{$idscaff});
            next;
        }
        $cpt++;
        if ($nb_ctg == 1)
        {
            $newid = $$ra_ctg[0]->GetName();
        }
        else
        {
            $newid = sprintf("Scaff%0" . ${lenid} . "d", $cpt);
        }
        my $anchor = $$rh_scaff{$idscaff}->GetAnchor();
        $anchor = "unk" if (!defined($anchor));
        $h_anchor{$newid} = $anchor;
        if ($nb_ctg > 1)
        {
#my ($r_seq,$rh_ctg_pos) = $$rh_scaff{$idscaff}->GetSequence($o_linkds,$rh_o_seqds,$ra_seqlist,$minmingaplen,$f_gff3,$gaplen_tolerance);
#$h_seq{$newid} = $$r_seq;
            my $path;
            ($path, $h_seq{$newid}) = split("\t", $sh_sequence{$idscaff});
            print $f_info "$newid previd=$idscaff nb_ctg=$nb_ctg nt_len="
              . length($h_seq{$newid})
              . " chr=$anchor lst=$path\n";
            for (my $i = 0; $i < scalar(@$ra_ctg); $i++)
            {
                $h_seqiddeja{$$ra_ctg[$i]->GetName()} = 1;
            }
            $h_anchor_sum{$anchor} += length($h_seq{$newid});
        }
        $$rh_scaff{$idscaff}->DeregisterContigs();    # meme si le scaff reste le meme, le id changera
        $h_newscaff{$cpt} = $$rh_scaff{$idscaff};
        $h_newscaff{$cpt}->SetName($cpt);
        $h_newscaff{$cpt}->SetContigs($$rh_scaff{$idscaff});
        if ($finalize)                                # je remets par defaut
        {
            $$rh_scaff{$idscaff}->_Set('_fast_getsequence', &TRUE);
            $$rh_scaff{$idscaff}->ResetGetSequenceRules();
        }
    }
    $f_info->close;
    $self->_Set('_rh_scaffolds', \%h_newscaff);

#$self->_Print(sprintf("DumpFasta: $nb_overlaps overlaps / $nb_gaps gaps = %.2f percent\n",$nb_overlaps*100/$nb_gaps)) if ( $nb_gaps );
    my $f_fa = &GetStreamOut($outfile);
    foreach my $seqid (sort {length($h_seq{$b}) <=> length($h_seq{$a})} keys(%h_seq))
    {
        my $len = length($h_seq{$seqid});
        print $f_fa ">$seqid len=$len chr=$h_anchor{$seqid}\n" . &FormatSeq(\$h_seq{$seqid}) . "\n";
    }

    foreach my $seqds (@$ra_seqlist)
    {
        my @a_ids =
          sort {$$rh_o_seqds{$seqds}->GetLen($b) <=> $$rh_o_seqds{$seqds}->GetLen($a)}
          $$rh_o_seqds{$seqds}->GetSeqIds($seqds);
        foreach my $id (@a_ids)
        {
            next if (exists($h_seqiddeja{$id}));
            my $seq = $$rh_o_seqds{$seqds}->GetSeq($id);
            if (!defined($seq) || $seq eq '')
            {
                Carp::confess("Error: unable to find the sequence of >$id<\n");
            }
            my $anchor =
              (    defined($$rh_contigs{$id})
                && defined($$rh_contigs{$id}->GetAnchors())
                && $$rh_contigs{$id}->GetAnchors() ne '') ? $$rh_contigs{$id}->GetAnchors() : "unk";
            print $f_fa ">$id len=" . length($seq) . " chr=$anchor\n" . &FormatSeq(\$seq) . "\n";
            $h_anchor_sum{$anchor} += length($seq);
        }
    }
    $f_fa->close;
	if ( scalar(keys(%h_anchor_sum)) > 1 )
	{
    	my $f_summary = &GetStreamOut("$outfile.scaff.anchoring");
		foreach my $id (sort keys(%h_anchor_sum))
		{
			print $f_summary "$id\t$h_anchor_sum{$id}\n";
		}
		$f_summary->close;
	}

    $f_gff3->close if (defined($f_gff3));
    $self->_Print("DumpFasta done, prefix=$outfile\n");

}

sub SliceGetSequenceScaffold
{
    my (
        $i,          $ra_tasks,   $rh_scaff,     $rsh_sequence, $o_linkds,
        $rh_o_seqds, $ra_seqlist, $minmingaplen, $f_gff3,       $gaplen_tolerance
      )
      = @_;

    my $idscaff = $$ra_tasks[$i];
    my ($r_seq, $rh_ctg_pos) =
      $$rh_scaff{$idscaff}->GetSequence($o_linkds, $rh_o_seqds, $ra_seqlist, $minmingaplen, $f_gff3, $gaplen_tolerance);
    $$rsh_sequence{$idscaff} = $$rh_scaff{$idscaff}->Get() . "\t" . $$r_seq;

    # il faudrait que je serialize aussi rh_ctg_pos mais je ne l'utilise pas pour le moment dans DumpFasta
}

# je cherche a valider le scaffold, non seulement avec un lien direct mais aussi via des contigs non adjacents
# Migr

sub ValidateScaffoldsIndirect
{
    my ($self, %h_param) = @_;

    print STDERR "ATTENTION: CODE A VERIFIER EN COMPARANT AVEC KEEP...\n";
    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    return if (!defined($rh_scaffolds));

    my $cutmode    = (exists($h_param{cutmode})    && $h_param{cutmode})    ? &TRUE                : &FALSE;
    my $validation = (exists($h_param{validation}) && $h_param{validation}) ? $h_param{validation} : 2;
    my $maxlen     = (exists($h_param{maxlen})     && $h_param{maxlen})     ? $h_param{maxlen}     : &MAXINT;
    my $noinit = (!exists($h_param{noinit})) ? &FALSE : &TRUE;

    $self->_Print("ValidateScaffoldsIndirect started $cutmode,$validation,$maxlen\n");
    my ($rh_graph) = $self->InitGraph(&FALSE, undef, &TRUE) if (!$noinit);    # je veux les weak
    my @a_ids = sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength();} keys(%$rh_scaffolds);

    #my %h_contig2scaff=();
    #$self->InitContig2Scaffold(\%h_contig2scaff);
    $self->ScaffoldsAnchoring();
    my $rh_new_scaff = undef;
    foreach my $idscaff (@a_ids)
    {
        my $rh_new_scaff =
          $$rh_scaffolds{$idscaff}->ValidateIndirect($o_linkds, $cutmode, $validation, $maxlen, $noinit);
        if (defined($rh_new_scaff) && $cutmode)
        {
            foreach my $scaff (keys(%$rh_new_scaff))
            {
                my $id = $self->AddScaffold($$rh_new_scaff{$scaff});

#				print STDERR "ValidateScaffoldsIndirect:/Split/Added ".$$rh_scaffolds{$id}->GetName()." ".$$rh_scaffolds{$id}->Get()."\n";
            }

#			print STDERR "ValidateScaffoldsIndirect:/Split/Remaining ".$$rh_scaffolds{$idscaff}->GetName()." ".$$rh_scaffolds{$idscaff}->Get()."\n";
        }

# je n'ai pas besoin de le faire car dans indirect je ne ralonge pas $self->ScaffoldsAnchoring($idscaff);
#$self->_Print("ValidateScaffoldsIndirect: ".$$rh_scaffolds{$idscaff}->GetName()." ".$$rh_scaffolds{$idscaff}->Get()."\n");
    }
    $self->_Print("ValidateScaffoldsIndirect done\n");
}

#
# je cherche un lien direct entre les contigs du chemin
# Migr
#

sub ValidateScaffoldsDirect
{
    my ($self, %h_param) = @_;

    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    return if (!defined($rh_scaffolds));

    my $o_linkds   = $self->_FastOneGet('_o_linkds');
    my $rh_contigs = $self->_FastOneGet('_rh_contigs');

    my $cutmode = (exists($h_param{cutmode}) && $h_param{cutmode}) ? &TRUE : &FALSE;
    my $length_decrease_tolerance =
        (exists($h_param{length_decrease_tolerance}) && $h_param{length_decrease_tolerance})
      ? ($h_param{length_decrease_tolerance} / 100)
      : 0.8;
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance})) ? $h_param{gaplen_tolerance} : $gaplen_tolerance_ori;
    my $validation       = (exists($h_param{validation}) && $h_param{validation}) ? $h_param{validation} : 2;
    my $maxlen           = (exists($h_param{maxlen}) && $h_param{maxlen}) ? $h_param{maxlen} : &MAXINT;

    $self->_Print("ValidateScaffoldsDirect started\n");
    my ($rh_graph, $o_graph) = $self->InitGraph(&FALSE, undef, &TRUE);    # je veux les weak
                                                                          #my %h_contig2scaff=();
                                                                          #$self->InitContig2Scaffold(\%h_contig2scaff);
    $self->ScaffoldsAnchoring();

    my @a_ids = sort {$$rh_scaffolds{$b}->GetLength() <=> $$rh_scaffolds{$a}->GetLength();} keys(%$rh_scaffolds);
    my $rh_new_scaff = undef;
    foreach my $idscaff (@a_ids)
    {
        my $rh_new_scaff =
          $$rh_scaffolds{$idscaff}
          ->ValidateDirect($o_graph, $o_linkds, $self, $cutmode, $validation, $maxlen, $gaplen_tolerance,
                           $length_decrease_tolerance);
        if (defined($rh_new_scaff) && $cutmode)
        {
            foreach my $scaff (keys(%$rh_new_scaff))
            {
                my $id = $self->AddScaffold($$rh_new_scaff{$scaff});

#				print STDERR "ValidateScaffoldsDirect:/Split/Added ".$$rh_scaffolds{$id}->GetName()." ".$$rh_scaffolds{$id}->Get()."\n";
            }

#			print STDERR "ValidateScaffoldsDirect:/Split/Remaining ".$$rh_scaffolds{$idscaff}->GetName()." ".$$rh_scaffolds{$idscaff}->Get()."\n";
        }
        $self->ScaffoldsAnchoring($idscaff)
          ;    # je le fais dans tous les cas car je peux avoir rallonge ou change le chemin
    }
    $self->SynchronizeContigRegistry();
    $self->_Print("ValidateScaffoldsDirect done\n");
}

# Migr: type=Scaffold
sub _SelectNonRedundantPaths
{
    my ($ra_path) = @_;

    my @a_nr = ();
    my @a_paths = sort {$b->GetLength() <=> $a->GetLength()} @$ra_path;
    foreach my $path (@a_paths)
    {
        my ($last) = $path->Get() =~ /(\S+)$/;

        # le debut etant identique
        # je regarde si le dernier contig fait parti d'un chemin plus long deja conserve
        for (my $i = 0; $i < scalar(@a_nr); $i++)
        {
            if (!$a_nr[$i]->IsContainingContig($last))
            {
                push(@a_nr, $path);
                last;
            }
        }
    }
    @$ra_path = @a_nr;
}

# Migr
# je valide dans les deux sens, a voir si je ne le fais que dans un ==> j'ai change je ne le fais que dans un sens
# j'aurais moins de redondance a virer

sub ValidatePaths
{
    my ($self, $outfile, %h_param) = @_;

    my $rh_o_seqds           = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist           = $self->_FastOneGet('_ra_seqlist');
    my $o_linkds             = $self->_FastOneGet('_o_linkds');
    my $rh_contigs           = $self->_FastOneGet('_rh_contigs');
    my $rh_valpaths          = $self->_FastOneGet('_rh_validatedpaths');
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance})) ? $h_param{gaplen_tolerance} : $gaplen_tolerance_ori;
    my $maxmingaplen = (exists($h_param{postprocess_maxgaplen_min})) ? $h_param{postprocess_maxgaplen_min} : &MAXINT;
    Carp::confess("gaplen_tolerance=$gaplen_tolerance<\n") if ($gaplen_tolerance eq '');

    $self->_PrintParams("ValidatePaths", %h_param);
    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux lien (et plus au graphe) je suis oblige de travailler sur un temporaire filtre
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();

#modifie le 20130331
#    my @a_filt_links = ();
#    $o_linkds->Filter(\@a_filt_links, %h_param);

    my $useweaks = (exists($h_param{maxweak})) ? $h_param{maxweak} : &FALSE;

    my ($rh_graph, $o_graph) = $self->InitGraph(&TRUE, undef, $useweaks)
      ;    # ici c'est fait expres que les liens soient filtres mais que le graph contienne tout
    my $f_path = &GetStreamOut($outfile);

    # il me faut les longueurs de sequences pour calculer la longueur du chemin

    my %h_validators = ();    # Scaffold

    my $key = '';
    my $cpt = 0;

    #@$ra_links_filt = ("src=seq1737 tgt=seq1899");

    #$self->ScaffoldsAnchoring();

    my @a_keys = sort {$b <=> $a} keys(%$rh_valpaths);
    my ($cpt_path) = (scalar(@a_keys)) ? ($a_keys[0] + 1) : 1;

    $self->_Print("ValidatePaths: $cpt_path-1 previously validated\n");

    my $nb_proc   = $self->_FastOneGet('_nb_cpus');
    my %sh_vpaths : shared;
	&Multithreading(\&SliceValidatePathThread, $nb_proc, $ra_links_filt, \%sh_vpaths, $o_linkds,$o_graph,$self,$gaplen_tolerance,$maxmingaplen);

    $cpt = scalar(keys(%sh_vpaths));

    foreach my $scaff (keys(%sh_vpaths))
    {
        my $o_scaff =
          New Scaffold('name' => "VPATH$scaff", 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
        next if (!$o_scaff->InitPathFromString(\$sh_vpaths{$scaff}, $self, &TRUE));
        my $ra_ctg = $o_scaff->GetArray();
        my ($src, $srcext, $next, $nextext) = $$ra_ctg[0]->ConvertPathLink2GraphLink($$ra_ctg[$#$ra_ctg], &FALSE);
        my %h_res = ();
        next if (!$o_linkds->GetKV2Oriented($src, $srcext, $next, $nextext, \%h_res));
        $h_res{hardlink} = 2;
        $o_linkds->SetKV2($src, $next, \%h_res);
        $$rh_valpaths{$cpt_path} = $o_scaff;
        $cpt_path++;
    }

    my $av = time;
    $self->_Print("ValidatePaths raw $cpt/$#$ra_links_filt+1\n");

    # je vais identifier les repeats et marquer les hardlinks
    # repeats, ce sont des contigs qui ont plusieurs couples de voisins immediats
    my %h_lstcontigs = ();
    my %h_neighbours = ();
    my $prev_src     = '';
    my $nb_same      = 0;
    my @a_path       = ();
    my @a_allpaths   = ();

    #foreach my $val (sort keys(%h_validators))
    foreach my $val (sort keys(%$rh_valpaths))
    {
        my ($src) = $$rh_valpaths{$val}->Get() =~ /^(\S+)/;
        if (!defined($src))
        {
            $self->_Print(  "Warning: the validator path >$val< is not in the expected format "
                          . $$rh_valpaths{$val}->Get()
                          . "\n");
            next;
        }
        if ($src ne $prev_src)
        {
            &_SelectNonRedundantPaths(\@a_path) if ($#a_path > 0);
            push(@a_allpaths, @a_path);
            @a_path = ();
        }
        push(@a_path, $$rh_valpaths{$val});
        $prev_src = $src;
    }
    &_SelectNonRedundantPaths(\@a_path) if ($#a_path > 0);
    push(@a_allpaths, @a_path);

    my $ap = time;

    print STDERR "# raw vpaths: " . scalar(@a_allpaths) . " " . ($ap - $av) . "s\n";

    # il faut encore supprimer les paths redondants
    @a_allpaths = sort {$b->GetLength() <=> $a->GetLength()} @a_allpaths;

    my @a_nrpaths    = ();
    my %h_nrstrpaths = ();
    for (my $i = 0; $i < scalar(@a_allpaths); $i++)
    {
        my $str_path = $a_allpaths[$i]->Get();
        my $found    = &FALSE;
        for (my $j = 0; $j < scalar(@a_nrpaths); $j++)
        {
            if ($h_nrstrpaths{$j} =~ /$str_path/)
            {
                $found = &TRUE;
                delete($a_allpaths[$i]);
                last;
            }
        }
        if (!$found)
        {
            my $j = scalar(@a_nrpaths);
            $h_nrstrpaths{$j} = $str_path;
            push(@a_nrpaths, $a_allpaths[$i]);
        }
    }

    my $ap2 = time;

    print STDERR "# nr vpaths: " . scalar(@a_nrpaths) . " " . ($ap2 - $ap) . "s\n";

    # il doit rester de la redondance mais sans doute pas trop
    #print STDERR "ValidatePaths #paths apres suppression de la redondance $#a_nrpaths\n";

    my %h_neigbours = ();
    my $nb_hardlink = 0;

    # les liens qui sont valides (externes) sont taggues hardlink=2
    # les liens internes qui ont servis a la validation sont taggues hardlink=1
    foreach my $path (sort @a_nrpaths)
    {
        #print STDERR "ValidatePaths: $path\n";
        my $ra_ctg = $path->GetArrayContigs();
        for (my $i = 0; $i < scalar(@$ra_ctg) - 1; $i++)
        {
            my ($src, $srcext, $next, $nextext) = $$ra_ctg[$i]->ConvertPathLink2GraphLink($$ra_ctg[$i + 1], &FALSE);
            my $srckey  = "$src $srcext";
            my $nextkey = "$next $nextext";
            $h_neigbours{$srckey} .= "$nextkey="
              if (!exists($h_neigbours{$srckey}) || $h_neigbours{$srckey} !~ /$nextkey=/);
            $h_neigbours{$nextkey} .= "$srckey="
              if (!exists($h_neigbours{$nextkey}) || $h_neigbours{$nextkey} !~ /$srckey=/);
            $h_lstcontigs{$next} = 1;
            $h_lstcontigs{$src}  = 1;
            #
            my %h_res = ();
            next if (!$o_linkds->GetKV2Oriented($src, $srcext, $next, $nextext, \%h_res));
            next if ($h_res{hardlink} >= 1);
            $nb_hardlink++;
            $h_res{hardlink} = 1;
            $o_linkds->SetKV2($src, $next, \%h_res);
        }
    }
    $self->_Print("# hardlinks=$nb_hardlink\n");

    # il faudrait que je fasse par gamme de gap
    if (0)    # je desactive je ne m'en sert pas
    {
        my $f_repeats = &GetStreamOut("$outfile.repeats.fa.gz");
        foreach my $ctg (sort keys(%h_lstcontigs))
        {
            next if (!exists($h_neigbours{"$ctg 5"}) || !exists($h_neigbours{"$ctg 3"}));
            chop($h_neigbours{"$ctg 5"});    # pour le =
            chop($h_neigbours{"$ctg 3"});
            my @a_5 = split('=', $h_neigbours{"$ctg 5"});
            my @a_3 = split('=', $h_neigbours{"$ctg 3"});
            next if ($#a_5 == 0 || $#a_3 == 0);    # un repeat doit avoir des voisins differents des deux cotes
            foreach my $seqds (@$ra_seqlist)
            {
                my $seq = $$rh_o_seqds{$seqds}->GetSeq($ctg);
                next if (!defined($seq) || $seq eq '');
                print $f_repeats ">$ctg-$#a_5-$#a_3 len=" . length($seq) . "\n" . &FormatSeq(\$seq) . "\n";
            }
        }
        $f_repeats->close;
    }
    foreach my $id (keys(%$rh_valpaths))
    {
        delete($$rh_valpaths{$id});
    }
    for (my $i = 1; $i <= scalar(@a_nrpaths); $i++)
    {
        $$rh_valpaths{$i} = $a_nrpaths[$i - 1];
        my $prevname = $$rh_valpaths{$i}->GetName();
        $$rh_valpaths{$i}->SetName($i);

        #		$$rh_valpaths{$i}->SetContigs($prevname); # ca a l'air bon, les contigs sont bien attaches aux scaffolds
        my $str_path = $$rh_valpaths{$i}->Get();
        my $ra_ctg   = $$rh_valpaths{$i}->GetArrayContigs();
        my ($src, $srcext, $tgt, $tgtext) = $$ra_ctg[0]->ConvertPathLink2GraphLink($$ra_ctg[$#$ra_ctg], &FALSE);
        my $key = "$src $srcext $tgt $tgtext";
        print $f_path "$key\t$$rh_graph{$src}{$srcext}{$tgt}{$tgtext}\t1\t"
          . $$rh_valpaths{$i}->GetInternalLength(&TRUE)
          . "\t$str_path\n";
    }
    $f_path->close;
    $self->_Print("ValidatePaths nr=" . scalar(@a_nrpaths) . "\n");

#print STDERR "VERIFIER que c'est une bonne idee le TrimRepeats a la fin de ValidatePaths\n";
#print STDERR "#vpaths=".scalar(keys(%$rh_valpaths))."\n";
#$self->TrimRepeats("ValidatePaths",$rh_valpaths);
#je ne le fais plus car a ce stade il y a des chemins chevauchants et les sequences dupliques ne sont pas forcement des repeats
#print STDERR Dumper($rh_valpaths);
}


sub SliceValidatePathThread
{
	#&Multithreading(\&SliceValidatePathThread, $nb_proc, $ra_links_filt, \%sh_vpaths, $o_linkds,$o_graph,$self,$gaplen_tolerance,$maxmingaplen);

    my ($i,$ra_tasks,$ra_links,$rsh_vpaths, $o_linkds, $o_graph, $o_scaffolder, $gaplen_tolerance, $maxmingaplen) = @_;

    # on ne les voit pas tous car ils sont ecrits en meme temps
    #print STDERR "SliceValidatePathThread: ($i,$ra_links,$rsh_vpaths, $o_linkds, $o_graph, $o_scaffolder, $gaplen_tolerance, $maxmingaplen)\n";

    my $rh_graph = $o_graph->GetGraph();

	my $link = $$ra_tasks[$i];

	#print STDERR "$link\n";

	#next if ($link !~ /seq01310.+seq27843/ ); # DEBUG
	my %h_val = ();
	return &FALSE if (!$o_linkds->GetKV($link, \%h_val));

	#JG 20120816 je desactive car je prefere le controler via le filtre
	#next if ( $h_val{weak} );
	#next if ( $h_val{hardlink} );
	my %h_key = split(/[=\s]+/, $link);

	my ($src, $srcext, $next, $nextext) = ($h_key{src}, $h_val{srcext}, $h_key{tgt}, $h_val{tgtext});
	my $key = "$src $srcext $next $nextext";

	# 201207 pourquoi je ne mets pas la taille du gap je sais qu'il existe
	my $gap = (&Exists($rh_graph, $src, $srcext, $next, $nextext)) ? $$rh_graph{$src}{$srcext}{$next}{$nextext} : undef;
	my $o_newpath = $o_graph->ValidateOneGraphPath( $o_linkds, $o_scaffolder, $src, $srcext, $next, $nextext, $gap, $gaplen_tolerance, &TRUE);

	return &FALSE if (!defined($o_newpath));

	my $o_longestnewpath = undef;
	$o_longestnewpath = $o_graph->FindLongestPath($o_newpath, $o_linkds, $o_scaffolder, $gaplen_tolerance, undef);    # ATTENTION 0 avant le 31/01/2012
	$o_newpath = $o_longestnewpath->Clone() if (defined($o_longestnewpath));

	return &FALSE if (!$o_newpath->IsAnchoringConsistent());

	#print STDERR Dumper($o_newpath);

	my ($gaplen) = $o_newpath->GetInternalLength(&FALSE); # lorsque je verifie la coherence je ne prend pas en compte les gaps
	my ($lmin, $lmax) = split(',', $gap);
	$lmax += (($lmax * $gaplen_tolerance) / 100);
	$lmin -= (($lmin * $gaplen_tolerance) / 100);
	my $ok = ($gaplen >= $lmin && $gaplen <= $lmax) ? &TRUE : &FALSE;

#	print STDERR "$ok = ($gaplen >= $lmin && $gaplen <= $lmax) $gap $gaplen_tolerance\n";

	return &FALSE if (!$ok);

	# apres la recherche du chemin le plus long c'est peu probable qu'il reste de tres long gaps
	my $suspicious_gaplen = &FALSE;
	my $ra_path           = $o_newpath->GetArray();
	for (my $i = 1; $i < scalar(@$ra_path); $i += 2)
	{
		my ($min, $max) = split(',', $$ra_path[$i]);
		if ($min > $maxmingaplen)
		{
			$suspicious_gaplen = &TRUE;
			last;
		}
	}
	if ($suspicious_gaplen)
	{
		print STDERR "Suspicious GAPLEN at " . $o_newpath->Get() . "\n";
		my $anchor_first = $$ra_path[0]->GetAnchorsFromRegistry();
		my $anchor_last  = $$ra_path[scalar(@$ra_path) - 1]->GetAnchorsFromRegistry();
		print STDERR $$ra_path[0]->Get()
		  . " anc=$anchor_first "
		  . $$ra_path[scalar(@$ra_path) - 1]->Get()
		  . " anc=$anchor_last \n";
		if (defined($anchor_first) && defined($anchor_last) && $anchor_first eq $anchor_last)
		{
			print STDERR "\tAnchors consistents, I kept the link\n";
			$suspicious_gaplen = &FALSE;
		}
		else
		{
			print STDERR "\tRejected\n";
		}
	}

	return &FALSE if ($suspicious_gaplen);

	my %h_res = ();
	return &FALSE if (!$o_linkds->GetKV2Oriented($src, $srcext, $next, $nextext, \%h_res));
	$o_newpath->SetName($i);
	$$rsh_vpaths{$i} = $o_newpath->Get();
	return &TRUE;
}

# Migr
sub AddNonRedundantPaths
{
    my ($self, $rh_paths, $tag, %h_param) = @_;

    my $rh_scaff       = $rh_paths;
    my $rh_scaffolds   = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds       = $self->_FastOneGet('_o_linkds');
    my %h_contig2scaff = ();
    my $idscaff;
    $self->InitContig2Scaffold(\%h_contig2scaff);    # updating mode

    #	$self->ScaffoldsAnchoring();

    my $o_lscaff = New Scaffolder('gaplen_tolerance'    => $self->_FastOneGet('_gaplen_tolerance'),
                                  'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
    if (defined($h_param{valid}) && $h_param{valid})
    {
        $o_lscaff->SetLinks($o_linkds);
        $o_lscaff->InitGraph(&TRUE, undef, &FALSE);
    }
    my $seedlen  = 0;
    my $nb_seeds = 0;
    foreach my $path (sort {$$rh_paths{$b}->GetLength() <=> $$rh_paths{$a}->GetLength()} keys(%$rh_paths))
    {
        #print STDERR "============ $path ============\n";
        #print STDERR "PRE=".$$rh_paths{$path}->GetName()." ".$$rh_paths{$path}->Get()."\n";
        my $ra_ctg      = $$rh_paths{$path}->GetArrayContigs();
        my $nb_ctg      = $$rh_paths{$path}->GetNumberOfContigs();
        my $firstctglen = $$ra_ctg[0]->GetLength();
        my $lastctglen  = $$ra_ctg[$nb_ctg - 1]->GetLength();
        if (   defined($h_param{minendpathlen})
            && $h_param{minendpathlen}
            && ($firstctglen < $h_param{minendpathlen} || $lastctglen < $h_param{minendpathlen}))
        {
            $$rh_paths{$path}->DeregisterContigs();
            next;
        }

       # je ne considere un scaffold que si aucun des contigs du path n'est deja dans un autre scaffold deja selectionne
        my $dup = undef;
        for (my $i = 0; $i < $nb_ctg; $i++)
        {
            if (exists($h_contig2scaff{$$ra_ctg[$i]->GetName()}))
            {
                $dup = $h_contig2scaff{$$ra_ctg[$i]->GetName()};
                #$self->_Print("Duplication found: ".$$ra_ctg[$i]->GetName()." found in $dup\n");
                last;
            }
        }
        if (defined($dup))
        {
            $$rh_paths{$path}->DeregisterContigs();
            next;
        }

        my $o_scafftmp = undef;
        if (defined($h_param{valid}) && $h_param{valid})
        {
            $o_lscaff->Reset();
            $o_scafftmp = $$rh_paths{$path}->Clone();
            $o_lscaff->AddScaffold($o_scafftmp);
            $o_lscaff->ValidateScaffoldsIndirect(%h_param)
              ;    # pourquoi je fais ce controle, pour les vpath ca n'a pas de sens, pour le microscaff plus
                   # meme si marginal on dirait que c'est utilise meme pour vpath, ca doit etre une histoire de seuil
            my %h_lscaff = ();
            %h_lscaff = $o_lscaff->GetAllScaffolds();
            my $nb = scalar(keys(%h_lscaff));

            if ($nb > 1)    # en cas de doute je ne garde pas
            {
                foreach my $scaff (keys(%h_lscaff))
                {
                    $h_lscaff{$scaff}->DeregisterContigs();
                }
                $self->_Print("$tag/Candidate scaffold cannot be validated : " . $$rh_paths{$path}->Get() . "\n");
                next;
            }
        }
        $o_scafftmp->DeregisterContigs();
        $o_lscaff->Reset();

        #print STDERR "POST=".$$rh_paths{$path}->GetName()."\n";

        my $len = $$rh_paths{$path}->GetLength();
        $self->_Print("$tag/Adding new validated path: " . $$rh_paths{$path}->Get() . " \n");

        #print STDERR Dumper($$rh_paths{$path});
        my $prevname = $$rh_paths{$path}->GetName();    # avant le convert

        #$self->UpdateContigRegistryAndAnchors($$rh_paths{$path},&FALSE);

        if (!$$rh_paths{$path}->IsAnchoringConsistent())
        {
            $$rh_paths{$path}->DeregisterContigs();
            $self->_Print("Warning: vpath rejected, inconsistent anchoring " . $$rh_paths{$path}->Get() . "\n");
            next;
        }

        $idscaff = $self->AddScaffold($$rh_paths{$path});

        #print STDERR "newname=$idscaff, prevname=$prevname\n";
        $self->InitContig2Scaffold(\%h_contig2scaff, undef, $idscaff);    # updating mode
                                                                          #$self->ScaffoldsAnchoring($idscaff);

        $seedlen += $len;
        $nb_seeds++;
    }

#$self->InitContig2Scaffold(\%h_contig2scaff);	# je remet tout au propre, car des contigs pourraient etre enregistres sur des scaffolds non conserves
    $self->ScaffoldsAnchoring();
    $self->SynchronizeContigRegistry();
    $self->_Print(
                  sprintf("$tag/AddNonRedundantPath $self seed len=$seedlen, mean=%d, nb seeds=$nb_seeds\n",
                          ($nb_seeds) ? ($seedlen / $nb_seeds) : 0)
                          );
}

# les repeats pourraient etre mieux traites si j'arrivais a gerer a ce stade la des contigs (internes) a plusieurs scaffolds
# pour le moment je ne fais rien car je n'arrive pas bien a savoir qui est repeat ou pas et j'ai peur de rajouter de la
# redondance artefactuellement

sub ScaffoldingValidatedPaths
{
    my ($self, %h_param) = @_;

    $h_param{cutmode} = 1     if (!exists($h_param{cutmode}));
    $h_param{maxlen}  = 50000 if (!exists($h_param{maxlen}));    # transmis a ValidateScaffoldsIndirect
    $h_param{minendpathlen} = 0
      if (!exists($h_param{minendpathlen}));                     # longueur min des deux extremites du chemin a evaluer
    $h_param{noinit} = 1;                                        # pour ValidateScaffoldsIndirect
    $h_param{valid}  = &TRUE;
    $self->_PrintParams("ScaffoldingValidatedPaths starting", %h_param);
    my $rh_valpaths = $self->_FastOneGet('_rh_validatedpaths');

    $self->AddNonRedundantPaths($rh_valpaths, 'ScaffoldingValidatedPaths', %h_param);

    # je ne le fais pas je n'arrive pas a valider
    #my %h_param2 = ('source'=>'ScaffoldingValidatedPaths');
    #$self->CheckGaps(%h_param2);
    $self->_Print("ScaffoldingValidatedPaths done\n");
}

sub RemovingRedundantContigs
{
    my ($self, %h_param) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    $self->_PrintParams("RemovingRedundantContigs", %h_param);
    my $rh_contigs = $self->_FastOneGet('_rh_contigs');
    my $o_linkds   = $self->_FastOneGet('_o_linkds');
    my $nb_cpus    = $self->_FastOneGet('_nb_cpus');
    $h_param{minovllen} = $h_param{minovllen} || 20;
    $h_param{minovlpci} = $h_param{minovlpci} || 95;
    $h_param{maxovlovh} = $h_param{maxovlovh} || 5;

    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux lien (et plus au graphe) je suis oblige de travailler sur un temporaire
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();
    my ($rh_graph,$o_localgraph) = $self->InitGraph(&FALSE, $ra_links_filt, &TRUE);    # ici je ne veux dans le graphe que ce qui passe le seuil

#modifie le 20130331
#    my @a_links_filt = ();
#    my $nb_ignored = $o_linkds->Filter(\@a_links_filt, %h_param);
#    my ($rh_graph, $o_localgraph) = $self->InitGraph(&FALSE, \@a_links_filt, &TRUE);

    my @a_ctg = sort {$$rh_graph{$a}{len} <=> $$rh_graph{$b}{len}} keys(%$rh_graph);
    printf STDERR "RemovingRedundantContigs: %d contigs in overlapping links\n", scalar(@a_ctg);

    if (scalar(@a_ctg) == 0)
    {
        print STDERR "RemovingRedundantContigs: stopped, no overlapping links\n";
        return;
    }
    my %h_redundantcontigs = ();
    my %h_val              = ();
    my (@a_pred, @a_succ);
    my %h_blacklist = ();
    for (my $i = 0; $i < scalar(@a_ctg); $i++)
    {
        my $code = "OK";
        my ($ctg, $ctglen) = ($a_ctg[$i], $$rh_graph{$a_ctg[$i]}{len});
        @a_pred = ();
        @a_succ = ();
        $o_localgraph->GetLongerNeighbours($ctg, 5, \@a_pred, \%h_redundantcontigs);
        $o_localgraph->GetLongerNeighbours($ctg, 3, \@a_succ, \%h_redundantcontigs);
        my $found = &FALSE;
        for (my $i = 0; $i < scalar(@a_pred) && !$found; $i++)
        {

            for (my $j = 0; $j < scalar(@a_succ) && !$found; $j++)
            {
                # je commence sans prendre en compte l'orientation
                if ($o_linkds->GetKV2($a_pred[$i], $a_succ[$j], \%h_val) && $h_val{ovlpci})
                {
                    $code = "$a_pred[$i] $a_succ[$j] ovllen=$h_val{ovllen} ovlpci=$h_val{ovlpci}";
                    $h_val{hardlink} = 1.5;
                    $o_linkds->SetKV2($a_pred[$i], $a_succ[$j], \%h_val);
                    $found = &TRUE;
                    $h_blacklist{$ctg} = 1;
                    print STDERR "RemovingRedundantContigs\t$i\t$ctg\t$ctglen\t$code\n";
                }
            }
        }
    }
    my $prefix = "lynx-RemovingRedundantContigs." . time . ".$$.nrctgs";

    # je supprime les contigs blacklistes
    $self->_Print("\tRemoving blacklisted contigs\n");
    my $f_fa = &GetStreamOut("$prefix.fa.gz");
    foreach my $seqds (@$ra_seqlist)
    {
        my @a_ids =
          sort {$$rh_o_seqds{$seqds}->GetLen($b) <=> $$rh_o_seqds{$seqds}->GetLen($a)}
          $$rh_o_seqds{$seqds}->GetSeqIds($seqds);
        foreach my $id (@a_ids)
        {
            next if (exists($h_blacklist{$id}));
            my $seq = $$rh_o_seqds{$seqds}->GetSeq($id);
            if (!defined($seq) || $seq eq '')
            {
                Carp::confess("Error: unable to find the sequence of >$id<\n");
            }
            print $f_fa ">$id len=" . length($seq) . "\n" . &FormatSeq(\$seq) . "\n";
        }
    }
    $f_fa->close;
    $self->_Print("\tRemoving links related to blacklisted contigs\n");
    my $ra_links_raw = $o_linkds->GetLinksRef();
    my $nb_raw       = scalar(@$ra_links_raw);
    my %h_key        = ();
    for (my $i = 0; $i < $nb_raw; $i++)
    {
        %h_key = split(/[=\s]/, $$ra_links_raw[$i]);
        if (exists($h_blacklist{$h_key{src}}) || exists($h_blacklist{$h_key{tgt}}))
        {
            $o_linkds->Del($$ra_links_raw[$i]);
        }
    }
    $o_linkds->Dump('dumpfile' => "$prefix.links.gz");
    $self->_Print("RemovingRedundantContigs done");
}



sub AlignLinks
{
    my ($self, %h_param) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    $self->_PrintParams("AlignLinks", %h_param);
    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $nb_cpus      = $self->_FastOneGet('_nb_cpus');

	my %h_param_tmp = %h_param;
	foreach my $k ('minovlpci','minovllen','maxovlovh')
	{
		delete($h_param_tmp{$k}) if ( exists($h_param_tmp{$k}) );
	}
    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux lien (et plus au graphe) je suis oblige de travailler sur un temporaire
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param_tmp);
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();

# modifie le 20130331
#    my @a_links_filt = ();
#    my $nb_ignored   = $o_linkds->Filter(\@a_links_filt, %h_param);

    my %h_val        = ();
    my %h_key        = ();

    #	my @a_links = @a_links_filt;
    #	my ($f_out, $multifasta) = tempfile();

    my $minpci      = $h_param{minovlpci} || 99;
    my $minlen      = $h_param{minovllen} || 20;
    my $maxoverhang = $h_param{maxovlovh} || 5;

    $self->_Print("HSP Filters = minpci=$minpci, minlen=$minlen, maxoverhang=$maxoverhang\n");

    my $nb_links      = scalar(@{$ra_links_filt});
    my $time_av_aaln5 = time;

    # au moins utile pour le debugging
    #$h_param{precomputedaalnfile} = "/home/gouzy/misc/test/lynx-AlignLinks.1354472235.79626.aaln.gz";
    my @sa_hits : shared = ();
    if (defined($h_param{precomputedaalnfile}))
    {
        my $f_aaln = &GetStreamIn("$h_param{precomputedaalnfile}");
        @sa_hits = <$f_aaln>;
        chomp(@sa_hits);
        $f_aaln->close;
        printf STDERR "precomputedaalnfile=$h_param{precomputedaalnfile} loaded, %d alignements\n", scalar(@sa_hits);
    }
    else
    {
        &Multithreading(\&SliceAlignLinks, $nb_cpus, $ra_links_filt, \@sa_hits, $o_linkds, $rh_o_seqds, $ra_seqlist);
    }

    my $time_ap_aaln5 = time;

    my $linkpersecond = $nb_links / (($time_ap_aaln5 - $time_av_aaln5) + 1);    # +1 au cas ou trop rapide!
    my $linkperhour = $linkpersecond * 3600;
    $self->_Print(
                  sprintf(
                          "AlignLinks: #cpu=$nb_cpus, #links=$nb_links, elapsed time(s)="
                            . ($time_ap_aaln5 - $time_av_aaln5)
                            . " %.4f links/s ==> %d links/h\n",
                          $linkpersecond, $linkperhour
                    )
                    );

    my $prefix         = "lynx-AlignLinks." . time . ".$$";
    my $f_outtrace     = &GetStreamOut("$prefix.trace.gz");
    my %h_hitsaaln     = ();
    my %h_links_compat = ();
    for (my $i = 0; $i < $nb_links; $i++)
    {
        my $link = $$ra_links_filt[$i];
        next if (!$o_linkds->GetKV($link, \%h_val));
        next if ($sa_hits[$i] eq '');
        &HSPSet::AALN::ParseAALNLine(\$sa_hits[$i], 1, \%h_hitsaaln);
        my $status = "COMPAT";
        $status = "NOK" if ($h_hitsaaln{pci} < $minpci);
        $status = "NOK" if ($h_hitsaaln{len} < $minlen);

# dans l'alignment je fais en sorte de reorienter pour exposer les extremites => le chev est sur fin de query et debut de target
        $status = "NOK" if ($h_hitsaaln{qend} < $h_val{srclen} - $maxoverhang);  # && $h_hitsaaln{qstart} > $maxoverhang
        $status = "NOK" if ($h_hitsaaln{tstart} > $maxoverhang); # && $h_hitsaaln{tend} < $h_val{tgtlen}-$maxoverhang );
        print $f_outtrace "$link\t" . $o_linkds->Get($link) . "\t$sa_hits[$i]\t$status\n";
        if ($status eq 'COMPAT')
        {
            $h_links_compat{$link} = "$h_hitsaaln{len} $h_hitsaaln{pci}";
        }
    }
    $f_outtrace->close;

    my $ra_links_raw = $o_linkds->GetLinksRef();
    my $nb_raw       = scalar(@$ra_links_raw);
    for (my $i = 0; $i < $nb_raw; $i++)
    {
        next if (!$o_linkds->GetKV($$ra_links_raw[$i], \%h_val));
        ($h_val{ovllen}, $h_val{ovlpci}) = (0, 0);
        if (defined($h_links_compat{$$ra_links_raw[$i]}))
        {
            ($h_val{ovllen}, $h_val{ovlpci}) = split(' ', $h_links_compat{$$ra_links_raw[$i]});
            $h_val{hardlink} = 1;
        }
        $o_linkds->SetKV($$ra_links_raw[$i],\%h_val);
		$o_linkds->TagDubiousLink($$ra_links_raw[$i]);
    }

    $o_linkds->Dump('dumpfile' => "$prefix.overlap.gz");

    #
    # pour le momment je vais faire le filtre ici comme ca les deux tableaux sont synchro
    #
    # je stocke les hits
    my $aalnhits = "$prefix.aaln.gz";
    my $f_out2   = &GetStreamOut($aalnhits);
    for (my $i = 0; $i < scalar(@sa_hits); $i++)
    {
        print $f_out2 "$sa_hits[$i]\n";
    }
    $f_out2->close;
    ## je les filtre
    #my %h_param_aaln = %h_param;
    #$h_param_aaln{infile}=$aalnhits;
    #$h_param_aaln{query_col}=1;
    #$h_param_aaln{outfile}="$prefix.m8.gz";
    #my $o_hspsettmp = New HSPSet::AALN(%h_param_aaln);
    #$o_hspsettmp->Load(%h_param_aaln);
    #$o_hspsettmp->PrintInfo();
    #$o_hspsettmp->DumpM8(&GetStreamOut($h_param_aaln{outfile}));

    $self->_Print("AlignLinks done");

    return "$prefix.overlap.gz";
}

use constant ALIGN_MAXLEN => 1000;
use constant BATCH_SIZE   => 10000;

# c'est trop stressant pour le filesystem lorsque les contigs sont petits et qu'il y en a beaucoup
# je vais faire des lots

sub SliceAlignLinks
{
    my ($i, $ra_tasks, $ra_inputs, $sa_hits, $o_linkds, $rh_o_seqds, $ra_seqlist) = @_;

    return if (($i % &BATCH_SIZE) != 0);    # je ne fais rien si ce n'est pas un multiple de &BATCH_SIZE
    my $seqfile  = "/tmp/lynx-align-links.$$.$i.fa";
    my $f_seq    = &GetStreamOut($seqfile);
    my %h_srclen = ();
    my $cpt      = 0;
    for (my $ij = $i; $ij < ($i + &BATCH_SIZE); $ij++)
    {
        my %h_val = ();
        my %h_key = ();
        last if (!exists($$ra_tasks[$ij]));
        $cpt++;
        my $link = $$ra_tasks[$ij];
        %h_key = split(/[=\s]+/, $link);
        %h_val = ();
        next if (!$o_linkds->GetKV($link, \%h_val));
        my ($srcstrand, $tgtstrand) = split(' ', $H_EXT2STRAND{"$h_val{srcext} $h_val{tgtext}"});
        my $srcseq = &LookupSeq($rh_o_seqds, $ra_seqlist, $h_key{src}, $srcstrand);
        my $tgtseq = &LookupSeq($rh_o_seqds, $ra_seqlist, $h_key{tgt}, $tgtstrand);
        my ($srcoffset, $tgtoffset) = (&FALSE, &FALSE);

        if ($h_val{srclen} > &ALIGN_MAXLEN)
        {
            $h_srclen{$ij} = $h_val{srclen};
            $srcseq = substr($srcseq, -1 * &ALIGN_MAXLEN);
        }
        if ($h_val{tgtlen} > &ALIGN_MAXLEN)
        {
            $tgtseq = substr($tgtseq, 0, &ALIGN_MAXLEN);
        }
		$srcseq =~ s/\s//g;
		$tgtseq =~ s/\s//g;
		$srcseq =~ s/[^ATGC]/N/ig;
		$tgtseq =~ s/[^ATGC]/N/ig;
        print $f_seq ">$h_key{src}$srcstrand$h_val{srclen}\n$srcseq\n>$h_key{tgt}$tgtstrand$h_val{tgtlen}\n$tgtseq\n";
    }
    $f_seq->close;
	system("aaln5 $seqfile --gi 30 --gx 5 | grep \\<  > $seqfile.aaln5.gi30");
	system("aaln5 $seqfile --gi 100 --gx 5 | grep \\< > $seqfile.aaln5.gi100"); # je le fais deux fois car sur les chevauchements longs il faut penaliser fortement les mismatchs sinon on a des hits longs mais pci faible

	# je charge en premier le moins stringent
	my $f_hit  = new IO::File("$seqfile.aaln5.gi30");
	my @a_hits = ();
	while (my $hit = <$f_hit>)
	{
		chomp($hit);
		push(@a_hits, $hit);
	}
    $f_hit->close;

	# j'essaye de remplacer le hit si j'ai un meilleur pci
	$f_hit  = new IO::File("$seqfile.aaln5.gi100");
	my $j      = 0;
	while (my $hit = <$f_hit>)
	{
		my @a_     = split(' ',$hit);
		my @a_prev = split(' ',$a_hits[$j]);
		if ( $a_prev[0] ne $a_[0] || $a_prev[3] ne $a_[3] )
		{
			Carp::croak("AlignLinks error, inconsistent number of lines between $seqfile.aaln5.gi30 and $seqfile.aaln5.gi100\n");
		}
		if ( $a_[7] > $a_prev[7] )
		{
			chomp($hit);
			$a_hits[$j]=$hit;
		}
		$j++;
	}

    if (scalar(@a_hits) != $cpt)
    {
        print STDERR "Warning: ===== $i ===== " . scalar(@a_hits) . " rows, " . &BATCH_SIZE . " expected\n";
        for (my $j = 0; $j < &BATCH_SIZE; $j++)
        {
            last if (!defined($$ra_tasks[$i + $j]));
            $$sa_hits[$i + $j] = '';
        }
    }
    else
    {
        for (my $j = 0; $j < &BATCH_SIZE; $j++)
        {
            my $ij = $i + $j;
            last if (!defined($$ra_tasks[$ij]));
            my @a_ = split(' ', $a_hits[$j]);
            if (defined($h_srclen{$ij}))
            {
                $a_[1] += ($h_srclen{$ij} - &ALIGN_MAXLEN);
                $a_[2] += ($h_srclen{$ij} - &ALIGN_MAXLEN);
            }
            $a_hits[$j] = join(' ', @a_);
            $$sa_hits[$ij] = $a_hits[$j];
        }
    }
    unlink $seqfile;
}

# 1. identify contigs highly connected
# 2. encapsulate repeats between two Neighbours
#            c1 - R - c2  and c1 - c2

sub InitRepeats
{
my ($self, %h_param) = @_;

	$self->_Print("InitRepeats started\n");
	my %rh_repeats = ();
    my ($rh_graph) = $self->InitGraph(&FALSE, undef, &TRUE);    # je fais un graphe avec les weak
    my $nb_removed = 0;
	my @a_contigs = keys(%{$rh_graph}); # je ne vais pas trier pour pouvoir echantillonner plutot que de tout analyser
	my %h_repeats =();
	my @a_lengths=();
	for(my $i=0;$i<scalar(@a_contigs);$i++)
	{
		push(@a_lengths,$$rh_graph{$a_contigs[$i]}{len});
	}
	use Statistics::Descriptive;
	my $stat = Statistics::Descriptive::Full->new();
	$stat->add_data(@a_lengths);
	my %h_length=();
	my %h_bylength=();
	my $step = ( scalar(@a_contigs) < 100 ) ? 100 : 10; # s'il y a peu de contigs les stats tous les 10 n'ont pas de sens, il y a des slot avec 1 seul contig
	for(my $i=$step;$i<=100;$i+=$step)
	{
		my $percentile = $i;
		$h_length{$percentile} = $stat->percentile($percentile);
		my @a_ =();
		$h_bylength{$percentile} = \@a_;
	}
#	print Dumper(%h_length);

	for(my $i=0;$i<scalar(@a_contigs);$i++)
	{
		my $c5 = $$rh_graph{$a_contigs[$i]}{count5} || 0;
		my $c3 = $$rh_graph{$a_contigs[$i]}{count3} || 0;
		my $tot_count = $c5 + $c3;
		my $percentile = $step;
		while ( $$rh_graph{$a_contigs[$i]}{len} > $h_length{$percentile} && $percentile < 100)
		{
			$percentile+=$step;
		}
#		print STDERR "ctg=$a_contigs[$i] tot_count = $tot_count, len=$$rh_graph{$a_contigs[$i]}{len}, pc=$percentile\n";
		push(@{$h_bylength{$percentile}},$c5,$c3);
	}

	my %h_repeat_cutoff=();
	foreach my $pc (sort { $a <=> $b } keys(%h_bylength) )
	{
		my $stat2 = Statistics::Descriptive::Full->new();
		$stat2->add_data(@{$h_bylength{$pc}});
		my $mean = $stat2->mean();
		my $sd   = $stat2->standard_deviation();
		my $var  = $stat2->variance();
		$h_repeat_cutoff{$pc}=$mean+2*$sd;
		#print STDERR "$pc $h_length{$pc} $h_repeat_cutoff{$pc}, mean=$mean, sd=$sd, var=$var\n";
	}

	my $repeat_len=0;
	my $nb_ctg=0;
	for(my $i=0;$i<scalar(@a_contigs);$i++)
	{
		my $c5 = $$rh_graph{$a_contigs[$i]}{count5} || 0;
		my $c3 = $$rh_graph{$a_contigs[$i]}{count3} || 0;
		my $percentile = $step;
		while ( $$rh_graph{$a_contigs[$i]}{len} > $h_length{$percentile} && $percentile < 100)
		{
			$percentile+=$step;
		}
		#print STDERR "$a_contigs[$i] pc=$percentile $c5,$c3\n";
		if ( $c5 > $h_repeat_cutoff{$percentile} && $c3 > $h_repeat_cutoff{$percentile} )
		{
			my $degree3 = scalar(keys(%{$$rh_graph{$a_contigs[$i]}{3}}));
			my $degree5 = scalar(keys(%{$$rh_graph{$a_contigs[$i]}{5}}));
			if ( $degree3 > 1 && $degree5 > 1 )
			{
				my $str = "count3=$c3 count5=$c5 (limit$percentile=mean+2*sd=$h_repeat_cutoff{$percentile}) degree3=$degree3 degree5=$degree5 len=$$rh_graph{$a_contigs[$i]}{len}";
				print STDERR "REPEAT detected $a_contigs[$i] $str\n";
				$repeat_len+=$$rh_graph{$a_contigs[$i]}{len};
				$h_repeats{$a_contigs[$i]}=$str;
				$nb_ctg++;
			}
		}
	}
	$self->_Set('_rh_repeats',\%h_repeats); # 20130502 je rajoute la contrainte du best => 20130503 c'est trop fort
	$self->_Print("InitRepeats done, #contigs=$nb_ctg, length=$repeat_len\n");
}

sub ScaffoldingRepeats
{
my ($self, %h_param) = @_;

	$self->_Print("ScaffoldingRepeats started NOT IMPLEMENTED\n");
	my $rh_repeats = $self->_FastOneGet('_rh_repeats');
	if ( !defined($rh_repeats) )
	{
		$self->InitRepeats(%h_param);
		$rh_repeats = $self->_FastOneGet('_rh_repeats');
	}
	$self->_Print("ScaffoldingRepeats done\n");
}

# $h_param{mode} = update : removal of all used links

sub Scaffolding
{
    my ($self, %h_param) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    $self->_PrintParams("Scaffolding", %h_param);
    my $o_linksscaff = $self->_FastOneGet('_o_linksscaff');
    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $rh_scaff     = $self->_FastOneGet('_rh_scaffolds');
    my $linkvslength = $self->_FastOneGet('_linkvslength');
	my $rh_best      = $self->_FastOneGet('_rh_bests');
	my $rh_repeats   = $self->_FastOneGet('_rh_repeats');
    $self->_Print("# Contigs: " . scalar(keys(%$rh_contigs)) . "\n");
    $self->_Print("# Scaffolds: " . scalar(keys(%$rh_scaff)) . "\n");
    $self->_Print("linkvslength=$linkvslength\n");

    # validation permet de n'agreger que si plusieurs liens entre contigs des 2 scaffolds sont trouves
    my $validation    = (exists($h_param{validation}) && $h_param{validation} > 0) ? $h_param{validation} : &FALSE;
    my $nb_scafflinks = scalar($o_linksscaff->GetLinks());
    my $usescafflinks = (exists($h_param{usescafflinks})) ? $h_param{usescafflinks} : &FALSE;
    my $useanchoredonly  = (exists($h_param{useanchoredonly}))  ? $h_param{useanchoredonly}  : &FALSE;
    my $merge_only    = (exists($h_param{merge_only})) ? $h_param{merge_only} : &FALSE;
    my $extend_only   = (exists($h_param{extend_only})) ? $h_param{extend_only} : &FALSE;
    my $usedtried     = (exists($h_param{usedtried}) && $nb_scafflinks && $h_param{usedtried} > 0) ? $h_param{usedtried} : &FALSE;
    my $userepeats    = (exists($h_param{userepeats}) && $h_param{userepeats} > 0) ? $h_param{userepeats} : &FALSE;
	if ( ! $userepeats && !defined($rh_repeats) )
	{
		$self->InitRepeats(%h_param);
		$rh_repeats = $self->_FastOneGet('_rh_repeats');
	}

    # expansion permet d'etendre (intra) a la volee les scaffolds (eventuellement avec des repeats)
    my $expansion        = (exists($h_param{expansion}) && $h_param{expansion} > 0) ? $h_param{expansion} : &FALSE;
    my $reciprocalbest   = (exists($h_param{reciprocalbest}) && $h_param{reciprocalbest} > 0) ? $h_param{reciprocalbest} : &FALSE;
    my $best             = (exists($h_param{best}) && $h_param{best} > 0) ? $h_param{best} : &FALSE;
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance}) && $h_param{gaplen_tolerance} =~ /^(\d+)$/) ? $h_param{gaplen_tolerance} : 0;    # c'est un pourcentage
    if (!$gaplen_tolerance)
    {
        $gaplen_tolerance = $self->_FastOneGet('_gaplen_tolerance');
        print STDERR "gaplen_tolerance not set, using default: $gaplen_tolerance\%\n";
    }
    my $anchoring_consistency = $self->_FastOneGet('_anchoring_consistency');

    print STDERR "usescafflinks=$usescafflinks   #links >" . scalar($o_linksscaff->GetLinks()) . "<\n";

    my $useweaks = (exists($h_param{maxweak}) && $h_param{maxweak} == 1) ? &TRUE : &FALSE;

	print STDERR "validation=$validation, usescafflinks=$usescafflinks, useanchoredonly=$useanchoredonly, merge_only=$merge_only, extend_only=$extend_only, usedtried=$usedtried, userepeats=$userepeats, expansion=$expansion, reciprocalbest=$reciprocalbest, best=$best, anchoring_consistency=$anchoring_consistency, useweaks=$useweaks\n";

    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my @a_links_filt = ();
    my $nb_ignored   = $o_linkds->Filter(\@a_links_filt, %h_param);
    $self->InitGraph(&TRUE, undef, $useweaks);
    my $o_graph = $self->_FastOneGet('_o_graph');    # apres le InitGraph

    if ($reciprocalbest)
    {
        $o_linkds->ReciprocalBestFilter(\@a_links_filt);
    }
    if ($best)
    {
        $o_linkds->BestFilter(\@a_links_filt,$rh_best);
    }

    my %h_contig2scaff = ();
    $self->InitContig2Scaffold(\%h_contig2scaff);
    $self->ScaffoldsAnchoring();

    my @a_links = ();
    my $nb_used = 0;

    if (exists($h_param{sortby}) && $h_param{sortby} eq 'count')
    {
        @a_links = sort {
            if ($usescafflinks)
            {
                my %h_linked = ();
                foreach my $l ($a, $b)
                {
                    my ($src, $tgt) = $l =~ /src=(\S+)\s*tgt=(\S+)/;
                    if (exists($h_contig2scaff{$src}) && exists($h_contig2scaff{$tgt}))
                    {
                        my %h_val = ();
                        if ($o_linksscaff->GetKV2N($h_contig2scaff{$src}, $h_contig2scaff{$tgt}, \%h_val))
                        {
							if ( $h_val{ambiguous_orient} == 0 )
							{
								$h_linked{$l} = $h_val{count};
								next;
							}
                        }
                    }
                    $h_linked{$l} = &FALSE;
                }
                return ($h_linked{$b} <=> $h_linked{$a}) if ($h_linked{$a} || $h_linked{$b} && $h_linked{$a} != $h_linked{$b});
            }
            my %h_a = ();
            $o_linkds->GetKV($a, \%h_a);
            my %h_b = ();
            $o_linkds->GetKV($b, \%h_b);
            return ($h_b{count}) <=> ($h_a{count});
        } @a_links_filt;
    }
    elsif (exists($h_param{sortby}) && $h_param{sortby} eq 'span')
    {
        @a_links = sort {
            if ($usescafflinks)
            {
                my %h_linked = ();
                foreach my $l ($a, $b)
                {
                    my ($src, $tgt) = $l =~ /src=(\S+)\s*tgt=(\S+)/;
                    if (exists($h_contig2scaff{$src}) && exists($h_contig2scaff{$tgt}))
                    {
                        my %h_val = ();
                        if ($o_linksscaff->GetKV2N($h_contig2scaff{$src}, $h_contig2scaff{$tgt}, \%h_val))
                        {
							if ( $h_val{ambiguous_orient} == 0 )
							{
								$h_linked{$l} = $h_val{count};
								next;
							}
                        }
                    }
                    $h_linked{$l} = &FALSE;
                }
                return ($h_linked{$b} <=> $h_linked{$a}) if ($h_linked{$a} || $h_linked{$b} && $h_linked{$a} != $h_linked{$b});
            }
            my %h_a = ();
            $o_linkds->GetKV($a, \%h_a);
            my %h_b = ();
            $o_linkds->GetKV($b, \%h_b);
            my $spana = ($h_a{srcspan} + $h_a{tgtspan});
            my $spanb = ($h_b{srcspan} + $h_b{tgtspan});
            return ($spanb <=> $spana);
        } @a_links_filt;
    }

    # traiter d'abord les gap les plus petits, et pour une gamme de taille ceux qui sont le plus supportes
    else
    {
        @a_links = sort {
            if ($usescafflinks)
            {
                my %h_linked = ();
                foreach my $l ($a, $b)
                {
                    my ($src, $tgt) = $l =~ /src=(\S+)\s*tgt=(\S+)/;
                    if (exists($h_contig2scaff{$src}) && exists($h_contig2scaff{$tgt}))
                    {
                        my %h_val = ();
                        if ($o_linksscaff->GetKV2N($h_contig2scaff{$src}, $h_contig2scaff{$tgt}, \%h_val))
                        {
							if ( $h_val{ambiguous_orient} == 0 )
							{
								$h_linked{$l} = $h_val{count};
								next;
							}
                        }
                    }
                    $h_linked{$l} = &FALSE;
                }
                return ($h_linked{$b} <=> $h_linked{$a}) if ($h_linked{$a} || $h_linked{$b} && $h_linked{$a} != $h_linked{$b});
            }
            my %h_a = ();
            $o_linkds->GetKV($a, \%h_a);
            my %h_b = ();
            $o_linkds->GetKV($b, \%h_b);
            $h_a{gaplen_max} =~ s/(\d)(\d+)/$1.'0' x length($2)/e;
            $h_a{gaplen_max} = 200 if ($h_a{gaplen_max} < 200);
            $h_b{gaplen_max} =~ s/(\d)(\d+)/$1.'0' x length($2)/e;
            $h_b{gaplen_max} = 200 if ($h_b{gaplen_max} < 200);
            return ($h_a{gaplen_max} - $h_a{count}) <=> ($h_b{gaplen_max} - $h_b{count});
        } @a_links_filt;
    }

    my $gaplen_fact = 1;
    $gaplen_fact += ($gaplen_tolerance / 100);
    my $nb_links = scalar(@a_links);

# je dois faire le filtre avant la boucle principale car les id de scaffolds peuvent changer pendant la boucle principale

	# 20130430: je vais chercher les liens correspondants aux liens entre scaffolds dans les liens bruts pas dans
	# ceux qui sont filtres
	if ( $usescafflinks )
	{
		$self->ScaffoldingLinkedScaffolds(%h_param);
	}
    if ($usescafflinks > 1)    # c'est le mode 2 qui force
    {
        print STDERR "usescafflinks = $usescafflinks\n";
        my @a_links_ok = ();
        for (my $i = 0; $i < $nb_links; $i++)
        {
            my %h_key = split(/[=\s]+/, $a_links[$i]);
            my ($src) = $h_key{src};
            my ($tgt) = $h_key{tgt};
            if (exists($h_contig2scaff{$src}) && exists($h_contig2scaff{$tgt}))
            {
                my %h_val = ();
                if ($o_linksscaff->GetKV2N($h_contig2scaff{$src}, $h_contig2scaff{$tgt}, \%h_val))
                {
                    print STDERR "$src,$tgt $h_contig2scaff{$src} $h_contig2scaff{$tgt}\n";
                    print STDERR "OK $a_links[$i]\n";
                    push(@a_links_ok, $a_links[$i]);
                }
            }
        }
        @a_links  = @a_links_ok;
        $nb_links = scalar(@a_links);
        print STDERR "usescafflinks=$usescafflinks, $nb_links=$nb_links\n";
    }

    my %h_val = ();
    my %h_key = ();
    for (my $i = 0; $i < $nb_links; $i++)
    {
        my $link      = $a_links[$i];
        my $link_used = &TRUE;
        %h_key = split(/[=\s]+/, $link);
        %h_val = ();
        next if (!$o_linkds->GetKV($link, \%h_val));

		if ( ! $userepeats && defined($$rh_repeats{$h_key{src}}) || defined($$rh_repeats{$h_key{tgt}}) )
		{
#        	print STDERR "REPEAT ignored: $i/$nb_links\t$link\t" . join(' ', %h_val) . "\n";
			next;
		}

        print STDERR "$i/$nb_links\t$link\t" . join(' ', %h_val) . "\n";

        my $o_ctgsrc = $self->GetRegisteredContig($h_key{src}, 'p', $h_val{srclen});
        my $o_ctgtgt = $self->GetRegisteredContig($h_key{tgt}, 'p', $h_val{tgtlen});

        $link_used = $self->_ScaffoldingContigPair(
                                                   $rh_scaff,   $o_ctgsrc,    $o_ctgtgt,
                                                   \%h_key,     \%h_val,      $o_linkds,
                                                   $validation, $gaplen_fact, $anchoring_consistency,
                                                   $linkvslength, $merge_only, $extend_only, $useanchoredonly
                                                   );
        if ($link_used)
        {
            $nb_used++;
            $h_val{used} = 1;
            $o_linkds->SetKV($link, \%h_val);
        }
    }
    $self->DumpLinks(%h_param) if (exists($h_param{dumpfile}));
    $self->SynchronizeContigRegistry();
    $self->_Print("Scaffolding done, Links: #ignored=$nb_ignored, #considered=" . scalar(@a_links) . ", #used=$nb_used\n");
}

# factoriser avec Scaffolding, il faudrait juste faire la selection et envoyer tout a scaffolding qui ne devrait pas trier

sub ScaffoldingLinkedScaffolds
{
    my ($self, %h_param) = @_;

    $self->_PrintParams("ScaffoldingLinkedScaffolds", %h_param);
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $o_linksscaff = $self->_FastOneGet('_o_linksscaff');

    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $rh_scaff     = $self->_FastOneGet('_rh_scaffolds');
    my $linkvslength = $self->_FastOneGet('_linkvslength');
    my $merge_only = (exists($h_param{merge_only})) ? $h_param{merge_only} : &FALSE;
    my $extend_only = (exists($h_param{extend_only})) ? $h_param{extend_only} : &FALSE;
    my $useanchoredonly = (exists($h_param{useanchoredonly})) ? $h_param{useanchoredonly} : &FALSE;
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance}) && $h_param{gaplen_tolerance} =~ /^(\d+)$/) ? $h_param{gaplen_tolerance} : 0;    # c'est un pourcentage
    if (!$gaplen_tolerance)
    {
        $gaplen_tolerance = $self->_FastOneGet('_gaplen_tolerance');
        print STDERR "gaplen_tolerance not set, using default: $gaplen_tolerance\%\n";
    }
    my $gaplen_fact = 1;
    $gaplen_fact += ($gaplen_tolerance / 100);
    my $anchoring_consistency = $self->_FastOneGet('_anchoring_consistency');
    $self->_Print("# Contigs: " . scalar(keys(%$rh_contigs)) . "\n");
    $self->_Print("# Scaffolds: " . scalar(keys(%$rh_scaff)) . "\n");

	my @a_linkedscaffolds=$o_linksscaff->GetLinks();
	my $validation = 2; # par defaut du LinkScaffold il y a plusieurs liens
	my %h_links_ok = ();
	my %h_val=();
	foreach my $linkscaff (@a_linkedscaffolds)
	{
		if ( $o_linksscaff->GetKV($linkscaff,\%h_val) )
		{
			my @a_lst1 = split(';',$h_val{srclst});
			my @a_lst2 = split(';',$h_val{tgtlst});
			foreach my $s1 (@a_lst1)
			{
				foreach my $s2 (@a_lst2)
				{
					my $key = sprintf("src=%s tgt=%s",sort($s1,$s2));
					my $link = $o_linkds->Get($key);
					if ( defined($link) )
					{
						$h_links_ok{$key}=$h_val{count};
					}
				}
			}
		}
		else
		{
			die "ScaffoldingLinkedScaffolds: Error impossible\n";
		}
	}

	my @a_links=sort { $h_links_ok{$b} <=> $h_links_ok{$a} } keys(%h_links_ok);

    my %h_contig2scaff = ();
    $self->InitContig2Scaffold(\%h_contig2scaff);
    $self->ScaffoldsAnchoring();
    my %h_key = ();
	my $nb_ignored=0;
	my $nb_used=0;
	my $nb_links=scalar(@a_links);
    for (my $i = 0; $i < $nb_links; $i++)
    {
        my $link      = $a_links[$i];
		print STDERR "LINK $link: ".$o_linkds->Get($link)."\n";
        my $link_used = &TRUE;
        %h_key = split(/[=\s]+/, $link);
        %h_val = ();
        next if (!$o_linkds->GetKV($link, \%h_val));

        print STDERR "ScaffoldingLinkedScaffolds $i/$nb_links\t$link\t" . join(' ', %h_val) . "\n";

        my $o_ctgsrc = $self->GetRegisteredContig($h_key{src}, 'p', $h_val{srclen});
        my $o_ctgtgt = $self->GetRegisteredContig($h_key{tgt}, 'p', $h_val{tgtlen});

        $link_used = $self->_ScaffoldingContigPair(
                                                   $rh_scaff,   $o_ctgsrc,    $o_ctgtgt,
                                                   \%h_key,     \%h_val,      $o_linkds,
                                                   $validation, $gaplen_fact, $anchoring_consistency,
                                                   $linkvslength, $merge_only, $extend_only, $useanchoredonly
                                                   );
        if ($link_used)
        {
            $nb_used++;
            $h_val{used} = 1;
            $o_linkds->SetKV($link, \%h_val);
        }
    }
    $self->_Print("Post # Contigs: " . scalar(keys(%$rh_contigs)) . "\n");
    $self->_Print("Post # Scaffolds: " . scalar(keys(%$rh_scaff)) . "\n");
    $self->_Print("ScaffoldingLinkedScaffolds done, Links: #ignored=$nb_ignored, #considered=" . scalar(@a_links) . ", #used=$nb_used\n");
}

sub ScaffoldingContigCentered
{
    my ($self, %h_param) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    $self->_PrintParams("ScaffoldingContigCentered", %h_param);
    my $o_linksscaff = $self->_FastOneGet('_o_linksscaff');
    my $rh_contigs   = $self->_FastOneGet('_rh_contigs');
    my $rh_scaff     = $self->_FastOneGet('_rh_scaffolds');
    my $linkvslength = $self->_FastOneGet('_linkvslength');
    my $o_linkds = $self->_FastOneGet('_o_linkds');
    $self->_Print("# Contigs: " . scalar(keys(%$rh_contigs)) . "\n");
    $self->_Print("# Scaffolds: " . scalar(keys(%$rh_scaff)) . "\n");
    $self->_Print("linkvslength=$linkvslength\n");

    # validation permet de n'agreger que si plusieurs liens entre contigs des 2 scaffolds sont trouves
    my $validation    = (exists($h_param{validation}) && $h_param{validation} > 0) ? $h_param{validation} : &FALSE;
    my $nb_scafflinks = scalar($o_linksscaff->GetLinks());
    my $usescafflinks = (exists($h_param{usescafflinks})) ? $h_param{usescafflinks} : &FALSE;
    my $merge_only    = (exists($h_param{merge_only})) ? $h_param{merge_only} : &FALSE;
    my $extend_only    = (exists($h_param{extend_only})) ? $h_param{extend_only} : &FALSE;
    my $useanchoredonly    = (exists($h_param{useanchoredonly})) ? $h_param{useanchoredonly} : &FALSE;
    my $usedtried =
      (exists($h_param{usedtried}) && $nb_scafflinks && $h_param{usedtried} > 0) ? $h_param{usedtried} : &FALSE;

    # expansion permet d'etendre (intra) a la volee les scaffolds (eventuellement avec des repeats)
    my $expansion = (exists($h_param{expansion}) && $h_param{expansion} > 0) ? $h_param{expansion} : &FALSE;
    my $reciprocalbest = (exists($h_param{reciprocalbest}) && $h_param{reciprocalbest} > 0) ? $h_param{reciprocalbest} : &FALSE;
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance}) && $h_param{gaplen_tolerance} =~ /^(\d+)$/) ? $h_param{gaplen_tolerance} : 0;    # c'est un pourcentage
    if (!$gaplen_tolerance)
    {
        $gaplen_tolerance = $self->_FastOneGet('_gaplen_tolerance');
        print STDERR "gaplen_tolerance not set, using default: $gaplen_tolerance\%\n";
    }
    my $anchoring_consistency = $self->_FastOneGet('_anchoring_consistency');

    my $minseedlen = (exists($h_param{minseedlen})) ? $h_param{minseedlen} : 0;	# ne considere que les longueurs > minseedlen comme graines
	delete($h_param{minseedlen}) if ( exists($h_param{minseedlen}) ); # pour eviter les faux msg d'erreurs dans Filter

    my $useweaks = (exists($h_param{maxweak}) && $h_param{maxweak} == 1) ? &TRUE : &FALSE;

    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux liens (et plus au graphe) je suis oblige de travailler sur un temporaire
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();
    my ($rh_graph,$o_graph) = $self->InitGraph(&FALSE, $ra_links_filt, $useweaks);    # ici je ne veux dans le graphe que ce qui passe le seuil

	my @a_links_filt_copy = @{$ra_links_filt};
	my $rh_rbh;
    ($nb_ignored,$rh_rbh) = $o_linkds_filt->ReciprocalBestFilter(\@a_links_filt_copy); # ca sert a renseigner rh_rbh

    my %h_contig2scaff = ();
    $self->InitContig2Scaffold(\%h_contig2scaff);
    $self->ScaffoldsAnchoring();
    my $gaplen_fact = 1;
    $gaplen_fact += ($gaplen_tolerance / 100);

    # 1) faire la liste des contigs qui sont dans les liens RBH
    my %h_val     = ();
    my %h_key     = ();
    my %h_contigs = ();
    my $nb_links  = scalar(@$ra_links_filt);
	my $cpt = 0;
    for (my $i = 0; $i < $nb_links; $i++)
    {
        my $link      = $$ra_links_filt[$i];
        my $link_used = &TRUE;
        %h_key = split(/[=\s]+/, $link);
        %h_val = ();
        next if (!$o_linkds_filt->GetKV($link, \%h_val));
        $h_contigs{$h_key{src}} = $h_val{srclen};
        $h_contigs{$h_key{tgt}} = $h_val{tgtlen};
    }

    %h_key = ();
    my $nb_used = 0;

    # ctg/ext-> nextext/next/othernextext -> nextnextext/nextnext
    # 2) pour chaque contig de longueur decroissante (on essaye de placer d'abord les plus longs)

	my %h_done = ();
	%h_val = ();
    foreach my $ctg (sort {$h_contigs{$b} <=> $h_contigs{$a}} keys(%h_contigs))
    {
		last if ( $h_contigs{$ctg} < $minseedlen );
        foreach my $ext ('5', '3')
        {
            my $rh_next_contigs = $o_graph->GetNextOrientedContigs($ctg, $ext);
			print STDERR "$ctg/$ext".Dumper($rh_next_contigs);
			my($rbh,$rbhext)=();

			# premier passage je cherche le rbh
            foreach my $next (keys(%$rh_next_contigs))
            {
                my ($nextgaplen_min, $nextgaplen_max, $nextext) = split(',', $$rh_next_contigs{$next});
				if ( $$rh_rbh{$ctg}{$ext}{best} eq $next && $$rh_rbh{$next}{$nextext}{best} eq $ctg )
				{
					($rbh,$rbhext) = ($next,$nextext);
					print STDERR "RBH $ctg/$ext => ($rbh,$rbhext)\n";
					last;
				}
			}

			next if ( ! defined($rbh) );

			my $rbh_fused = &FALSE;

			my %h_used_to_validate = ();
            foreach my $next (
                sort
                {
                    my @A = split(',', $$rh_next_contigs{$a});
                    my @B = split(',', $$rh_next_contigs{$b});
                    return $A[1] <=> $B[1];
                } keys(%$rh_next_contigs)
              )
            {
                my ($nextgaplen_min, $nextgaplen_max, $nextext) = split(',', $$rh_next_contigs{$next});
				print STDERR "$ctg/$next ctg -> next: $nextgaplen_min, $nextgaplen_max, $nextext\n";
				if ( $next eq $rbh  && $nextext eq $rbhext ) 
				{
					if ( $self->_ScaffoldingContigPairFull($o_linkds_filt,$ctg,$ext,$next,$nextext,$rh_scaff,\%h_contigs,$validation, $gaplen_fact,$anchoring_consistency,$linkvslength,$merge_only,$extend_only,$useanchoredonly))
					{
						$rbh_fused=&TRUE;
					}
				}

                my $othernextext = ($nextext eq '5') ? 3 : 5;
                my $rh_nextnext_contigs = $o_graph->GetNextOrientedContigs($next, $othernextext);

#foreach my $nextnext (sort { my @A=split(',',$$rh_nextnext_contigs{$a}); my @B=split(',',$$rh_nextnext_contigs{$b});return (($A[1]+$A[0])/2 <=> ($B[1]+$B[0])/2);} keys(%$rh_nextnext_contigs))
                foreach my $nextnext (
                    sort
                    {
                        my @A = split(',', $$rh_nextnext_contigs{$a});
                        my @B = split(',', $$rh_nextnext_contigs{$b});
                        return $A[1] <=> $B[1];
                    } keys(%$rh_nextnext_contigs)
                  )
                {
                    my ($nextnextgaplen_min, $nextnextgaplen_max, $nextnextext) = split(',', $$rh_nextnext_contigs{$nextnext});

					if ($o_linkds_filt->GetKV2Oriented($ctg, $ext, $nextnext, $nextnextext, \%h_val))
					{
						#next if ( $nextnext ne $rbh || $nextnextext ne $rbhext); 

						if ( $self->_ScaffoldingContigPairFull($o_linkds_filt,$ctg,$ext,$next,$nextext,$rh_scaff,\%h_contigs,$validation, $gaplen_fact,$anchoring_consistency,$linkvslength,$merge_only,$extend_only,$useanchoredonly))
						#if ( $self->_ScaffoldingContigPairFull($o_linkds,$next,$nextext,$nextnext,$nextnextext,$rh_scaff,\%h_contigs,$validation, $gaplen_fact,$anchoring_consistency,$linkvslength) )
						{
							$h_used_to_validate{$nextnext}=$nextnextext;
							last;
						}
						else
						{
							$self->_Print("$ctg,$ext,$next,$nextext not merged\n");
						}
					}
                }
            }

			# je fais un deuxieme passage pour rajouter les next, next qui ont servis a valider des next
            foreach my $next (
                sort
                {
                    my @A = split(',', $$rh_next_contigs{$a});
                    my @B = split(',', $$rh_next_contigs{$b});
                    return $A[1] <=> $B[1];
                } keys(%$rh_next_contigs)
              )
            {
				next if ( ! defined($h_used_to_validate{$next}) );
                my ($nextgaplen_min, $nextgaplen_max, $nextext) = split(',', $$rh_next_contigs{$next});
				if ( $self->_ScaffoldingContigPairFull($o_linkds_filt,$ctg,$ext,$next,$nextext,$rh_scaff,\%h_contigs,$validation, $gaplen_fact,$anchoring_consistency,$linkvslength,$merge_only,$extend_only,$useanchoredonly))
				{
					#$nb_used++;
				}

			}
			# pour finir je verifie si c'est ok avec le rbh
			if ( defined($rbh) && ! $rbh_fused )
			{
				# pour le rbh je ne veux pas de seuil sur la taille des gaps
				if ( ! $self->_ScaffoldingContigPairFull($o_linkds_filt,$ctg,$ext,$rbh,$rbhext,$rh_scaff,\%h_contigs,$validation,2*$gaplen_fact,$anchoring_consistency,$linkvslength,$merge_only,$extend_only,$useanchoredonly) )
				{
					print STDERR "##### RBH Failure ######## ($ctg, $ext, $rbh, $rbhext)\n";
				}
			}
        }
    }
    $self->DumpLinks(%h_param) if (exists($h_param{dumpfile}));
    $self->SynchronizeContigRegistry();
    $self->_Print("ScaffoldingContigCentered done #used=$nb_used\n");
}

sub _ScaffoldingContigPairFull
{
my ($self,$o_linkds,$ctg,$ext,$next,$nextext,$rh_scaff,$rh_contigs,$validation, $gaplen_fact,$anchoring_consistency,$linkvslength,$merge_only,$extend_only,$useanchoredonly)=@_;

	my %h_key=();
	($h_key{src}, $h_key{tgt}) = ($ctg lt $next) ? ($ctg, $next) : ($next, $ctg);
	my ($lsrc, $lsrcext, $lnext, $lnextext) =
	  ($ctg lt $next) ? ($ctg, $ext, $next, $nextext) : ($next, $nextext, $ctg, $ext);
	my %h_val = ();
	return &FALSE if (!$o_linkds->GetKV2Oriented($lsrc, $lsrcext, $lnext, $lnextext, \%h_val));
	#print STDERR "###### 1 ####### $lsrc,  'p', $h_contigs{$lsrc} // $lnext, 'p', $h_contigs{$lnext}\n";
	my $o_ctgsrc = $self->GetRegisteredContig($lsrc,  'p', $$rh_contigs{$lsrc});
	my $o_ctgtgt = $self->GetRegisteredContig($lnext, 'p', $$rh_contigs{$lnext});
	my $link_used =
	  $self->_ScaffoldingContigPair(
									$rh_scaff,   $o_ctgsrc,    $o_ctgtgt,
									\%h_key,     \%h_val,      $o_linkds,
									$validation, $gaplen_fact, $anchoring_consistency,
									$linkvslength,$merge_only,$extend_only,$useanchoredonly
									);
	if ( $link_used )
	{
		$h_val{used} = 1;
		$o_linkds->SetKV2($lsrc, $lnext, \%h_val);
	}
	return $link_used;
}

sub _ScaffoldingContigPair
{
    my (
        $self,        $rh_scaff,              $o_ctgsrc, $o_ctgtgt,
        $rh_key,      $rh_val,                $o_linkds, $validation,
        $gaplen_fact, $anchoring_consistency, $linkvslength, $merge_only,$extend_only,$useanchoredonly
      )
      = @_;

    my $link_used = &TRUE;

    my $nb_ctgsrcscaff = $o_ctgsrc->GetNumberOfScaffoldsFromRegistry();
    my $nb_ctgtgtscaff = $o_ctgtgt->GetNumberOfScaffoldsFromRegistry();

    if ($nb_ctgsrcscaff > 1 || $nb_ctgtgtscaff > 1)
    {
        print STDERR "$$rh_key{src}=>#scaff=$nb_ctgsrcscaff "
          . $o_ctgsrc->GetScaffoldsIds()
          . ", $$rh_key{tgt}=>#scaff=$nb_ctgtgtscaff "
          . $o_ctgtgt->GetScaffoldsIds() . "\n";
        return &FALSE;
    }

	if ( $merge_only && (! $nb_ctgsrcscaff || !$nb_ctgtgtscaff) )
	{
		print STDERR "merge_only, $$rh_key{src} $$rh_key{tgt} rejected\n";
		return &FALSE;
	}

	if ( $extend_only && ($nb_ctgsrcscaff+$nb_ctgtgtscaff != 1) )
	{
		print STDERR "extend_only, $$rh_key{src} $$rh_key{tgt} rejected\n";
		return &FALSE;
	}

    if (!$nb_ctgsrcscaff)
    {
        my $o_scaff = New Scaffold('name' => 'TMP0src', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
        my $idscaff = $self->AddScaffold($o_scaff);
        $o_scaff->SetName($idscaff);
        $o_ctgsrc->SetScaffold($o_scaff, $idscaff);
        $o_scaff->Push($o_ctgsrc);
        $self->ScaffoldsAnchoring($idscaff)
          ; # le AddScaffold le fait mais je le refais pour declencher l'ancrage car avant il n'y a pas de contigs au moment du Add
        $o_scaff->InitAnchor(&TRUE);
        print STDERR Dumper($o_scaff) if (defined($o_scaff->GetAnchor()));
    }
    if (!$nb_ctgtgtscaff)
    {
        my $o_scaff = New Scaffold('name' => 'TMP0tgt', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
        my $idscaff = $self->AddScaffold($o_scaff);
        $o_scaff->SetName($idscaff);
        $o_ctgtgt->SetScaffold($o_scaff, $idscaff);
        $o_scaff->Push($o_ctgtgt);
        $self->ScaffoldsAnchoring($idscaff);
        $o_scaff->InitAnchor(&TRUE);
        print STDERR Dumper($o_scaff) if (defined($o_scaff->GetAnchor()));
    }

    # les repeats sont elimines plus haut
    my $scaffsrcname = $o_ctgsrc->GetScaffoldsIds();
    my $scafftgtname = $o_ctgtgt->GetScaffoldsIds();

    my $o_src = $$rh_scaff{$scaffsrcname};
    my $o_tgt = $$rh_scaff{$scafftgtname};

    if ($linkvslength)
    {
        my $minscafflen = &Min($o_src->GetLength(), $o_tgt->GetLength());
        if ($minscafflen < $linkvslength && $minscafflen < $$rh_val{gaplen_min})
        {
            print STDERR "( $minscafflen < $$rh_val{gaplen_min} ) REJECTED\n";
            next;
        }
    }

    if (!defined($o_src) || !defined($o_tgt))
    {
        print STDERR "$scaffsrcname,$$rh_scaff{$scaffsrcname} $scafftgtname,$$rh_scaff{$scafftgtname}\n";
        Carp::confess("At that stage, a contig must be assigned to a scaffold\n");
    }

    my $scaffsrcanc = $o_src->GetAnchor() || '';
    my $scafftgtanc = $o_tgt->GetAnchor() || '';

	if ( $useanchoredonly && $scaffsrcanc eq '' && $scafftgtanc eq '' )
	{
		return &FALSE;
	}

    if (0)
    {
        if (defined($o_ctgsrc->GetAnchors()) && ($scaffsrcanc eq '' || $o_ctgsrc->GetAnchors() !~ /$scaffsrcanc/))
        {
            if ($o_ctgsrc->GetLength() > 1000)
            {
                print STDERR "INCONSISTENT CTG / SCAFF anchoring\n";
                print STDERR "GetAnchors= " . $o_ctgsrc->GetAnchors() . "\n";
                print STDERR "GetAnchorsFromRegistry= " . $o_ctgsrc->GetAnchorsFromRegistry() . "\n";
                print STDERR Dumper($o_ctgsrc);
                print STDERR Dumper($o_src);
                $o_src->InitAnchor(&TRUE, &TRUE);

                #Carp::confess("ERROR: the scaffold must contain the contig anchor\n");
            }
        }
        if (defined($o_ctgtgt->GetAnchors()) && ($scafftgtanc eq '' || $o_ctgtgt->GetAnchors() !~ /$scafftgtanc/))
        {
            if ($o_ctgtgt->GetLength() > 1000)
            {
                print STDERR "INCONSISTENT CTG / SCAFF anchoring\n";
                print STDERR "GetAnchors= " . $o_ctgtgt->GetAnchors() . "\n";
                print STDERR "GetAnchorsFromRegistry= " . $o_ctgtgt->GetAnchorsFromRegistry() . "\n";
                print STDERR Dumper($o_ctgtgt);
                print STDERR Dumper($o_tgt);
                $o_src->InitAnchor(&TRUE, &TRUE);

                #Carp::confess("ERROR: the scaffold must contain the contig anchor\n");
            }
        }
    }

    if ($scaffsrcname ne $scafftgtname && $o_src->AreAnchoringCompatibles($o_tgt))
    {
    	print STDERR "src=" . $o_ctgsrc->GetName() . " (scaff=$scaffsrcname, scaffanchor=$scaffsrcanc), tgt=" . $o_ctgtgt->GetName() . " (scaff=$scafftgtname, scaffanchor=$scafftgtanc)\n";
        #$self->_Print("5. merging scaffolds\n");
        if ($$rh_key{src} =~ /$DEBUG_LINK_PATTERN/ || $$rh_key{tgt} =~ /$DEBUG_LINK_PATTERN/)
        {
            print STDERR "_MergeScaffolds src=$$rh_key{src} tgt=$$rh_key{tgt} " . join(' ', %$rh_val) . "\n";
        }

        #print STDERR Dumper($o_src);
        #print STDERR Dumper($o_tgt);
		$self->_TryMerging($rh_scaff,$o_src,$o_tgt, $o_ctgsrc,$o_ctgtgt, $rh_key, $rh_val, $o_linkds, $validation, $gaplen_fact, $anchoring_consistency,$scaffsrcanc,$scafftgtanc,$scaffsrcname,$scafftgtname,\$link_used);
    }
    return $link_used;
}

sub _TryMerging
{
my($self,$rh_scaff,$o_src,$o_tgt, $o_ctgsrc,$o_ctgtgt, $rh_key, $rh_val, $o_linkds, $validation, $gaplen_fact, $anchoring_consistency,$scaffsrcanc,$scafftgtanc,$scaffsrcname,$scafftgtname,$r_link_used)=@_;

	if (!$o_src->Merge($o_tgt, $rh_key, $rh_val, $o_linkds, $validation, $gaplen_fact, $anchoring_consistency,&TRUE))
	{
		if ($scaffsrcanc eq $scafftgtanc && $scaffsrcanc ne '')
		{
			print STDERR "\tMerging failed\n";
			print STDERR "FAILED src=" . $o_ctgsrc->GetName() . " scaff=$scaffsrcname " . $o_src->Get() . "\n";
			print STDERR "FAILED tgt=" . $o_ctgtgt->GetName() . " scaff=$scafftgtname " . $o_tgt->Get() . "\n";
		}

		#print STDERR "\tsrc=$h_key{src} tgt=$h_key{tgt} ".join(' ',%h_val)."\n";
		$$r_link_used = &FALSE;    # if ( ! $usedtried );
	}
	else
	{
		my ($kept, $obsolete) = ($o_src->GetNumberOfContigs() > 0) ? ($o_src, $o_tgt) : ($o_tgt, $o_src);
		my $id = $kept->GetName();
		if ($id ne $scafftgtname && $id ne $scaffsrcname)
		{
			Carp::confess("ERROR: current scaffid ($id) not in $scafftgtname,$scaffsrcname\n");
		}
		$$rh_scaff{$id} = $kept;
		$$rh_scaff{$id}->Check();

		#$self->InitContig2Scaffold(\%h_contig2scaff);
		$self->ScaffoldsAnchoring($id);
		my ($anchor) = $$rh_scaff{$id}->GetAnchor() || '';
		$self->_Print("Scaffolding Merge: scaffid=$id anchor=" . $anchor . " deleting " . $obsolete->GetName() . "\n");
		if ($anchor eq '' && ($scaffsrcanc ne '' || $scafftgtanc ne ''))
		{
			print STDERR Dumper($$rh_scaff{$id});
			Carp::confess("ERROR: a merged anchored scaffold must be anchored post merging\n");
		}

		#$$rh_scaff{$id}->RecomputeGapLength($o_linkds); ce n'est pas une bonne idee
		#if ( $anchor eq '' )
		#{
		#	print STDERR Dumper($$rh_scaff{$id});
		#}
		delete($$rh_scaff{$obsolete->GetName()});
	}
}



sub LoadOpticalMap
{
    my ($self, %h_param) = @_;

    $self->_PrintParams("LoadOpticalMap", %h_param);
    my $o_omap = $self->_FastOneGet('_o_omap');
    $o_omap->Init(%h_param);
    $o_omap->Load(%h_param);
}

sub ContigAnchoring
{
    my ($self, %h_param) = @_;

    $self->_PrintParams("ContigAnchoring", %h_param);
    my $rh_o_seqds   = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist   = $self->_FastOneGet('_ra_seqlist');
    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    my $o_omap       = $self->_FastOneGet('_o_omap');
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $minmingaplen = $self->_FastOneGet('_mingaplen');

    $self->SynchronizeContigRegistry();

    if ($o_omap->GetNumberOfMaps() == 0)
    {
        $self->_Print("Warning: no optical map found, unable to run ContigAnchoring\n");
        return;
    }

    foreach my $idscaff (keys(%$rh_scaffolds))
    {
        my @a_results = ();
        my ($r_scaffseq, $rh_ctg_pos) =
          $$rh_scaffolds{$idscaff}->GetSequence($o_linkds, $rh_o_seqds, $ra_seqlist, $minmingaplen);
        my @a_res = $o_omap->MapSequence($idscaff, $r_scaffseq, 'full', &MAXINT, &TRUE)
          ;    # format long, pas de limite dans le nombre de hits, chevauchement < min_scaffold_len/2
        push(@a_results, @a_res);
        next if (scalar(@a_results) == 0);

        if (scalar(@a_results) > 1)
        {
            print STDERR "Warning: chimeric signal in scaffold $idscaff len="
              . $$rh_scaffolds{$idscaff}->GetLength() . "\n";
        }

        for (my $i = 0; $i < scalar(@a_results); $i++)
        {
            print STDERR "$a_results[$i]\n";
            my (
                $idmap,        $nb_sites,   $om_start, $om_end,        $omlen,
                $om_pos_start, $om_pos_end, $strand,   $idseq,         $seq_sites,
                $seq_start,    $seq_stop,   $seqlen,   $seq_pos_start, $seq_pos_end,
                $ali_len,      $score,      $densite,  $span
              )
              = split(/\t/, $a_results[$i]);

            next if (!defined($idmap));

            $self->_Print(
                         "ScaffoldsAnchoring $idscaff: $idmap,$seq_pos_start,$seq_pos_end,$om_start,$om_end,$strand\n");
            my $ra_ctg = $$rh_scaffolds{$idscaff}->GetArrayContigs();
            for (my $i = 0; $i < scalar(@$ra_ctg); $i++)
            {
                my @a_pos = split(/\t/, $$rh_ctg_pos{$$ra_ctg[$i]->GetName()});
                if (scalar(@$ra_ctg) == 1
                    || ($a_pos[0] >= $seq_pos_start && $a_pos[1] <= $seq_pos_end)
                  )    # je ne veux ancrer que les contigs qui sont dans le hit avec l'om
                       # avec un cas particulier lorsqu'il n'y a qu'un contig qui ne serait jamais inclus
                {
                    $self->_Print(  "ScaffoldsAnchoring $idscaff: contig="
                                  . $$ra_ctg[$i]->GetName()
                                  . " anchor=$idmap $om_pos_start:$om_pos_end:$strand\n");
                    $$ra_ctg[$i]->SetAnchor($idmap, "$om_pos_start:$om_pos_end:$strand");
                }
            }
        }
    }
}

sub BuildPseudomolecules
{
    my ($self, %h_param) = @_;

    my $o_linkds         = $self->_FastOneGet('_o_linkds');
    my $rh_contigs       = $self->_FastOneGet('_rh_contigs');
    my $rh_o_seqds       = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist       = $self->_FastOneGet('_ra_seqlist');
    my $rh_scaffolds     = $self->_FastOneGet('_rh_scaffolds');
    my $o_omap           = $self->_FastOneGet('_o_omap');
    my $min_scaffold_len = $o_omap->_FastOneGet('_min_scaffold_len');
    my $maxvlen          = $self->_FastOneGet('_maxvalidationlength');
    my $vallen           = (defined($maxvlen) && $maxvlen ne '') ? $maxvlen : &MAXVALIDATION_LENGTH;
    my $validation       = (exists($h_param{validation}) && $h_param{validation}) ? $h_param{validation} : 2;
    my $minmingaplen     = $self->_FastOneGet('_mingaplen');
    my $unknowngaplen    = &MIN_UNKNOWN_GAPSIZE;
	my $scaffoldsarecontigs = (exists($h_param{scaffoldsarecontigs}) && $h_param{scaffoldsarecontigs}) ? &TRUE: &FALSE;
	# $h_param{mappingfile}         : uses an preexisting mapping file
	# $h_param{scaffoldsarecontigs} : creates a scaffold for each contig in the mapping file

    $self->_PrintParams("BuildPseudomolecules", %h_param);

    #	$self->ScaffoldsAnchoring(); # si je le fais je ne peux plus bypasser via la lecture de mappingfile
    $self->SynchronizeContigRegistry();

    if ($o_omap->GetNumberOfMaps() == 0)
    {
        $self->_Print("Warning: no optical map found, unable to run BuildPseudomolecules\n");
        return;
    }

    #
    # MAPPING
    #

#idmap   nb_sites        om_start        om_end  omlen   om_pos_start    om_pos_end      strand  idseq   seq_sites       seq_start       seq_stop        seqlen  seq_pos_start   seq_pos_end     ali_len score   density
    my @a_results = ();
    if (exists($h_param{mappingfile}) && -e $h_param{mappingfile})
    {
        my $f_in = &GetStreamIn($h_param{mappingfile});
        while (my $l = <$f_in>)
        {
            chomp($l);
            next if ($l =~ /#*idmap/);
			if( $scaffoldsarecontigs )
			{
				my @a_ = split("\t",$l);
				my $ctg = $a_[8];
				my $ctgseq = &LookupSeq($rh_o_seqds, $ra_seqlist, $ctg, 'p');
				my $o_ctgsrc = $self->GetRegisteredContig($ctg,  'p', length($ctgseq));
				my $o_scaff = New Scaffold('name' => 'TMP0src', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
				my $idscaff = $self->AddScaffold($o_scaff);
				$o_scaff->SetName($idscaff);
				print STDERR "$a_[8] ==> scaff id = $idscaff\n";
				$a_[8] = $idscaff;
				$o_ctgsrc->SetScaffold($o_scaff, $idscaff);
				$o_scaff->Push($o_ctgsrc);
				$self->ScaffoldsAnchoring($idscaff); # le AddScaffold le fait mais je le refais pour declencher l'ancrage car avant il n'y a pas de contigs au moment du Add
				$o_scaff->InitAnchor(&TRUE);
				#print STDERR Dumper($o_scaff) if (defined($o_scaff->GetAnchor()));
				$l = join("\t",@a_);
			}
            push(@a_results, $l);
        }
        $f_in->close;
    }
    else
    {
        my $outfile = (exists($h_param{mappingfile})) ? "$h_param{mappingfile}.$$" : "lynx-omap-mapping.$$";
        my $f_out = &GetStreamOut($outfile);
        foreach my $idscaff (sort {$a <=> $b} keys(%$rh_scaffolds))
        {
            #next if ( $idscaff < 20 || $idscaff > 30 );
            my ($r_scaffseq, $rh_ctg_pos) =
              $$rh_scaffolds{$idscaff}->GetSequence($o_linkds, $rh_o_seqds, $ra_seqlist, $minmingaplen);
            my @a_res = $o_omap->MapSequence($idscaff, $r_scaffseq, 'full', &MAXINT, &TRUE)
              ;    # format long, pas de limite dans le nombre de hits, chevauchement < min_scaffold_len/2
            foreach my $hit (@a_res)
            {
                print $f_out "$hit\n";
            }
            push(@a_results, @a_res);
        }
        $f_out->close;
    }

    # a partir de maintenant je ne veux pas faire de l'ancrage automatique (dans le AddScaffold)
    $self->_Set('_ontheflyanchoring', &FALSE);

#
# identifier les scaffolds qui doivent etre coupes
#
# attention quand je lis a partir d'un fichier les id peuvent etre Scaff00000 et non pas 0, ca veut dire que cela va etre dur sans refaire le mapping
    my @a_sorted_results = sort {
        my @A = split(/\t/, $a);
        my @B = split(/\t/, $b);
        return $B[8] cmp $A[8] if ($A[8] ne $B[8]);    # 8=idseq
        return $A[13] <=> $B[13];                      # 13=$seq_pos_start
    } @a_results;
    my $previdseq = '';
    my @a_hits    = ();

	print STDERR "$scaffoldsarecontigs\n";
    @a_results = ();
    foreach my $hit (@a_sorted_results)
    {
        my (
            $idmap,       $nb_sites, $om_start,  $om_end,    $omlen,    $om_pos_start, $om_pos_end,
            $strand,      $idseq,    $seq_sites, $seq_start, $seq_stop, $seqlen,       $seq_pos_start,
            $seq_pos_end, $ali_len,  $score,     $densite,   $span
          )
          = split(/\t/, $hit);
        $idseq =~ s/Scaff0*//;
        if ($idseq ne $previdseq)
        {
            if (scalar(@a_hits) > 1  && ! $scaffoldsarecontigs )
            {
                $self->SplitScaffold($previdseq, \@a_hits, $validation, $vallen);
            }
            push(@a_results, @a_hits);
            @a_hits = ();
        }
        push(@a_hits, $hit) if ( $seq_pos_end - $seq_pos_start > $min_scaffold_len );
        $previdseq = $idseq;
    }


    if (scalar(@a_hits) > 1 && ! $scaffoldsarecontigs )
    {
        $self->SplitScaffold($previdseq, \@a_hits, $validation, $vallen);    # ATTENTION pas fini ne coupe pas
    }
    push(@a_results, @a_hits);

    #
    # Pour chaque scaffold je garde sa position
    #
    # a ce stade la, apres le split je ne considere que le meilleur hit par scaffold (si egalite je garde tout)
    my %h_scaff2best = ();
    for (my $i = 0; $i < scalar(@a_results); $i++)
    {
        my (
            $idmap,       $nb_sites, $om_start,  $om_end,    $omlen,    $om_pos_start, $om_pos_end,
            $strand,      $idseq,    $seq_sites, $seq_start, $seq_stop, $seqlen,       $seq_pos_start,
            $seq_pos_end, $ali_len,  $score,     $densite,   $span
          )
          = split(/\t/, $a_results[$i]);

        # je ne dois pas faire le test de consistence tant que les scaffolds peuvent etre chimeriques
        my $scaffanchor = $$rh_scaffolds{$idseq}->GetAnchor();
        if (defined($scaffanchor))
        {
            my ($chr, $pos) = split(':', $scaffanchor);
            if ($chr ne $idmap)
            {
                $self->_Print("BuildPseudomolecule warning: inconsistency anchoring $idseq, previously=$scaffanchor, now=$idmap, skipped\n");
                next;
            }
        }
        if (!defined($h_scaff2best{$idseq}{SCORE}) || $score >= $h_scaff2best{$idseq}{SCORE})
        {
            $h_scaff2best{$idseq}{SCORE}   = $score;
            $h_scaff2best{$idseq}{DENSITY} = $densite;
        }
    }
	print STDERR Dumper(%h_scaff2best);
    my @a_filtresults = ();
    for (my $i = 0; $i < scalar(@a_results); $i++)
    {
        my (
            $idmap,       $nb_sites, $om_start,  $om_end,    $omlen,    $om_pos_start, $om_pos_end,
            $strand,      $idseq,    $seq_sites, $seq_start, $seq_stop, $seqlen,       $seq_pos_start,
            $seq_pos_end, $ali_len,  $score,     $densite,   $span
          )
          = split(/\t/, $a_results[$i]);
        if (   defined($h_scaff2best{$idseq})
            && $score == $h_scaff2best{$idseq}{SCORE}
            && $densite == $h_scaff2best{$idseq}{DENSITY})
        {
            push(@a_filtresults, $a_results[$i]);
        }
    }

    # trier le tableau selon les chr et les positions sur l'OM
    @a_sorted_results = sort {
        my @A = split(/\t/, $a);
        my @B = split(/\t/, $b);
        return $A[0] cmp $B[0] if ($A[0] ne $B[0]);
        return $A[5] <=> $B[5];
    } @a_filtresults;

    my $prev_end           = 0;
    my $prev_om            = '';
    my $prev_om_pos_end    = 0;
    my $prev_om_start      = 0;
    my $prev_om_end        = 0;
    my $prev_seq_pos_end   = 0;
    my $prev_seq_pos_start = 0;
    my $prev_seqlen        = 0;
    my $prev_strand        = '+';
    my $o_pseudo           = undef;
    for (my $i = 0; $i < scalar(@a_sorted_results); $i++)
    {
        my (
            $idmap,       $nb_sites, $om_start,  $om_end,    $omlen,    $om_pos_start, $om_pos_end,
            $strand,      $idseq,    $seq_sites, $seq_start, $seq_stop, $seqlen,       $seq_pos_start,
            $seq_pos_end, $ali_len,  $score,     $densite,   $span
          )
          = split(/\t/, $a_sorted_results[$i]);
        if ($strand eq '-')
        {
            my $seq_pos_end_bk = $seq_pos_end;
            $seq_pos_end   = $seqlen - $seq_pos_start + 1;
            $seq_pos_start = $seqlen - $seq_pos_end_bk + 1;
            $strand        = '+';
            $$rh_scaffolds{$idseq}->Reverse();
        }
        $idseq =~ s/Scaff0*//;
        if ($om_pos_end < $prev_om_pos_end && $idmap eq $prev_om)    # inclusion
        {
            print STDERR "# $a_sorted_results[$i]\n";

            # pour le moment, les hits chevauchants je les laisse de cote
            next;
        }

        print STDERR "$a_sorted_results[$i]\n";
        if ($prev_om ne $idmap)
        {
            if (defined($o_pseudo))
            {
                $self->AddScaffold($o_pseudo);
            }
            $o_pseudo =
              New Scaffold('name' => "PSEUDO$idmap", 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
        }
        else
        {
            # estimation de la taille du gap sur l'OM
            my $om_delta = $om_pos_start - $prev_om_pos_end;

            # estimation de la taille du gap sur la sequence
            my $prev_ovh  = ($prev_seqlen - $prev_seq_pos_end + 1);
            my $curr_ovh  = $seq_pos_start;
            my $seq_delta = ($prev_ovh + $curr_ovh);
            print STDERR "# OM  GAP=$om_delta\n";
            print STDERR "# SEQ OVH=$seq_delta\n";

            if ($om_delta < 0 || $seq_delta < 0)
            {
                my $max_overlap =
                  &Max(abs($om_delta), abs($seq_delta))
                  ;    # attention ca peut etre gros il faudra etre stringent dans Merge
                       # cela peut etre du a un chevauchement/entrelacement et plutot que de mettre un gap,
                       # je pourrais essayer de voir si je n'ai pas des contigs communs entre les deux scaffolds et ne
                       # commencer a rajouter le second qu'au dernier contig commun
                       # si je n'arrive pas a detecter le chevauchement, je rajoute un gap
                if (
                    $o_pseudo->MergeOverlapping($$rh_scaffolds{$idseq},           $max_overlap, $self->_FastOneGet('_gaplen_tolerance'), undef)
                  )
                {
                    print STDERR "\tMERGED\n";
                    delete($$rh_scaffolds{$idseq});
                }
                else
                {
                    $o_pseudo->Push("$unknowngaplen,$unknowngaplen");
                }
            }
            else
            {
                my $delta = $om_delta - $seq_delta;
                print STDERR "# OM-SEQ: $delta";
                if ($delta < 0)
                {
                    print STDERR "\t==> $unknowngaplen\n";
                    $o_pseudo->Push("$unknowngaplen,$unknowngaplen");
                }
                else
                {
                    print STDERR "\t==> $delta\n";
                    $o_pseudo->Push("$delta,$delta");
                }
            }
        }
        if (exists($$rh_scaffolds{$idseq}))
        {
            $$rh_scaffolds{$idseq}->DeregisterContigs();
            my $ra_path = $$rh_scaffolds{$idseq}->GetArray();
            for (my $i = 0; $i < scalar(@$ra_path); $i++)
            {
                $o_pseudo->Push($$ra_path[$i]);
            }
            delete($$rh_scaffolds{$idseq}) ;    # si je mappe a plusieurs endroit avec le meme score c'est peut etre dangeureux
        }
        $prev_om            = $idmap;
        $prev_om_pos_end    = $om_pos_end;
        $prev_om_start      = $om_start;
        $prev_om_end        = $om_end;
        $prev_seq_pos_end   = $seq_pos_end;
        $prev_seq_pos_start = $seq_pos_start;
        $prev_seqlen        = $seqlen;
        $prev_strand        = $strand;
    }
    $self->AddScaffold($o_pseudo);
    $self->ScaffoldsAnchoring();
    $self->_Set('_ontheflyanchoring', &TRUE);
}

# les hits sont tries par position croissante sur la sequence du scaffold
# attention ce n'est pas fini, je ne coupe pas

sub SplitScaffold
{
    my ($self, $idscaff, $ra_hits, $validation, $vallen) = @_;

    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $rh_scaffolds = $self->_FastOneGet('_rh_scaffolds');
    my $rh_o_seqds   = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist   = $self->_FastOneGet('_ra_seqlist');
    my $minmingaplen = $self->_FastOneGet('_mingaplen');

    my ($r_scaffseq, $rh_ctg_pos) =
      $$rh_scaffolds{$idscaff}->GetSequence($o_linkds, $rh_o_seqds, $ra_seqlist, $minmingaplen);
    my $ra_path = $$rh_scaffolds{$idscaff}->GetArray();

    print STDERR "ID=$idscaff\n";

    #print STDERR $$rh_scaffolds{$idscaff}->Get()."\n";

    # recupere les liens non valides par au moins 2
    my @a_notvalidated = $$rh_scaffolds{$idscaff}->GetListOfNotValidatedLinks($o_linkds, $validation, $vallen);
    print STDERR "NOTVALIDATED: \n";
    print STDERR join(" // ", @a_notvalidated);
    print STDERR "\n";

    #
    my @a_linkcoverage = $$rh_scaffolds{$idscaff}->GetLinkCoverage($o_linkds, $vallen);

    # calcule le lien le plus faible
    for (my $i = 0; $i < $#$ra_hits; $i++)
    {
        my @a_1 = split(/\t/, $$ra_hits[$i]);
        my @a_2 = split(/\t/, $$ra_hits[$i + 1]);
        print STDERR "$$ra_hits[$i]\n$$ra_hits[$i+1]\n";
        my ($min, $max) = &MinMax($a_1[14], $a_2[13]);    # au cas ou il y ait un chev
        print STDERR "===>$min,$max\n";
        my %h_weak = ();
        for (my $j = 0; $j < scalar(@$ra_path) - 1; $j += 2)
        {
            my @a_pos = split(/\t/, $$rh_ctg_pos{$$ra_path[$j]->GetName()});
            if (   ($a_pos[0] >= $min && $a_pos[1] <= $max)
                || ($a_pos[0] >= $min && $a_pos[0] <= $max)
                || ($a_pos[1] >= $min && $a_pos[1] <= $max))
            {
                my $src  = $$ra_path[$j]->GetName();
                my $next = $$ra_path[$j + 2]->GetName();
                $self->_Print(
                    "SplitScaffold $idscaff: $j src=$src (coverage=$a_linkcoverage[$j]) next=$next (coverage=$a_linkcoverage[$j+2])\n"
                    );
                my %h_val = ();
                if (!$o_linkds->GetKV2($src, $next, \%h_val))
                {
                    $self->_Print("SplitScaffold $idscaff: $j src=$src next=$next unlinked WEAK\n");
                    $h_weak{UNLINKED} .= "$j ";
                }
                else
                {
                    $self->_Print(join(' ', %h_val) . "\n");
                    if (!defined($h_weak{MINCOUNT}) || $h_val{count} < $h_weak{MINCOUNT})
                    {
                        $h_weak{MINCOUNT}    = $h_val{count};
                        $h_weak{MINCOUNTPOS} = $j;
                        $h_weak{MINGAP}      = $$ra_path[$j + 1];
                        $h_weak{MINCOVERAGE} = $a_linkcoverage[$j];
                    }
                    elsif (defined($h_weak{MINCOUNT}) && $h_val{count} == $h_weak{MINCOUNT})
                    {
                        $h_weak{MINCOUNTPOS} .= " $j";
                        $h_weak{MINGAP}      .= " $$ra_path[$j+1]";
                        $h_weak{MINCOVERAGE} .= " $a_linkcoverage[$j]";
                    }
                }
                if ($j != 0 && $j != scalar(@$ra_path) - 1)
                {
                    if (!defined($h_weak{COVMIN}) || $a_linkcoverage[$j] < $h_weak{COVMIN})
                    {
                        if ($$ra_path[$j]->GetLength() < $vallen / 2)
                        {
                            $h_weak{COVMIN}    = $a_linkcoverage[$j];
                            $h_weak{COVMINPOS} = $j;
                        }
                    }
                    elsif (defined($h_weak{COVMIN}) && $a_linkcoverage[$j] == $h_weak{COVMIN})
                    {
                        if ($$ra_path[$j]->GetLength() < $vallen / 2)
                        {
                            $h_weak{COVMINPOS} .= " $j";
                        }
                    }
                }
            }
            elsif ($a_pos[0] <= $min && $a_pos[1] >= $max)
            {
                $self->_Print("SplitScaffold $idscaff: $j INTRA CONTIG " . $$ra_path[$j]->GetName() . "\n");
            }
        }
        print STDERR Dumper(%h_weak);
    }
}

#=========================================================================================================
#=========================================================================================================
#=========================================================================================================
#=========================================================================================================

# extrait tous les liens des paths chevauchants et lance le scaffolding
# il y a des trucs en dur, ok si beaucoup de banques d'inserts differents sinon est ce que l'on peut valider?

sub ScaffoldingOverlappingPaths
{
    my ($self, $rh_paths, $rh_lscaff) = @_;

    $self->_Print("ScaffoldingOverlappingPaths starting\n");
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $o_locallinks = New LinkSet();
    my $o_lscaff = New Scaffolder('gaplen_tolerance'    => $self->_FastOneGet('_gaplen_tolerance'),
                                  'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
    my %h_contig2scaff = ();
    foreach my $scaff (sort {&GetPathLength($$rh_paths{$b}) <=> &GetPathLength($$rh_paths{$a})} keys(%$rh_paths))
    {
        my @a_path = split(' ', $$rh_paths{$scaff});
        for (my $i = 0; $i <= $#a_path - 2; $i += 2)
        {
            for (my $j = $i + 2; $j <= $#a_path; $j += 2)
            {
                my ($src, $srcext, $tgt, $tgtext) = &_Scaffold2Path($a_path[$i], $a_path[$j], &TRUE);
                my %h_val = ();
                if ($o_linkds->GetKV2Oriented($src, $srcext, $tgt, $tgtext, \%h_val))
                {
                    $o_locallinks->SetKV2($src, $tgt, \%h_val);
                }
            }
        }

        # je veux quand meme garder les scaffolds initiaux (les plus grands) independants pour eviter de degrader
        # mais apres je vais rajouter les autres liens via les metthodes Scaffolding
        my $dup  = undef;
        my $path = $$rh_paths{$scaff};
        while ($path =~ /#(\S+)([pm])(\d+)/g)
        {
            my ($ctg, $ori, $len) = ($1, $2, $3);
            if (exists($h_contig2scaff{$ctg}))
            {
                $dup = $ctg;
                last;
            }
        }
        if (!defined($dup))
        {
            #print STDERR "NO RED $o_lscaff $path\n";
            $o_lscaff->AddScaffold($path);
            $o_lscaff->InitContig2Scaffold(\%h_contig2scaff);
        }
    }
    $o_lscaff->SetLinks($o_locallinks);
    $o_lscaff->InitGraph(&TRUE, undef, &TRUE);    # attention je prends les weak mais en principe j'ai initie le vpath sans les weak
    my %h_scaff_param = ();

    #$h_scaff_param{dumpfile} = "FPC0170.dump" if ( $cluster eq 'FPC0170'  );
    %h_scaff_param = split(
        /[=\s]/,
        "mincount=50 maxgaplen_max=5000 maxweak=0 validation=0 mingaplen_max=1 minsrclen=10000 mintgtlen=10000 minsrcspan=5000 mintgtspan=5000"
        );
    $o_lscaff->Scaffolding(%h_scaff_param);    # 1er  passage j'assemble les gros
    %h_scaff_param = ();

    #$h_scaff_param{validation}=2;  #en principe le lien englobant est donne, change le 20120205
    $o_lscaff->Scaffolding(%h_scaff_param);
    %$rh_lscaff = $o_lscaff->GetAllScaffolds();
    $self->_Print("ScaffoldingOverlappingPaths done\n");
}

#
#

sub _UpdateScaffLinks
{
my ($self,$rh_graph, $o_linkds, $o_linksscaff, $rh_contig2scaff, $rh_ctg, $o_ctg, $idscaff, $scaffext) = @_;

    my ($src, $ori, $srclen) = $o_ctg->GetArray();

	my $rh_best  = $self->_FastOneGet('_rh_bests'); # 20130502 je rajoute la contrainte du best => 20130503 c'est trop fort

    my $exposed_srcend = $H_CTG2SCAFFEXT{"$ori $scaffext"};
    my @a_successors   = keys(%{$$rh_graph{$src}{$exposed_srcend}});
    foreach my $succ (@a_successors)
    {
        next if (!exists($$rh_ctg{$succ}));    # ou il n'est pas sur une extremite (encore moins dans un scaff)

        next if ($$rh_contig2scaff{$succ} <= $$rh_contig2scaff{$src});    # ou bien deja fait ou dans le meme scaff

        my %h_val = ();
        next if (!$o_linkds->GetKV2($src, $succ, \%h_val));
        my ($s, $srcext, $t, $succext) =
            ($src lt $succ)
          ? ($src, $h_val{srcext}, $succ, $h_val{tgtext})
          : ($succ, $h_val{tgtext}, $src, $h_val{srcext});

		#if ( ( &Exists($rh_best,$s,$srcext,'best') && $$rh_best{$s}{$srcext}{best} =~ /\b$t\b/ ) ||
		#     ( &Exists($rh_best,$t,$succext,'best') && $$rh_best{$t}{$succext}{best} =~ /\b$s\b/ ) )
		#{
			my $succscaff = $$rh_contig2scaff{$succ};
			next if (!&Exists($rh_ctg, $succ, $succscaff));

			foreach my $succscaffext (keys(%{$$rh_ctg{$succ}{$succscaff}}))
			{
				my $connexion = $$rh_ctg{$succ}{$succscaff}{$succscaffext} . " " . $succscaffext;

				#my $exposed_succend = $H_CTG2SCAFFEXT{$$rh_ctg{$succ}{$succscaff}};
				my $exposed_succend = $H_CTG2SCAFFEXT{$connexion};
				next if ($exposed_succend ne $succext || $exposed_srcend ne $srcext);

				my %h_res = ();
				$h_res{count}  = $h_val{count};
				$h_res{srclst} = "$src;";
				$h_res{tgtlst} = "$succ;";
				$h_res{srcext} = $scaffext;
				$h_res{tgtext} = $succscaffext;
				$o_linksscaff->AddKV("src=$idscaff tgt=$succscaff", \%h_res);
			}
		#}
    }

    return $srclen;
}

# je ne considere pas les gap entre scaff pour le calcul de la longueur

sub LinkScaffolds
{
    my ($self, $outfile, $mode, %h_param) = @_;

    $self->_PrintParams("LinkScaffolds starting", %h_param);
    my $o_linkds     = $self->_FastOneGet('_o_linkds');
    my $rh_scaff     = $self->_FastOneGet('_rh_scaffolds');
    my $o_linksscaff = $self->_FastOneGet('_o_linksscaff');
    my $idscaff      = scalar(keys(%$rh_scaff));

    my $maxvlen = $self->_FastOneGet('_maxvalidationlength');
    my $vallen = (defined($maxvlen) && $maxvlen ne '') ? $maxvlen : &MAXVALIDATION_LENGTH;
    print STDERR "LinkScaffolds: Validation length: $vallen\n";

    my $minnbctg = (exists($h_param{minlinkedcontigs})) ? $h_param{minlinkedcontigs} : $MINCTGLINKED_BETWEEN_SCAFFOLDS;
	delete($h_param{minlinkedcontigs}) if ( exists($h_param{minlinkedcontigs}) ); # je la supprime car sinon j'ai des msg a cause du controle dans Filter

    if (defined($mode) && $mode eq 'update')
    {
        # je quitte si j'ai deja les liens ou s'il n'y a pas de scaffolds
        #printf STDERR "SCA: %d\n",scalar(keys(%$rh_scaffoldslinks));
        #printf STDERR "IDS: %d\n",$idscaff;
        return if (scalar($o_linksscaff->GetLinks()) || !$idscaff);
        $self->_Print("LinkScaffolds update mode starting\n");
    }

    $o_linksscaff->Reset();

    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux lien (et plus au graphe) je suis oblige de travailler sur un temporaire filtre
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();
    my ($rh_graph,$o_graph) = $self->InitGraph(&FALSE, $ra_links_filt, &TRUE);    # ici je ne veux dans le graphe que ce qui passe le seuil

#remplace le 20130331
# j'initialise le graphe avec les lien filtres donc tous les chemins doivent etre au dessus du filtre
#    my @a_filt_links = ();
#    $o_linkds->Filter(\@a_filt_links, %h_param);    # ici je ne veux que les liens qui passent le seuil
#    my ($rh_graph) = $self->InitGraph(&FALSE, \@a_filt_links, &TRUE)
#      ;    # je confirme que je veux les liens filtres dans LinkScaffolds, je veux les weaks mais je peux filtrer avant

    my %h_contig2scaff = ();
    $self->InitContig2Scaffold(\%h_contig2scaff);    # updating mode
    $self->ScaffoldsAnchoring();

    #print STDERR "idscaff=$idscaff $rh_graph\n";
    # premier passage je fais la liste de l'orientation de tous les contigs dans les scaffolds
    my %h_ctg               = ();
    my %h_maxctgatscaffends = ();
    foreach my $idscaff (sort keys(%$rh_scaff))
    {
        #my $pathlen=&GetPathLength($$rh_scaff{$idscaff});
        my $len    = 0;
        my $ra_ctg = $$rh_scaff{$idscaff}->GetArrayContigs();

        #print STDERR "$idscaff 5\n";
        for (my $i = 0; $i < scalar(@$ra_ctg) && $len < $vallen; $i++)
        {
            $h_maxctgatscaffends{$idscaff}{5}++;
            my ($ctg, $ori, $ctglen) = $$ra_ctg[$i]->GetArray();
            $h_ctg{$ctg}{$idscaff}{5} = $ori;
            $h_ctg{$ctg}{len} = $ctglen;
            $len += $ctglen;

            #print STDERR "\t$i-$a_[$i]";
        }

# 20120724 je ne comprends plus ce delete($h_maxctgatscaffends{$idscaff}{5}) if ( $len < $vallen );	# je ne le garde que si je depasse
#print STDERR "\t$len\n";
        $len = 0;

        #print STDERR "$idscaff 3\n";
        for (my $i = scalar(@$ra_ctg) - 1; $i >= 0 && $len < $vallen; $i--)
        {
            $h_maxctgatscaffends{$idscaff}{3}++;
            my ($ctg, $ori, $ctglen) = $$ra_ctg[$i]->GetArray();
            $h_ctg{$ctg}{$idscaff}{3} = $ori;
            $h_ctg{$ctg}{len} = $ctglen;
            $len += $ctglen;

            #print STDERR "\t$i-$a_[$i]";
        }

        #  20120724 je ne comprends plus  ce delete($h_maxctgatscaffends{$idscaff}{3}) if ( $len < $vallen );
        #print STDERR "\t$len\n";
    }

    # deuxieme passage, je compte les liens entre scaffolds
    my %h_scaffgraph = ();
    foreach my $idscaff (sort {$a <=> $b} keys(%$rh_scaff))    # bien faire le sort
    {
        my $ra_ctg = $$rh_scaff{$idscaff}->GetArrayContigs();
        my $len    = 0;
        for (my $i = 0; $i < scalar(@$ra_ctg) && $len < $vallen; $i++)
        {
            $len += $self->_UpdateScaffLinks($rh_graph, $o_linkds_filt, $o_linksscaff, \%h_contig2scaff, \%h_ctg, $$ra_ctg[$i], $idscaff, "5");
        }
        $len = 0;
        for (my $i = scalar(@$ra_ctg) - 1; $i >= 0 && $len < $vallen; $i--)
        {
            $len += $self->_UpdateScaffLinks($rh_graph, $o_linkds_filt, $o_linksscaff, \%h_contig2scaff, \%h_ctg, $$ra_ctg[$i], $idscaff, "3");
        }
    }

    #my %h_param_tmp = %h_param;
    #$h_param_tmp{dumpfile} = "$outfile.tmp.$$";
    #$o_linksscaff->Dump(%h_param_tmp);
    $o_linksscaff->CleanUp();    # 20120410 je ne comprends pas pourquoi je fais ca ici
                                 # j'ai compris, c'est au cas ou il y ait un soucis d'orientation des scaffolds
    my @a_scafflinks = $o_linksscaff->GetLinks();
    my %h_key        = ();
    my %h_val        = ();
    foreach my $scafflink (@a_scafflinks)
    {
        %h_key = split(/[=\s]+/, $scafflink);

        my $scafflinkdetails = $o_linksscaff->Get($scafflink);

        #print STDERR "$scafflink\n\t$scafflinkdetails\n";

        %h_val = ();
        $o_linksscaff->GetKV($scafflink, \%h_val);

        my $nb_1N = 0;
        foreach my $tag ('src', 'tgt')
        {
            my $ext = $tag . "ext";
            my $lst = $tag . "lst";
            next if (!&Exists(\%h_maxctgatscaffends, $h_key{$tag}, $h_val{$ext}));
            my $minnbc = &Min($h_maxctgatscaffends{$h_key{$tag}}{$h_val{$ext}}, $minnbctg);
            $nb_1N++ if ($minnbc < $minnbctg);
            my @a_ = split(/[\;]/, $h_val{$lst});

            # j'ai rajoute $h_val{count} < $minnbctg pour ne pas accepter de tres gros contigs lies par un seul lien
            # j'ai ensuite rajoute $nb_1N >  pour eviter de considerer les liens 1-1
            if (scalar(@a_) < $minnbc || $h_val{count} < $minnbctg || $nb_1N > 1)
            {
                #		print STDERR "\t$#a_+1 < $minnbc // $h_val{count} < $minnbctg || $nb_1N > 1 )\n";
                $o_linksscaff->Del($scafflink);
                last;
            }
        }
    }
    my %h_param_end = ('dumpfile' => $outfile);
    $o_linksscaff->Dump(%h_param_end);
    $self->_Print("LinkScaffolds done\n");
}

# pour chaque contig d'un path je cherche s'il y a un successeur
#    1. plus proche que l'actuel
#    2. lie/compatible avec les contigs aval
# il faut faire la manip dans les deux sens

sub Finishing
{
    my ($self, %h_param) = @_;

    $self->_Print("Finishing starting\n");
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance})) ? $h_param{gaplen_tolerance} : $gaplen_tolerance_ori;
    my $minlen           = (exists($h_param{minlen})) ? $h_param{minlen} : 0;
    my $rh_scaff         = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds         = $self->_FastOneGet('_o_linkds');
    my $vallen = $self->_FastOneGet('_maxvalidationlength');

    my $o_linkds_filt = New LinkSet();	# ATTENTION comme je passe sur o_linkds pour avoir access aux lien (et plus au graphe) je suis oblige de travailler sur un temporaire
    my $nb_ignored   = $o_linkds_filt->FilterAndLoad($o_linkds, %h_param);
    my $ra_links_filt = $o_linkds_filt->GetLinksRef();
    my ($rh_graph,$o_graph_filt) = $self->InitGraph(&FALSE, $ra_links_filt, &TRUE);    # ici je ne veux dans le graphe que ce qui passe le seuil

# modifie le 20130331
#    my @a_filt_links     = ();
#    $o_linkds->Filter(\@a_filt_links, %h_param);
#    # un graphe filtre pour les candidats
#    my ($rh_graph_filt, $o_graph_filt) = $self->InitGraph(&FALSE, \@a_filt_links, &TRUE);

    $self->SynchronizeContigRegistry();
    $self->ScaffoldsAnchoring();

    # un graphe avec tout pour valider
    # my ($rh_graph,$o_graph) = $self->InitGraph(&TRUE,undef,&TRUE);


    my @a_sorted_scaff = sort {$$rh_scaff{$b}->GetLength() <=> $$rh_scaff{$a}->GetLength()} keys(%$rh_scaff);
    for (my $m = 0; $m < scalar(@a_sorted_scaff); $m++)
    {
        my $idscaff = $a_sorted_scaff[$m];
        next if ($$rh_scaff{$idscaff}->GetNumberOfContigs() < 2);

        # $$rh_scaff{$idscaff}->RecomputeGapLength($o_linkds);

        my $edited = &TRUE;
        while ($edited)
        {
            #print STDERR "\n$idscaff len=".$$rh_scaff{$idscaff}->GetLength()."\n";
            $edited = &FALSE;
            for (my $reverse = 0; $reverse <= 1; $reverse++)
            {
                if ($reverse)
                {
                    $$rh_scaff{$idscaff}->Reverse();
                }
                my $ra_path    = $$rh_scaff{$idscaff}->GetArray();
                my @a_new_path = ();
                my $o_left = New Scaffold(name => 'TMPL', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
                my $o_right = New Scaffold(name => 'TMPR', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
                for (my $i = 0; $i < scalar(@$ra_path); $i++)
                {
                    $o_right->Push($$ra_path[$i]);
                }
                $o_right->DeregisterContigs();
                my $prev_gap = undef;
                my $o_ctg    = undef;
                for (my $i = 0; $i < (scalar(@$ra_path) - 2); $i += 2)
                {
                    $o_left->Push($prev_gap) if (defined($prev_gap));
                    ($o_ctg, $prev_gap) = $o_right->Shift();    # je transfere un contig de droite a gauche
                    $o_left->Push($o_ctg);
                    $o_left->DeregisterContigs();
                    push(@a_new_path, $$ra_path[$i]);
                    my ($src, $srcext, $next, $nextext) = $$ra_path[$i]->ConvertPathLink2GraphLink($$ra_path[$i + 2], &FALSE);
                    my $ra_cand = $o_graph_filt->GetContigsCompatiblesWithGap($o_linkds_filt, $src, $srcext, $next, $nextext, $gaplen_tolerance, $$ra_path[$i + 1]);
                    if (scalar(@$ra_cand) == 0)
                    {
                        push(@a_new_path, $$ra_path[$i + 1]);
                        next;
                    }

                    #print STDERR "=====$src,$srcext,$next,$nextext=====\n";
                    #print STDERR Dumper($ra_cand);
                    my $found = &FALSE;
                    foreach my $cand (@$ra_cand) # 'ctg [pm] len' # je verifie que le successeur est lie a un autre contig du scaffold
                    {
                        my ($ctg, $strand, $ctglen, $gapupstream, $gapdownstream) = split(' ', $cand);
                        next if ($$rh_scaff{$idscaff}->IsContainingContig($ctg)); # pas deja dans le scaffold, A ACTIVER
                             #print STDERR "($ctg,$strand,$ctglen)\n";
                        my $o_ctg = New Contig('name' => $ctg, 'strand' => $strand, 'len' => $ctglen);

                        #print STDERR "AJOUT1: ".$o_ctg->Get()."\n";
                        my $o_scaff = New Scaffold(name => $idscaff, 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
                        $o_scaff->Push($o_ctg);
                        $o_scaff->DeregisterContigs();

#print STDERR "AJOUT1: ".$o_scaff->Get()."\n";
#if ( $o_left->IsLinkValidated($o_scaff,$o_linkds,2,&TRUE) ) # si je valide avec l'aval je risque l'entrelacement, c'est pas mal mais ca n'empeche pas tout
                        if (   $o_left->IsLinkValidated($o_scaff, $o_linkds, 2, &TRUE)
                            || $o_scaff->IsLinkValidated($o_right, $o_linkds, 2, &TRUE))
                        {

                            # je vais verifier qu'aucun contig prevu en 3' n'est lie en 5'
                            my $o_vscaff = New Scaffold(name => 'TMP', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
                            my $len   = 0;
                            my @a_tmp = ();
                            for (my $v = $i + 2; $v < scalar(@$ra_path) && $len < $vallen; $v++)
                            {
                                if ($v % 2 == 0)
                                {
                                    $len += $$ra_path[$v]->GetLength();
                                }
                                push(@a_tmp, $$ra_path[$v]);
                            }
                            $o_vscaff->Push(@a_tmp);
                            $o_vscaff->DeregisterContigs();
                            if ($o_vscaff->IsLinkValidated($o_scaff, $o_linkds, 1, &TRUE))
                            {
                                print STDERR "$idscaff/$i/$#$ra_path strand=$reverse src=" . $$ra_path[$i]->Get() . " srcext=$srcext next=" . $$ra_path[$i + 2]->Get() . " nextext=$nextext GAP=$$ra_path[$i+1]\n";
                                print STDERR "REJECTED: potential interleaving\n";
                                next;
                            }

                            print STDERR "$idscaff/$i/$#$ra_path strand=$reverse src=" . $$ra_path[$i]->Get() . " srcext=$srcext next=" . $$ra_path[$i + 2]->Get() . " nextext=$nextext GAP=$$ra_path[$i+1]\n";
                            print STDERR "\tLinkValidated ($ctg,$strand,$ctglen,$gapupstream,$gapdownstream) GAP=$$ra_path[$i+1], tolerance=$gaplen_tolerance\n";

                            my $o_ctg = $self->GetNewRegisteredContig($ctg, $strand, $ctglen);
                            if (!$$rh_scaff{$idscaff}->IsContigAnchoringCompatible($o_ctg))
                            {
                                print STDERR "REJECTED: contig anchoring >" . $o_ctg->GetAnchorsFromRegistry() . "< is not compatible with current scaffold anchoring >" . $$rh_scaff{$idscaff}->GetAnchor() . "<\n";
                                next;
                            }
                            push(@a_new_path, $gapupstream);

                            #print STDERR "AJOUT2: ".$o_ctg->Get()."\n";
                            push(@a_new_path, $o_ctg);
                            push(@a_new_path, $gapdownstream);
                            $found = &TRUE;
                            last;
                        }
                    }
                    if (!$found)
                    {
                        push(@a_new_path, $$ra_path[$i + 1]);
                    }
                }
                push(@a_new_path, $$ra_path[$#$ra_path]);
                if (scalar(@a_new_path) > scalar(@$ra_path))
                {
                    #print STDERR "\n=======================================================\n";
                    #print STDERR "\tUPDATE ".scalar(@a_new_path)." AV=".scalar(@$ra_path)."\n";
                    $$rh_scaff{$idscaff}->ResetPath();
                    $$rh_scaff{$idscaff}->Push(@a_new_path);
                    $$rh_scaff{$idscaff}->SetContigs();
                    $$rh_scaff{$idscaff}->InitAnchor(&TRUE);

                    #print STDERR Dumper($$rh_scaff{$idscaff});
                    $edited = &TRUE;
                }
            }
            $$rh_scaff{$idscaff}->Reverse();
        }

        #last; # je ne fais que le premier pour pouvoir analyser les resultats
    }
    $self->ScaffoldsAnchoring();
    $self->TrimRepeats("Finishing");
    $self->SynchronizeContigRegistry();
    $self->_Print("Finishing done\n");
}

sub FillGapsWithScaffolds
{
    my ($self, %h_param) = @_;

    $self->_Print("FillGapsWithScaffolds starting\n");
    my $gaplen_tolerance_ori = $self->_FastOneGet('_gaplen_tolerance');
    my $gaplen_tolerance = (exists($h_param{gaplen_tolerance})) ? $h_param{gaplen_tolerance} : $gaplen_tolerance_ori;
    my $rh_scaff         = $self->_FastOneGet('_rh_scaffolds');
    my $o_linkds         = $self->_FastOneGet('_o_linkds');

    my @a_sorted_scaff    = sort {$$rh_scaff{$a}->GetLength() <=> $$rh_scaff{$b}->GetLength()} keys(%$rh_scaff);
    my @a_revsorted_scaff = reverse @a_sorted_scaff;
    my %h_scafflen        = ();

    $self->ScaffoldsAnchoring();

    #foreach my $idscaff (@a_revsorted_scaff)
    for (my $m = 0; $m < scalar(@a_sorted_scaff); $m++)
    {
        my $idscaff = $a_sorted_scaff[$m];
        print STDERR "\n$idscaff\n";
        my $ra_path = $$rh_scaff{$idscaff}->GetArray();
        next if (scalar(@$ra_path) == 1);
        for (my $i = 1; $i < (scalar(@$ra_path) - 1); $i += 2)    # gaps, en plus je fais un i+1
        {
            my $max    = scalar(@$ra_path) - 1;
            my @a_     = split(',', $$ra_path[$i]);
            my $gaplen = $a_[1] + int($a_[1] * $gaplen_tolerance / 100);
            my $o_left = New Scaffold(name => 'TMPL', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
            $o_left->Push(@$ra_path[0 .. $i - 1]);
            $o_left->DeregisterContigs();
            my $o_left_border = New Scaffold(name => 'TMPLB', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
            $o_left_border->Push($$ra_path[$i - 1]);
            my $o_right = New Scaffold(name => 'TMPR', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
            $o_left_border->DeregisterContigs();
            $o_right->Push(@$ra_path[$i + 1 .. $max]);
            my $o_right_border = New Scaffold(name => 'TMPRB', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
            $o_right_border->Push($$ra_path[$i + 1]);
            $o_right_border->DeregisterContigs();

            for (my $k = 0; $k < $i; $k++)
            {
                my $smallerscaff = $a_sorted_scaff[$k];
                if (!defined($h_scafflen{$smallerscaff}))
                {
                    $h_scafflen{$smallerscaff} = $$rh_scaff{$smallerscaff}->GetLength();
                }
                my $smallerlen = $h_scafflen{$smallerscaff};
                last if ($smallerlen > $gaplen);

                next if (!$$rh_scaff{$idscaff}->AreAnchoringCompatibles($$rh_scaff{$smallerscaff}));

                #printf STDERR " $smallerscaff,$smallerlen,%d,%d ",length($upstreampath),length($downstreampath);

                my $insert = undef;
                my ($previd, $prevstrand) = $$ra_path[$i - 1]->GetArray();
                my ($nextid, $nextstrand) = $$ra_path[$i + 1]->GetArray();
                for (my $i = 0; $i < 2; $i++)
                {
                    $$rh_scaff{$smallerscaff}->Reverse() if ($i);

                    my $ra_smallerpath = $$rh_scaff{$smallerscaff}->GetArray();

# il faut un lien au niveau des contigs des extremites du gap sinon plusieurs gaps proches peuvent etre possibles ==> entrelacement
                    if (   $o_left_border->IsLinkValidated($$rh_scaff{$smallerscaff}, $o_linkds, 1, &TRUE)
                        && $$rh_scaff{$smallerscaff}->IsLinkValidated($o_right_border, $o_linkds, 1, &TRUE))
                    {
                        # ensuite je cherche le sous chemin reelement ancre dans l'insert
                        my %h_val = ();
                        my $imin  = 0;
                        for ($imin = 0; $imin < scalar(@$ra_smallerpath); $imin += 2)
                        {
                            my ($id, $strand) = $$ra_smallerpath[$imin]->GetArray();
                            print STDERR "imin=$imin / " . scalar(@$ra_smallerpath) . " $previd,$prevstrand,$id,$strand \n";
                            last if ($o_linkds->GetKV2OrientedContigs($previd, $prevstrand, $id, $strand, \%h_val));

                            # je ne veux pas risquer l'entrelacement
                            # ca ne marche pas car le point commun peut etre interne
                        }
                        next if ($imin >= scalar(@$ra_smallerpath));
                        my $imax = scalar(@$ra_smallerpath) - 1;
                        while ($imax >= 0)
                        {
                            my ($id, $strand) = $$ra_smallerpath[$imax]->GetArray();
                            print STDERR "imax=$imax / "
                              . scalar(@$ra_smallerpath)
                              . " $id,$strand,$nextid,$nextstrand \n";
                            last if ($o_linkds->GetKV2OrientedContigs($id, $strand, $nextid, $nextstrand, \%h_val));
                            $imax -= 2;
                        }

                        #$self->_Print("imin=$imin, imax=$imax $#a_insert\n");

                        next
                          if ($imax <= $imin || $imax == 0 || $imin == scalar(@$ra_smallerpath))
                          ; # pourquoi, meme si je ne recupere qu'un contig, je peux le faire (mais en principe c'est fait par AddRepeats

                        my $o_scaffgap =
                          New Scaffold(name => 'GAP', 'maxvalidationlength' => $self->_FastOneGet('_maxvalidationlength'));
                        $o_scaffgap->Push(@$ra_smallerpath[$imin .. $imax]);
                        $o_scaffgap->DeregisterContigs();

                        # et en plus il faut plusieurs liens indirects
                        if (   $o_left->IsLinkValidated($o_scaffgap, $o_linkds, 2, &TRUE)
                            && $o_scaffgap->IsLinkValidated($o_right, $o_linkds, 2, &TRUE))
                        {
                            print STDERR "$idscaff.$m.$gaplen strand=$i " . $o_scaffgap->Get() . " ==============\n";
                            $self->_Print(  "path insert validated candidate for scaff $idscaff pos=$m ($imin-$imax / "
                                          . scalar(@$ra_smallerpath)
                                          . ")\n");

                            # supprimer le scaff integre // attention seulement une partie a ete integree
                            last;
                        }
                    }
                }
                if (defined($insert))
                {
                    $self->_Print(
                          "FillGapsWithScaffolds $idscaff at $m, insertion of $smallerscaff len=$smallerlen $insert\n");
                }
            }
        }
    }
    $self->TrimRepeats("FillGapsWithScaffolds");
    $self->SynchronizeContigRegistry();
    $self->_Print("FillGapsWithScaffolds done\n");
}

sub MicroScaffolding
{
    my ($self, %h_param) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    Carp::confess("MicroScaffolding - a migrer\n");
}

sub ReassembleScaffolds
{
    my ($self, %h_param) = @_;

    Carp::croak("ReassembleScaffolds - a migrer\n");
}

sub SeedScaffolding
{
    my ($self, %h_param) = @_;

    my $rh_o_seqds = $self->_FastOneGet('_rh_o_seqds');
    my $ra_seqlist = $self->_FastOneGet('_ra_seqlist');
    Carp::confess("SeedScaffolding - a migrer\n");
}

sub ValidateScaffoldsSpan
{
    my ($self, $outfile, %h_param) = @_;

    Carp::confess("ValidateScaffoldsSpan - a migrer\n");
}

sub TrackNextScaffolds
{
    my ($self, %h_param) = @_;

    Carp::confess("TrackNextScaffolds - a migrer \n");
}

##############################################################################################################################################

#sub Exists
#{
#my($rh_,@a_keys) = @_;
#
#	foreach my $k (@a_keys)
#	{
#		return &FALSE if ( ! exists($$rh_{$k}) );
#		$rh_ = $$rh_{$k};
#	}
#	return &TRUE;
#}

# existe aussi dans Scaffold.pm

sub LookupSeq
{
    my ($rh_o_seqds, $ra_seqlist, $id, $strand) = @_;

    foreach my $seqds (@$ra_seqlist)
    {
        my $seq = $$rh_o_seqds{$seqds}->GetSeq($id);
        next if (!defined($seq) || $seq eq '');
        if ($strand eq 'm' || $strand eq '-')
        {
            return &RevCompSeq(\$seq);
        }
        return $seq;
    }
    print STDERR "WARNING: unable to find >$id<\n";
    return 'N' x 111;
}

sub _PrintParams
{
    my ($self, $fct, %h_param) = @_;

    my $param = '';
    foreach my $p (sort keys(%h_param))
    {
        $param .= "$p=$h_param{$p} ";
    }
    $self->_Print("$fct starting\n\tparams=$param\n");
}

sub _Print
{
    my ($self, $msg) = @_;

    print STDERR "$msg";
}

1;
