#
# $Id: ContigRegistry.pm 236 2013-12-27 17:25:23Z gouzy $
#

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

package ContigRegistry;

use strict;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use strict;
use warnings;
use Digest::MD5 qw(md5 md5_hex md5_base64);

my %H_SCAFF2EXT    = ( "p p" => "3 5", "p m" => "3 3", "m p" => "5 5", "m m" => "5 3" );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 236 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
my ($self,%h_param) = @_;

	my @a_expected = ('len','name');
	$self->AssertParams(\%h_param,\@a_expected);

	$self->SUPER::_Init();

	$self -> _Set ('_len', 0);
	$self -> _Set ('_name', '');
	$self -> _Set ('_repeat', &FALSE);
	my @a_ = ();
	$self -> _Set ('_ra_contigs', \@a_);	  

	foreach my $param (keys(%h_param))
	{
		$self -> _Set ("_$param", $h_param{$param});
	}
}


sub GetName
{
my($self)=@_;

	return $$self{_name};
}


sub SetName
{
my($self,$str)=@_;

	$$self{_name}=$str;
}


sub GetLength
{
my($self)=@_;

	return $$self{_len};
}


sub SetLength
{
my($self,$len)=@_;

	$$self{_len}=$len;
}

sub GetContigArray
{
my($self)=@_;

	return $self->_FastOneGet('_ra_contigs');
}


sub AddContig
{
my($self,$o_ctg)=@_;

	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	return if ( $self->IsContainingContig($o_ctg) ); # peut etre mettre a jour les scaffolds ?
	$o_ctg->_Set('_contigregistry',$self);
	push(@$ra_contigs,$o_ctg);
}

sub IsRepeat
{
my($self)=@_;

	return $self->_FastOneGet('_repeat');
}

sub SetRepeat
{
my($self)=@_;

	if ( $self->GetNumberOfScaffolds() > 1)
	{
		$self->_Set('_repeat',&TRUE);
	}
	else
	{
		$self->_Set('_repeat',&FALSE);
	}
}

sub IsContainingContig
{
my($self,$o_ctg)=@_;

	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	foreach my $ctg (@$ra_contigs)
	{
		return &TRUE if ( $o_ctg eq $ctg );
	}
	return &FALSE;
}

# retourne le contig principal, le cree s'il n'existe pas
sub GetMainContig
{
my($self,$ctg,$strand,$len)=@_;

	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	if ( ! scalar(@$ra_contigs) )
	{
		my $o_ctg = New Contig(name=>$ctg,strand=>$strand,len=>$len,contigregistry=>$self);
		push(@$ra_contigs,$o_ctg);
		return $o_ctg;
	}
	foreach my $o_ctg (@$ra_contigs)
	{
		# je renvoie le premier qui soit associe a un scaffold (en principe celui dont les liens avec les voisins sont les plus forts
		return $o_ctg if ( defined($o_ctg) && defined($o_ctg->GetScaffoldsIds()) && $o_ctg->GetScaffoldsIds() ne '' );
	}
	# il peut y a voir que des undef
	my $o_ctg = New Contig(name=>$ctg,strand=>$strand,len=>$len,contigregistry=>$self);
	push(@$ra_contigs,$o_ctg);
	return $o_ctg;
}

# cree et retourne une nouvelle instance du contig
sub GetNewContigInstance
{
my($self,$ctg,$strand,$len)=@_;

	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	my $o_ctg = New Contig(name=>$ctg,strand=>$strand,len=>$len,contigregistry=>$self);
	push(@$ra_contigs,$o_ctg);
	return $o_ctg;
}


sub GetAnchors
{
my($self)=@_;

	my %h_lst=();
	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	return undef if ( ! scalar(@$ra_contigs) );
	foreach my $o_ctg (@$ra_contigs)
	{
		next if ( ! defined($o_ctg) );
		my $anchors=$o_ctg->GetAnchors();
		next if ( !defined($anchors) || $anchors eq '' );
		$h_lst{$anchors}=1;
	}
	return join(' ',sort keys(%h_lst));
}


sub GetAnchor
{
my($self,$str)=@_;

	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	return undef if ( ! scalar(@$ra_contigs) );
	foreach my $o_ctg (@$ra_contigs)
	{
		next if ( ! defined($o_ctg) );
		my $anchor = $o_ctg->GetAnchor($str);
		return $anchor if ( defined($str) && defined($anchor) );
	}
	return undef;
}



sub GetNumberOfAnchors
{
my($self)=@_;

	my $anchors = $self->GetAnchors();
	return 0 if ( ! defined($anchors) );
	return scalar(split(' ',$anchors));
}


sub GetScaffoldsIds
{
my($self)=@_;

	my %h_lst = ();
	my $str = '';
	my $ra_contigs = $self->_FastOneGet('_ra_contigs');
	return undef if ( ! scalar(@$ra_contigs) );
	foreach my $o_ctg (@$ra_contigs)
	{
		next if ( ! defined($o_ctg) ); # des contigs peuvent etre undef
		my $scaffolds=$o_ctg->GetScaffoldsIds();
		next if ( !defined($scaffolds) || $scaffolds eq '');
		$h_lst{$scaffolds}=1;
	}
	return join(' ',sort keys(%h_lst));
}

sub GetNumberOfScaffolds
{
my($self)=@_;

	my $scaffolds = $self->GetScaffoldsIds();
	return 0 if ( ! defined($scaffolds) );
	return scalar(split(' ',$scaffolds));
}



sub AssertParams
{
my ($self,$rh_param,$ra_expected) = @_;
	
	foreach my $param (@{$ra_expected})
	{
		if ( !exists($$rh_param{$param}) )
		{
			print Dumper(%$rh_param);
			Carp::confess("Fatal error in constructor: $param must be provided\n") 
		}
	}
}


sub Dump
{
my($self)=@_;

	print Dumper($self);
}


sub _Print
{
my($self,$msg)=@_;

	print STDERR "$msg";
}



1;
