#
# $Id: M8.pm 440 2014-10-17 14:13:47Z gouzy $
#

package HSPSet::M8;

use HSPSet;
use base qw( HSPSet );

use General;
use GeneralBioinfo;
use strict;
use warnings;
use HSPSet::ArrayRef;
use Devel::Size qw(size total_size);
use Data::Dumper;

use constant MAX_PER_SLICE => 5000000;

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 440 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	my @a_expected = ('query_col');
	$self->AssertParams(\%h_param,\@a_expected);

	$self->SUPER::_Init(%h_param);
}


sub SpeedupPE
{
my ($self) = @_;

	my $ontheflypattern = "Best|LinkByPE|OneSpecificEnd";
	my $onthefly = &FALSE;
	my $allfilters = $self->_Get('_filters');
	my $remaining_filters = $allfilters;
	if ( defined($allfilters) && $allfilters =~ /$ontheflypattern/ ) # je ne mets pas NonRedundantPE car on peut l'utiliser pour autre chose
	{
		if ( $self->FastFilteringPE($allfilters) )
		{
			$onthefly=&TRUE;
			$remaining_filters=~s/($ontheflypattern)//g;
			# je ne le fais pas car le filtre rapide ne verifie pas les score des hsp $self->_Set('_filters',$remaining_filters);
			$self->_Print("FastFilteringPE done: when needed filters $ontheflypattern have been performed in fast mode (more accurate filtering will be applied later)\n");
		}
	}
#	$self->_Print("Filters to be applied: $allfilters; post loading remaining filters:$remaining_filters\n") if ( defined($allfilters));
}

=head2 procedure Load

 Title        : Load
 Usage        : $o_hspds->Load;

=cut

sub Load 
{
my ($self,%h_param) = @_;

	$self->SpeedupPE() if (! defined($h_param{nofast}) );

	my %h_min = ();
	my %h_max = ();

	$self->InitMinMax(\%h_min,\%h_max,'pci','len','gap','mm');

	my %h_hsp=();
	my $filein = $self->_FastOneGet('_infile');
	$filein =~ s/\s//g;
	my @a_filein = split(',',$filein);	# infile=file1 ou infile=file1,file2
	$self->_Print("\n======================================\nLoading $filein starting\n");
	my $query_col = $self->_FastOneGet('_query_col');
	my $bestperquery  =  ( defined($h_param{onlybestperquery}) && $h_param{onlybestperquery} ) ? &TRUE: &FALSE;
	my $chain =          ( defined($h_param{chain} ) && $h_param{chain} ) ? &TRUE: &FALSE;
	my $splicedtarget = ( defined($h_param{spliced_target}   ) && $h_param{spliced_target} ) ? &TRUE: &FALSE;
	my ($cpt_in,$cpt_out)=(0,0);
	foreach my $fin (@a_filein)
	{
		my $f_inf = &GetStreamIn($fin);
		while (my $l = <$f_inf>)
		{	
			chomp($l);
			&ParseM8Line(\$l,$query_col,\%h_hsp);
			$cpt_in++;
			my $skip = 0;
			foreach my $criterium ('pci','len')
			{
				if ( $h_hsp{$criterium} < $h_min{$criterium} || $h_hsp{$criterium} > $h_max{$criterium} )
				{
					$skip = 1;
					last;
				}
			}
			next if ( $skip );
			my $val = $self->GetEntry(\%h_hsp);
			if ( $bestperquery )
			{
				my $ret = $self->AddIfBest("query=$h_hsp{query} target=$h_hsp{target}",$val,$h_hsp{len}*$h_hsp{pci});
				$cpt_out++ if ( $ret != 2 );
			}
			else
			{
				if ( $chain )
				{
					# if current hit is the best score, it replaces the previous one
					# otherwise, when it is compatible with the previous hit, it is "merged" with it (as its pci) 
					my $ret = $self->MergeWithOrReplaceBest("query=$h_hsp{query} target=$h_hsp{target}",$val,$h_hsp{len}*$h_hsp{pci},$splicedtarget);
					$cpt_out++ if ( $ret == 0 );
				}
				else
				{
					$self->Add("query=$h_hsp{query} target=$h_hsp{target}",$val);
					$cpt_out++;
				}
			}
		}
		$f_inf->close;
	}
	$self->_Set('count',$cpt_out);
	$self->_Print("HSP in: $cpt_in\nHSP out: $cpt_out\n");
	$self->_Print("Loading $filein done\n");
	return $h_param{filters};
}


# c'est un filtre pas tres intelligent
# je garde pour des analyses ulterieures tous les inserts qui matchent plusieurs query
# desormais Best OneSpecificEnd et LinkByPE sont fait ici
# 

sub FastFilteringPE
{
my($self,$filters) = @_;

	my $filein             = $self->_FastOneGet('_infile');
	#if ( $filein =~ /\.PE-prefiltered.gz$/ ) # je ne le fais pas si j'ai en entree un fichier deja pretraite
	#{
		# surtout ne pas passer, je fais unlink a la fin
	#	return;
	#}
	my $max_per_slice = &MAX_PER_SLICE;
	my $cpt_slice = 1;
	my $filein_sorted      = $self->_FastOneGet('_infile').".sorted_by_target.gz";
	my $filein_prefiltered = $self->_FastOneGet('_infile').".PE-prefiltered.gz".sprintf(".part%04d.gz",$cpt_slice++);
	my $listoffiles="$filein_prefiltered";
	my $colsq              = $self->_FastOneGet('_query_col');
	my $pattern            = $self->_FastOneGet('_target')->_FastOneGet('_pattern');
	my $paired             = $self->_FastOneGet('_target')->_FastOneGet('_paired');
	if ( !defined($paired) || $paired !~ /^(no|0)$/ )
	{
		if (  !defined($pattern) || $pattern eq '' )
		{
			Carp::croak("FastFilteringPE skipped: can't find a pattern for the target database\n");
		}
	}
	else
	{
		return &FALSE;
	}

	my $coltg =  ($colsq ==  2 ) ? 1 : 2;
	my $tmpdir = "FastFilteringPE.$$.".time;
	mkdir $tmpdir;
	#chdir $tmpdir;
	my %h_opt = ('tmpdir' => $tmpdir);
	&UnixSort($filein,$filein_sorted,$coltg,\%h_opt);
	my $f_filt = &GetStreamOut($filein_prefiltered);
	my $f_in   = &GetStreamIn($filein_sorted);
	my $filtBest = ( $filters =~ /Best/ )  ? &TRUE : &FALSE;
	my $filtOneSpe = ( $filters =~ /OneSpecificEnd/ )  ? &TRUE : &FALSE;
	my %h_res=();
	my $prev = '';
	my %h_query=();
	my %h_ext=();
	my @a_hsps=();
	# le filtre LinkByPE est systematique
	my %h_hsp =();
	my $count_slice=0;
	while(my $l=<$f_in>)
	{
		&ParseM8Line(\$l,$colsq,\%h_res);
		my ($insert,$ext) = $h_res{target} =~ /$pattern/;
		if ( $insert ne $prev )
		{
			# on doit tomber sur plusieurs query et les deux ext doivent etre mappees, _Filter sera plus fin
			if ( scalar(keys(%h_query)) > 1 && scalar(keys(%h_ext)) > 1 )
			{
				$count_slice += &_Filter(\@a_hsps,$colsq,$f_filt,$filtBest,$filtOneSpe);
				if ($count_slice>=$max_per_slice)
				{
					$f_filt->close;
					$filein_prefiltered = $self->_FastOneGet('_infile').".PE-prefiltered.gz".sprintf(".part%04d.gz",$cpt_slice++);
					$listoffiles.=",$filein_prefiltered";
					$f_filt = &GetStreamOut($filein_prefiltered);
					$count_slice=0;
				}
			}
			%h_query=();
			%h_ext=();
			@a_hsps=();
		}
		$h_query{$h_res{query}}++;
		$h_ext{$ext}++;
		push(@a_hsps,$l);
		$prev=$insert;
	}
	if ( scalar(keys(%h_query)) > 1 && scalar(keys(%h_ext)) > 1 )
	{
		$count_slice += &_Filter(\@a_hsps,$colsq,$f_filt,$filtBest,$filtOneSpe);
	}
	$f_in->close;
	$f_filt->close;
	unlink $filein_sorted;
	#chdir "..";
	system("rm -rf $tmpdir") if ( -d $tmpdir );
	#filtrage
	$self->_Set('_infile',$listoffiles);
	return &TRUE;
}

sub _Filter
{
my($ra_hsps,$colsq,$f_filt,$filtBest,$filtOneSpe) = @_;

	my %h_hsp =();
	my %h_best=();
	my %h_score=();
	my %h_target2query=();
	my %h_insert=();
	my $cpt_line=0;
	for(my $i=0;$i<scalar(@$ra_hsps);$i++)
	{
		&ParseM8Line(\$$ra_hsps[$i],$colsq,\%h_hsp);
		my $score = sprintf("%.02f",$h_hsp{score}*$h_hsp{pci});
		if ( $filtBest && ( ! defined($h_best{$h_hsp{target}}) || $score > $h_best{$h_hsp{target}} ) )
		{
			$h_best{$h_hsp{target}} = $score;
			$h_score{$h_hsp{target}} = '';	
			$h_target2query{$h_hsp{target}} = '';	
		}
		# je repasse si je suis le meilleur ou dans tous les cas si je ne cherche pas le Best
		if ( ! $filtBest || $score == $h_best{$h_hsp{target}} )
		{
			$h_score{$h_hsp{target}} .= "$i ";
			$h_target2query{$h_hsp{target}} .= "$h_hsp{query} ";
		}
	}		
	# ici je pars de l'hypothese que le meilleur des 2 extremites sur un meme contig suffit pour l'eliminer
	my $same_ctg = &FALSE;
	my $min_nb_query = &MAXINT;
	foreach my $ext (keys(%h_target2query)) 
	{
		my @a_queries = split(' ',$h_target2query{$ext});
		my $nb_queries = scalar(@a_queries);
		$min_nb_query= &Min($min_nb_query,$nb_queries);
		for(my $i=0;$i<$nb_queries;$i++)
		{
			my $vv = $a_queries[$i];
			if ( ! defined($h_insert{$vv}) )
			{
				$h_insert{$vv} = "$ext";
			}
			elsif ( $h_insert{$vv} ne $ext )
			{
				$same_ctg=&TRUE;
			}
			last if ( $same_ctg );
		}
		last if ( $same_ctg );
	}

	return 0 if ( $min_nb_query > 1 && $filtOneSpe );

	# il faut verifier apres le filtre qu'il reste plusieurs queries
	if ( ! $same_ctg && scalar(keys(%h_insert)) > 1 )
	{
		my $best = join(' ',values(%h_score));	# je recupere les indices pour les deux extremites
		foreach my $ind (split(' ',$best))
		{
			print $f_filt $$ra_hsps[$ind];
			$cpt_line++;
		}
	}
	return $cpt_line;
}


1;
