#
# $Id$
#

package KeyGene;

use strict;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use File::Basename;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use Scaffold;
use threads;
use threads::shared;

use constant DEF_MINCTGLEN => 5000;
use constant DEF_MINNBTAGS => 3;

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 187 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	$self->SUPER::_Init();

	$self->_Set('_nb_cpus', 1);
	$self->_Set('_debug', 1);

	foreach my $param (keys(%h_param))
	{
		$self -> _Set ("_$param", $h_param{$param});
	}

	$self->_Set('_rh_graph', undef);
	$self->_Set('_rh_bacs', {});
}


=head2 procedure Load

 Title        : Load
 Usage        : $o_hspds -> Load;

=cut

sub Load 
{
my ($self,%h_param) = @_;

	my $rh_bacs = $self->_FastOneGet('_rh_bacs');
	my @a_expected=('file');
	&AssertParams(\%h_param,\@a_expected);
	my $f_in = &GetStreamIn($h_param{file});
	while(my $l = <$f_in>)
	{
		chomp($l);
		my ($bacid,$nbtags,$nbctgs,$lstctgs) = split('\t',$l,4);
		$$rh_bacs{$bacid}{nbtags}=$nbtags;
		$$rh_bacs{$bacid}{lstctgs}=$lstctgs;
		$$rh_bacs{$bacid}{nbctgs}=$nbctgs;
	}
	$f_in->close;

	my $nb = scalar(keys(%$rh_bacs));
	if ( $self->_FastOneGet('_debug') )
	{
		print STDERR "KeyGene: $nb bacs loaded\n";
		print STDERR "KeyGene: number of threads =".$self->_FastOneGet('_nb_cpus')."\n";
	}
}

# je garde l'info bac -> ctgs et ctg -> bacs
sub InitGraph
{
my($self)=@_;

	$self->_Print("KeyGene: InitGraph started\n");

	my %h_graph=();
	my $rh_bacs = $self->_FastOneGet('_rh_bacs');
	foreach my $bac (keys(%$rh_bacs))
	{
		my @a_lst = split(' ',$$rh_bacs{$bac}{lstctgs});
		for(my $i=0;$i<scalar(@a_lst);$i++)
		{
			my ($idctg,$tagcount) = split("##",$a_lst[$i]);
			$h_graph{$idctg}{$bac}=$tagcount;
		}
	}
	$self->_Set('_rh_graph',\%h_graph);
	$self->_Print("KeyGene: InitGraph done\n");
	return \%h_graph;
}

sub _Print
{
my($self,$msg)=@_;

	if ( $self->_FastOneGet('_debug') )
	{
		print STDERR "$msg";
	}
}

# retourne le nombre de bacs qui font le lien entre deux scaffolds

sub GetBacLinksBetweenScaffolds
{
my($self,$o_src,$o_tgt,%h_param)=@_;
	
	#print STDERR "\n";
	#print STDERR $o_src->Get()." LEN=".$o_src->GetLength()."\n";
	#print STDERR $o_tgt->Get()." LEN=".$o_tgt->GetLength()."\n";
	my $minctglen = $h_param{minctglen} || &DEF_MINCTGLEN;
	my $minnbtags = $h_param{minnbtags} || &DEF_MINNBTAGS;
	my $rh_graph = $self->_FastOneGet('_rh_graph');
	my %h_ctg_src=();
	&_ParsePath($o_src->Get(),\%h_ctg_src);
	my %h_ctg_tgt=();
	&_ParsePath($o_tgt->Get(),\%h_ctg_tgt);
	my $nb_bacs = 0;
	foreach my $ctg (keys(%h_ctg_src))	# il faudrait celui avec le moins de liens
	{
		# suppression des contigs qui sont dans les deux listes
		if ( exists($h_ctg_tgt{$ctg}) )
		{
			#print STDERR "Removing duplicated $ctg\n";
			delete($h_ctg_src{$ctg});
			delete($h_ctg_tgt{$ctg});
			next;
		}
	}
	my %h_linkedbacs_src=();
	my %h_linkedbacs_tgt=();
	$self->GetLinkedBacs(\%h_ctg_src,\%h_linkedbacs_src);
	$self->GetLinkedBacs(\%h_ctg_tgt,\%h_linkedbacs_tgt);

	my $nb=0;
	foreach my $bac (keys(%h_linkedbacs_src))
	{
		if ( exists($h_linkedbacs_tgt{$bac}) )
		{
			#print STDERR "BAC $bac src(name=".$o_src->GetName().",len=$h_linkedbacs_src{$bac}{LEN},#tags=$h_linkedbacs_src{$bac}{NBTAGS}) tgt(name=".$o_tgt->GetName().",len=$h_linkedbacs_tgt{$bac}{LEN},#tags=$h_linkedbacs_tgt{$bac}{NBTAGS})\n";
			if ( $h_linkedbacs_src{$bac}{LEN}    >= $minctglen    && 
			     $h_linkedbacs_tgt{$bac}{LEN}    >= $minctglen    &&
				 $h_linkedbacs_src{$bac}{NBTAGS} >= $minnbtags &&
				 $h_linkedbacs_tgt{$bac}{NBTAGS} >= $minnbtags
			   )
			{
				$nb++;
			}
		}
	}
	#if ( $nb )
	#{
	#	print STDERR "$nb bridging BAC for ".$o_src->GetName()." and ".$o_tgt->GetName()."\n";
	#}
	return $nb;
}


sub GetLinkedBacs
{
my($self,$rh_ctgs,$rh_linkbacs)=@_;

	my $rh_graph = $self->_FastOneGet('_rh_graph');
	foreach my $ctg (keys(%$rh_ctgs))
	{
		foreach my $bac (keys(%{$$rh_graph{$ctg}}))
		{
			$$rh_linkbacs{$bac}{LEN}+=$$rh_ctgs{$ctg};
			$$rh_linkbacs{$bac}{NBTAGS}+=$$rh_graph{$ctg}{$bac};
		}
	}
	#print STDERR "==========================\n";
	#print STDERR Dumper($rh_linkbacs);
}



sub _ParsePath
{
my($str,$rh_ctgs)=@_;

	while($str=~/#(\S+)([pm])(\d+)/g)
	{
		$$rh_ctgs{$1}=$3;
	}
}


sub AssertParams
{
my ($self,$rh_param,$ra_expected) = @_;

        foreach my $param (@{$ra_expected})
        {
                if ( !defined($$rh_param{$param}) )
                {
                        print STDERR Dumper(%$rh_param);
                        Carp::croak("Fatal error in constructor: $param must be provided\n") 
                }
        }
}


sub _Print
{
my($self,$msg)=@_;

	print STDERR "$msg" if ( defined($$self{'_debug'}) );
}


1;
