#
# $Id: LinkSet.pm 519 2016-08-26 05:40:19Z gouzy $
#

package LinkSet;

use strict;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use Statistics::Descriptive;

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

my %H_SCAFF2EXT    = ( "p p" => "3 5", "p m" => "3 3", "m p" => "5 5", "m m" => "5 3" );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 519 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	my @a_expected = ();
	$self->AssertParams(\%h_param,\@a_expected);

	$self->SUPER::_Init();

	foreach my $param (keys(%h_param))
	{
		$self -> _Set ("_$param", $h_param{$param});
	}

	$self -> _Set ('_rh_linkds', {});
	$self -> _Set ('_rh_stats', {});
	$self -> _Set ('_nb_links', 0);
}

sub AssertParams
{
my ($self,$rh_param,$ra_expected) = @_;
	
	foreach my $param (@{$ra_expected})
	{
		if ( !defined($$rh_param{$param}) )
		{
			print Dumper(%$rh_param);
			Carp::croak("Fatal error in constructor: $param must be provided\n") 
		}
	}
}

sub IsEmpty
{
my($self)=@_;

	return 1 if ( ! $self -> _FastOneGet ('_nb_links') );
	return 0;
}

sub Reset
{
my($self)=@_;

	$self -> _Set ('_rh_linkds', {});
	$self -> _Set ('_nb_links',0);
}

sub BuildKey
{
my($self,$id1,$ext1,$id2,$ext2)=@_;

	if ($id1 lt $id2)
	{
		return "src=$id1 tgt=$id2";
	}
	else
	{
		return "src=$id2 tgt=$id1";
	}
}

sub Del
{
my($self,$key)=@_;

	if ( defined($self->{'_rh_linkds'}->{$key}) )
	{
		delete ($self->{'_rh_linkds'}->{$key});
		$self->{'_nb_links'}--;
	}
}

sub Del2
{
my($self,$id1,$id2)=@_;

	my ($src,$tgt)=sort($id1,$id2);
	my $key = "src=$src tgt=$tgt";
	if ( defined($self->{'_rh_linkds'}->{$key}) )
	{
		delete ($self->{'_rh_linkds'}->{$key});
		$self->{'_nb_links'}--;
	}
}

sub Add
{
my($self,$key,$val) = @_;

	if ( ! defined($self->{'_rh_linkds'}->{$key}) )
	{
		$self->{'_rh_linkds'}->{$key} = $val;
		$self->{'_nb_links'}++;
	}
	else
	{
		$self->{'_rh_linkds'}->{$key} .= "##$val";
	}
}


sub AddKV
{
my($self,$key,$rh_val) = @_;

	my $str = '';
	foreach my $k (sort keys(%$rh_val))
	{
		$str .= "$k=$$rh_val{$k} ";
	}
	chop($str);
	if ( ! defined($self->{'_rh_linkds'}->{$key}) )
	{
		$self->{'_rh_linkds'}->{$key} = $str;
		$self->{'_nb_links'}++;
	}
	else
	{
		$self->{'_rh_linkds'}->{$key} .= "##$str";
	}
}


sub Set
{
my($self,$key,$val) = @_;

	$self->{'_nb_links'}++ if ( ! defined($self->{'_rh_linkds'}->{$key}) );
	$self->{'_rh_linkds'}->{$key} = $val;
}

sub SetKV
{
my($self,$key,$rh_val)=@_;

	$self->{'_nb_links'}++ if ( ! defined($self->{'_rh_linkds'}->{$key}) );
	my $str = '';
	foreach my $k (sort keys(%$rh_val))
	{
		$str .= "$k=$$rh_val{$k} ";
	}
	chop($str);
	$self->{'_rh_linkds'}->{$key}=$str;
}

sub SetKV2
{
my($self,$id1,$id2,$rh_val)=@_;

	my ($src,$tgt)=sort($id1,$id2);
	my $key = "src=$src tgt=$tgt";
	$self->{'_nb_links'}++ if ( ! defined($self->{'_rh_linkds'}->{$key}) );
	$self->SetKV($key,$rh_val);
}

# returns, key-value hash
# ATTENTION: je regarde dans tous les liens charges pas dans les filtres

sub GetKV
{
my($self,$key,$rh_val)=@_;

	if (defined($self->{'_rh_linkds'}->{$key}))
	{
		%{$rh_val} = split(/[\s=]+/,$self->{'_rh_linkds'}->{$key});
		# implicit return
	}
	else
	{
		return &FALSE;
	}
}

# ATTENTION: je regarde dans tous les liens charges pas dans les filtres

sub GetKV2
{
my($self,$id1,$id2,$rh_val)=@_;

	my ($src,$tgt)=sort($id1,$id2);
	my $key = "src=$src tgt=$tgt";
	return $self->GetKV($key,$rh_val);
}

# ATTENTION: je regarde dans tous les liens charges pas dans les filtres

sub GetKV2Oriented
{
my($self,$id1,$ext1,$id2,$ext2,$rh_val)=@_;

	my $key = '';
	my ($srcext,$tgtext);
	if ( $id1 lt $id2 )
	{
		$key = "src=$id1 tgt=$id2";
		($srcext,$tgtext) = ($ext1,$ext2);
	}
	else
	{
		$key = "src=$id2 tgt=$id1";
		($srcext,$tgtext) = ($ext2,$ext1);
	}
	return &FALSE if ( ! $self->GetKV($key,$rh_val) );

	if ( $$rh_val{srcext} != $srcext || $$rh_val{tgtext} != $tgtext )
	{
#		print STDERR "FALSE: $key $$rh_val{srcext} != $srcext || $$rh_val{tgtext} != $tgtext \n";
		return &FALSE;
	}
	return %{$rh_val};
}

# ATTENTION: je regarde dans tous les liens charges pas dans les filtres

sub GetKV2OrientedContigs
{
my($self,$id1,$strand1,$id2,$strand2,$rh_val)=@_;

	my ($ext1,$ext2) = split(' ',$H_SCAFF2EXT{"$strand1 $strand2"});

	Carp::croak("GetKV2OrientedContigs, undefined ends: Input($id1,$strand1,$id2,$strand2) Expected(id1,[pm],id2,[pm])\n") if ( !defined($ext1) || ! defined($ext2) );

	#print STDERR "$id1,$strand1,$id2,$strand2 => ($ext1,$ext2)\n";

	return $self->GetKV2Oriented($id1,$ext1,$id2,$ext2,$rh_val);
}

# idem mais tri numerique (pour les idscaff)

sub GetKV2N
{
my($self,$id1,$id2,$rh_val)=@_;

	my ($src,$tgt)=sort { $a<=>$b }($id1,$id2);
	my $key = "src=$src tgt=$tgt";
	return $self->GetKV($key,$rh_val);
}

sub GetLinks
{
my($self) = @_;

	return keys(%{$self->{_rh_linkds}});
}

sub GetLinksRef
{
my($self) = @_;

	my @a_links = keys(%{$self->{_rh_linkds}});
	return \@a_links;
}


sub Get
{
my($self,$key) = @_;

	return $self->{'_rh_linkds'}->{$key};
}

sub Get2
{
my($self,$id1,$id2)=@_;

	my ($src,$tgt)=sort { $a cmp $b }($id1,$id2);
	return $self->Get("src=$src tgt=$tgt");
}

# pour identification des seuils de counts par insert_size
# il faut le lancer sur des liens les plus propres possibles

sub InitStatistics
{
my ($self,%h_param) = @_;

	my $tag = $h_param{tag} || '';
	my @a_links = $self->GetLinks();
	my %h_insert=();
	my %h_val=();
	for(my $i=0;$i<=$#a_links;$i++)
	{
		if ( $self->GetKV($a_links[$i],\%h_val) )
		{
			if (! defined($h_insert{$h_val{insert_size}}) )
			{
				my @a_ = ();
				$h_insert{$h_val{insert_size}}=\@a_;
			}
			push(@{$h_insert{$h_val{insert_size}}},$h_val{count});
		}
	}

	foreach my $insert (keys(%h_insert))
	{
		my $stat = Statistics::Descriptive::Full->new();
		$stat->add_data(@{$h_insert{$insert}});
		print STDERR "$tag $insert\n";
		print STDERR "\tmean=".$stat->mean()."\n";
		print STDERR "\tmin=".$stat->min()."\n";
		print STDERR "\tmax=".$stat->max()."\n";
       	print STDERR "\tvar=".$stat->variance()."\n";
       	print STDERR "\tmedian=".$stat->median()."\n";
       	print STDERR "\tmode=".$stat->mode()."\n";
       	print STDERR "\tsd=".$stat->standard_deviation()."\n";
		if ( 0 )
		{
		my $cutoff = 1;
		my $prev   = 0;
		my $x = 0;
		my $index = 0;
		my @a_count = ();
		while ( $cutoff < 100 )
		{
			($x, $index) = $stat->percentile($cutoff);
			push(@a_count,$x);
			$cutoff++;
		}
		print STDERR join(' ',@a_count)."\n";
		}
	}
}

# pour generation rapport
sub ComputeStats
{
my ($self,$rh_seq,%h_param) = @_;

	my @a_counts = ('weak','hardlink','insert_size','used','srcext','tgtext','ambiguous_orient');
	my $outprefix = ( defined($h_param{statfile}) ) ? $h_param{statfile} : "LynxStats.$$.".time;
	my @a_links = ();
	$self->Filter(\@a_links,%h_param);
	my $f_out = &GetStreamOut("$outprefix.log");
	my %h_stats = ();
	my %h_ctg   = ();
	my @a_ratiospan = ();
	my @a_span = ();
	my @a_count = ();
	my @a_count_spansupinsert = ();
	my @a_ovlpci = ();
	my @a_ovllen = ();
	my @a_gaplen_max = ();
	my @a_gaplen_min = ();
	my @a_nlib = ();
	my $spansupinsert=0;
	my $nb_links = 0;
	my $max_insert_size=0;
	for(my $i=0;$i<=$#a_links;$i++)
	{
		my $link=$a_links[$i];
		my %h_key=split(/[=\s]/,$link);
		my $lk = $self->Get($link);
		$lk=~s/##$//;
		my @a_lk = split(/#+/,$lk);
		$h_stats{nb_orientation}{scalar(@a_lk)}++;
		foreach my $lk (@a_lk)
		{
			my %h_val=split(/[=\s]+/,$lk);
			$nb_links++;
			foreach my $tag (@a_counts)
			{
				next if ( ! defined($h_val{$tag}) );
				$h_stats{$tag}{$h_val{$tag}}++;
			}
			$h_ctg{$h_key{src}}{len}=$h_val{srclen};
			$h_ctg{$h_key{tgt}}{len}=$h_val{tgtlen};
			$h_ctg{$h_key{src}}{$h_val{srcext}}++;
			$h_ctg{$h_key{src}}{"len$h_val{srcext}"}+=$h_val{tgtlen};
			$h_ctg{$h_key{tgt}}{"len$h_val{tgtext}"}+=$h_val{srclen};
			$h_ctg{$h_key{tgt}}{$h_val{tgtext}}++;
			my $ratiospan = ( $h_val{srcspan} > $h_val{tgtspan} ) ? ($h_val{srcspan}/$h_val{tgtspan}) : ($h_val{tgtspan}/$h_val{srcspan});
			if( ($h_val{srcspan}+$h_val{tgtspan}) > $h_val{insert_size}*1.5 )
			{
				$spansupinsert++;
				push(@a_count_spansupinsert,$h_val{count});
			}
			push(@a_count,$h_val{count});
			push(@a_ratiospan,$ratiospan);
			push(@a_span,$h_val{srcspan},$h_val{tgtspan});
			push(@a_gaplen_max,$h_val{gaplen_max});
			push(@a_gaplen_min,$h_val{gaplen_min});
			push(@a_ovllen,$h_val{ovllen});
			push(@a_ovlpci,$h_val{ovlpci});
			push(@a_nlib,$h_val{nlib});
			$max_insert_size=&Max($h_val{insert_size},$max_insert_size);
		}
	}
	my @a_class = (1,2,3,4,5,6,7,8,9,10);
	&Histo(\@a_ratiospan,\@a_class,&GetStreamOut("$outprefix.ratiospan.tsv"));
	&Histo(\@a_nlib,\@a_class,&GetStreamOut("$outprefix.nlib.tsv"));
	@a_class = (100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 2000, 3000, 4000, 5000, 10000, 20000);
	&Histo(\@a_span,\@a_class,&GetStreamOut("$outprefix.span.tsv"));
	&Histo(\@a_gaplen_max,\@a_class,&GetStreamOut("$outprefix.gaplen_max.tsv"));
	&Histo(\@a_gaplen_min,\@a_class,&GetStreamOut("$outprefix.gaplen_min.tsv"));
	@a_class = (1,2,3,4,5,6,7,8,9,10,20,30,40,50,60,70,80,90,100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 2000, 3000, 4000, 5000, 10000, 20000);
	&Histo(\@a_count,\@a_class,&GetStreamOut("$outprefix.count.tsv"));
	&Histo(\@a_count_spansupinsert,\@a_class,&GetStreamOut("$outprefix.count_spansupinsert.tsv"));
	@a_class = (10,20,30,40,50,60,70,80,90,101);
	&Histo(\@a_ovllen,\@a_class,&GetStreamOut("$outprefix.ovllen.tsv"));
	@a_class = (80,90,91,92,93,94,95,96,97,98,99,100);
	&Histo(\@a_ovlpci,\@a_class,&GetStreamOut("$outprefix.ovlpci.tsv"));
	@a_class = (1,2,3,4,5,6,7,8,9,10,20,30,40,50,60,70,80,90,100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 2000, 3000, 4000, 5000, 10000, 20000);
	my @a_ctg = sort keys(%{$rh_seq});
	my $f_unlinked=&GetStreamOut("$outprefix.unlinked.ctg.fa");
	my $f_linkedthreeonly=&GetStreamOut("$outprefix.linkedthreeonly.ctg.fa");
	my $f_linkedfiveonly=&GetStreamOut("$outprefix.linkedfiveonly.ctg.fa");
	my $nb_unlinked=0;
	my $len_unlinked=0;
	my @a_five_lk = ();
	my @a_three_lk = ();
	$max_insert_size *= 1.1;
	foreach my $ctg (@a_ctg)
	{	
		if ( defined($h_ctg{$ctg}) )
		{
			if ( defined($h_ctg{$ctg}{5}) && defined($h_ctg{$ctg}{3}) )
			{
				$h_stats{connected}{both}++;
				push(@a_five_lk,$h_ctg{$ctg}{5});
				push(@a_three_lk,$h_ctg{$ctg}{3});
				#print "$ctg  both $h_ctg{$ctg}{5} $h_ctg{$ctg}{3}\n" if ( $h_ctg{$ctg}{5} > 10 || $h_ctg{$ctg}{3} > 10 );
				#print "$ctg  lenboth $h_ctg{$ctg}{len5} $h_ctg{$ctg}{len3}\n" if ( $h_ctg{$ctg}{len5} > $max_insert_size || $h_ctg{$ctg}{len3} > $max_insert_size );
			}
			elsif( defined($h_ctg{$ctg}{5}) )
			{
				$h_stats{connected}{fiveonly}++;
				print $f_linkedfiveonly ">$ctg len=$$rh_seq{$ctg}{len}\n".&FormatSeq(\$$rh_seq{$ctg}{sequence})."\n";
				push(@a_five_lk,$h_ctg{$ctg}{5});
				#print "$ctg five $h_ctg{$ctg}{5}\n"       if ( $h_ctg{$ctg}{5} > 100 );
				#print "$ctg lenfive $h_ctg{$ctg}{len5}\n" if ( $h_ctg{$ctg}{len5} > $max_insert_size );
			}
			elsif( defined($h_ctg{$ctg}{3}) )
			{
				$h_stats{connected}{threeonly}++;
				print $f_linkedthreeonly ">$ctg len=$$rh_seq{$ctg}{len}\n".&FormatSeq(\$$rh_seq{$ctg}{sequence})."\n";
				push(@a_three_lk,$h_ctg{$ctg}{3});
				#print "$ctg three $h_ctg{$ctg}{3}\n"       if ( $h_ctg{$ctg}{3} > 100 );
				#print "$ctg lenthree $h_ctg{$ctg}{len3}\n" if ( $h_ctg{$ctg}{len3} > $max_insert_size );
			}
		}
		else
		{
			print $f_unlinked ">$ctg len=$$rh_seq{$ctg}{len}\n".&FormatSeq(\$$rh_seq{$ctg}{sequence})."\n";
			$len_unlinked+=$$rh_seq{$ctg}{len};
			$nb_unlinked++;
		}
	}
	$f_unlinked->close;
	&Histo(\@a_five_lk,\@a_class,&GetStreamOut("$outprefix.links5.tsv"));
	&Histo(\@a_three_lk,\@a_class,&GetStreamOut("$outprefix.links3.tsv"));
	print $f_out "Nb contigs: ".scalar(@a_ctg)."\n";
	print $f_out "Nb unlinked contigs: $nb_unlinked\n";
	print $f_out "Size represented in unlinked contigs: $len_unlinked\n";
	print $f_out "Nb links: ".scalar(@a_links)." mult=$nb_links\n";
	print $f_out "Nb links span > insert_size: $spansupinsert\n";
	foreach my $tag (sort keys(%h_stats))
	{	
		foreach my $k (keys(%{$h_stats{$tag}}))
		{
			print $f_out "$tag\t$k\t$h_stats{$tag}{$k}\n";
		}
	}
	$self -> _Set ('_rh_stats', \%h_ctg);
}

# lorsque plusieurs sources ont parl, on garde celle d'insert_size minimum mais on augmente le compte avec les tailles d'inserts superieur
# il y a des cas ou le lien est donne dans plusieurs orientations, je prend la majoritaire
# 20120417: je rajoute l'option secondbest pour garder un orientation alternative
# 20130505: je rajoute le filtre sur le nlib > 1 lorsque la taille des contigs est petite

sub CleanUp
{
my ($self,%h_param) = @_;

	$h_param{limit} = &MAXINT if ( ! defined($h_param{limit}) );
	$h_param{ambiguous_percent_cutoff}=30 if ( ! defined($h_param{ambiguous_percent_cutoff}) ); # a link is declared ambigous if the second ranked orientation is higher than this limit
	my $rank = ( defined($h_param{rank}) ) ? $h_param{rank} : 1; # par defaut je garde l'orientation la plus frequente
	my $ambiguous_percent_cutoff = $h_param{ambiguous_percent_cutoff};
	print STDERR "rank=$rank\n" if ( $rank > 1 );
	my $rh_stats = $self->_FastOneGet('_rh_stats');
	$self->_Print("limit=$h_param{limit} (# of links for the second best orientation), ambiguous_percent_cutoff=$ambiguous_percent_cutoff\n");
	if ( ! defined($rh_stats) )
	{
		$self->_Print("CleanUp: WARNINGS: statistics of connection not computed, some unique conections can be lost during cleaning\n");
	}
	my @a_links = $self->GetLinks();
	my @a_sorted = ();
	foreach my $link (@a_links)
	{
		my $linkdetails = $self->Get($link);
		my @a_ = split(/##/,$linkdetails);
		# JG 20120824 je cherche a comprendre le LinkScaffolds next if ( $#a_ == 0 && $rank == 1);	
		# first pass, I look for the most probable orientation
		my %h_orient = ();
		my %h_orient_sources = ();
		my $max=0;
		foreach my $l (@a_)
		{
			my %h_val = split(/[=\s]+/,$l);
			my $key = "$h_val{srcext} $h_val{tgtext}";
			$h_orient{$key}+=$h_val{count};
			$h_orient_sources{$key}++;
			$max = $h_val{count} if ( $h_val{count} > $max);
		}
		@a_sorted = sort { $h_orient{$b} <=> $h_orient{$a} } keys(%h_orient);
		my %h_key = split(/[=\s]+/,$link);
		my $selected_orientation=$a_sorted[0];
		my $ambiguous_orientation=&FALSE;
		if ( $rank > 1 )
		{
			if ( ($rank-1) > $#a_sorted )
			{
				$self->Del($link);
				next;
			}
			$selected_orientation=$a_sorted[$rank-1];
		}
		else
		{
			$ambiguous_orientation = &TRUE if ( $h_orient{$a_sorted[1]} > ($h_orient{$a_sorted[0]} * $ambiguous_percent_cutoff / 100 ) );
			if ( 0 )	# avant le 20 mars 2013
			{
				#if ( $max < $limit && $h_orient_sources{$a_sorted[1]} >= $h_orient_sources{$selected_orientation} )
				if ( $h_orient{$a_sorted[1]} == $h_orient{$selected_orientation} )
				{
					#=> tenir compte du span, verifier qu'il n'y a pas un desequilibre pour eviter les repeats?
					#$self->_Print("\tambiguous orientation $link\n\t$linkdetails $h_orient_sources{$a_sorted[1]} $a_sorted[1] == $h_orient_sources{$selected_orientation} $selected_orientation\n");
					$self->Del($link);
					next;
				}
				elsif ( $h_orient{$a_sorted[1]} >= $h_param{limit} )
				{
					# prendre le majoritaire peut poser des problemes sur les cas limites 
					# mais bon, s'il n'y a pas d'autre connections pour le contig il faut le garder quand meme
					my $nodel = &FALSE;
					if ( defined($$rh_stats{$h_key{src}}) && defined($$rh_stats{$h_key{tgt}}) )
					{
						my($srcext,$tgtext) = split(' ',$selected_orientation);
						if ( $$rh_stats{$h_key{src}}{$srcext} == 1 ||  $$rh_stats{$h_key{tgt}}{$tgtext} == 1 )
						{
							$nodel=&TRUE;
							$self->_Print("CleanUp: Link protection >$link< $h_key{src} $srcext has $$rh_stats{$h_key{src}}{$srcext} connections, and $h_key{tgt} $tgtext has $$rh_stats{$h_key{tgt}}{$tgtext}\n\t$linkdetails\n");
						}
					}
					if (! $nodel)
					{
						$self->Del($link);
						next;
					}
				}
			}
		}
#		print STDERR "$link\n\t$linkdetails\n" if ( scalar(@a_sorted) > 1 );
		
		my $mingaplen=&MAXINT;
		my $maxgaplen=&MAXINT; # pour les deux gaplen je cherche le min car je vais prendre de preference l'insert le plus court
		my %h_newhit = ();
		$h_newhit{count}=$h_orient{$selected_orientation};
		my %h_nlib=();

		$h_newhit{ambiguous_orient}=$ambiguous_orientation;
		foreach my $l (@a_)
		{
			my %h_val = split(/[=\s]+/,$l);
			my $key = "$h_val{srcext} $h_val{tgtext}";
			next if ( $key ne $selected_orientation );
			foreach my $tag ('srcext','tgtext','srclen','tgtlen','ovllen','ovlpci')
			{
				next if ( ! defined($h_val{$tag}) );
				$h_newhit{$tag} = $h_val{$tag};
			}
			foreach my $tag ('insert_size','srcstart','tgtstart','gaplen_min','gaplen_max','weak','used')
			{
				next if ( ! defined($h_val{$tag}) );
				$h_newhit{$tag} = $h_val{$tag} if ( ! defined($h_newhit{$tag}) || $h_newhit{$tag} > $h_val{$tag} );
				$h_nlib{$h_val{$tag}}=1 if ( $tag eq 'insert_size' );
			}
			foreach my $tag ('srcend','tgtend','hardlink','nlib','ambiguous_orient')
			{
				next if ( ! defined($h_val{$tag}) );
				$h_newhit{$tag} = $h_val{$tag} if ( ! defined($h_newhit{$tag}) || $h_newhit{$tag} < $h_val{$tag} );
			}
			foreach my $tag ('srclst','tgtlst')
			{
				next if ( ! defined($h_val{$tag}) );
				$h_newhit{$tag} .= $h_val{$tag} if ( ! defined($h_newhit{$tag}) || $h_newhit{$tag} !~ /$h_val{$tag}/ );
			}
		}
		# ca peut ne pas etre defini lorsque ce sont des scaffolds
		if ( defined($h_newhit{srcend}) && $h_newhit{srcstart} )
		{
			$h_newhit{srcspan} = $h_newhit{srcend}-$h_newhit{srcstart}+1;
			$h_newhit{tgtspan} = $h_newhit{tgtend}-$h_newhit{tgtstart}+1;
		}
		foreach my $tag ('srclst','tgtlst')
		{
			next if ( ! defined($h_newhit{$tag}) );
			chop($h_newhit{$tag}); # c'est pour virer le ; final
			my @a_2 = split(';',$h_newhit{$tag});
			$h_newhit{$tag."count"} = scalar(@a_2);
		}
		my $nlib = scalar(keys(%h_nlib));
		if ( $nlib )
		{
			if ( defined($h_newhit{nlib}) )
			{
				$h_newhit{nlib}=&Max($nlib,$h_newhit{nlib});
			}
			else
			{
				$h_newhit{nlib}=$nlib;
			}
		}
		$self->SetKV($link,\%h_newhit);
	}
}

sub FilterAndDump
{
my ($self,%h_param) = @_;

	my $outprefix = ( defined($h_param{statfile}) ) ? $h_param{statfile} : "LynxStats.$$.".time;
	my @a_links = ();
	$self->Filter(\@a_links,%h_param);
	$self->_Print("Dumping links >$h_param{dumpfile}< starting\n");
	my $f_links = &GetStreamOut($h_param{dumpfile});
	foreach my $link (@a_links)
	{
		my $linkdetails = $self->Get($link);
		print $f_links "$link\t$linkdetails\n";
	}
	$f_links->close;
	$self->_Print("Dumping links: done\n");
}

sub Dump
{
my($self,%h_param)=@_;

	if ( ! defined($h_param{dumpfile}) )
	{
		print STDERR "Warning: dumping links failed, dumpfile parameter undefined\n";
		return;
	}
	$self->_Print("Dumping links >$h_param{dumpfile}< starting\n");
	my $f_links = &GetStreamOut($h_param{dumpfile});
	my @a_links = $self->GetLinks();
	foreach my $link (@a_links)
	{
		my $linkdetails = $self->Get($link);
		print $f_links "$link\t$linkdetails\n";
	}
	$f_links->close;
	$self->_Print("Dumping links: done\n");
}


sub TagDubiousLink
{
my($self,$link)=@_;

	
	my %h_val=();
	return &FALSE if (!$self->GetKV($link, \%h_val));

	# si les deux sont identiques je ne suis pas dans le mode standard de lien donc je ne veux pas les tagguer
	if ( $h_val{gaplen_min} != $h_val{gaplen_max} )
	{
		if ( $h_val{gaplen_min} < 0 && $h_val{gaplen_max} < 0 && defined($h_val{ovllen}) && $h_val{ovllen} == 0 )
		{
			$h_val{weak}=2;
		}
		elsif ( defined($h_val{ovllen}) && $h_val{ovllen} > 0 && $h_val{gaplen_min} > 0 )
		{
			$h_val{weak}=2;
		}
	}
	if ( $h_val{count} > &Min($h_val{tgtspan},$h_val{srcspan}) )  # 20130330 / repeats
	{
		$h_val{weak}=2;
	}
	if ( $h_val{weak} == 2 )
	{
		$self->SetKV($link,\%h_val);
		return &TRUE;
	}
	return &FALSE;
}

sub Load
{
my($self,%h_param)=@_;

	if ( ! defined($h_param{dumpfile}) )
	{
		print STDERR "Warning: loading links failed, dumpfile parameter undefined\n";
		return;
	}
	$self->_Print("Loading links >$h_param{dumpfile}< starting\n");
	my $f_links = &GetStreamIn($h_param{dumpfile});
	my %h_val=();
	my $nb_notloaded = 0;
	while(my $l=<$f_links>)
	{
		chomp($l);
		my ($link,$details) = split("\t",$l,2);
		$details =~ s/^\s//;
		$self->Set($link,$details);
		$self->TagDubiousLink($link);
	}
	$f_links->close;
	$self->_Print("Loading links: done\n");
}


sub RemoveUnexpectedOverlappingLinks
{
my($self,%h_param)=@_;

	my @a_links = $self->GetLinks();
	my @a_sorted = ();
	foreach my $link (@a_links)
	{
		my $linkdetails = $self->Get($link);
		my %h_val = split(/[=\s]+/,$linkdetails);
		if ( defined($h_val{ovllen}) && $h_val{ovllen} > 0 && $h_val{gaplen_min} > 0 )
		{
			print STDERR "Removing link: $link, $linkdetails\n";
			$self->Del($link);
		}
	}
}


# je cree un lien lorsque deux sequences chevauchent

# prerequis: all vs all dans les deux sens => ne gere pas les chev entre deux assembly separes

sub LoadHSPSetByOverlap
{
my ($self,$o_hspds,$mode)=@_;

	my $rh_hspds = $o_hspds->{_rh_hspds};
	my @a_keys = keys(%{$rh_hspds});
	my $nb_links=0;
	my %h_flag = ();
	$h_flag{oriented}=1 if ( $mode =~ /oriented/ );
	foreach my $key (@a_keys)
	{
		my ($target)= $key =~ /target=(\S+)/;
		my ($query)=  $key =~ /query=(\S+)/;

		next if ( $target lt $query );

		my @a_ = split(/##/,$rh_hspds->{$key});

		next if ( scalar(@a_) != 1 ); # je ne gere pas les multihits pour le moment

		print STDERR "$key, $a_[0]\n";

		my %h_hsp = split(/[\s=]/,$a_[0]);

		next if( exists($h_flag{oriented}) && $h_hsp{strand} eq '-' );

		my $tgtlen  = $o_hspds->_FastOneGet('_query')->GetLen($target); # ici le dataset query = dataset target
		my $srclen  = $o_hspds->_FastOneGet('_query')->GetLen($query);

		my ($srcext,$tgtext)=();
		my $delta_src3 = $srclen-$h_hsp{qend};
		my $delta_src5 = $h_hsp{qstart}-1;


		$srcext = ( $delta_src5 <= $delta_src3 ) ? 5 : 3;

		#print STDERR "$delta_src5 $delta_src3 srclen=$srclen => $srcext\n";

		if  ( $h_hsp{strand} eq '+' )
		{
			$tgtext = ( $srcext eq '3' )  ? 5 : 3;
		}
		else 
		{
			$tgtext = ( $srcext eq '3' )  ? 3 : 5;
		}

		my $lkey    = "src=$query tgt=$target";
		my $gaplen  = -1 * $h_hsp{len};
		my $count   = int($h_hsp{len}/25)+1;         # the count is the number of 25nt slices
		my $val     = "srcext=$srcext srclen=$srclen tgtext=$tgtext tgtlen=$tgtlen gaplen_max=$gaplen gaplen_min=$gaplen srcstart=$h_hsp{qstart} srcend=$h_hsp{qend} tgtstart=$h_hsp{tstart} tgtend=$h_hsp{tend} count=$count insert_size=1 hardlink=0 weak=0 used=0 tgtspan=$h_hsp{len} srcspan=$h_hsp{len} nlib=1 ambiguous_orient=0";

		print STDERR "$val\n";
		$nb_links++;
		$self->Add($lkey,join(' ',sort split(' ',$val)));
	}
	return $nb_links;
}

# je cree un lien lorsqu'une sequence target fait le pont avec deux queries (e.g pacbio)

sub LoadHSPSetBySeq
{
my ($self,$o_hspds,$mode)=@_;

	$self->_Print("Loader: LoadHSPSetBySeq started\n");
	my $gaplen_default = 0;
	my $insert_size = $o_hspds->_FastOneGet('_target')->_FastOneGet('_insert_size');
	if ( $mode eq 'mrna' )
	{
		$gaplen_default = 1000;
	}
	elsif ( $insert_size > 0 ) # un insert_size n'a pas de sens mais je veux l'utiliser pour incrementer le nlib
	{
		$gaplen_default = $insert_size;
	}
	print STDERR "gaplen_default=$gaplen_default\n";
	my %h_flag=();
	my $rh_hspds = $o_hspds->{_rh_hspds};
	my %h_target=();
	my @a_keys = keys(%{$rh_hspds});
	foreach my $key (@a_keys)
	{
		my ($target)=$key =~ /target=(\S+)/;
		foreach my $entry (split(/##/,$rh_hspds->{$key}))
		{
			$h_target{$target} .= "$entry##";
		}
	}
	my %h_target2query=();
	my %h_tmp=();
	my $nb_links=0;
	foreach my $target (sort keys(%h_target))
	{
		$h_target{$target}=~s/##$//;
		my @a_   = split(/##/,$h_target{$target});
		for (my $i=0; $i<$#a_;$i++)
		{
			for (my $j=$i+1; $j<=$#a_;$j++)
			{
				# 20160724 comme je peux permutter je dois faire le split du i dans la boucle ou alors je dois le remettre comme je l'ai trouve
				my %h_i = split(/[\s=]/,$a_[$i]);
				$h_i{qlen}  = $o_hspds->_FastOneGet('_query')->GetLen($h_i{query});

				my %h_j = split(/[\s=]/,$a_[$j]);
				next if ( $h_j{query} eq $h_i{query} );
				$h_j{qlen}  = $o_hspds->_FastOneGet('_query')->GetLen($h_j{query});

				%h_tmp=();
				if ( $h_j{query} lt $h_i{query} )
				{
					%h_tmp = %h_i;
					%h_i   = %h_j;
					%h_j   = %h_tmp;
				}

				my $src = $h_i{query};
				my $tgt = $h_j{query};
			
				my ($srcext,$tgtext)=();
				# j'ordonne sur la target
				my ($tmin,$tmax)=(&MAXINT,0);
				if ( $h_i{tstart} < $h_j{tstart} )
				{
					if ( $h_i{tend} > $h_j{tend} )
					{
						# bizarre, inclusion je ne fais rien
						next;
					}
					$srcext = ( $h_i{strand} eq '+' ) ? 3 : 5;
					$tgtext = ( $h_j{strand} eq '+' ) ? 5 : 3;
					$tmin = &Min($tmin,$h_i{tstart});
					$tmax = &Max($tmax,$h_j{tend});
				}
				else
				{
					if ( $h_j{tend} > $h_i{tend} )
					{
						# bizarre, inclusion je ne fais rien
						next;
					}
					$tgtext = ( $h_j{strand} eq '+' ) ? 3 : 5;
					$srcext = ( $h_i{strand} eq '+' ) ? 5 : 3;
					$tmin = &Min($tmin,$h_j{tstart});
					$tmax = &Max($tmax,$h_i{tend});
				}

				my $key     = "src=$src tgt=$tgt";
				my $tregion = $tmax-$tmin+1;
				my $gaplen_max  = $gaplen_default;
				my $gaplen_min  = 0; # 20160723, je vais essayer de prendre en compte lorsque c'est bien disjoint en prenant en compte les overhangs 
				if ( $srcext eq '5' )
				{
					$gaplen_min += $h_i{qstart}-1;
				}
				else
				{
					$gaplen_min += $h_i{qlen}-$h_i{qend};
				}
				if ( $tgtext eq '5' )
				{
					$gaplen_min += $h_j{qstart}-1;
				}
				else
				{
					$gaplen_min += $h_j{qlen}-$h_j{qend};
				}
				if ( ! $gaplen_max )
				{	
					$gaplen_max = &Max(1,$tregion-($h_i{len}+$h_j{len})); # region couverte sur target - la sommes des chev
				}
				if ( $gaplen_min > $gaplen_max )
				{
					$gaplen_min = &Min(-1,$gaplen_max);
				}
				# 20160723, plus tres adapte my $count   = int(&Min($h_j{len},$h_i{len})/100)+1;         # le count est le nb de 100 nt 
				my $count   = int(($h_j{len}+$h_i{len})/1000)+1;         # je favorise le span total
				# j'utilise insert_size=$gaplen plutot que insert_size=$tregion
				# 20160723 my $val     = "srcext=$srcext srclen=$h_i{qlen} tgtext=$tgtext tgtlen=$h_j{qlen} gaplen_max=$gaplen_max gaplen_min=$gaplen_min srcstart=$h_i{qstart} srcend=$h_i{qend} tgtstart=$h_j{tstart} tgtend=$h_j{tend} count=$count insert_size=$gaplen_max hardlink=0 weak=0 used=0 tgtspan=$h_j{len} srcspan=$h_i{len} nlib=1 ambiguous_orient=0";
				# j'utilisais tstart/tend au lieu de qstart et qend sur h_j
				my $val     = "srcext=$srcext srclen=$h_i{qlen} tgtext=$tgtext tgtlen=$h_j{qlen} gaplen_max=$gaplen_max gaplen_min=$gaplen_min srcstart=$h_i{qstart} srcend=$h_i{qend} tgtstart=$h_j{qstart} tgtend=$h_j{qend} count=$count insert_size=$gaplen_max hardlink=0 weak=0 used=0 tgtspan=$h_j{len} srcspan=$h_i{len} nlib=1 ambiguous_orient=0";
				$nb_links++;
				if ( 0 && exists($h_flag{oriented}) )
				{
					my $k = "$srcext$tgtext";
					next if ( $k eq '55' || $k eq '33' );
				}
				my $valtmp = join(' ',sort split(' ',$val));
				if ( $self->Get($key) =~ /$valtmp/ )
				{
					print STDERR "WARNING DEJA: $key : $valtmp\n";
					print STDERR "WARNING $h_target{$target}\n";
				}
				else
				{
					$self->Add($key,$valtmp);
				}
			}
		}	
	}
	$self->_Print("Loader: LoadHSPSetBySeq done ($nb_links links)\n");
	return $nb_links;
}


# utilise le placement sur une reference pour orienter et creer des liens

sub LoadHSPSetByReference
{
my ($self,$o_hspds,$mode)=@_;

	$self->_Print("Filter: LoadHSPSetByReference started\n");
	my $rh_hspds = $o_hspds->{_rh_hspds};
	my $max_insertion_length = 1000; # valeur arbitraire de ce que l'on va autoriser comme insertion si l'on utilise les liens PE + AddRepeats
	my %h_hits=();
	my @a_keys = keys(%{$rh_hspds});
	foreach my $key (@a_keys)
	{
		my ($query,$target)=$key =~ /query=(\S+).+target=(\S+)/;
		foreach my $entry (split(/##/,$rh_hspds->{$key}))
		{
			my %h_i = split(/[\s=]/,$entry);
			$h_hits{$target}{$h_i{score}} .= "$entry##";
		}
	}

	my %h_hits_ok = (); # non redondant, triable par pos
	foreach my $target (sort keys(%h_hits))
	{
		foreach my $score (sort { $b <=> $a } keys(%{$h_hits{$target}})) # score decroissant
		{
			my @a_  = split(/##/,$h_hits{$target}{$score});
			for (my $i=0; $i<scalar(@a_);$i++)	# je regarde si cela chevauche beaucoup quelque chose que j'ai deja garde
			{
				my $incl=&FALSE;
				my %h_cand = split(/[\s=]/,$a_[$i]);
				foreach my $qstart (sort { $a <=> $b } keys(%{$h_hits_ok{$target}}))
				{
					last if ( $qstart > $h_cand{tend} ); # j'arrete des que je depasse
					my %h_ok = split(/[\s=]/,$h_hits_ok{$target}{$qstart});
					if ( $h_cand{tstart} >= $h_ok{tstart} && $h_cand{tend} <= $h_ok{tend} )
					{	
						#print STDERR "$target\n\tremoving\n\t$a_[$i]\n\t$h_hits_ok{$target}{$qstart}\n";
						$incl=&TRUE;
						last;
					}
				}
				if ( ! $incl )
				{
					$h_hits_ok{$target}{$h_cand{tstart}}=$a_[$i];
				}
			}
		}
	}

	%h_hits=();
	%h_hits = %h_hits_ok;

	my $nb_links=0;
	my $max_step = 10;
	foreach my $target (sort keys(%h_hits))
	{
		my @a_tilling = sort { $a <=> $b } keys(%{$h_hits{$target}});
		for(my $i=0;$i<scalar(@a_tilling);$i++)
		{
			my %h_i = split(/[\s=]/,$h_hits{$target}{$a_tilling[$i]});
			$h_i{qlen}  = $o_hspds->_FastOneGet('_query')->GetLen($h_i{query});
			my $count   = $max_step;
			for(my $j=$i+1;( $j<$i+$max_step) && ($j < scalar(@a_tilling));$j++,$count--)
			{
				my %h_j = split(/[\s=]/,$h_hits{$target}{$a_tilling[$j]});
				$h_j{qlen}  = $o_hspds->_FastOneGet('_query')->GetLen($h_j{query});
				my %h_src = ();
				my %h_tgt = ();
				if ( $h_i{query} lt $h_j{query} )
				{
					%h_src = %h_i;
					%h_tgt = %h_j;
				}
				else
				{
					%h_src = %h_j;
					%h_tgt = %h_i;
				}
				my $src = $h_src{query};
				my $tgt = $h_tgt{query};
				my ($srcext,$tgtext)=();
				if ( $h_src{tstart} < $h_tgt{tstart} )
				{
					$srcext = ( $h_src{strand} eq '+' ) ? 3 : 5;
					$tgtext = ( $h_tgt{strand} eq '+' ) ? 5 : 3;
				}
				else
				{
					$tgtext = ( $h_tgt{strand} eq '+' ) ? 3 : 5;
					$srcext = ( $h_src{strand} eq '+' ) ? 5 : 3;
				}
				my $span = &Min($h_tgt{qlen},$h_src{qlen});
				($h_src{qstart},$h_src{qend}) = ( $srcext eq '3' ) ? ($h_src{qlen}-$span,$h_src{qend}) : ($h_src{qstart},$h_src{qstart}+$span);
				($h_tgt{qstart},$h_tgt{qend}) = ( $tgtext eq '3' ) ? ($h_tgt{qlen}-$span,$h_tgt{qend}) : ($h_tgt{qstart},$h_tgt{qstart}+$span);
				my $key     = "src=$src tgt=$tgt";
				my $gaplen  = $h_j{tstart}-$h_i{tend}+1; # c'est bien curr et prev
				my $gaplen_max = ( $gaplen > 0 ) ? ($gaplen+$max_insertion_length) : $gaplen;
				my $insert_size = &Max($gaplen,1);
				# le pb des paires de repeats c'est que cela double le count au moment de la fusion des hits
				if ( $span > $gaplen  || $count == $max_step )
				{# je ne veux pas des liens lointains avec des petits contigs
					$span /= ($max_step-$count+1);
					my $lcount = 1000+$count; # je mets 1000 pour ne pas que cela devienne moins prioritaire que d'autres liens lorsque l'on utilise le count
					my $val     = "srcext=$srcext srclen=$h_src{qlen} tgtext=$tgtext tgtlen=$h_tgt{qlen} gaplen_max=$gaplen_max gaplen_min=$gaplen srcstart=$h_src{qstart} srcend=$h_src{qend} tgtstart=$h_tgt{qstart} tgtend=$h_tgt{qend} count=$lcount insert_size=$gaplen hardlink=1 weak=0 used=0 tgtspan=$span srcspan=$span nlib=1 ambiguous_orient=0";
					print STDERR "span=$span len=".length($span)."\n$val\n";
					$nb_links++;
					$self->Add($key,join(' ',sort split(' ',$val)));
				}
			}
		}
	}
	$self->_Print("Filter: LoadHSPSetByReference done\n");
}



sub LoadHSPSetByPE
{
my ($self,$o_hspds,$mode)=@_;

	my %h_flag=();
	$h_flag{oriented}=1 if ( $mode =~ /oriented/ );
	$h_flag{update}=1   if ( $mode =~ /update/ );
	my $rh_hspds = $o_hspds->{_rh_hspds};
	my $nb_links=0;
	if ( ! scalar(keys(%{$rh_hspds})) )
	{
		$self->_Print("LoadHSPSetByPE No HSP found\n");
		return $nb_links;
	}
	$o_hspds->InitEndLabels();
	my $end1 = $o_hspds->_FastOneGet('_end1');
	my $end2 = $o_hspds->_FastOneGet('_end2');

	if ( !defined($end1) || ! defined($end2) || $end1 eq '' || $end2 eq '' )
	{
		Carp::croak("Fatal error: LoadHSPSetByPE cannot work when paired-end labels are not defined (see InitEndLabels() and CheckEndLabels())\n");
	}
	my $insert_size = $o_hspds->_FastOneGet('_target')->_FastOneGet('_insert_size');
	my $insert_sd = $o_hspds->_FastOneGet('_target')->_FastOneGet('_insert_sd');
	$insert_sd = 50 if ( $insert_sd eq '' );
	$insert_sd  = int($insert_size * $insert_sd / 100);
	my ($min_insert_size,$max_insert_size) = ($insert_size-$insert_sd,$insert_size+$insert_sd);
	if ( $min_insert_size <= 0 )
	{
		Carp::croak("Fatal error: insert_size not set for the library\n");
	}
	my $keep_ambiguous = ( defined($o_hspds->_FastOneGet('_keepambiguous')) ) ? $o_hspds->_FastOneGet('_keepambiguous') : &FALSE;
#	print STDERR "keep_ambiguous=$keep_ambiguous\n";

	my $orientation = $o_hspds->_FastOneGet('_target')->_FastOneGet('_paired');
	my ($revstrand) = ( defined($orientation) && $orientation eq 'opposite' ) ? &TRUE : &FALSE;
	my $pattern  = $o_hspds->_FastOneGet('_target')->_FastOneGet('_pattern');

	# trying to remove paired-end contamination
	my $MAX_INSERT_SIZE_FOR_CONTAMINANTS = 500;
	my $remove_putative_paired_end = ( $insert_size >= $MAX_INSERT_SIZE_FOR_CONTAMINANTS && $o_hspds->_FastOneGet('_target')->_FastOneGet('_source') eq 'illumina_mp' ) ? &TRUE: &FALSE;
	print STDERR "remove_putative_paired_end_contamination = $remove_putative_paired_end\n";

	my %h_insert=();
	foreach my $key (sort keys(%{$rh_hspds}))
	{
		my ($target)=$key =~ /target=(\S+)/;
		my ($query) =$key =~ /query=(\S+)/;
		my ($insert,$ext) = $target =~ /$pattern/;
		next if ( $ext ne $end1 && $ext ne $end2 );
		$h_insert{$insert}{$ext} .= $o_hspds->Get($key)."##";
		$h_insert{$insert}{$ext}=~ s/##$//;
	}

	my $internal_links=0;
	my $inconsistant_insert_length=0;
	my %h_span = ();
	my $tot_hsp = 0;
	my $tot_contaminant = 0;
	foreach my $insert (keys(%h_insert))
	{
		next if (! defined($h_insert{$insert}{$end1}) || ! defined($h_insert{$insert}{$end2}) );
		$tot_hsp++;
		my @a_1 = split(/##/,$h_insert{$insert}{$end1});
		my @a_2 = split(/##/,$h_insert{$insert}{$end2});
		if ( !$keep_ambiguous && $#a_1 > 0 && $#a_2 > 0 )	# je n'essais pas de gerer si les deux extremites sont ambigues ==> j'essaye
		{
			delete($h_insert{$insert});
			next;
		}
		my $ok = &FALSE;
		for(my $i1=0;$i1<=$#a_1;$i1++)
		{
			my %h_1 = split(/[=\s]+/,$a_1[$i1]);
			$h_1{qlen}  = $o_hspds->_FastOneGet('_query')->GetLen($h_1{query});
			for(my $i2=0;$i2<=$#a_2;$i2++)
			{
				my %h_2 = split(/[=\s]+/,$a_2[$i2]);
				next if ( $h_1{query} eq $h_2{query} );

				$h_2{qlen}  = $o_hspds->_FastOneGet('_query')->GetLen($h_2{query});

				if ( $revstrand ) 
				{
					$h_1{strand} = (  $h_1{strand} eq '+' ) ? '-' : '+';
					$h_2{strand} = (  $h_2{strand} eq '+' ) ? '-' : '+';
				}

				# suppression des liens trop loin des extremites

				if (  $h_1{qstart} > $max_insert_size && $h_1{qend} < ($h_1{qlen}-$max_insert_size) )
				{
					#$self->_Print("Internal link (repeat): $h_1{qstart} > $max_insert_size && $h_1{qend} < ($h_1{qlen}-$max_insert_size)\n");
					$internal_links++;
					next;
				}
				elsif ( $h_2{qstart} > $max_insert_size && $h_2{qend} < ($h_2{qlen}-$max_insert_size) )
				{
					#$self->_Print("Internal link (repeat) : $h_2{qstart} > $max_insert_size && $h_2{qend} < ($h_2{qlen}-$max_insert_size)\n");
					$internal_links++;
					next;
				}
				my $src1 = $h_1{query};
				my $src2 = $h_2{query};
				my $ext1 = '';
				my $ext2 = '';
				my $gap = '';
				if ( $h_1{strand} eq '+' && $h_2{strand} eq '+' )
				{
					$gap = ($h_1{qlen} - $h_1{qstart}) + ($h_2{qlen} - $h_2{qend} );
					$ext1 = 3;
					$ext2 = 3;
				}
				elsif ( $h_1{strand} eq '-' && $h_2{strand} eq '-' )
				{
					$gap = $h_1{qend} + $h_2{qend};
					$ext1 = 5;
					$ext2 = 5;
				}
				else
				{	# le 20120220, j'ai cru m'etre trompe en voyant pas mal d'ambiguites sur l'orientation de lib cacao 454
					# sur papier, j'ai valide que on pouvait traiter pareil lorsque les deux brins etaient identiques
					# mais lorsqu'ils sont differents, le contig amont et celui qui a le + (il est donc branche sur le 3')
					# je l'ecris un peu diffrent mais c'est pareil qu'avant, en fait ce qui compte c'est la connection et pas l'ordre des contigs
			
					if ( $h_1{strand} eq '+' )
					{
						$gap = ($h_1{qlen} - $h_1{qend}) + $h_2{qstart};
						$ext1 = 3;
						$ext2 = 5;
					}
					elsif ( $h_2{strand} eq '+')
					{
						$gap = ($h_2{qlen} - $h_2{qend}) + $h_1{qstart};
						$ext1 = 5;
						$ext2 = 3;
					}
				}

				if ( $gap > $max_insert_size )
				{
					#$self->_Print("Too far: $gap >= $max_insert_size\n");
					$inconsistant_insert_length++;
					next;
				}
				#elsif ( $gap < $min_insert_size ) c'est des conneries on ne connait pas la taille d'un gap
				#{
				#	$self->_Print("Too short: $gap < $min_insert_size\n");
				#	$inconsistant_insert_length++;
				#	next;
				#}
				elsif (  $remove_putative_paired_end )
				{
					if ( $gap < $MAX_INSERT_SIZE_FOR_CONTAMINANTS )
					{
						$tot_contaminant++;
						next;
					}
				}

				if ( $src1 lt $src2 )
				{
					$self->_UpdateSpan(\%h_span,"$src1 $ext1","$src2 $ext2",\%h_1,\%h_2);
				}
				else
				{
					$self->_UpdateSpan(\%h_span,"$src2 $ext2","$src1 $ext1",\%h_2,\%h_1);
				}
				$ok = &TRUE;
			}
		}
		if ( ! $ok )
		{
			delete($h_insert{$insert});
			next;
		}	
	}
	print STDERR "$tot_contaminant / $tot_hsp \n";
	if (exists($h_flag{update}))
	{
		&__RemoveFromListOfInsert($rh_hspds,\%h_insert,$pattern);
	}

	foreach my $srcspan (keys (%h_span))
	{
		my($src,$srcext)=split(' ',$srcspan);
		my $srclen = $o_hspds->_FastOneGet('_query')->GetLen($src);
		foreach my $tgtspan (keys(%{$h_span{$srcspan}}))
		{
			my ($tgt,$tgtext) = split(' ',$tgtspan);
			my $tgtlen = $o_hspds->_FastOneGet('_query')->GetLen($tgt);
			my $pair = join(' ',sort($tgt,$src));
			my $delta = &MAXINT;
			if ( $srcext eq '3' && $tgtext eq '5' )
			{
				#----------+---|--   --+---|--------
				$delta = $srclen-$h_span{$srcspan}{$tgtspan}{MEANSRC}+$h_span{$srcspan}{$tgtspan}{MEANTGT};
			}
			elsif ( $srcext eq '5' && $tgtext eq '5' )
			{
				#-|--+--------   --+---|--------
				$delta = $h_span{$srcspan}{$tgtspan}{MEANSRC}+$h_span{$srcspan}{$tgtspan}{MEANTGT};
			}
			elsif ( $srcext eq '5' && $tgtext eq '3' )
			{
				#-|--+--------   ----------|-----+--
				$delta = $h_span{$srcspan}{$tgtspan}{MEANSRC}+($tgtlen-$h_span{$srcspan}{$tgtspan}{MEANTGT});
			}
			elsif ( $srcext eq '3' && $tgtext eq '3' )
			{
				#--------+--|-   ----------|-----+--
				$delta = $srclen-$h_span{$srcspan}{$tgtspan}{MEANSRC}+($tgtlen-$h_span{$srcspan}{$tgtspan}{MEANTGT});
			}
			if( $delta == &MAXINT )
			{
				$self->_Print("Warning: $srcspan, $tgtspan $h_span{$srcspan}{$tgtspan}{MEANSRC} $h_span{$srcspan}{$tgtspan}{MEANTGT}\n");
				print STDERR Dumper(%{$h_span{$srcspan}});
				next;
			}
			my $gaplen_max = int($max_insert_size-$delta); # c'est clairement une surestimation
			my $gaplen_min = int($min_insert_size-$delta);

			# surtout ne pas faire ca
			#my %h_val = split(/[=\s]+/,"srcext=$srcext srclen=$srclen tgtext=$tgtext tgtlen=$tgtlen gaplen_max=$gaplen_max gaplen_min=$gaplen_min srcstart=$h_span{$srcspan}{$tgtspan}{MINSRC} srcend=$h_span{$srcspan}{$tgtspan}{MAXSRC} tgtstart=$h_span{$srcspan}{$tgtspan}{MINTGT} tgtend=$h_span{$srcspan}{$tgtspan}{MAXTGT} count=$h_span{$srcspan}{$tgtspan}{COUNT} insert_size=$insert_size hardlink=0 weak=0 used=0");
			# on peut rajouter les liens
			#$self->SetKV2($src,$tgt,\%h_val);

			my $tottgtspan = $h_span{$srcspan}{$tgtspan}{MAXTGT}-$h_span{$srcspan}{$tgtspan}{MINTGT}+1;
			my $totsrcspan = $h_span{$srcspan}{$tgtspan}{MAXSRC}-$h_span{$srcspan}{$tgtspan}{MINSRC}+1;
			my $key = "src=$src tgt=$tgt";
# 20121216 je garde les negatifs pour gerer les chevauchements
#			$gaplen_max=&Max(1,$gaplen_max);
#			$gaplen_min=&Max(1,$gaplen_min);
			($gaplen_min,$gaplen_max)=&MinMax($gaplen_min,$gaplen_max);
			my $val = "srcext=$srcext srclen=$srclen tgtext=$tgtext tgtlen=$tgtlen gaplen_max=$gaplen_max gaplen_min=$gaplen_min srcstart=$h_span{$srcspan}{$tgtspan}{MINSRC} srcend=$h_span{$srcspan}{$tgtspan}{MAXSRC} tgtstart=$h_span{$srcspan}{$tgtspan}{MINTGT} tgtend=$h_span{$srcspan}{$tgtspan}{MAXTGT} count=$h_span{$srcspan}{$tgtspan}{COUNT} insert_size=$insert_size hardlink=0 weak=0 used=0 tgtspan=$tottgtspan srcspan=$totsrcspan nlib=1 ambiguous_orient=0";
			$nb_links++;
			if ( exists($h_flag{oriented}) )
			{
				my $k = "$srcext$tgtext";
				next if ( $k eq '55' || $k eq '33' );
			}
			$self->Add($key,join(' ',sort split(' ',$val)));
		}
	}
	return $nb_links;
}

# filter out a given object to fill in the current one
sub FilterAndLoad
{
my($self,$o_linkds_raw,%h_param)=@_;

	my $ra_links_raw = $o_linkds_raw->GetLinksRef();
	my $nb_raw = scalar(@$ra_links_raw);
	my %h_val = ();
	my @a_params = keys(%h_param);
	my $nb_ignored=0;
	my $link;
	for(my $i=0;$i<$nb_raw;$i++)
	{
		next if ( !$o_linkds_raw->GetKV($$ra_links_raw[$i],\%h_val) );
		my $ignored=&FALSE;
		foreach my $k (@a_params)		# si on arrive avec un mincount=100, on vire tous les links a < 100
										# si on arrive avec un maxinsert_size=4000, on vire les insert_size > 4000
		{	
			if ( $k =~ /^min(\S+)/ )
			{
				if ( ! defined($h_val{$1}) )
				{
					print STDERR "WARNING: Filter $1 filter/key undefined in $$ra_links_raw[$i]\n";
				}
				else
				{
					$ignored = &TRUE if ( $h_val{$1} < $h_param{$k} );
				}
			}
			elsif ( $k =~ /^max(\S+)/ )
			{
				if ( ! defined($h_val{$1}) )
				{
					print STDERR "WARNING: Filter $1 filter/key undefined in $$ra_links_raw[$i]\n";
				}
				else
				{
					$ignored = &TRUE if ( $h_val{$1} > $h_param{$k} );
				}
			}
			#print STDERR "Ignored $k: $link, $linkdetails \n" if ($ignored);
			last if ($ignored);
		}
		my $key = "insert_size_$h_val{insert_size}_mincount";
		if ( defined($h_param{$key}) && $h_val{count} < $h_param{$key} )
		{	
#			print STDERR "$key=$h_param{$key} $h_val{count} < $h_param{$key}\n";
			$ignored = &TRUE;
		}
		if ( $ignored )
		{
			$nb_ignored++;
		}
		else
		{
			$self->Set($$ra_links_raw[$i],$o_linkds_raw->Get($$ra_links_raw[$i]));
		}
	}
	$self->_Print("FilterAndLoad: $nb_ignored links ignored / $nb_raw\n");
	return $nb_ignored;
}


sub Filter
{
my($self,$ra_filt_links,%h_param)=@_;

	my $ra_links_raw = $self->GetLinksRef();
	@$ra_filt_links=();
	my $nb_ignored=0;
	my %h_val = ();
	my $nb_raw = scalar(@$ra_links_raw);
	my @a_params = keys(%h_param);
	for(my $i=0;$i<$nb_raw;$i++)
	{
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		if ( exists($h_param{sameprefix}) && $h_param{sameprefix} )
		{
			my($srcpref,$tgtpref)=$$ra_links_raw[$i]=~/src=(\S+)\.\d+\s+tgt=(\S+)\.\d+/;
			if (!defined($srcpref) || !defined($tgtpref) || $srcpref ne $tgtpref)
			{
				$nb_ignored++;
				next;
			}
		}
		my $ignored=&FALSE;
		foreach my $k (@a_params)		# si on arrive avec un mincount=100, on vire tous les links a < 100
										# si on arrive avec un maxinsert_size=4000, on vire les insert_size > 4000
		{	
			if ( $k =~ /^min(\S+)/ )
			{
				if ( ! defined($h_val{$1}) )
				{
					print STDERR "WARNING: Filter $1 filter/key undefined in $$ra_links_raw[$i]\n";
				}
				else
				{
					$ignored = &TRUE if ( $h_val{$1} < $h_param{$k} );
				}
			}
			elsif ( $k =~ /^max(\S+)/ )
			{
				if ( ! defined($h_val{$1}) )
				{
					print STDERR "WARNING: Filter $1 filter/key undefined in $$ra_links_raw[$i]\n";
				}
				else
				{
					$ignored = &TRUE if ( $h_val{$1} > $h_param{$k} );
				}
			}
			#print STDERR "Ignored $k: $link, $linkdetails \n" if ($ignored);
			last if ($ignored);
		}
		if ( $ignored )
		{
			$nb_ignored++;
		}
		else
		{
			push(@$ra_filt_links,$$ra_links_raw[$i]);
		}
	}
	$self->_Print("Filter: $nb_ignored links ignored / $nb_raw\n");
	return $nb_ignored;
}




sub ReciprocalBestFilter
{
my($self,$ra_links_raw)=@_;

	$self->_Print("ReciprocalBestFilter: starting\n");
	my @a_filt_links=();
	my $nb_ignored=0;
	my %h_val = ();
	my %h_key = ();
	my %h_best = ();
	my $nb_raw = scalar(@$ra_links_raw);
	# 1er passage je calcule les meilleurs voisins
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		foreach my $id ( 'src','tgt' )
		{
			my $end = $id."ext";
			my $score = $h_val{count};
			$score += $h_val{ovllen} if ( defined($h_val{ovllen}) );
#			print STDERR "$id,$end $h_key{$id} >$h_val{$end}< $h_best{$h_key{$id}}{$h_val{$end}}{score}\n";
			if ( ! defined($h_best{$h_key{$id}}{$h_val{$end}}{score}) || $score > $h_best{$h_key{$id}}{$h_val{$end}}{score} )
			{
				$h_best{$h_key{$id}}{$h_val{$end}}{best} = '' if ( $score > $h_best{$h_key{$id}}{$h_val{$end}}{score} );
				$h_best{$h_key{$id}}{$h_val{$end}}{score} = $score;
				$h_best{$h_key{$id}}{$h_val{$end}}{best} .= ( $id eq 'src' ) ? $h_key{tgt}  : $h_key{src}; # je concatene comme ca en cas d'egalite j'elimine car ce n'est pas eq
#				print STDERR "\tbest=$h_best{$h_key{$id}}{$h_val{$end}}{best}\n";
			}
		}
	}
	# 2e passage, je supprime tous les liens qui ne concernent pas deux meilleurs voisins
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		my $rbh=&FALSE;
#		print STDERR "$h_key{src} $h_val{srcext} $h_best{$h_key{src}}{$h_val{srcext}}{best} vs $h_key{tgt}\n";
#		print STDERR "$h_key{tgt} $h_val{tgtext} $h_best{$h_key{tgt}}{$h_val{tgtext}}{best} vs $h_key{src}\n";
#		print STDERR "\n";
		if ( $h_best{$h_key{src}}{$h_val{srcext}}{best} eq $h_key{tgt} && $h_best{$h_key{tgt}}{$h_val{tgtext}}{best} eq $h_key{src} )
		{
			$rbh = &TRUE;
		}
		if ( ! $rbh )
		{
			#print STDERR "$$ra_links_raw[$i] removed\n";
			$nb_ignored++;
		}
		else
		{
			push(@a_filt_links,$$ra_links_raw[$i]);
		}
	}
	@$ra_links_raw = ();
	@$ra_links_raw = @a_filt_links;
	$self->_Print("ReciprocalBestFilter: $nb_ignored links ignored / $nb_raw\n");
	return  ($nb_ignored,\%h_best);
}


# base sur le count + sur le chevauchement
sub Best
{
my($self,$rh_best) = @_;

	$self->_Print("Best: starting\n");
	my $ra_links_raw = $self->GetLinksRef();
	my $nb_raw = scalar(@$ra_links_raw);
	my %h_val=();
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		foreach my $id ( 'src','tgt' )
		{
			my $end = $id."ext";
			my $score = $h_val{count};
			$score += $h_val{ovllen} if ( defined($h_val{ovllen}) );
			if ( ! defined($$rh_best{$h_key{$id}}{$h_val{$end}}{score}) || $score > $$rh_best{$h_key{$id}}{$h_val{$end}}{score} )
			{
				$$rh_best{$h_key{$id}}{$h_val{$end}}{best} = '' if ( $score > $$rh_best{$h_key{$id}}{$h_val{$end}}{score} );
				$$rh_best{$h_key{$id}}{$h_val{$end}}{score} = $score;
				$$rh_best{$h_key{$id}}{$h_val{$end}}{best} .= ( $id eq 'src' ) ? $h_key{tgt}  : $h_key{src};
			}
		}
	}
	$self->_Print("Best: done\n");
}

# base sur la somme des deux % de span
sub BestSpan
{
my($self,$rh_best) = @_;

	$self->_Print("BestSpan: starting\n");
	my $ra_links_raw = $self->GetLinksRef();
	my $nb_raw = scalar(@$ra_links_raw);
	my %h_val=();
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		foreach my $id ( 'src','tgt' )
		{
			my $end = $id."ext";
			my $score = ($h_val{srcspan}*100)/$h_val{srclen}+($h_val{tgtspan}*100)/$h_val{tgtlen};
			if ( ! defined($$rh_best{$h_key{$id}}{$h_val{$end}}{score}) || $score > $$rh_best{$h_key{$id}}{$h_val{$end}}{score} )
			{
				$$rh_best{$h_key{$id}}{$h_val{$end}}{best} = '' if ( $score > $$rh_best{$h_key{$id}}{$h_val{$end}}{score} );
				$$rh_best{$h_key{$id}}{$h_val{$end}}{score} = $score;
				$$rh_best{$h_key{$id}}{$h_val{$end}}{best} .= ( $id eq 'src' ) ? $h_key{tgt}  : $h_key{src};
			}
		}
	}
	$self->_Print("BestSpan: done\n");
}


sub BestFilter
{
my($self,$ra_links_raw,$rh_best)=@_;

	$self->_Print("BestFilter: starting\n");
	my @a_filt_links=();
	my %h_key=();
	my %h_val=();
	my $kept = 0;
	my $tot = scalar(@$ra_links_raw);
	for(my $i=0;$i<$tot;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		if ( (&Exists($rh_best,$h_key{src},5,'best') && $$rh_best{$h_key{src}}{5}{best} =~ /\b$h_key{tgt}\b/) ||
		 (&Exists($rh_best,$h_key{src},3,'best') && $$rh_best{$h_key{src}}{3}{best} =~ /\b$h_key{tgt}\b/) ||
		 (&Exists($rh_best,$h_key{tgt},5,'best') && $$rh_best{$h_key{tgt}}{5}{best} =~ /\b$h_key{src}\b/) ||
		 (&Exists($rh_best,$h_key{tgt},3,'best') && $$rh_best{$h_key{tgt}}{3}{best} =~ /\b$h_key{src}\b/) )
		{
			push(@a_filt_links,$$ra_links_raw[$i]);
			$kept++;
		}
	}
	@$ra_links_raw = ();
	@$ra_links_raw = @a_filt_links;
	$self->_Print("BestFilter: done, $kept/$tot links kept\n");
}


# a partir d'une selection de contigs, je fais la liste des contigs lies
# je selectionne les contigs qui sont lies a plusieurs contigs de la liste
# j'extrais tous les liens impliquant ces contigs selectionnes

sub DensifyingRegion
{
my($self,$fileofcontignames)=@_;

	$self->_Print("Densifying region: starting\n");
	my $ra_links_raw = $self->GetLinksRef();
	my @a_filt_links=();
	my $nb_selected=0;
	my %h_val = ();
	my %h_key = ();
	my %h_best = ();
	#0, je charge les contigs selectionnes
	my %h_selection =();
	my $f_in = &GetStreamIn($fileofcontignames);
	while(my $l=<$f_in>)
	{
		chomp($l);
		$h_selection{$l}=1;
	}
	$f_in->close;
	# 1: je cree un graphe a partir des liens issus de contigs selectionnes
	my $nb_raw = scalar(@$ra_links_raw);
	my @a_links_selected=();
	my %h_graph = ();
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		if ( defined($h_selection{$h_key{src}}) || defined($h_selection{$h_key{tgt}}) )
		{
			$h_graph{$h_key{src}}{$h_key{tgt}}=1;
			$h_graph{$h_key{tgt}}{$h_key{src}}=1;
		}
	}
	# 2: je fais un 2e passage pour creer les liens entre contigs qui sont dans le graphe
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		if ( exists($h_graph{$h_key{src}}) && exists($h_graph{$h_key{tgt}} ))
		{
			$h_graph{$h_key{src}}{$h_key{tgt}}=1;
			$h_graph{$h_key{tgt}}{$h_key{src}}=1;
		}
	}
	# 3: je selectionne les contigs a plusieurs arretes (ce sont des ponts encore contigs de la selection initiale
	foreach my $id (sort keys(%h_graph))
	{
		$h_selection{$id}=1 if ( scalar($h_graph{$id}) > 1 );
	}
	# 4: je ne conserve que les liens avec entre contigs selectionnes
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		if ( (defined($h_selection{$h_key{src}}) && defined($h_selection{$h_key{tgt}})) )
		{
			push(@a_links_selected,$$ra_links_raw[$i]);
			$nb_selected++;
		}
	}
	$self->_Print("Densifying region: done, $nb_selected selected links\n");
	return(\@a_links_selected);
}
  
sub BestNeighbours
{
my($self,$fileofcontignames,$fileofselectedcontignames)=@_;

	$self->_Print("BestNeighbours: starting\n");
	my $ra_links_raw = $self->GetLinksRef();
	my @a_filt_links=();
	my $nb_ignored=0;
	my %h_val = ();
	my %h_key = ();
	my %h_best = ();
	#0, je charge les contigs selectionnes
	my %h_selection =();
	my $f_in = &GetStreamIn($fileofcontignames);
	while(my $l=<$f_in>)
	{
		chomp($l);
		$h_selection{$l}=2;
	}
	$f_in->close;
	# 1er passage je calcule les meilleurs voisins, a factoriser avec ReciprocalBest
	my $nb_raw = scalar(@$ra_links_raw);
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		foreach my $id ( 'src','tgt' )
		{
			my $end = $id."ext";
			my $score = $h_val{count};
			$score += $h_val{ovllen} if ( defined($h_val{ovllen}) );
#			print STDERR "$id,$end $h_key{$id} >$h_val{$end}< $h_best{$h_key{$id}}{$h_val{$end}}{score}\n";
			if ( ! defined($h_best{$h_key{$id}}{$h_val{$end}}{score}) || $score > $h_best{$h_key{$id}}{$h_val{$end}}{score} )
			{
				$h_best{$h_key{$id}}{$h_val{$end}}{best} = '' if ( $score > $h_best{$h_key{$id}}{$h_val{$end}}{score} );
				$h_best{$h_key{$id}}{$h_val{$end}}{score} = $score;
				$h_best{$h_key{$id}}{$h_val{$end}}{best} .= ( $id eq 'src' ) ? $h_key{tgt}  : $h_key{src}; # je concatene comme ca en cas d'egalite j'elimine car ce n'est pas eq
#				print STDERR "\tbest=$h_best{$h_key{$id}}{$h_val{$end}}{best}\n";
			}
		}
		if ( 0 ) # Ms
		{
		# 20130321: j'en profite pour rajouter les les liens a deux banques dans la selection potentielle
		if ( defined($h_val{nlib}) && $h_val{nlib} > 1 && defined($h_val{ambiguous_orient}) && $h_val{ambiguous_orient} == 0 )
		{
			if ( defined($h_selection{$h_key{src}}) && $h_selection{$h_key{src}} == 2 && ! defined($h_selection{$h_key{tgt}}) )
			{
				$h_selection{$h_key{tgt}}=1;
			}
			if ( defined($h_selection{$h_key{tgt}}) && $h_selection{$h_key{tgt}} == 2 && ! defined($h_selection{$h_key{src}}) )
			{
				$h_selection{$h_key{src}}=1;
			}
		}
		}
	}
	if ( 1 ) # pour Ms
	{
		# 3, je rajoute a la selection tous les contigs qui sont le meilleur hit d'un contig selectionne
		foreach my $id (keys(%h_best))
		{
			foreach my $ext ('5','3')
			{
				# ici je prenais les contigs non selectionne dont le meilleur etait selectionne
				#if ( defined($h_selection{$h_best{$id}{$ext}{best}}) && $h_selection{$h_best{$id}{$ext}{best}} == 2 )
				#{
				#	$h_selection{$id} = 1 if ( ! defined($h_selection{$id}) );
				#}
				if ( defined($h_selection{$id}) && $h_selection{$id} == 2 )	# je suis parmi les selectionnes
				{
					my $besthit = $h_best{$id}{$ext}{best};
					print STDERR "$id => BEST $ext = $besthit $h_best{$id}{$ext}{score}\n";
					if ( exists($h_best{$id}{$ext}{best}) && $h_selection{$besthit} != 2 )  # si j'ai un besthit et qu'il n'est pas selectionne
					{
						$h_selection{$besthit}=1;
					}
				}
			}
		}
		# 4, je supprime de la selection, tous les nouveaux selectionnes qui ne sont pas lies des deux cotes par leur best avec un contig selectionne
		foreach my $id (keys(%h_best))
		{
			if ( defined($h_selection{$id}) && $h_selection{$id} == 1 )	# je suis un neo selectionne
			{
				if ( !exists($h_best{$id}{5}{best}) || !exists($h_best{$id}{3}{best}) )
				{
					delete($h_selection{$id});
					next;
				}
				next if ( exists($h_selection{$h_best{$id}{5}{best}}) && exists($h_selection{$h_best{$id}{3}{best}}) );
				delete($h_selection{$id});
			}
		}
	}

	# 5, je ne garde que les liens concernant des contigs selectionnes
	for(my $i=0;$i<$nb_raw;$i++)
	{
		my %h_key = split(/[=\s]+/,$$ra_links_raw[$i]);
		next if ( !$self->GetKV($$ra_links_raw[$i],\%h_val) );
		if ( exists($h_selection{$h_key{src}}) && exists($h_selection{$h_key{tgt}}) && defined($h_val{ambiguous_orient}) && $h_val{ambiguous_orient} == 0 ) 
		{
			next if ( $h_selection{$h_key{src}} == 2 || $h_selection{$h_key{tgt}} == 2 );
			next if ( defined($h_val{nlib}) && $h_val{nlib} > 1 );
		}
		$self->Del($$ra_links_raw[$i]);
		$nb_ignored++;
	}
	if (defined($fileofselectedcontignames))
	{
		my $f_out = &GetStreamOut($fileofselectedcontignames);
		foreach my $id (sort keys(%h_selection))
		{
			print $f_out "$id\n";
		}
		$f_out->close;
	}
	$self->_Print("BestNeighbours: $nb_ignored links ignored / $nb_raw\n");
}

sub __RemoveFromListOfInsert
{
my($rh_hspds,$rh_insert,$pattern) = @_;

	foreach my $key (keys(%{$rh_hspds}))
	{
		my ($target)=$key =~ /target=(\S+)/;
		my ($insert,$ext) = $target =~ /$pattern/;
		if ( !defined($$rh_insert{$insert}) )
		{
			delete($rh_hspds->{$key});
		}
	}
}

sub _UpdateSpan
{
my ($self,$rh_span,$src,$tgt,$rh_src,$rh_tgt) = @_;

	#if (! defined($$rh_src{qstart}) || ! defined($$rh_src{qend}) )
	#{
	#	print STDERR ")))))\n\t$src, ".join('=',%$rh_src)."\n\t$tgt ".join('=',%$rh_tgt)."\n\n";
	#}
	# sur la source
	if ( ! defined($$rh_span{$src}{$tgt}{MINSRC}) || $$rh_src{qstart} < $$rh_span{$src}{$tgt}{MINSRC} )
	{
		$$rh_span{$src}{$tgt}{MINSRC} = $$rh_src{qstart};
	}
	if ( ! defined($$rh_span{$src}{$tgt}{MAXSRC}) || $$rh_src{qend} > $$rh_span{$src}{$tgt}{MAXSRC} )
	{
		$$rh_span{$src}{$tgt}{MAXSRC} = $$rh_src{qend};
	}
	$$rh_span{$src}{$tgt}{MEANSRC}=($$rh_span{$src}{$tgt}{MAXSRC}+$$rh_span{$src}{$tgt}{MINSRC})/2;

	# sur la cible
	if ( ! defined($$rh_span{$src}{$tgt}{MINTGT}) || $$rh_tgt{qstart} < $$rh_span{$src}{$tgt}{MINTGT} )
	{
		$$rh_span{$src}{$tgt}{MINTGT} = $$rh_tgt{qstart};
	}
	if ( ! defined($$rh_span{$src}{$tgt}{MAXTGT}) || $$rh_tgt{qend} > $$rh_span{$src}{$tgt}{MAXTGT} )
	{
		$$rh_span{$src}{$tgt}{MAXTGT} = $$rh_tgt{qend};
	}
	$$rh_span{$src}{$tgt}{MEANTGT}=($$rh_span{$src}{$tgt}{MAXTGT}+$$rh_span{$src}{$tgt}{MINTGT})/2;
	$$rh_span{$src}{$tgt}{COUNT}++;
}


sub _UpdHSP
{
my($r_str,$hsp)=@_;

	if ( ! defined($r_str) || $$r_str eq '')
	{
		$$r_str = $hsp;
	}
	else
	{
		$$r_str .= "##$hsp";
	}
}


sub _Print
{
my($self,$msg)=@_;

	print STDERR "$msg";
}


sub PrintInfo
{
my($self) = @_;

}


1;
