#
# $Id: OpticalMap.pm 519 2016-08-26 05:40:19Z gouzy $
#

package OpticalMap;

use strict;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use File::Basename;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use threads;
use threads::shared;

my $DEBUGJG = 0;

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 519 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	$self->SUPER::_Init();

	$self->_Set('_nb_cpus', 1);
	$self->_Set('_debug', 1);
	$self->_Set('_min_sites',5);
	$self->_Set('_min_scaffold_len',0);
	$self->_Set('_min_scaffold_span',0);
	$self->_Set('_min_scaffold_hitlen',0); 
	$self->_Set('_min_density',0);
	$self->_Set('_palendromic',1);

	foreach my $param (keys(%h_param))
	{
		$self -> _Set ("_$param", $h_param{$param});
	}

	$self->_Set('_rh_omap', {});
	$self->_Set('_rh_silico', {}); # contient la sequence digeree que je mappe, c'est une variable temp
	$self->_Set('_ra_hits', undef); 
	$self->_Set('_rh_memo', {});
}

sub Init
{
	my ($self,%h_param) = @_;

	foreach my $param (keys(%h_param))
	{
		$self -> _Set ("_$param", $h_param{$param});
	}
}


=head2 procedure Load

 Title        : Load
 Usage        : $o_hspds -> Load;

=cut

sub Load 
{
my ($self,%h_param) = @_;

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my @a_expected=('files','format','fact');
	&AssertParams(\%h_param,\@a_expected);

	my $min_frag_len=0;
	if ( $h_param{format} =~ /XML2/i )
	{
		$min_frag_len = $self->LoadXML2(%h_param); # format MvSl
	}
	elsif ( $h_param{format} =~ /XML/i )
	{
		$min_frag_len = $self->LoadXML(%h_param); # format Medicago
	}
	elsif ( $h_param{format} =~ /SOMA/i )
	{
		$min_frag_len = $self->LoadSOMA(%h_param);
	}
	elsif ( $h_param{format} =~ /FASTA/i )
	{
		$min_frag_len = $self->LoadFASTA(%h_param);
	}
	elsif ( $h_param{format} =~ /BIONANO_RCMAP/i )
	{
		$min_frag_len = $self->LoadBIONANO_RCMAP(%h_param);
	}
	else
	{
		Carp::croak("Load: format $h_param{format} not in XML, SOMA, BIONANO_RCMAP, XML2, FASTA\n");
	}
	$min_frag_len = &Max($h_param{min_frag_len},$min_frag_len);
	$self->_Set('_min_frag_len',$min_frag_len);
	my $nb = scalar(keys(%$rh_omaps));
	if ( $self->_FastOneGet('_debug') )
	{
		print STDERR "OpticalMap: $nb maps loaded\n";
		print STDERR "OpticalMap: min fragment length=$min_frag_len\n";
		print STDERR "OpticalMap: number of threads =".$self->_FastOneGet('_nb_cpus')."\n";
	}
}

sub GetNumberOfMaps
{
my($self)=@_;
	
	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $nb = scalar(keys(%$rh_omaps));
	return $nb;
}

sub LoadXML
{
my($self,%h_param) = @_;

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $fact = $h_param{fact};
	my @a_lof=`ls $h_param{files}`;
	chomp(@a_lof);
	my $min_frag_len = &MAXINT;
	my $enzyme='';
	for(my $i=0;$i<scalar(@a_lof);$i++)
	{
		my $f_in = &GetStreamIn($a_lof[$i]);
		$a_lof[$i] =~ s/\.\w+$//; # suppression de l'extansion
		$a_lof[$i] = basename($a_lof[$i]);
		my $ok = &FALSE;
		while(my $l=<$f_in>)
		{
			if ( $l =~ /<name>OM_/ )
			{
					$ok = &TRUE;
			}
			if ( $ok )
			{
				if ( $l =~ /<enzymes>(.+)<\/enzymes>/ )
				{
						$enzyme = $1;
				}
				if ( $l =~ /<map_block>(.+)<\/map_block>/ )
				{
					my $totlen = 0;
					my @a_length=split(' ',$1);
					for(my $j=0;$j<scalar(@a_length);$j++)
					{
						my @a_ = ($fact*$a_length[$j],$fact*(0.02*$a_length[$j]+0.2));
						$totlen+=$fact*$a_length[$j];
						$$rh_omaps{$a_lof[$i]}[$j] = \@a_;
						$min_frag_len = $fact*$a_length[$j] if ( $fact*$a_length[$j] < $min_frag_len );
					}
					$ok = &FALSE;
					printf STDERR "$a_lof[$i]: nt=$totlen #frags=".scalar(@a_length)." meanfraglen=%.2f\n",$totlen/scalar(@a_length);
					last;
				}
			}
		}
	}
	print STDERR "min_frag_len = $min_frag_len\n";
	return $min_frag_len;
}

sub LoadXML2
{
my($self,%h_param) = @_;

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $fact = $h_param{fact};
	my @a_lof=`ls $h_param{files}`;
	chomp(@a_lof);
	my $min_frag_len = &MAXINT;
	my $enzyme='';
	my $totlen = 0;
	my $genomelen = 0;
	for(my $i=0;$i<scalar(@a_lof);$i++)
	{
		my $f_in = &GetStreamIn($a_lof[$i]);
		$a_lof[$i] =~ s/\.\w+$//; # suppression de l'extension
		$a_lof[$i] = basename($a_lof[$i]);
		my $ok = &FALSE;
		my $idmap = '';
		while(my $l=<$f_in>)
		{
			if ( $l =~ /RESTRICTION_MAP ID="([^\"]+)"/ )
			{
				$idmap = $1;
				$idmap =~ s/\s/-/g;
				$idmap = "$a_lof[$i]-$idmap";
				$ok = &TRUE;
			}
			if ( $ok )
			{
				if ( $l =~ /ENZYME="(.+)"/i )
				{
						$enzyme = $1;
				}
				if ( $l =~ /<FRAGMENTS/ )
				{
					my $j=0;
					while ( ($l=<$f_in>) && $l !~ /<\/FRAGMENT/ )
					{
						my ($len) = $l =~ /S="(\d+)"/;
						my @a_ = ($len,$fact*(0.02*($len/$fact)+0.2));
						#print join(' ',@a_),"#$fact,$len\n";
						$totlen+=$a_[0];
						$$rh_omaps{$idmap}[$j] = \@a_;
						$min_frag_len = $len if ( $len < $min_frag_len );
						$j++;
					}
					$ok = &FALSE;
					printf STDERR "$idmap: nt=$totlen #frags=$j meanfraglen=%.2f\n",$totlen/$j;
					$genomelen+=$totlen;
					$totlen=0;
				}
			}
		}
	}
	print STDERR "Genome size: $genomelen nt\n";
	return $min_frag_len;
}

sub LoadBIONANO_RCMAP
{
my($self,%h_param) = @_;

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $fact = $h_param{fact};
	my @a_lof=`ls $h_param{files}`;
	chomp(@a_lof);
	my $delta=0;
	my $min_frag_len = &MAXINT;
	my $totlen=0;
	for(my $i=0;$i<scalar(@a_lof);$i++)
	{
	#	print "$$ra_lof[$i]\n";
		my $f_in = &GetStreamIn($a_lof[$i]);
		$a_lof[$i] =~ s/\.\w+$//; # suppression de l'extension
		$a_lof[$i] = basename($a_lof[$i]);
		my @a_ = ();
		$$rh_omaps{$a_lof[$i]} = \@a_;
		my $cpt=0;
		my $prev=0;
		while(my $l=<$f_in>)
		{
			chomp($l);
			next if ( $l =~ /^#/ );
#   988  741983.2        52      1       1       20.0    0.0     12      12
			$l =~ s/^\s+//;
			my @a_=split(' ',$l);
			$a_[5] =~ s/\.\d*//;
			$delta = $a_[5] - $prev;
			$prev = $a_[5];
			next if ( $delta < 50 ); # il y a des tres petits sites aux extremites
			@a_ = ($delta,$fact*(0.02*($delta/$fact)+0.2));
			$totlen+=$delta;
			$$rh_omaps{$a_lof[$i]}[$cpt++] = \@a_;
			$min_frag_len = $delta if ( $delta < $min_frag_len );
		}
		$f_in->close;
	}
	print STDERR "Genome size: $totlen nt\n";
	return $min_frag_len;
}


sub LoadSOMA
{
my($self,%h_param) = @_;

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $fact = $h_param{fact};
	my @a_lof=`ls $h_param{files}`;
	chomp(@a_lof);
	my $min_frag_len = &MAXINT;
	for(my $i=0;$i<scalar(@a_lof);$i++)
	{
	#	print "$$ra_lof[$i]\n";
		my $f_in = &GetStreamIn($a_lof[$i]);
		$a_lof[$i] =~ s/\.\w+$//; # suppression de l'extansion
		$a_lof[$i] = basename($a_lof[$i]);
		my @a_ = ();
		$$rh_omaps{$a_lof[$i]} = \@a_;
		my $cpt=0;
		while(my $l=<$f_in>)
		{
			chomp($l);
			my @a_ = split(' ',$l);
			$a_[0] *= $fact;
			$a_[1] *= $fact;
			next if ( (2*$a_[1]) > $a_[0] ); # je ne prend pas en compte les fragments dont la taille est inferieure a 2*sd
			$min_frag_len = $a_[0] if ( $a_[0] < $min_frag_len );
			$$rh_omaps{$a_lof[$i]}[$cpt++] = \@a_;
		}
		$f_in->close;
	}
	return $min_frag_len;
}


sub LoadFASTA
{
my($self,%h_param) = @_;

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $fact = $h_param{fact};
	my @a_lof=`ls $h_param{files}`;
	chomp(@a_lof);
	my $min_frag_len = 1000;
	my %h_seq = ();
	if ( ! scalar(@a_lof) )
	{
		Carp::croak("$0 ERROR: unable to find file(s) using >$h_param{files}<\n");
	}
	for(my $i=0;$i<scalar(@a_lof);$i++)
	{
		&FastaToHash($a_lof[$i],\%h_seq);
	}

	foreach my $id ( sort keys(%h_seq) )
	{
		next if ( $h_seq{$id}{len} < $self->_FastOneGet('_min_scaffold_len') );
		$self->CutSequence(\$h_seq{$id}{sequence});
		my $rh_silico = $self->_FastOneGet('_rh_silico');
		my $ra_sites = $$rh_silico{forward};
		for(my $i=0;$i<scalar(@$ra_sites); $i++)
		{
			my @a_ = ($$ra_sites[$i],(0.02*$$ra_sites[$i]+0.2));
			$$rh_omaps{$id}[$i] = \@a_;
		}
		$rh_silico = undef;
	}
	return $min_frag_len;
}



sub MapSequence
{
my($self,$idseq,$r_seq,$format,$max_nb_hits,$nooverlappinghits)=@_;

	my $min_scaffold_len = $self->_FastOneGet('_min_scaffold_len');
	my $min_scaffold_span = $self->_FastOneGet('_min_scaffold_span');
	my $min_scaffold_hitlen = $self->_FastOneGet('_min_scaffold_hitlen');
	my $min_density = $self->_FastOneGet('_min_density');
	my $min_sites = $self->_FastOneGet('_min_sites');
	my $rh_dejavu   = $self->_FastOneGet('_rh_memo');

	return () if ( length($$r_seq)  < $min_scaffold_len );

	$format = 'short' if (! defined($format) );

	my $digest = md5_hex("$$r_seq#$format#$max_nb_hits#");
	if ( defined($$rh_dejavu{$digest}) )
	{
		return @{$$rh_dejavu{$digest}};
	}


	$self->CutSequence($r_seq);

	my $rh_omaps = $self->_FastOneGet('_rh_omap');
	my $ra_hits  = $self->_FastOneGet('_ra_hits');
	@$ra_hits = ();
	my $rh_silico = $self->_FastOneGet('_rh_silico');
	my $current_sd = $self->_FastOneGet('_sd');
	my $gap_score = $self->_FastOneGet('_gap_score');
	my $match_score = $self->_FastOneGet('_match_score');

	my $seq_sites = scalar(@{$$rh_silico{forward}});
	$seq_sites = 1 if ( ! $seq_sites );
	$self->_Print(sprintf("OpticalMap: mapping $idseq len=".length($$r_seq)." #frags=$seq_sites meanfraglen=%.2f\n",length($$r_seq)/$seq_sites)) if ( $self->_FastOneGet('_debug') );

	# build the list of atomic computations
	
	my %sh_results :shared = ();

	foreach my $omap (keys(%$rh_omaps))
	{
		foreach my $orient ('forward','reverse')
		{
			my $key = "$omap $orient";
			$sh_results{$key}='';
		}
	}

	my @a_keys = keys(%sh_results);


	my %h_threads = ();
	my $nb_proc = $self->_FastOneGet('_nb_cpus');
	$nb_proc = &Min(scalar(@a_keys),$nb_proc);
	my $step   = int(scalar(@a_keys)/$nb_proc);
	$step++ if ( int(scalar(@a_keys)/$nb_proc) != scalar(@a_keys)/$nb_proc );
	for (my $i=0;$i<scalar(@a_keys);$i+=$step)
	{
		my $key = $a_keys[$i];
		my ($omap,$orient) = split(' ',$key);
		my $imax = &Min(scalar(@a_keys)-1,$i+$step-1);
		if ( $nb_proc > 1 )
		{
			$h_threads{$i} = threads->create(\&_MatchThread,$max_nb_hits,\@a_keys,$i,$imax,\%sh_results,$idseq,$min_sites,$min_scaffold_len,$min_scaffold_span,$min_density,$min_scaffold_hitlen,$current_sd,$gap_score,$match_score,$rh_omaps,$rh_silico);
		}
		else
		{
			                                  &_MatchThread($max_nb_hits,\@a_keys,$i,$imax,\%sh_results,$idseq,$min_sites,$min_scaffold_len,$min_scaffold_span,$min_density,$min_scaffold_hitlen,$current_sd,$gap_score,$match_score,$rh_omaps,$rh_silico);
		}
	}
	for (my $i=0;$i<scalar(@a_keys);$i+=$step)
	{
		next if ( $nb_proc <= 1 );
		$h_threads{$i}->join();
		if ( my $err = $h_threads{$i}->error() )
		{
			Carp::croak("MapSequence: thread error >$err<\n");
		}
	}

	# collect results

	foreach my $key (@a_keys)
	{
		$sh_results{$key}=~s/###$//;
		push(@$ra_hits,split(/###/,$sh_results{$key})) if ( $sh_results{$key} ne '' );
	}

	my @a_nrhits = ();

	my $cpt=0;

	foreach my $hit (sort {my @A=split(/\t/,$a); my @B=split(/\t/,$b); return $B[16] <=> $A[16]; } @$ra_hits)
	{
		my ($idmap,$nb_sites,$om_start,$om_end,$omlen,$om_pos_start,$om_pos_end,$strand,$idseq,$seq_sites,$seq_start,$seq_stop,$seqlen,$seq_pos_start,$seq_pos_end,$ali_len,$score,$densite,$span) = split(/\t/,$hit);
		my $overlap = &FALSE;
		if ( $nooverlappinghits )
		{
			# je vais au plus rapide meme si la complexite n'est pas bonne
			for(my $i=0;$i<scalar(@a_nrhits);$i++)
			{
				my ($idmap2,$nb_sites2,$om_start2,$om_end2,$omlen2,$om_pos_start2,$om_pos_end2,$strand2,$idseq2,$seq_sites2,$seq_start2,$seq_stop2,$seqlen2,$seq_pos_start2,$seq_pos_end2,$ali_len2,$score2,$densite2,$span2) = split(/\t/,$a_nrhits[$i]);
				next if ( $score >= $score2 ); # si les meilleurs de score max chevauchent je veux les garder
				#print STDERR "$seq_pos_start >= ($seq_pos_start2-$min_scaffold_len/2) && $seq_pos_end <= ($seq_pos_end2+$min_scaffold_len/2)\n";
				if ( $seq_pos_start >= ($seq_pos_start2-$min_scaffold_len/2) && $seq_pos_end <= ($seq_pos_end2+$min_scaffold_len/2) )
				{
					#print STDERR "Overlap found: $seq_pos_start >= ($seq_pos_start2-$min_scaffold_len/2) && $seq_pos_end <= ($seq_pos_end2+$min_scaffold_len/2)\n";
					#print STDERR "$hit\n";
					$overlap=&TRUE 
				}
			}
		}
		next if ( $overlap );
		$cpt++;
		if ( !defined($format) || $format ne 'full' )
		{
			push(@a_nrhits,"$idmap\t$seq_pos_start\t$seq_pos_end\t$om_pos_start\t$om_pos_end\t$strand");
		}
		else
		{
			push(@a_nrhits,$hit);
		}
		last if ($cpt >= $max_nb_hits );
	}

	$$rh_dejavu{$digest}=\@a_nrhits;

	return @a_nrhits;
}

sub _MatchThread
{
my($max_nb_hits,$ra_keys,$imin,$imax,$rsh_results,$idseq,$min_sites,$min_scaffold_len,$min_scaffold_span,$min_density,$min_scaffold_hitlen,$current_sd,$gap_score,$match_score,$rh_omaps,$rh_silico) = @_;

	#print STDERR "THREAD: $imin <= i <= $imax\n";
	for(my $i=$imin;$i<=$imax;$i++)
	{
		my $key=$$ra_keys[$i];
		my ($omap,$orient) = split(' ',$key);
		print STDERR "pid=$$ $key ($omap,$orient)\n" if ( $DEBUGJG );
		my $idmap = $omap;
		my $nb_sites = scalar(@{$$rh_omaps{$omap}});
		my $omlen = &_FastOneGetOMPosition($$rh_omaps{$omap},$nb_sites-1);
		my $seq_sites = scalar(@{$$rh_silico{$orient}});
		my $seqlen = &_FastOneGetSeqPosition($$rh_silico{$orient},$seq_sites);
		my $fini = &FALSE;
		my @a_interval = ("0,".($seq_sites-1));
		my $str_sites = 0 x $seq_sites;
		my $nb_hits;
		while(! $fini)
		{
			my $found = &FALSE;
			foreach my $interval (@a_interval)
			{
				my ($minseq,$maxseq)=split(',',$interval);
				#print STDERR ">$orient $interval<\n";
				my($score,$seq_start,$seq_stop,$om_start,$om_end) = &_Match($current_sd,$gap_score,$match_score,$$rh_omaps{$omap},$$rh_silico{$orient},$minseq,$maxseq);
				my $hitlen = $seq_stop-$seq_start+1;
				my $hit = 1 x $hitlen;
				substr($str_sites,$seq_start,$hitlen,$hit);
				#print STDERR "$orient $score,$seq_start,$seq_stop,$om_start,$om_end,$str_sites,$minseq,$maxseq\n";
				
				# pour l'affichage je passe en coordonnees sur le brin + de la sequence
				my $strand='+';
				if ($orient eq 'reverse' )
				{
					my $tmp    = $seq_stop;
					$seq_stop  = ($seq_sites-1) - $seq_start;
					$seq_start = ($seq_sites-1) - $tmp;
					$strand    = '-';
				}
				my($seq_pos_start) = &_FastOneGetSeqPosition($$rh_silico{forward},$seq_start);
				my($seq_pos_end)   = &_FastOneGetSeqPosition($$rh_silico{forward},$seq_stop);

				my($om_pos_start)  = &_FastOneGetOMPosition($$rh_omaps{$omap},$om_start);
				my($om_pos_end)    = &_FastOneGetOMPosition($$rh_omaps{$omap},$om_end);
				#print STDERR "####$seq_pos_start,$seq_pos_end $om_pos_start,$om_pos_end  $om_start,$om_end\n";
				my $ali_len = $om_end-$om_start+1;
				next if ( $ali_len == 0 || $seqlen == 0 );
				my $densite = sprintf("%.04f",$score/$ali_len);
				my $pcs = int(($seq_pos_end-$seq_pos_start)*100/$seqlen);
				#print STDERR "$imin/$imax $pcs (($seq_pos_end-$seq_pos_start)*100/$seqlen) >= $min_scaffold_span && $densite >= $min_density )\n";
				next if ( ($seq_pos_end-$seq_pos_start) < $min_scaffold_hitlen );
				if ( (($seq_pos_end-$seq_pos_start)*100/$seqlen) >= $min_scaffold_span && $densite >= $min_density && $ali_len >= $min_sites)
				{
					$seq_start++;
					$seq_stop++;
					$om_start++;
					$om_end++;
					$seq_pos_start++;
					$seq_pos_end++;
					$om_pos_start++;
					$om_pos_end++;
					$$rsh_results{$key} .= "$idmap\t$nb_sites\t$om_start\t$om_end\t$omlen\t$om_pos_start\t$om_pos_end\t$strand\t$idseq\t$seq_sites\t$seq_start\t$seq_stop\t$seqlen\t$seq_pos_start\t$seq_pos_end\t$ali_len\t$score\t$densite\t$pcs###";
					$nb_hits++;
					$found=&TRUE;
				}
			}
			last if ( ! $found );
			last if ( $nb_hits >= $max_nb_hits );

			# calcule les fragments pour lesquels il n'y a pas encore de hits
			@a_interval = ();
			while($str_sites=~ /(0{$min_sites,})/g)
			{
				my $imin = length($`);
				my $imax = $imin+length($1)-1;
				my($seq_pos_start) = &_FastOneGetSeqPosition($$rh_silico{$orient},$imin);
				my($seq_pos_end)   = &_FastOneGetSeqPosition($$rh_silico{$orient},$imax);
				next if ( ($seq_pos_end-$seq_pos_start+1) < $min_scaffold_len );
				#print STDERR "$imin-$imax=>$seq_pos_start-$seq_pos_end\n";
				push(@a_interval,"$imin,$imax");
			}
			last if ( scalar(@a_interval) == 0 );
			#print STDERR "$orient\n";
			#print STDERR Dumper(@a_interval);
		}
	}
}


# pour le moment je compte comme SOMA meme si apres je calcule la taille des fragments
sub CutSequence
{
my($self,$r_seq)=@_;

	my $min_frag_len = $self->_FastOneGet('_min_frag_len');
	my $rh_silico    = $self->_FastOneGet('_rh_silico');
	my $enz          = $self->_FastOneGet('_enzyme_site');
	my $palendromic  = $self->_FastOneGet('_palendromic');

#	print STDERR "genome: $min_frag_len\n" if ( $DEBUGJG );

	my $seq = "\U".$$r_seq;
	my @a_pos = (); 
	my $pos = 0;
	while(($pos = index($seq, $enz, $pos+1)) != -1) 	# je calcule comme SOMA les positions des sites
	{
		push(@a_pos, $pos+1);
	}
	if( ! $palendromic )
	{
		my $enzrc = &RevCompSeq(\$enz);
		my $pos = 0;
		while(($pos = index($seq, $enzrc, $pos+1)) != -1) 	
		{
			push(@a_pos, $pos+1);
		}
		my @a_sort_pos = sort { $a <=> $b } @a_pos;
		@a_pos = @a_sort_pos;
	}

	for(my $i=$#a_pos;$i>=1;$i--)	 # et apres je calule les tailles, c'est debile mais je garde le format SOMA pour le moment
	{
		$a_pos[$i] -= $a_pos[$i-1];
	}

	# je merge les fragments qui sont inferieurs a la taille min (fragment dont la taille est > 2*sd) observee en lisant l'OM
	# le dernier fragment peut etre < a la taille min mais bon dans ce cas on ne peut rien dire de toute facon)
	my @a_merged = ();
	my $size = 0;
	for(my $i=0;$i<scalar(@a_pos);$i++)
	{
		$size=$a_pos[$i];
		while($i<$#a_pos && $size<$min_frag_len)
		{
			$i++;
			$size += $a_pos[$i];
		}
#		print STDERR "$size\n" if ( $DEBUGJG );
		push(@a_merged,$size);
	}

	$$rh_silico{len} = length($seq);
	$$rh_silico{forward}= \@a_merged;
	my @a_rev = reverse(@a_merged);
	$$rh_silico{reverse}= \@a_rev;
}

sub _FastOneGetMax
{
my($min_seq,$max_seq,$min_om,$max_om,$ra_mat)=@_;

	my @a_max = ();
	my $max = 0;
	for(my $i=$min_seq;$i<=$max_seq;$i++)
	{
		for (my $j=$min_om;$j<=$max_om;$j++)
		{
			if ( $$ra_mat[$i][$j] == $max )
			{
				push(@a_max,"$i,$j");
			}
			if ( $$ra_mat[$i][$j] > $max )
			{
				$max = $$ra_mat[$i][$j];
				@a_max = ("$i,$j");
			}
		}
	}
	return ($max,@a_max);
}

sub _FastOneGetSeqPosition
{
my($ra_seq,$pos)=@_;

	my $spos = 0;
	for (my $i=0;$i<$pos;$i++)
	{
		$spos += $$ra_seq[$i];
	}
	return $spos;
}

sub _FastOneGetOMPosition
{
my($ra_omap,$pos)=@_;

	my $mpos = 0;
	for (my $i=0;$i<$pos;$i++)
	{
		$mpos += $$ra_omap[$i][0];
	}
	return $mpos;
}


sub _Match
{
my($current_sd,$gap_score,$match_score,$ra_omap,$ra_seq,$minseq,$maxseq) = @_;

	# http://www.biorecipes.com/DynProgBasic/code.html
	# 1ere ligne et 1ere colonne reserve et a 0

	# i = sequence
	# j = omap

#	print STDERR "gap_score=$gap_score\n" if ( $DEBUGJG );

	my @a_mat = ();

	my $nb_seq = $#$ra_seq+1;
	my $nb_om = $#$ra_omap+1;

	for(my $i=0;$i<=$nb_seq;$i++)
	{
		for (my $j=0;$j<=$nb_om;$j++)
		{
			$a_mat[$i][$j]=0;
		}
	}

	# print "nb_seq=$nb_seq, nb_om=$nb_om\n";

	my %h_sim = ();
	for (my $i=$minseq, my $ii=$minseq+1;$i<=$maxseq;$i++,$ii++)
	{
		for(my $j=0, my $jj=1;$j<=$#$ra_omap;$j++,$jj++)
		{
			my ($omlen,$omsd)=($$ra_omap[$j][0],$$ra_omap[$j][1]);
			# il y a des soucis d'arrondis (en principe regles), d'ou sprintf pour remplir la matrice et comparaison en numerique pour rechercher le chemin
			my $pos_score = sprintf("%.04f",((-1.00 * $match_score)*1));
			for(my $sd=1;$sd<=$current_sd;$sd++)
			{
				if ( $$ra_seq[$i] > ($omlen-$sd*$omsd) && $$ra_seq[$i] < ($omlen+$sd*$omsd) )
				{
					$pos_score = sprintf("%.04f",$match_score-(0.01*$match_score)*$sd+0.01*$match_score); # =2-(0.02*B3)+0.02
					last;
				}
			}
			my $score = &Max($a_mat[$ii-1][$jj-1]+$pos_score,&Max($a_mat[$ii-1][$jj]+$gap_score,$a_mat[$ii][$jj-1]+$gap_score));
			$score = 0 if ( $score < 0 );
			$a_mat[$ii][$jj]=sprintf("%.04f",$score);
			$h_sim{"$ii,$jj"}=$pos_score if ( $pos_score > 0 );
		}
	}
	
	#print STDERR "$minseq,$maxseq 0,$nb_seq\n";
	my @a_int = split(',',"$minseq,$maxseq,0,$nb_om");

	my ($max,@a_max) = &_FastOneGetMax(@a_int,\@a_mat);

#	printf STDERR "MAX score $max at %d,%d   ($current_sd,$gap_score,$match_score,$minseq,$maxseq)\n",@a_max;

	my($starti,$startj)=();
	for ( my $i=0; $i <= 0 ; $i++ )	# je fais une boucle au cas ou un jour je traite plusieurs max
	{
		($starti,$startj) = split(',',$a_max[$i]);
	}

	my $ii = $starti;
	my $jj = $startj;
	while ( $ii > 0 && $jj > 0 )	# au cas ou je remonte dans la diagonale tant que le score est identique au max
	{
		last if ($a_mat[$ii][$jj] != $a_mat[$ii-1][$jj-1] );
		$ii--;
		$jj--;
	}
	$starti = $ii;
	$startj = $jj;
	my $lasti=$ii;
	my $lastj=$jj;
	while ( $ii > 0 && $jj > 0 )
	{
		last if ( $a_mat[$ii][$jj] == 0 );
		$lasti = $ii;
		$lastj = $jj;
		my $isupmatch = (defined($h_sim{"$ii,$jj"})) ? $h_sim{"$ii,$jj"} : (-1 * $match_score);
		#print "$isupmatch, $a_mat[$ii][$jj] $a_mat[$ii-1][$jj-1]\n";
		my $delta_match = sprintf("%.04f",$a_mat[$ii][$jj]-$isupmatch);
		my $delta_gap   = sprintf("%.04f",$a_mat[$ii][$jj]-$gap_score);
		if ( $delta_match == $a_mat[$ii-1][$jj-1] ) # si je compare en numerique j'ai des problemes sans doute a cause d'arrondis
		{
			$ii--;
			$jj--;
		}
		elsif ( $delta_gap == $a_mat[$ii][$jj-1])
		{
			$jj--;
		}
		elsif ( $delta_gap == $a_mat[$ii-1][$jj])
		{
			$ii--;
		}
		else
		{
			printf STDERR "%s %s %s - %s == %s\n", $delta_match, $delta_gap, $a_mat[$ii][$jj],$isupmatch,$a_mat[$ii-1][$jj-1];
			die "$0 FATAL ERROR $ii, $jj $starti,$startj #($a_mat[$ii][$jj]-$isupmatch)#==#$a_mat[$ii-1][$jj-1]#\n";
		}
	}
	#&PrintMat($nb_seq,$nb_om,\@a_mat,$lasti,$starti,$lastj,$startj,$ra_omap,$ra_seq);
	return ($max,--$lasti,--$starti,--$lastj,--$startj);
}


sub PrintMat
{
my($nb_seq,$nb_om,$ra_mat,$is,$ie,$js,$je,$ra_omap,$ra_seq)=@_;

	$is = 0 if ( ! defined($is) );
	$js = 0 if ( ! defined($js) );
	$ie = $nb_seq if ( ! defined($ie) );
	$je = $nb_om  if ( ! defined($je) );

	printf STDERR "%d, %d, $js, $je $nb_seq,$nb_om,$ra_mat,$is,$ie,$js,$je,$ra_omap,$ra_seq\n",&Max($js-1,0),&Min($je+1,$nb_om-1);
	print STDERR "\t\t";
	for (my $j=&Max($js-1,0);$j<=&Min($je+1,$nb_om-1);$j++)
	{
		#printf STDERR " %-5s",$j;
		printf STDERR "\t%s",$j;
	}
	print STDERR "\n\t\t";
	for (my $j=&Max($js-1,0);$j<=&Min($je+1,$nb_om-1);$j++)
	{
		#printf STDERR " %-5s",$$ra_omap[$j-1][0];
		printf STDERR "\t%s",$$ra_omap[$j-1][0] if ( $j >= 1 );
	}
	print STDERR "\n";
	for(my $i=$is;$i<=$ie;$i++)
	{
		#printf STDERR " %-5s $$ra_seq[$i-1]\t",$i;
		printf STDERR "\t%s\t$$ra_seq[$i-1]\t",$i if ( $i >= 1 );
		for (my $j=$js;$j<=$je;$j++)
		{
			#printf STDERR " %-5s",$$ra_mat[$i][$j];
			printf STDERR "\t%s",$$ra_mat[$i][$j];
		}
		print STDERR "\n";
	}
}


sub AssertParams
{
my ($self,$rh_param,$ra_expected) = @_;

        foreach my $param (@{$ra_expected})
        {
                if ( !defined($$rh_param{$param}) )
                {
                        print STDERR Dumper(%$rh_param);
                        Carp::croak("Fatal error in constructor: $param must be provided\n") 
                }
        }
}


sub _Print
{
my($self,$msg)=@_;

	print STDERR "$msg" if ( defined($$self{'_debug'}) );
}


1;
