#
# $Id: RamBdbHash.pm 246 2014-01-03 09:43:50Z gouzy $
#

package RamBdbHash;

use strict;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use BerkeleyDB;

use File::Temp qw/ tempfile tempdir /;


# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 246 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
	my ($self,%h_param) = @_;

	#$self->AssertParams(\%h_param,\@a_expected);

	$self->SUPER::_Init();

	foreach my $param (keys(%h_param))
	{
		$self -> _Set ("_$param", $h_param{$param});
	}

	if ( ! defined($ENV{TMPDIR}) )
	{
			Carp::croak("Error: TMPDIR is required for using BerkeleyDB storage\n");
	}
	my $ftmp = tempdir ( TEMPLATE => 'RamBdbHash-XXXX', DIR => $ENV{TMPDIR}, CLEANUP=> 0);
	my $dir = $ftmp;

	my $cache = $h_param{cache} || 1000000000;

	my $env = new BerkeleyDB::Env 
		-Cachesize => $cache,
		-TxMax     => 1000,
		-MaxLocks  => 10000,
		-MaxObjects => 10000,
		-Home   => $dir,
		# rajouter DB_THREAD est tres problematique
		-Flags  => DB_CREATE| DB_INIT_CDB | DB_INIT_MPOOL or die "cannot open environment: $BerkeleyDB::Error\n";

	my %h_db=();

	my ($fseq) = File::Temp->new( TEMPLATE => 'RamBdbHash-XXXX', DIR => $dir, SUFFIX => '.db');
	$fseq = $fseq->filename;

	my $o_bdb = new BerkeleyDB::Hash(
		-Flags  => DB_CREATE,
		-Env       => $env,
		-Filename  => $fseq) or die "ERROR, unable to create >$fseq\n";

	my %h_hash=();

	$self->_Set('_o_bdb', $o_bdb);
	$self->_Set('_rh_hash', \%h_hash);
}

sub AssertParams
{
my ($self,$rh_param,$ra_expected) = @_;
	
	foreach my $param (@{$ra_expected})
	{
		if ( !defined($$rh_param{$param}) )
		{
			print STDERR Dumper(%$rh_param);
			Carp::croak("Fatal error in constructor: $param must be provided\n") 
		}
	}
}


sub Add
{
my($self,$key,$r_seq) = @_;

	$self->Set($key,$$r_seq);
}


sub Reset
{
my($self)=@_;

	my $o_bdb = $self->_FastOneGet('_o_bdb');
	my ($k, $v) = ("", "") ;
	my $cursor = $o_bdb->db_cursor() ;
	while ($cursor->c_get($k, $v, DB_NEXT) == 0)
	{
		$o_bdb->db_del($k);
	}
	undef $cursor ;
	my $rh_hash = $self->_FastOneGet('_rh_hash');
	%{$rh_hash}=();
}

# je regarde d'abord dans le hash puis dans la bdb

sub Get
{
my($self,$key) = @_;

	my $val = $self->{_rh_hash}->{$key};
	if ( ! defined($val) )
	{
		my $o_bdb = $$self{'_o_bdb'};
		$o_bdb->db_get($key,$val);
	}
	return $val;
}

sub Exists
{
my($self,$key) = @_;

	if ( ! exists($self->{_rh_hash}->{$key}) )
	{
		my $o_bdb = $$self{'_o_bdb'};
		my $status = ( $o_bdb->db_exists($key) eq '' ) ? 1 : 0; # c'est bizarre le status est '' lorsqu'il trouve
		return $status;
	}
	else
	{
		return 1;
	}
}


# si la cle du hash existe je l'utilise sinon bdb par defaut

sub Set
{
my($self,$key,$val) = @_;

	if ( exists($self->{_rh_hash}->{$key}) )
	{
		$self->{_rh_hash}->{$key} = $val;
	}
	else
	{
		my $o_bdb = $$self{'_o_bdb'};
		$o_bdb->db_put($key,$val);
	}
	return 1;
}

# si la cle du hash existe je l'utilise sinon
# je recupere la valeur dans la bdb et je cree la cle dans le hash

sub ConcatNolock
{
my($self,$key,$val) = @_;

	if ( exists($self->{_rh_hash}->{$key}) )
	{
		$self->{_rh_hash}->{$key} .= $val;
	}
	else
	{
		my $prev;
		my $o_bdb = $$self{'_o_bdb'};
		$o_bdb->db_get($key,$prev);
		if ( ! defined($prev) )
		{
			$self->Set($key,$val);
		}
		else
		{
			$self->Del($key);
			$self->{_rh_hash}->{$key} = "$prev$val";
		}
	}
	#return 1;
	return;
}

sub Concat
{
my($self,$key,$val) = @_;

	$self->ConcatNolock($key,$val);
}

# ex: chargement d'une banque HD

sub ConcatFast
{
my($self,$key,$val) = @_;

	$self->{_rh_hash}->{$key} .= $val;
}

# si la cle du hash existe sinon bdb pour limiter les acces a bdb

sub Del
{
my($self,$key) = @_;

	if ( exists($self->{_rh_hash}->{$key}) )
	{
		delete($self->{_rh_hash}->{$key});
	}
	else
	{
		my $o_bdb = $$self{'_o_bdb'};
		$o_bdb->db_del($key);
	}
	return 1;
}


sub GetKeys
{
my($self,$which) = @_;

	my @a_keys = ();

	if (! defined($which) || $which =~ /hash/i )
	{
		my $rh_hash = $self->_FastOneGet('_rh_hash');
		@a_keys=keys(%{$rh_hash});
	}
	#print STDERR scalar(@a_keys)." keys in hash\n";

	if (! defined($which) || $which =~ /bdb/i )
	{
		my $o_bdb  = $self->_FastOneGet('_o_bdb');
		my ($k, $v) = ("", "") ;
		my $cursor = $o_bdb->db_cursor() ;
		while ($cursor->c_get($k, $v, DB_NEXT) == 0)
		{ 
			push(@a_keys,$k);
		}
		undef $cursor ;
	}
	#print STDERR scalar(@a_keys)." keys in total\n";
	return @a_keys;
}



1;
