#
# $Id: Scaffold.pm 519 2016-08-26 05:40:19Z gouzy $
#

# Copyright INRA
# Jerome.Gouzy@toulouse.inra.fr

package Scaffold;

use strict;
use General;
use GeneralBioinfo;
use base qw( LipmObject );
use Data::Dumper;
use warnings;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use lib "../lib-dev";
use Contig;
use ContigRegistry;

use constant MAXVALIDATION_LENGTH => 200000;
use constant MAXVALIDATION_STEPS  => 5;
use constant LARGE_INSERT_LENGTH  => 30000;
use constant MAX_DISTANCE         => 200000; # MicroScaffoling ValidateScaffoldsSpan
use constant MAXGAPLENSTEPS => 5;
use constant ANCHOR_MIN_CTGLEN => 500;
use constant ALIGN_MAXLEN => 1000;

my $OVL_MARGIN= 0.95; # % de la longueur

my $DEBUG_LINK_PATTERN = 'TTOO'; # cf constructeur
my $GAPLEN_MASKABLE    = 2000; # pour le tagging weak avec gaplen vs insert_size

my %H_SCAFF2EXT    = ( "p p" => "3 5", "p m" => "3 3", "m p" => "5 5", "m m" => "5 3" );
my %H_CTG2SCAFFEXT = ( "p 5" => 5, "p 3" => 3, "m 5" => 3, "m 3" => 5 );
my %H_SCAFFEXT2CTG = %H_CTG2SCAFFEXT;
my $SEP = '###';

use constant DEFAULT_OVLPCI => 95;
use constant DEFAULT_OVLLEN => 30;
use constant DEFAULT_OVLOVH => 10;

use Alnpack;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 519 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, initialize the object, do not use directly, use New method instead

=cut

sub _Init
{
my ($self,%h_param) = @_;

	my @a_expected = ('name','maxvalidationlength');
	$self->AssertParams(\%h_param,\@a_expected);

	$self->SUPER::_Init();

	my @a_ = ();
	$self->_Set('_ra_path', []); # un chemin contient des contigs et des gaps, c'est un array mais ca foire si j'initialise avec ()
	$self->_Set('_anchor', undef);	# contrairement a contig, un scaffold ne peut avoir qu'une ancre
	$self->_Set('_name', undef);
	$self->_Set('_anchor_min_len', &ANCHOR_MIN_CTGLEN);
	$self->_Set('_maxvalidationlength',&MAXVALIDATION_LENGTH);

	foreach my $param (keys(%h_param))
	{
		$self->_Set("_$param", $h_param{$param});
	}

	$self->ResetGetSequenceRules(\%h_param);

	$self->_Set('_rh_cache_overlaps', {});
}

sub ResetGetSequenceRules
{
my($self,$rh_param)=@_;

	$self->_Set('_minovlpci',&DEFAULT_OVLPCI) if ( !exists($$rh_param{minovlpci}));
	$self->_Set('_minovllen',&DEFAULT_OVLLEN) if ( !exists($$rh_param{minovllen}));
	$self->_Set('_maxovlovh',&DEFAULT_OVLOVH) if ( !exists($$rh_param{maxovlovh}));
	$self->_Set('_fast_getsequence',&TRUE);
}

# rajoute une liste d'element de type string (pour les gaps) et Contig pour les contigs

# TEST OK
sub Push
{
my($self,@a_elements)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	my $name = $self->GetName();
	for(my $i=0;$i<scalar(@a_elements);$i++)
	{
		next if ( ref($a_elements[$i]) ne 'Contig' );
		$a_elements[$i]->SetScaffold($self,$name);
	}
	push(@$ra_path,@a_elements);
	$self->_Set('_ra_path',$ra_path);	# je fais ca pour la premiere fois ou il n'est pas initialise
	$self->CheckFormat();
}

# TEST ========== NOK
sub Pop
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return undef if ( !defined($ra_path) );
	my $contig = pop(@$ra_path);
	my $gap = pop(@$ra_path);
	$self->_Set('_ra_path',$ra_path);	# je fais ca pour la premiere fois ou il n'est pas initialise
	$self->CheckFormat();
	return ($contig,$gap);
}

# rajoute au debut du tableau une liste d'element de type string (pour les gaps) et Contig pour les contigs
# TEST OK
sub Unshift
{
my($self,@a_elements)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	my $name = $self->GetName();
	for(my $i=0;$i<scalar(@a_elements);$i++)
	{
		next if ( ref($a_elements[$i]) ne 'Contig' );
		$a_elements[$i]->SetScaffold($self,$name);
	}
	my @a_newpath = (@a_elements,@$ra_path);
	$self->_Set('_ra_path',\@a_newpath);
	$self->CheckFormat();
}

# TEST =========== NOK
sub Shift
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return undef if ( !defined($ra_path) );
	my $contig = shift(@$ra_path);
	my $gap = shift(@$ra_path);
	$self->CheckFormat();
	return ($contig,$gap);
}


sub Last
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return undef if ( !defined($ra_path) );
	return $$ra_path[scalar(@$ra_path)-1]->GetArray();  # renvoi name,strand,len
}

sub First
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return undef if ( !defined($ra_path) );
	return $$ra_path[0]->GetArray();
}


# TEST OK
sub CheckFormat
{
my($self)=@_;

	my $ra_path = $self->GetArray();
	for(my $i=0;$i<scalar(@$ra_path);$i++)	# contig gap contig
	{
		if ( $i % 2 == 0 )
		{
			if ( ref($$ra_path[$i]) ne 'Contig' )
			{
				print STDERR Dumper($ra_path);
				print STDERR "$i ".$self->Get()."\n";
				Carp::confess("PB contig format: a Contig is expected at position $i, found >$$ra_path[$i]<\n");
			}
			elsif ( $$ra_path[$i]->Get() !~ /^#\S+[pm]\d+/ )
			{
				print STDERR Dumper($ra_path);
				print STDERR "$i ".$self->Get()."\n";
				Carp::confess("PB contig format at position $i ".$self->Get()." >".$$ra_path[$i]->Get()."<\n");
			}
		}
		elsif ($$ra_path[$i] !~ /^[-\d]+,[-\d]+$/ )
		{
			#print STDERR Dumper($ra_path);
			Carp::confess("PB gap format at position $i in ".$self->Get()." found=>$$ra_path[$i]<\n")    
		}
	}
}


sub Reset
{
my($self)=@_;

	$self->ResetPath();
	$self->_Set('_anchor', undef);
}

# meme si j'undef le path, les contigs existent toujours car le ContigRegistry (meme s'ils ne pointent plus vers le scaff)

sub ResetPath
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return if ( ! defined($ra_path) );
	for(my $i=0;$i<scalar(@$ra_path)-1;$i+=2)	# contig gap contig
	{
		next if (! defined($$ra_path[$i]) ); # ca ne doit pas arriver
		$$ra_path[$i]->DelScaffold($self); # je desenregistre le scaffold dans le contig
	}
	$self->_Set('_ra_path',undef);
}

sub DeregisterContigs
{
my($self,$imin,$imax)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	if (! defined($ra_path) )
	{
		print STDERR "DeregisterContigs: on an empty path scaffid=".$self->GetName()."\n";
		print STDERR Dumper($self);
		return;
	}
	$imin = 0                   if ( !defined($imin) );
	$imax = scalar(@$ra_path)-1 if ( !defined($imax) );
	if ( $imin < 0 || $imax > scalar(@$ra_path)-1 )
	{
		Carp::confess("DeregisterContigs error: imin=$imin, imax=$imax, #len path=".scalar(@$ra_path)."\n");
	}
	elsif ( $imin > $imax )
	{
		print STDERR "DeregisterContigs RAS: imin=$imin, imax=$imax\n";
		# rien a faire
	}
	else
	{
		#print STDERR "$imin / $imax ".$self->Get()."\n";
		$imin++ if ( $imin % 2 != 0 ); # je suis arrive sur un gap 
		for(my $i=$imin;$i<=$imax;$i+=2)	# contig gap contig
		{
		#	print STDERR "$i $$ra_path[$i]\n";
			$$ra_path[$i]->DelScaffold($self); # je desenregistre le scaffold dans le contig
		}
	}
}

# TEST OK
sub GetName
{
my($self)=@_;

	return $$self{_name};
}

sub GetOverlapParams
{
my($self)=@_;

	return ($$self{_minovlpci},$$self{_minovllen},$$self{_maxovlovh});
}

sub SetOverlapParams
{
my($self,@a_param)=@_;

	$$self{_minovlpci} = $a_param[0] if( defined($a_param[0]) );
	$$self{_minovllen} = $a_param[1] if( defined($a_param[1]) );
	$$self{_maxovlovh} = $a_param[2] if( defined($a_param[2]) );
}

sub SetName
{
my($self,$str)=@_;

	$$self{_name} = $str;
}

# TEST OK
sub Get
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	if ( ! defined($ra_path) )
	{
		print STDERR Dumper($self);
		Carp::confess("UNEXPECTED ERROR\n");
	}
	my $str = '';
	for(my $i=0;$i<scalar(@$ra_path);$i++)
	{
		if ( $i % 2 == 0 )
		{
			if ( ref($$ra_path[$i]) ne 'Contig' )
			{
				print STDERR Dumper($self);
				Carp::confess("UNEXPECTED ERROR\n");
			}
			$str .= $$ra_path[$i]->Get();
		}
		else
		{
			$str .= $$ra_path[$i];
		}
		$str.=" ";
	}
	$str =~ s/\s+$//;
	return $str;
}


# TEST OK
sub GetArray
{
my($self)=@_;

	return $self->_FastOneGet('_ra_path');
}

# TEST OK
sub GetArrayContigs
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return undef if ( !defined($ra_path) );
	my @a_ctg = ();
	for(my $i=0;$i<scalar(@$ra_path);$i+=2)
	{
		push(@a_ctg,$$ra_path[$i]);
	}
	return \@a_ctg;
}

# TEST OK
sub GetNumberOfContigs
{
my($self)=@_;

	my $ra_path = $self->_FastOneGet('_ra_path');
	return 0 if ( ! defined($ra_path) );
	return int((scalar(@$ra_path)-1)/2+1); # le int() pour le cas scalar => 0
}

# TEST OK
sub IsContainingContig
{
my($self,$ctg)=@_;

	my $ra_ctg = $self->GetArrayContigs();
	return &FALSE if ( !defined($ra_ctg) );
	for(my $i=0;$i<scalar(@$ra_ctg);$i++)
	{
		if ( ref($ctg) eq 'Contig' )
		{
			return &TRUE if ( $$ra_ctg[$i] eq $ctg );
		}
		else
		{
			return &TRUE if ( $$ra_ctg[$i]->GetName() eq $ctg );
		}
	}
	return &FALSE;
}


# TEST OK
sub IsAnchored
{
my($self)=@_;
	
	return ( defined($$self{_anchor}) && $$self{_anchor} ne '') ? &TRUE : &FALSE;
}


# TEST OK
sub GetAnchor
{
my($self)=@_;
	
	return ( defined($$self{_anchor}) && $$self{_anchor} ne '') ? $$self{_anchor} : undef;
}


# TEST OK
sub AreAnchoringCompatibles
{
my($self,$o_scaff2)=@_;

	my $anchor1 = $self->GetAnchor();
	my $anchor2 = $o_scaff2->GetAnchor();
	my $ok = ( !defined($anchor1) || ! defined($anchor2) || $anchor1 eq '' || $anchor2 eq '' || $anchor1 eq $anchor2 ) ? &TRUE : &FALSE;
	$self->_Print("AreAnchoringCompatibles ".$self->GetName()." vs ".$o_scaff2->GetName()." ==> $ok\n") if ( ! $ok );
	return $ok;
}

# l'ancrage d'un scaffold est forcement calcule a partir de l'ancrage des contigs participants
# dans le cas de l'OM, le scaffold sera utilise pour l'alignement mais ce seront les contigs sur le segment aligne qui seront ancres

# TEST OK
sub InitAnchor
{
my($self,$stop_if_inconsistent,$debug)=@_;

	$debug = &FALSE if ( ! defined($debug) );
	my %h_seed =();
	$self->BuildListOfAnchors(\%h_seed,$stop_if_inconsistent,$debug);
	my @a_new= keys(%h_seed);
	if ( $debug )
	{
		print STDERR "InitAnchor ".$self->GetName()."\n";
		print STDERR Dumper(%h_seed);
		print STDERR Dumper(@a_new);
	}
	return undef if ( scalar(@a_new) != 1 );  # si necessaire c'est BuildListOfAnchors qui crashe si pas consistent 
	my $prev_anchor = $self->GetAnchor();
	if ( defined($prev_anchor) && $prev_anchor ne $a_new[0] )
	{
		$self->_Print("Warning: anchoring modification for ".$self->GetName()." previous anchoring=$prev_anchor, current anchoring=$a_new[0]\n");
	}
	$$self{_anchor} = $a_new[0];
}



sub BuildListOfAnchors
{
my($self,$rh_seed,$stop_if_inconsistent,$debug)=@_;

	my $anchor_min_len = $self->_FastOneGet('_anchor_min_len');
	my $ra_ctg = $self->GetArrayContigs();
	Carp::confess("BuildListOfAnchors\n") if ( ! defined($ra_ctg) );
	for(my $i=0;$i<scalar(@$ra_ctg);$i++)
	{
		if ( $debug )
		{
			$self->_Print("len=".$$ra_ctg[$i]->GetLength()." < $anchor_min_len GetNumberOfAnchorsFromRegistry=".$$ra_ctg[$i]->GetNumberOfAnchorsFromRegistry()." GetNumberOfScaffoldsFromRegistry=".$$ra_ctg[$i]->GetNumberOfScaffoldsFromRegistry()."\n");
		}
		next if ( $$ra_ctg[$i]->GetLength() < $anchor_min_len );
		next if ( $$ra_ctg[$i]->GetNumberOfAnchorsFromRegistry() != 1 );
		next if ( $$ra_ctg[$i]->GetNumberOfScaffoldsFromRegistry() > 1 ); # je ne considere pas les ancres sur des contigs dupliques 
																		  # par contre prend les ancres sur des contigs non dans des scaff (c'est le cas lorsque je fais loadanchors)
		my $anchor = $$ra_ctg[$i]->GetAnchorsFromRegistry();
		$$rh_seed{$anchor}.= $$ra_ctg[$i]->Get()." ";
	}
	if ( $stop_if_inconsistent && scalar(keys(%$rh_seed)) > 1 )
	{
		$self->_Print("BuildListOfAnchors: inconsistencies found during anchoring scaff=".$self->GetName()."\n".$self->Get()."\n");
		for(my $i=0;$i<scalar(@$ra_ctg);$i++)
		{
			next if ( $$ra_ctg[$i]->GetLength() < $anchor_min_len );
			next if ( $$ra_ctg[$i]->GetNumberOfAnchorsFromRegistry() != 1 );
			next if ( $$ra_ctg[$i]->GetNumberOfScaffoldsFromRegistry() != 1 ); # je ne considere pas les ancres sur des contigs dupliques
			$self->_Print("ctg=".$$ra_ctg[$i]->GetName()." len=".$$ra_ctg[$i]->GetLength()." anchor=".$$ra_ctg[$i]->GetAnchorsFromRegistry()."\n");
			print STDERR Dumper($$ra_ctg[$i]);
		}
		#$self->Dump();
		$self->_Print("BuildListOfAnchors: warning: anchor_min_len=$anchor_min_len, contigs shorter than anchor_min_len are not taken into account for anchoring the scaffold\n");
		Carp::confess("BuildListOfAnchors: inconsistencies found during anchoring\n");
	}
}

# TEST OK
#sub IsPathAnchoringConsistent
sub IsAnchoringConsistent
{
my($self,$stop_if_inconsistent)=@_;

	$stop_if_inconsistent = &FALSE if ( !defined($stop_if_inconsistent) );
	my %h_seed =();
	$self->BuildListOfAnchors(\%h_seed,$stop_if_inconsistent,&FALSE);
	if ( scalar(keys(%h_seed)) > 1 )
	{
		return &FALSE;
	}
	return &TRUE;
}

sub IsContigAnchoringCompatible
{
my($self,$o_ctg)=@_;

	my $anchor_min_len = $self->_FastOneGet('_anchor_min_len');
	return &TRUE if ( $o_ctg->GetLength() < $anchor_min_len );
	return &TRUE if ( $o_ctg->GetNumberOfAnchorsFromRegistry() != 1 ); # si c'est vide ou deja ambigu c'est ok
	return &TRUE if ( $o_ctg->GetNumberOfScaffoldsFromRegistry() > 1 );
	my $ctg_anchor = $o_ctg->GetAnchorsFromRegistry();
	my $scaff_anchor = $self->GetAnchor();
	return &TRUE  if ( !defined($scaff_anchor) || $scaff_anchor eq '' || $ctg_anchor eq $self->GetAnchor() );
	return &FALSE;
}

# TEST OK
sub Reverse
{
my($self)=@_;

	my $ra_path = $self->GetArray();
	for(my $i=0;$i<scalar(@$ra_path);$i+=2)
	{
		$$ra_path[$i]->Reverse();
	}
	@$ra_path = reverse(@$ra_path);
}

# TEST OK
sub GetLength
{
my($self,$full)=@_;

	my $ra_path = $self->GetArray();
	if ( ! defined($ra_path) )
	{
		print STDERR Dumper($self);
		Carp::confess("GetLength\n") 
	}
	my $totnt=0;
	for(my $i=0;$i<scalar(@$ra_path);$i++)
	{
		if ( $i % 2 == 0 )
		{ 
			$totnt+=$$ra_path[$i]->GetLength();
		}
		elsif ( $full )	# on prend en compte les gaps
		{
			my @a_=split(',',$$ra_path[$i]);
			$totnt+=$a_[1]; # on renvoie le max
		}
	}
	return $totnt;
}


# TEST OK 
sub GetInternalLength
{
my($self,$full)=@_;

	my $ra_path = $self->GetArray();
	my $totnt=0;
	for(my $i=1;$i<scalar(@$ra_path)-1;$i++)
	{
		if ( $i % 2 == 0 )
		{ 
			$totnt+=$$ra_path[$i]->GetLength();
		}
		elsif ( $full )	# on prend en compte les gaps
		{
			my @a_=split(',',$$ra_path[$i]);
			$totnt+=$a_[1];	# on renvoie le max
		}
	}
	return $totnt;
}



# TEST OK
# pour le moment je le garde mais cela deviendra sans doute inutile
sub GetLengthByString
{
my($self)=@_;

	my $path = $self->Get();  # cf commentaire suivant
	my $totnt=0;
	while($path =~ /#(\S+)([pm])(\d+)/g )	# laisser le while car j'utilise la fonction ave un path sans gap
	{
		my ($id,$strand,$len) = ($1,$2,$3);
		$totnt+=$len;
	}
	return $totnt;
}


# TEST OK
#sub IsTandemRepeatInPath
sub IsContainingTandemRepeat
{
my($self) = @_;

	my $ra_ctg = $self->GetArrayContigs();
	for(my $i=1;$i<scalar(@$ra_ctg);$i++)
	{
		if ($$ra_ctg[$i]->GetName() eq $$ra_ctg[$i-1]->GetName())
		{
			return &TRUE;
		}
	}
	return &FALSE;
}

# supprime les repeats en tandem, verifie le format en sortie
# TEST OK
sub Check
{
my($self)=@_;

	my $prevctg ='';
	my $ra_path = $self->GetArray();
	my $len_prev = scalar(@$ra_path);
	return if ( !defined($ra_path) );
	my @a_newpath=();
	my $ok=&FALSE;
	for(my $i=0;$i<scalar(@$ra_path);$i++)
	{
		if ( $i % 2 == 0 )	
		{
			$ok = ( $$ra_path[$i]->GetName() ne $prevctg ) ? &TRUE : &FALSE;
			$prevctg = $$ra_path[$i]->GetName();
			$self->DeregisterContigs($i,$i) if ( ! $ok );
		}
		if ( $ok )
		{
			push(@a_newpath,$$ra_path[$i]);
		}
	}
	if ( ref($a_newpath[$#a_newpath]) ne 'Contig' )	# lorsque le repeat est a la fin il faut enlever le gap
	{
		pop(@a_newpath);
	}
	my $len_curr = scalar(@a_newpath);
	if ( $len_curr != $len_prev )
	{
		$self->_Print("Check/Tandem repeat detected in ".$self->GetName()." len_prev=$len_prev ".$self->Get()."\n");
	}
	$self->_Set('_ra_path',\@a_newpath);
	if ( $len_curr != $len_prev )
	{
		$self->_Print("Check/Tandem \t\t".$self->GetName()." len_curr=$len_curr ".$self->Get()."\n");
	}
	$self->CheckFormat();
}

# Utile OUI

sub GetPrePostLengthInPath
{
my($self,$ctg)=@_;

	$ctg=$1 if (  $ctg =~ /^#(\S+)[pm]\d+/ );
	my $ra_path = $self->GetArray();
	for(my $i=0;$i<scalar(@$ra_path);$i+=2)
	{
		my ($lctg,$len) = ($$ra_path[$i]->GetName(),$$ra_path[$i]->GetLength());
		if ( $lctg eq $ctg )
		{
			my $o_prev = New Scaffold(name=>'TMPL','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
			$o_prev->Push(@$ra_path[0..$i-1]);
			my $o_post = New Scaffold(name=>'TMPR','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
			$o_post->Push(@$ra_path[$i+1..$#$ra_path]);
			my $prevlen = ( $i != 0 )       ? $o_prev->GetLength() : 0;
			my $postlen = ( $i != $#$ra_path)? $o_post->GetLength() : 0;
			return ($prevlen,$postlen);
		}
	}
	Carp::confess("_FastOneGetPrePostLenInPath: $ctg not found in ".$self->Get()."\n");
}


# gaplen: c'est le max qui arrive pour definir la contrainte mais quand je teste 
# j'utilise le min pour calculer la distance et favoriser la compatibilite

# Migre sans rien changer (a part l'init de $scaff, anciennement $r_scaff)

sub IsCompatible
{
my($self,$ctgin,$gaplen,$r_dist,$end_scaff,$expected_end) = @_;

	my $scaff = $self->Get();
	my $debug = &TRUE;
	if ( 0 && $scaff =~ /$DEBUG_LINK_PATTERN/ || $ctgin =~ /$DEBUG_LINK_PATTERN/ )
	{
		$debug = &TRUE;
		print STDERR "IsCompatible >$scaff< $ctgin,$gaplen,$$r_dist,$end_scaff,$expected_end\n";
	}
	#  #MERGE.pe19m981  981 is the sequence length
	#  5555 <-gaps
	my @a_ = split(' ',$scaff);

	#return &TRUE if ( $#a_ == 0 );	# toujours compatible s'il n'y a qu'un scaffold

	@a_ = reverse(@a_) if ($end_scaff == 3 );
	for(my $i=0; $i<scalar(@a_) ; $i++)
	{
		print STDERR "\t\t$i  $a_[$i]\n" if ( $debug );
		if ( $a_[$i] =~ /#(\S+)([pm])(\d+)$/ )
		{
			my($id,$strand,$len)=($1,$2,$3);
			if ( $1 eq $ctgin )
			{
				my $compatible_end = '';
				if ( $end_scaff == 5 )
				{
					$compatible_end=($strand eq 'p') ? 5 : 3 ;
				}
				else
				{
					$compatible_end=($strand eq 'p') ? 3 : 5 ;
				}
				if ( ($i==0 || $$r_dist < $gaplen) && $compatible_end eq $expected_end )
				{
					return &TRUE;
				}
				else
				{
					print STDERR "\t$ctgin dist=$$r_dist, strand=$strand gaplen=$gaplen compat_end=$compatible_end expected_end=$expected_end\n\tend scaffold=$end_scaff - $scaff\n" if ( $debug );
					return &FALSE;
				}
			}
			$$r_dist += $len;
		}
		else
		{
			my ($mingaplen,$maxgaplen) = split(',',$a_[$i]);
			#my $delta = ( $mingaplen < 0 ) ? 1 : (($mingaplen+$maxgaplen)/2);
			#$delta = 1 if ( $delta < 0 );
			my $delta = &Max($mingaplen,1); # change le 20130406
			#$maxgaplen = 1 if ( $maxgaplen < 1 );
			#$mingaplen = 1 if ( $mingaplen < 1 );
			#$$r_dist += $mingaplen; # j'utilise le min pour favoriser la compatibilite
			#$$r_dist += int($maxgaplen/2);
			$$r_dist += $delta; # change le 20130307
		}
		if ( $$r_dist && $$r_dist > $gaplen )
		{
			print STDERR "\t\t$i   $$r_dist > $gaplen\n" if ( $debug );
			return &FALSE;
		}
	}
	# je ne dois pas arriver ici
	print STDERR "IsCompatible ($self,$ctgin,$gaplen,$$r_dist,$end_scaff,$expected_end)\n";
	print STDERR Dumper($self);
	Carp::confess("IsCompatible Assert\n");
}



# ca va etre super lourd 
# je ne vais pas l'utiliser car on ne peut pas valider les liens lorsque les contigs sont trop longs cad > taille d'insert max
# apres le probleme c'est que l'on n'a pas forcement suffisament de profondeur
# en plus lorsque l'on a plusieurs repeats qui se suivent cela ne sert a rien
# bon je la remets, on peut la desactiver avec $validation_cutoff == 0 ou 1

# Migr

sub IsLinkValidated
{
my($self,$o_rightscaff,$o_linkds,$validation_cutoff,$use_weak) = @_;

	my $vallen = $self->_FastOneGet('_maxvalidationlength');

	my $ra_left  = $self->GetArray();
	my $ra_right = $o_rightscaff->GetArray();

	my $leftscaff  = $self->Get();
	my $rightscaff = $o_rightscaff->Get();

	my $debug = &TRUE;
	if ( 0 && $leftscaff =~ /$DEBUG_LINK_PATTERN/ || $rightscaff =~ /$DEBUG_LINK_PATTERN/ )
	{
		$debug = &TRUE;
		print STDERR "$leftscaff <> $rightscaff\n";
	}

	return &TRUE if ( $validation_cutoff < 1 );
	return &TRUE if ( scalar(@$ra_left) == 1 && scalar(@$ra_right) == 1); # cas particulier de deux contigs
	print STDERR "$leftscaff <> $rightscaff plus de deux contigs\n" if ($debug);

	my $leftlen = 0;
	my $nb_links = 0;

	#le 15/03/2012 je recupere la longueur max en fonction des inserts utilises et je supprime les STEP
	#my $step_left  = &Min(&MAXVALIDATION_STEPS*2,scalar(@a_left));
	#my $step_right = &Min(&MAXVALIDATION_STEPS*2,scalar(@a_right));
	#for(my $l=0;$l<$step_left && $leftlen<$vallen;$l++)

	for(my $l=scalar(@$ra_left)-1;$l >= 0 && $leftlen<$vallen;$l--)
	{
		if ( $l % 2 == 0 )
		{
			my $rightlen = 0;
			my ($lid,$lstrand,$llen) = ($$ra_left[$l]->GetName(),$$ra_left[$l]->GetStrand(),$$ra_left[$l]->GetLength());
			for(my $r=0;$r < scalar(@$ra_right) && $rightlen<$vallen;$r++)
			{
				if ( $r % 2 == 0 )
				{
					my ($rid,$rstrand,$rlen) = ($$ra_right[$r]->GetName(),$$ra_right[$r]->GetStrand(),$$ra_right[$r]->GetLength());

					my %h_val = ();
					if ( $o_linkds->GetKV2OrientedContigs($lid,$lstrand,$rid,$rstrand,\%h_val) )
					{
						# cas particulier ou a l'extremite je ne peux pas valider parce que j'ai un gros contig en bout
						if ( (scalar(@$ra_left) == 1 && $r == 0 && $rlen > ($vallen/2)) || (scalar(@$ra_right) == 1 && ($l==scalar(@$ra_left)-1) && $llen > ($vallen/2))  )
						{
							print STDERR "Warning: big contig at one end vs singleton, unable to validate => OK : $l ($lid,$lstrand,$llen) - $r ($rid,$rstrand,$rlen)\n" if ( $debug );
							return &TRUE;
						}
						# cas particulier ou a l'extremite je ne peux pas valider parce que j'ai deux gros contigs en bout
						if ( ($r == 0 && $rlen > ($vallen/2)) && (($l==scalar(@$ra_left)-1) && $llen > ($vallen/2))  )
						{
							print STDERR "Warning: two big contigs at the ends, unable to validate => OK : $l ($lid,$lstrand,$llen) - $r ($rid,$rstrand,$rlen)\n" if ( $debug );
							return &TRUE;
						}

						if ( $use_weak || ! $h_val{weak} )
						{
							if ( $h_val{hardlink} >= 1 )	# comme c'est un hardlink c'est sur que par ailleurs il y a un pont donc c'est bien valide aussi 
							{
								my $spansrc=$h_val{srcspan};
								my $spantgt=$h_val{tgtspan};
								# si la taille du span est superieure des deux cote a la taille de l'insert, ca veut dire que le lien est 
								# donne par plusieurs banques sur une grande region (c'est le cas des grands contigs)
								if ( $spansrc > $h_val{insert_size} && $spantgt > $h_val{insert_size} )
								{
									return &TRUE;
								}
								# si la sequence la plus courte est aussi pleinement couverte c'est bon (rajoute le 20120202)
								if ( 0 && $lid =~ /$DEBUG_LINK_PATTERN/ || $rid =~ /$DEBUG_LINK_PATTERN/ )
								{
									print STDERR "$lid, $lid $h_val{srclen} < $h_val{tgtlen} $spansrc $spantgt \n";
								}
								# je veux des liens surs mais je n'arrive pas a trouver de regles pour garantir alors en dur!
								if ( $h_val{count} >= 50 && $spansrc >= 5000 && $spantgt >= 5000 )
								{
									return &TRUE;
								}	
							}
							$nb_links++;
						}
						print STDERR "$leftscaff <> $rightscaff $l,$r,$nb_links    $lid,$lstrand,$rid,$rstrand, $h_val{count}\n" if ( $debug );
					}
					return &TRUE if ( $nb_links >=  $validation_cutoff);
					$rightlen+=$rlen;
				}
				else
				{
					my($minlen,$maxlen)=split(',',$$ra_right[$r]);
					$minlen=1 if ( $minlen < 1 );
					$rightlen+=$minlen;
				}
			}
		}
		else
		{
			my($minlen,$maxlen)=split(',',$$ra_left[$l]);
			$minlen=1 if ( $minlen < 1 );
			$leftlen+=$minlen;
		}
	}
	return &FALSE;
}

# affecter les contigs du scaffold merge au scaffold courant

sub SetContigs
{
my($self,$mergedscaff)=@_;

	my $ra_ctg=$self->GetArrayContigs();
	for(my $i=0;$i<scalar(@$ra_ctg);$i++)
	{
		if ( $$ra_ctg[$i]->IsContainingScaffold($mergedscaff) &&  $mergedscaff ne $self)
		{
			$$ra_ctg[$i]->DelScaffold($mergedscaff);
		}
		$$ra_ctg[$i]->RemoveTemporaryScaffolds();
		# je le sort de la boucle pour etre sur que les contigs sont bien rattaches au scaffolds courant
		$$ra_ctg[$i]->SetScaffold($self,$self->GetName());
	}
}


# Migr
sub Merge
{
my($self,$o_scafftgt,$rh_key,$rh_val,$o_linkds,$validation,$gaplen_fact,$seed_consistency,$reload_gaps)=@_;

	my $srcid = $self->GetName();
	my $tgtid = $o_scafftgt->GetName();
	my $srcscaff = $self->Get();
	my $tgtscaff = $o_scafftgt->Get();

	return &FALSE if ( $seed_consistency && ! $self->AreAnchoringCompatibles($o_scafftgt) );

	my $debug = &FALSE;
	if ( 0 && ($srcscaff =~ /$DEBUG_LINK_PATTERN/ || $tgtscaff =~ /$DEBUG_LINK_PATTERN/) )
	{
		$debug = &TRUE;
		print STDERR "\n_MergeScaffolds $srcscaff <> $tgtscaff\n";
	}
	my $ok = &TRUE;
	my $observedstrand='';
	my $distancefive=0;
	my $distancethree=0;
	my $notadistissue=4;

	$$rh_val{gaplen_max} = 1 if ( $$rh_val{gaplen_max} < 0 ); # 20130307 je rajoute pour que les gap negatifs ne posent pas souci 
	my $compatfive  = $self->IsCompatible($$rh_key{src},$$rh_val{gaplen_max}*$gaplen_fact,\$distancefive,5,$$rh_val{srcext});
	my $compatthree = $self->IsCompatible($$rh_key{src},$$rh_val{gaplen_max}*$gaplen_fact,\$distancethree,3,$$rh_val{srcext});
	if ( $debug )
	{
		print STDERR "======\n";
		print STDERR "$$rh_key{src},$$rh_val{gaplen_max}*$gaplen_fact,\$distancefive,5,$$rh_val{srcext} $distancefive\n";
		print STDERR "$$rh_key{src},$$rh_val{gaplen_max}*$gaplen_fact,\$distancethree,3,$$rh_val{srcext} $distancethree\n";
		print STDERR "_MergeScaffolds val=$validation $srcid<=>$tgtid $gaplen_fact five=$compatfive, three=$compatthree  src=$$rh_key{src},$$rh_val{srcext} tgt=$$rh_key{tgt},$$rh_val{tgtext} $$rh_val{gaplen_max}\n";
		print STDERR ">".join(' ',%$rh_key),"\n";
		print STDERR ">".join(' ',%$rh_val),"\n";
		print STDERR "\n";
	}
	if( $compatfive && $compatthree )
	{
		print STDERR "Warning: $$rh_key{tgt} is compatible with both ends of $$rh_key{src}, likely a repeat, I do nothing\n";
		return &FALSE;
	}
	my ($gmin,$gmax)=();
	my ($gmin_tol,$gmax_tol)=(); # 20150105, je rajoute pour tester avec tolerance mais je veux garder le trou sans tolerance
	my %h_val=();
	if ( $compatfive)
	{
		my $dist = $distancefive;
		if ( $o_scafftgt->IsCompatible($$rh_key{tgt},$$rh_val{gaplen_max}*$gaplen_fact,\$dist,5,$$rh_val{tgtext}) && $self->NewOverlapConsistent($dist,$gaplen_fact,$rh_val) )
		{
			my $gapok = &TRUE;
			$o_scafftgt->Reverse();
			($gmin,$gmax) = &GetGapLen($dist,$$rh_val{gaplen_min},$$rh_val{gaplen_max});
			$gmax_tol = $$rh_val{gaplen_max}*$gaplen_fact-$dist;
			if ( $gmax_tol < 0 )
			{
				my (@a_lastprev)  = $o_scafftgt->Last();
				my (@a_firstnext) = $self->First();
				if ( ! $o_linkds->GetKV2OrientedContigs($a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1],\%h_val) )
				{
					print STDERR "compatfive Negative gap rejected $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
					$gapok=&FALSE;
				}
				else
				{
					print STDERR "Negative gap OK $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
				}
			}

			if ( $gapok )
			{
				if ( ! $validation || $o_scafftgt->IsLinkValidated($self,$o_linkds,$validation,&FALSE) )	# au moins 2 sequences d'un cote donnent le lien
				{
					my $gap = sprintf("%d,%d",$gmin,$gmax);
					$o_scafftgt->Push($gap,@{$self->GetArray()});
					$o_scafftgt->Check();
					#$o_scafftgt->CheckGapsOnePath($o_linkds);
					$o_scafftgt->SetContigs($self);
					$o_scafftgt->InitAnchor(&TRUE);
					$o_scafftgt->ReloadGapsOnePath($o_linkds) if ( $reload_gaps );
					$self->Reset();
					return &TRUE;
				}
			}
			# je remets comme c'etait
			$o_scafftgt->Reverse();
		}
		printf STDERR "dist=$dist $gaplen_fact %f\n",$$rh_val{gaplen_max}*$gaplen_fact if ( $debug );
		$notadistissue-- if ( $dist < $$rh_val{gaplen_max} );
		$dist = $distancefive;
		if( $o_scafftgt->IsCompatible($$rh_key{tgt},$$rh_val{gaplen_max}*$gaplen_fact,\$dist,3,$$rh_val{tgtext}) && $self->NewOverlapConsistent($dist,$gaplen_fact,$rh_val) )
		{
			($gmin,$gmax) = &GetGapLen($dist,$$rh_val{gaplen_min},$$rh_val{gaplen_max});
			$gmax_tol = $$rh_val{gaplen_max}*$gaplen_fact-$dist;
			my $gapok = &TRUE;
			if ( $gmax_tol < 0 )
			{
				my (@a_lastprev)  = $o_scafftgt->Last();
				my (@a_firstnext) = $self->First();
				if ( ! $o_linkds->GetKV2OrientedContigs($a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1],\%h_val) )
				{
					print STDERR "Negative gap rejected $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
					$gapok=&FALSE;
				}
				else
				{
					print STDERR "Negative gap OK $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
				}
			}

			if ( $gapok )
			{
				if ( ! $validation || $o_scafftgt->IsLinkValidated($self,$o_linkds,$validation,&FALSE) )	# au moins 2 sequences d'un cote donnent le lien
				{
					my $gap = sprintf("%d,%d",$gmin,$gmax);
					$o_scafftgt->Push($gap,@{$self->GetArray()});
					$o_scafftgt->Check();
					#$self->CheckGapsOnePath($o_linkds);
					$o_scafftgt->SetContigs($self);
					$o_scafftgt->InitAnchor(&TRUE);
					$o_scafftgt->ReloadGapsOnePath($o_linkds) if ( $reload_gaps );
					$self->Reset();
					return &TRUE;
				}
			}
		}
		printf STDERR "dist=$dist $gaplen_fact %f\n",$$rh_val{gaplen_max}*$gaplen_fact if ( $debug );
		$notadistissue-- if ( $dist < $$rh_val{gaplen_max} );
	}
	if ( $compatthree )
	{
		my $dist = $distancethree;
		if ( $o_scafftgt->IsCompatible($$rh_key{tgt},$$rh_val{gaplen_max}*$gaplen_fact,\$dist,5,$$rh_val{tgtext}) && $self->NewOverlapConsistent($dist,$gaplen_fact,$rh_val) )
		{
			($gmin,$gmax) = &GetGapLen($dist,$$rh_val{gaplen_min},$$rh_val{gaplen_max});
			$gmax_tol = $$rh_val{gaplen_max}*$gaplen_fact-$dist;
			my $gapok = &TRUE;
			if ( $gmax_tol < 0 )
			{
				my (@a_lastprev)  = $self->Last();
				my (@a_firstnext) = $o_scafftgt->First();
				if ( ! $o_linkds->GetKV2OrientedContigs($a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1],\%h_val) )
				{
					print STDERR "Negative gap rejected $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
					$gapok=&FALSE;
				}
				else
				{
					print STDERR "Negative gap OK $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
				}
			}

			if ( $gapok )
			{
				if ( ! $validation || $self->IsLinkValidated($o_scafftgt,$o_linkds,$validation,&FALSE) )	# au moins 2 sequences d'un cote donnent le lien
				{
					my $gap = sprintf("%d,%d",$gmin,$gmax);
					$self->Push($gap,@{$o_scafftgt->GetArray()});
					$self->Check();
					#$self->CheckGapsOnePath($o_linkds);
					$self->SetContigs($o_scafftgt);
					$self->InitAnchor(&TRUE);
					$self->ReloadGapsOnePath($o_linkds) if ( $reload_gaps );
					$o_scafftgt->Reset();
					return &TRUE;
				}
			}
		}
		printf STDERR "dist=$dist $gaplen_fact %f\n",$$rh_val{gaplen_max}*$gaplen_fact if ( $debug );
		$notadistissue-- if ( $dist < $$rh_val{gaplen_max} );
		$dist = $distancethree;
		if( $o_scafftgt->IsCompatible($$rh_key{tgt},$$rh_val{gaplen_max}*$gaplen_fact,\$dist,3,$$rh_val{tgtext}) && $self->NewOverlapConsistent($dist,$gaplen_fact,$rh_val) )
		{
			$o_scafftgt->Reverse();
			($gmin,$gmax) = &GetGapLen($dist,$$rh_val{gaplen_min},$$rh_val{gaplen_max});
			$gmax_tol = $$rh_val{gaplen_max}*$gaplen_fact-$dist;
			my $gapok = &TRUE;
			if ( $gmax_tol < 0 )
			{
				my (@a_lastprev)  = $self->Last();
				my (@a_firstnext) = $o_scafftgt->First();
				if ( ! $o_linkds->GetKV2OrientedContigs($a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1],\%h_val) )
				{
					print STDERR "Negative gap rejected $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
					$gapok=&FALSE;
				}
				else
				{
					print STDERR "Negative gap OK $a_lastprev[0],$a_lastprev[1],$a_firstnext[0],$a_firstnext[1]\n";
				}
			}

			if ( $gapok )
			{
				if ( ! $validation || $self->IsLinkValidated($o_scafftgt,$o_linkds,$validation,&FALSE) )	# au moins 2 sequences d'un cote donnent le lien
				{
					($gmin,$gmax) = &GetGapLen($dist,$$rh_val{gaplen_min},$$rh_val{gaplen_max});
					my $gap = sprintf("%d,%d",$gmin,$gmax);
					$self->Push($gap,@{$o_scafftgt->GetArray()});
					$self->Check();
					#$self->CheckGapsOnePath($o_linkds);
					$self->SetContigs($o_scafftgt);
					$self->InitAnchor(&TRUE);
					$self->ReloadGapsOnePath($o_linkds) if ( $reload_gaps );
					$o_scafftgt->Reset();
					return &TRUE;
				}
			}
			$o_scafftgt->Reverse();
		}
		printf STDERR "dist=$dist $gaplen_fact %f\n",$$rh_val{gaplen_max}*$gaplen_fact if ( $debug );
		$notadistissue-- if ( $dist < $$rh_val{gaplen_max} );
	}
	if ( $notadistissue && $$rh_val{srclen} > 10000 && $$rh_val{tgtlen} > 10000 && ! $$rh_val{weak} && $$rh_val{count} > 10)
	{
		print STDERR "Warning: not a distance issue involving large contigs > 10000 $$rh_key{src} $$rh_key{tgt} $notadistissue, (validation=$validation)\n";
		print STDERR join(' ',%$rh_val)."\n";
	}
	return &FALSE;
}

sub NewOverlapConsistent
{
my($self,$dist,$fact,$rh_val)=@_;

	if ( ($$rh_val{gaplen_max}*$fact-$dist) < 0 && defined($$rh_val{ovllen}) && $$rh_val{ovllen} == 0 )
	{
		print STDERR "NewOverlapConsistent: $$rh_val{gaplen_max}-$dist) < 0 // $$rh_val{ovllen} == 0 \n";
		return &FALSE;
	}
	return &TRUE;
}

#
# Edition des scaffolds, post scaffolding
#

sub ValidateIndirect
{
my($self,$o_linkds,$cutmode,$validation,$maxlen,$noinit)=@_;

	my $self_copy = undef;
	$self_copy    = $self->Clone() if ( $cutmode );
	my $ra_path   = ($cutmode) ? $self_copy->GetArray() : $self->GetArray();

	return undef if ( scalar(@$ra_path) < 5 ); # cannot be used on scaffolds with less than 3 contigs
	my $lastinit=0;

	my$newscaff=1;
	my %h_new_scaffolds=();

	for(my $i=2;$i<@$ra_path-2;$i+=2)	# contig gap contig
	{
		if ( $i % 2 == 0  )
		{
			if ( ref($$ra_path[$i]) ne 'Contig' )
			{
				Carp::confess("ValidateIndirect: unexpected contig format ".$self->GetName()." >".$$ra_path[$i]->Get()."<\n");
			}
			next if ( $$ra_path[$i]->GetLength()  > $maxlen );
		}
		# pas optimal de reallouer en permanence
		my $o_left=New Scaffold(name=>'TMPL','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
		$o_left->Push(@$ra_path[$lastinit..$i-2]);
		my $o_right=New Scaffold(name=>'TMPR','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
		$o_right->Push(@$ra_path[$i..$#$ra_path]);
		$o_right->DeregisterContigs();
		$o_left->DeregisterContigs();

		if ( ! $o_left->IsLinkValidated($o_right,$o_linkds,$validation,&TRUE) )	# au moins 2 sequences d'un cote donnent le lien
		{
			$self->_Print("ValidateIndirect ".$self->GetName()." NOK $i/$#$ra_path ".$$ra_path[$i-2]->Get().",".$$ra_path[$i]->Get()."\n");
			print STDERR "LEFT: $lastinit--$i-2\t".$o_left->Get()."\n";
			print STDERR "RIGHT: $i--$#$ra_path\t".$o_right->Get()."\n";
			if ( $cutmode )	 # je mets la partie amont dans un autre scaffold
			{
				$o_left->InitAnchor(&TRUE);
				$h_new_scaffolds{$newscaff++} = $o_left;
				# je supprime autant de Contig gap que j'en ai dans o_left
				my $ra_ctg = $o_left->GetArrayContigs(); # shift enleve le gap
				for(my $j=0;$j<scalar(@$ra_ctg);$j++)
				{
					$self->Shift();
				}
				$self->InitAnchor(&TRUE);	# je refais l'ancrage
				$lastinit = $i;
			}
		}
	}
	return \%h_new_scaffolds;
}

sub KeepOnlyOverlappingContigs
{
my($self,$o_linkds,$rh_o_seqds,$ra_seqlist,$rh_param)=@_;

	my $ra_path   = $self->GetArray();
	my @a_copypath = @$ra_path;
	return undef if ( scalar(@$ra_path) < 3 ); # cannot be used on scaffolds with less than 2 contigs
	$self->_Set('_fast_getsequence',&FALSE);
	my %h_new_scaffolds=();
	my$newscaff=1;
	my($prev_id,$prev_strand)=($$ra_path[0]->GetName(),$$ra_path[0]->GetStrand());
	my $prev_seq    = &LookupSeq($rh_o_seqds,$ra_seqlist,$prev_id,$prev_strand);
	my $curr_seq    = '';
	my($curr_id,$curr_strand);
	my $lastinit=0;
	my %h_val=();
	my $cut=&FALSE;
	for(my $i=2;$i<scalar(@a_copypath);$i+=2)	# contig gap contig
	{
		if ( $i % 2 == 0  )
		{
			if ( ref($a_copypath[$i]) ne 'Contig' )
			{
				Carp::confess("ValidateIndirect: unexpected contig format ".$self->GetName()." >".$a_copypath[$i]->Get()."<\n");
			}
		}
		my $o_left=New Scaffold(name=>'TMPL','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
		$o_left->Push(@a_copypath[$lastinit..$i-2]);
		my $o_right=New Scaffold(name=>'TMPR','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
		$o_right->Push(@a_copypath[$i..$#a_copypath]);
		$o_right->DeregisterContigs();
		$o_left->DeregisterContigs();

		($curr_id,$curr_strand) = ($a_copypath[$i]->GetName(),$a_copypath[$i]->GetStrand());
		$curr_seq    = &LookupSeq($rh_o_seqds,$ra_seqlist,$curr_id,$curr_strand);
		my %h_hitsaaln=();
		my $ovl = $self->GetOverlap(\$prev_seq,\$curr_seq,\%h_hitsaaln,$rh_param);
		print STDERR "$i-2 ".$a_copypath[$i-2]->GetName()." $i ".$a_copypath[$i]->GetName()." ovl=$ovl ".join(' ',%h_hitsaaln)."\n";
		if ( ! $ovl )
		{
			$o_left->InitAnchor(&TRUE);
			$h_new_scaffolds{$newscaff++} = $o_left;
			$lastinit = $i;
			$cut++;
		}

		# dans tous les cas je dois mettre a jour les liens sinon le GetSequence va passer automatiquement dans le mode Fast

		if ( $o_linkds->GetKV2OrientedContigs($prev_id,$prev_strand,$curr_id,$curr_strand,\%h_val) )
		{
			($h_val{ovllen},$h_val{ovlpci})=($h_hitsaaln{len},$h_hitsaaln{pci});
			$h_val{hardlink}=1;
			$h_val{gaplen_min}=1 if  ( $ovl );
			print STDERR "$prev_id => $curr_id ". join(' ',%h_val)."\n";
			$o_linkds->SetKV2($prev_id,$curr_id,\%h_val)
		}

		$prev_seq = $curr_seq;
		($prev_id,$prev_strand)=($curr_id,$curr_strand);
	}
	if ($cut)
	{
		$self->DeregisterContigs(0,$lastinit-2);
		for(my $j=0;$j<$cut;$j+=2)
		{
			$self->Shift();
		}
		$self->InitAnchor(&TRUE);	# je refais l'ancrage
	}
	#print STDERR Dumper(%h_new_scaffolds);
	return \%h_new_scaffolds;
}


# largement inspire de ValidationScaffoldIndirect

sub GetListOfNotValidatedLinks
{
my($self,$o_linkds,$validation,$maxlen)=@_;

	my $ra_path   = $self->GetArray();

	return () if ( scalar(@$ra_path) < 5 ); # cannot be used on scaffolds with less than 3 contigs
	my $lastinit=0;

	my$newscaff=1;
	my @a_notvalidated=();

	for(my $i=2;$i<@$ra_path-2;$i+=2)	# contig gap contig
	{
		if ( $i % 2 == 0  )
		{
			if ( ref($$ra_path[$i]) ne 'Contig' )
			{
				Carp::confess("ValidateIndirect: unexpected contig format ".$self->GetName()." >".$$ra_path[$i]->Get()."<\n");
			}
			next if ( $$ra_path[$i]->GetLength()  > $maxlen );
		}
		# pas optimal de reallouer en permanence
		my $o_left=New Scaffold(name=>'TMPL','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
		$o_left->Push(@$ra_path[$lastinit..$i-2]);
		my $o_right=New Scaffold(name=>'TMPR','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
		$o_right->Push(@$ra_path[$i..$#$ra_path]);
		$o_right->DeregisterContigs();
		$o_left->DeregisterContigs();

		if ( ! $o_left->IsLinkValidated($o_right,$o_linkds,$validation,&TRUE) )	# au moins 2 sequences d'un cote donnent le lien
		{
			push(@a_notvalidated,($i-2)."-".$i);
		}
	}
	return @a_notvalidated;
}

sub GetLinkCoverage
{
my($self,$o_linkds,$maxlen)=@_;

	my @a_coverage =();
	my $ra_path   = $self->GetArray(); # je le fait sur le path et pas sur les contigs seuls

	$a_coverage[0] = 0;
	$a_coverage[scalar(@$ra_path)] = 0;
	for(my $i=2;$i<scalar(@$ra_path)-2;$i+=2)
	{
		my $ctglen = $$ra_path[$i]->GetLength();
		#print STDERR "pos=$i len=$ctglen\n";
		my ($mingap,$maxgap) = split(',',$$ra_path[$i-1]);
		my $upstreamlen =  $mingap;
		my $k = $i-2;
		$a_coverage[$i]=0;
		while ( $upstreamlen < $maxlen && $k >= 0 )
		{
			#print STDERR "\t$k $upstreamlen\n";
			($mingap,$maxgap) = split(',',$$ra_path[$i+1]);
			my $downstreamlen =  $mingap;
			my $l = $i+2;
			while (($upstreamlen+$ctglen+$downstreamlen) < $maxlen && $l < scalar(@$ra_path) )
			{
				#print STDERR "\t\t$l $upstreamlen+$ctglen+$downstreamlen $a_coverage[$i] $i,$k,$l\n";
				my ($src,$srcext,$tgt,$tgtext) = $$ra_path[$k]->ConvertPathLink2GraphLink($$ra_path[$l],&FALSE);
				my %h_val=();
				if ( $o_linkds->GetKV2Oriented($src,$srcext,$tgt,$tgtext,\%h_val) )
				{
					$a_coverage[$i]+=$h_val{count};
				}
				$downstreamlen += $$ra_path[$l]->GetLength();
				$l+=2;
				if ( $l < scalar(@$ra_path) )
				{
					my ($mingap,$maxgap) = split(',',$$ra_path[$l-1]);
					$downstreamlen +=  $mingap;
				}
			}
			$upstreamlen += $$ra_path[$k]->GetLength();
			$k-=2;
			if ( $k >= 0 )
			{
				my ($mingap,$maxgap) = split(',',$$ra_path[$k+1]);
				$upstreamlen +=  $mingap;
			}

		}
	}
	return @a_coverage;

}


#
#  je cherche un lien direct entre les contigs du chemin
#

sub ValidateDirect
{
my($self,$o_graph,$o_linkds,$o_scaffolder,$cutmode,$validation,$maxlen,$gaplen_tolerance,$length_decrease_tolerance)=@_;

	#my $self_copy = undef;
	#$self_copy    = $self->Clone(); # if ( $cutmode );
	#my $ra_path   = ($cutmode) ? $self_copy->GetArray() : $self->GetArray();
	#my $ra_path   = $self_copy->GetArray();
	my $ra_path   = $self->GetArray();

	return undef if ( scalar(@$ra_path) < 3 ); # cannot be used on scaffolds with less than 3 contigs

	my $prev_ok = -1;
	my @a_edited_path=();

	my$newscaff=1;
	my %h_new_scaffolds=();

	push(@a_edited_path,$$ra_path[0]);
	for(my $i=0;$i<scalar(@$ra_path)-2;$i+=2)	# contig gap contig
	{
		my($src,$srcext,$tgt,$tgtext)=$$ra_path[$i]->ConvertPathLink2GraphLink($$ra_path[$i+2],&FALSE);
		if ( ! $o_graph->Exists($src,$srcext,$tgt,$tgtext) )
		{
			if ( $i != 0 )
			{
				next if ( $$ra_path[$i]->GetName() ne $a_edited_path[$#a_edited_path]->GetName() );
			}
			$self->_Print("ValidateScaffoldsDirect ".$self->GetName()." NOK $src,$srcext,$tgt,$tgtext $i/$#$ra_path ".$$ra_path[$i]->GetName().",".$$ra_path[$i+2]->GetName()."\n");
			if ( $o_graph->Exists($src,$tgtext,$tgt,$srcext) )
			{
				$self->_Print("ValidateScaffoldsDirect ".$$ra_path[$i]->GetName()." PB $i/$#$ra_path inversion\n");
			}
			my $inext = &Min($i+2+2,scalar(@$ra_path)-1);
			my $iprev = &Max(0,$prev_ok);
			my($asrc,$asrcext,$atgt,$atgtext)=$$ra_path[$iprev]->ConvertPathLink2GraphLink($$ra_path[$inext],&FALSE);
			#print STDERR "$iprev ".$$ra_path[$iprev]->GetName()." $asrc,$asrcext -> $inext ".$$ra_path[$inext]->GetName()." $atgt,$atgtext\n";
			my $o_newslicepath = undef;
			if ( $o_graph->Exists($asrc,$asrcext,$atgt,$atgtext) )
			{
				$o_newslicepath = $o_graph->ValidateOneGraphPath($o_linkds,$o_scaffolder,$asrc,$asrcext,$atgt,$atgtext,undef,$gaplen_tolerance,&FALSE);
			}
			if( defined($o_newslicepath)  )
			{
				#print STDERR "NEWPATH= ".$o_newslicepath->Get()."\n";
				my $o_longestnewpath = $o_graph->FindLongestPath($o_newslicepath,$o_linkds,$o_scaffolder,$gaplen_tolerance,undef); 
				if (defined($o_longestnewpath))
				{
					$o_newslicepath->DeregisterContigs();
					#$o_newslicepath = $o_longestnewpath->Clone();
					$o_newslicepath = $o_longestnewpath;
				}
				$o_newslicepath->InitAnchor(&FALSE);
				$o_newslicepath->DeregisterContigs();

				my $o_prevslicepath = New Scaffold('name'=>'TMP99999999','maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
				for(my $c=$iprev;$c<=$inext;$c++)
				{
					$o_prevslicepath->Push($$ra_path[$c]);
				}
				$o_prevslicepath->DeregisterContigs();

				$self->_Print("ValidateScaffoldsDirect ".$self->GetName()." PB: precedent chemin ".$o_prevslicepath->Get().": len=".$o_prevslicepath->GetLength()."\n");
				$self->_Print("ValidateScaffoldsDirect ".$self->GetName()." PB: nouveau chemin   ".$o_newslicepath->Get().": len=".$o_newslicepath->GetLength()."\n");
				# change le 20120403, je veux editer dans tous les cas, y compris en perdant des sequences
				# car je pense les remettre lorsque je rajouterais les repeats
				# j'ai rechange pour remettre pour tolerer une legere diminution

				print STDERR "Anchor current=".$self->GetAnchor()." new=".$o_newslicepath->GetAnchor()." compat=".$self->AreAnchoringCompatibles($o_newslicepath)."\n";

				if ( (($o_newslicepath->GetLength() > ($o_prevslicepath->GetLength()*$length_decrease_tolerance)) || $o_prevslicepath->IsContainingTandemRepeat()) && $self->AreAnchoringCompatibles($o_newslicepath) )
				{
					print STDERR "o_newslicepath\n";
					print STDERR Dumper($o_newslicepath);
					print STDERR "o_prevslicepath\n";
					print STDERR Dumper($o_prevslicepath);
					my $ra_new = $o_newslicepath->GetArray();
					my $found = &FALSE;
					my $nb = scalar(@a_edited_path);
					for(my $ii=$nb-1;$ii>=0 && ! $found;$ii--) 
					{
						if ( ref($a_edited_path[$ii]) eq 'Contig' )
						{
							$found = &TRUE if ( $a_edited_path[$ii]->GetName() eq $$ra_new[0]->GetName() );
							if ( $found )
							{
								#print STDERR "found ".$a_edited_path[$ii]->GetName()." eq ".$$ra_new[0]->GetName()." $ii/$nb\n";
							}
						}
						pop(@a_edited_path);
					}
					push(@a_edited_path,@$ra_new);
					if ( 0 )
					{
						$self->_Print("ValidateScaffoldsDirect ".$self->GetName()." AV: ".$self->Get()."\n");
						$self->_Print("ValidateScaffoldsDirect ".$self->GetName()." AP: ");
						for(my $jj=0;$jj<scalar(@a_edited_path);$jj++)
						{
							if ( $jj % 2 == 0 )
							{
								$self->_Print($a_edited_path[$jj]->Get()." ");
							}
							else
							{
								$self->_Print("$a_edited_path[$jj] ");
							}
						}
						$self->_Print(".....\n");
					}
					#&_CheckScaffoldFormat(join(' ',@a_edited_path));	
					next;
				}
			}
			elsif ( $cutmode )	 # je mets la partie amont deja valide dans un autre scaffold, et je recommence a zero (pas le meme algo que Indirect
			{
				my $o_newscaff = New Scaffold('name'=>"TMP$newscaff",'maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
				$o_newscaff->Push(@a_edited_path);
				$o_newscaff->InitAnchor(&TRUE); # je dois le faire avant le deregister sinon je ne peux pas
				$o_newscaff->DeregisterContigs();
				$h_new_scaffolds{$newscaff++}=$o_newscaff;
				#print STDERR "ValidateScaffoldsDirect CUTMODE creation scaffold ".$self->GetName()." --> TMP".$o_newscaff->GetName()." ".$o_newscaff->Get()."\n";
				@a_edited_path=();
				push(@a_edited_path,$$ra_path[$i+2]);
			}
			else
			{
				$self->_Print("ValidateScaffoldsDirect ".$self->GetName()." PB $i/$#$ra_path ".$$ra_path[$i]->GetName().",".$$ra_path[$i+2]->GetName()." unlinked\n");
			}
		}
		if ( $$ra_path[$i+2]->GetName() ne $a_edited_path[$#a_edited_path]->GetName() )
		{
			push(@a_edited_path,$$ra_path[$i+1],$$ra_path[$i+2]);
		}
		$prev_ok = $i;
	}
	$self->ResetPath(); # j'ai l'impression que ceux qui sont supprimes du path lors d'edition ne sont pas desenregistres alors qu'ils devraient
	$self->Push(@a_edited_path);
	$self->SetContigs();
	#print STDERR $self->Get()."\n";
	$self->InitAnchor(&TRUE);
	return \%h_new_scaffolds;
}


sub InitPathFromString
{
my($self,$r_newpath,$o_scaffolder,$consistency) = @_;

	my @a_path = split(/\s+/,$$r_newpath);
	for(my $i=0;$i<scalar(@a_path);$i++)
	{
		if ( $a_path[$i] =~ /^#(\S+)([pm])(\d+)$/ ) # ne le supprime plus c'est une string qui arrive!
		{
			my($name,$strand,$len)=($1,$2,$3);
			my $o_ctg = $o_scaffolder->GetNewRegisteredContig($name,$strand,$len);
			$self->Push($o_ctg);
		}
		else
		{
			$self->Push($a_path[$i]);
		}
	}
	return &FALSE if ( ! $self->IsAnchoringConsistent() );
	$self->InitAnchor($consistency);
	$self->Check();
	return &TRUE;
}



# supprime les contigs des extremites qui ne sont pas relies au suivant

sub Trim
{
my($self,$o_linkds) = @_;

	my @a_newpath = ();
	my $ra_path = $self->GetArray();
	my $imin=0;
	my %h_val=();
	for($imin=0;$imin<=(scalar(@$ra_path)-3);$imin+=2)
	{
		my ($src,$lstrand) = ($$ra_path[$imin]->GetName(),$$ra_path[$imin]->GetStrand());
		my ($tgt,$rstrand) = ($$ra_path[$imin+2]->GetName(),$$ra_path[$imin+2]->GetStrand());
		# tant que je n'ai pas de lien, je continue
		next if ( ! $o_linkds->GetKV2OrientedContigs($src,$lstrand,$tgt,$rstrand,\%h_val) );
		last;
	}
	my $imax = scalar(@$ra_path)-1;
	for(;$imax>=($imin+2);$imax-=2)
	{
		my ($src,$lstrand) = ($$ra_path[$imax-2]->GetName(),$$ra_path[$imax-2]->GetStrand());
		my ($tgt,$rstrand) = ($$ra_path[$imax]->GetName(),$$ra_path[$imax]->GetStrand());
		# tant que je n'ai pas de lien, je continue
		next if ( ! $o_linkds->GetKV2OrientedContigs($src,$lstrand,$tgt,$rstrand,\%h_val) );
		last;
	}	
	for(my $i=$imin;$i<=$imax;$i++)
	{
		push(@a_newpath,$$ra_path[$i]);
	}
	return if ( scalar(@a_newpath) ==  scalar(@$ra_path) );

	$self->_Print("Trim: ".$self->GetName()." Edit imin=$imin, imax=$imax / ".scalar(@$ra_path)."\n");

	if ( scalar(@a_newpath) >= 3 ) # au moins 2 contigs 
	{
		$self->DeregisterContigs(0,$imin-1);
		$self->DeregisterContigs($imax+1,scalar(@$ra_path)-1);
		$self->_Set('_ra_path',\@a_newpath);
		$self->Check();
		return 1;
	}
	else
	{
		$self->Reset();
		return -1;
	}
	return 1;
}

# je vais verifier que la longeur qu'un gap est infrieure aux longueurs donnees par les hits englobants
# AVERIFIER, pas utilise

sub RecomputeGapLength
{
my($self,$o_linkds) = @_;

	return if ( $self->GetNumberOfContigs() < 3 );
	my $vallen = $self->_FastOneGet('_maxvalidationlength');
	for (my $r=0;$r<2;$r++)
	{
		if ( $r )
		{
			$self->Reverse();
		}
		my $ra_path = $self->GetArray();
		my $nb = scalar(@$ra_path);
		return if  ( !defined($ra_path) || $nb == 0 ); 	# cas des scaffolds avec une seule sequence
		my $idscaff=$self->GetName();
		#print STDERR "RecomputeGapLength $idscaff, $nb\n";
		for(my $i=0;$i<$nb-4;$i+=2)	# CONTIGS
		{
			my ($mingaplen,$maxgaplen) = split(',',$$ra_path[$i+1]);
			my $step = 0;
			my $len=0;
			for(my $j=$i+4;$j<$nb && $step<&MAXGAPLENSTEPS && $len<$vallen;$j+=2,$step++)	# CONTIG
			{
				$len += $$ra_path[$j-2]->GetLength();
				my %h_val=();
				my($src,$srcext,$tgt,$tgtext) = $$ra_path[$i]->ConvertPathLink2GraphLink($$ra_path[$j],&TRUE);
				if (  $o_linkds->GetKV2Oriented($src,$srcext,$tgt,$tgtext,\%h_val) )
				{
					if ( $maxgaplen > ($h_val{gaplen_max}-$len) )
					{
						#print STDERR "PB ============= scaff=$idscaff $r pos $i -> $j ".$$ra_path[$i]->GetName()." next($mingaplen,$maxgaplen),len=$len nextnext ($h_val{gaplen_min},$h_val{gaplen_max})\n";
			            # 20121216 j'accepte les gap negatifs pour pourvoir gerer le chev
						#$maxgaplen = &Max(1,&Min($maxgaplen,$h_val{gaplen_max}-$len));
						#$mingaplen = &Max(1,&Min($mingaplen,$maxgaplen));
						#print STDERR "\t$mingaplen,$maxgaplen\n";
						$maxgaplen = &Min($maxgaplen,$h_val{gaplen_max}-$len);
						$mingaplen = &Min($mingaplen,$maxgaplen);
						$$ra_path[$i+1] = "$mingaplen,$maxgaplen";
					}
				}
			}
		}
	}
	$self->Reverse();
}

# Migre
# Attention ca semble degrader les resultats si on l'utilise ResetGapsOnePath en amont ou pendant le scaffolding

sub CheckGapsOnePath
{
my($self,$o_linkds,$reset) = @_;

	my $ra_path = $self->GetArray();
	my $nb = scalar(@$ra_path);
	return if  ( !defined($ra_path) || $nb == 0 ); 	# cas des scaffolds avec une seule sequence

	# 1ere etape, je mets la longeur lue dans le graphe

	for(my $i=1;$i<$nb;$i+=2)	# GAPS
	{
		my ($mingaplen,$maxgaplen) = split(',',$$ra_path[$i]);
		my %h_val = ();	
		if ( $mingaplen > $maxgaplen )	# ne devrait jamais arriver
		{
			# 20121216 j'accepte les gaps negatifs $$ra_path[$i]=&Max(1,$maxgaplen).",".&Max(1,$mingaplen);
			$$ra_path[$i]="$maxgaplen,$mingaplen";
		}
		my ($src,$lstrand) = ($$ra_path[$i-1]->GetName(),$$ra_path[$i-1]->GetStrand());
		my ($tgt,$rstrand) = ($$ra_path[$i+1]->GetName(),$$ra_path[$i+1]->GetStrand());
		if ( $o_linkds->GetKV2OrientedContigs($src,$lstrand,$tgt,$rstrand,\%h_val) ) # ca n'ameliore pas
		{
			# 20121216 j'accepte les gaps negatifs $$ra_path[$i]=&Max(1,$maxgaplen).",".&Max(1,$mingaplen);
			$$ra_path[$i]="$h_val{gaplen_min},$h_val{gaplen_max}";
		}
	}

	# 2e etape, j'utilise les liens englobant pour recalculer la longueur des gaps

	if ( defined($reset) && $reset)
	{
		$self->ResetGapsOnePath($o_linkds);
	}
}


# reinit de la longueur des gaps a partir du graphe, uniquement les liens direct
sub ReloadGapsOnePath
{
my($self,$o_linkds) = @_;

	my $ra_path = $self->GetArray();
	my $nb = scalar(@$ra_path);
	print STDERR "###".$self->Get()."\n";
	return if ( $nb < 3 );
	for(my $i=0;$i<$nb-2;$i+=2)	# CONTIGS
	{
		my %h_val=();
		my($src,$srcext,$tgt,$tgtext) = $$ra_path[$i]->ConvertPathLink2GraphLink($$ra_path[$i+2],&TRUE);
		if ( $o_linkds->GetKV2Oriented($src,$srcext,$tgt,$tgtext,\%h_val) )
		{
			$$ra_path[$i+1] = "$h_val{gaplen_min},$h_val{gaplen_max}";
		}
	}
 	$self->CheckFormat();
}

#
# Migre
#

sub ResetGapsOnePath
{
my($self,$o_linkds) = @_;

	my $ra_path = $self->GetArray();
	my $nb = scalar(@$ra_path);
	for(my $i=0;$i<$nb;$i+=2)	# CONTIG
	{
		my $step = 0;
		my $reset = &FALSE;
		my $j=$i+4;
		next if ($j >= $nb);
		#for(my $j=$i+4;$j<$nb && $step < &MAXGAPLENSTEPS && ! $reset;$j+=2,$step++)	# CONTIG
		if ( 1 )
		{
			my %h_val=();
			my($src,$srcext,$tgt,$tgtext) = $$ra_path[$i]->ConvertPathLink2GraphLink($$ra_path[$j],&TRUE);
			if ( $o_linkds->GetKV2Oriented($src,$srcext,$tgt,$tgtext,\%h_val) && ! $h_val{weak} )
			{
				# 20121216 j'accepte les gaps negatifs
				#$reset = $self->ResetGapLen($i,$j,&Max(1,$h_val{gaplen_min}).",".&Max(1,$h_val{gaplen_max}));
				$reset = $self->ResetGapLen($i,$j,"$h_val{gaplen_min},$h_val{gaplen_max}");
			}
		}
	}
#	$self->CheckFormat();
}

#
# Migre
#

sub ResetGapLen
{
my($self,$istart,$iend,$gaplen) = @_;

	my $ra_path = $self->GetArray();
	my $nb = scalar(@$ra_path);
	$istart = 0     if ( !defined($istart) );
	$iend   = $nb-1 if ( !defined($iend) );
	my($min,$max)=split(',',$gaplen);
	my $totgaps=0;
	my $totctgs=0;
	for(my $i=$istart+1;$i<$iend;$i++)	# contig gap contig, je ne veux que les gaps et les contigs internes
	{
		if ( $i % 2 == 0 )
		{
			$totctgs+=$$ra_path[$i]->GetLength();
		}
		else
		{
			my($mingaplen,$maxgaplen)=split(',',$$ra_path[$i]);
			$totgaps+=$maxgaplen;
		}
	}
	# methode 20130425: j'edite seulement maxgaplen pour qu'il ne soit pas superieur a la taille du lien englobant moins la taille des sequences incluses
	for(my $i=$istart+1;$i<$iend;$i+=2)	# contig gap contig, je ne veux que les gaps
	{
		my($mingaplen,$maxgaplen)=split(',',$$ra_path[$i]);
		my $newmaxgaplen = &Max(&Max($mingaplen+1,&Min($maxgaplen,$max-$totctgs)),1);
		$$ra_path[$i] = "$mingaplen,"."$newmaxgaplen";
		#print STDERR "\tAV ($mingaplen,$maxgaplen) => $$ra_path[$i]\n";
	}

	if ( 0 )
	{	# methode avant le 20130425
		my $fact = $totgaps/&Max(1,$max-$totctgs);
		#print STDERR "ResetGapLen $istart,$iend,$gaplen $fact = $totgaps/&Max(1,$max-$totctgs)\n";
		return &FALSE if ($fact <= 1 ); # je n'augmente pas

		for(my $i=$istart+1;$i<$iend;$i+=2)	# contig gap contig, je ne veux que les gaps
		{
			my($mingaplen,$maxgaplen)=split(',',$$ra_path[$i]);
			$maxgaplen = int($maxgaplen/$fact);
			$mingaplen = int($mingaplen/$fact);
			$maxgaplen = &Max($maxgaplen,1);
			$mingaplen = &Max($mingaplen,1);
			$$ra_path[$i] = "$mingaplen,$maxgaplen";
		}
	}
	return &TRUE;
}

# attention, il y a aussi GetGapLen, je sais c'est nul
# longueur du gap cad entre les deux extremites

sub IsGapLengthConsistent
{
my($self,$gap,$tolerance)=@_;

	my $gaplen = $self->GetInternalLength(&TRUE);
	my ($lmin,$lmax) = split(',',$gap);
	$lmax += ( ($lmax*$tolerance)/100 );
	$lmin -= ( ($lmin*$tolerance)/100 );
	my $res = ( $gaplen >= $lmin && $gaplen <= $lmax ) ? &TRUE: &FALSE;
	return $res;
}

# 
sub GetGapLen
{
my($dist,$min,$max)=@_;

	#$min = &Max(1,$min);
	#$max = &Max(1,$max/1);
	$min -= $dist;
	$max -= $dist;
	#print STDERR "GetGapLen $dist, $min, $max\n";
	#if ( $max < 0 )
	#{
	#	Carp::croak("$min, $max\n");
	#}
	return ($min,$max);
}

# Migre
sub TrimRepeats
{
my($self,$tag,$rh_contig2scaff) = @_;

	my @a_newpath = ();
	my $ra_path = $self->GetArray();
	my $imin=0;
	my $name = $self->GetName();
	for($imin=0;$imin<scalar(@$ra_path);$imin+=2)
	{
		my $nm = $$ra_path[$imin]->GetName();
		next if ( exists($$rh_contig2scaff{$nm}) && $name ne $$rh_contig2scaff{$nm} ); # REPEAT
		last;
	}
	my $imax = scalar(@$ra_path)-1;
	for(;$imax>=$imin;$imax-=2)
	{
		my $nm = $$ra_path[$imax]->GetName();
		next if ( exists($$rh_contig2scaff{$nm}) && $name ne $$rh_contig2scaff{$nm} ); # REPEAT
		last;
	}
	my $lendup=0;
	# deuxieme passe, je supprime les sequences des extremites des scaffolds qui sont deja dans des scaffolds plus grands
	# par contre je garde en double des sequences qui sont internes a deux scaffolds
	for(my $i=$imin;$i<=$imax;$i++)
	{
		push(@a_newpath,$$ra_path[$i]);
		if ( $i % 2 == 0 )
		{
			my $nm = $$ra_path[$i]->GetName();
			if ( $name ne $$rh_contig2scaff{$nm} )
			{
				$lendup+= $$ra_path[$i]->GetLength()  ;
			}
		}
	}
	#print STDERR "lendup=$lendup\n";
	return 0 if ( scalar(@a_newpath) == scalar(@$ra_path) );

	$self->_Print("TrimRepeats/$tag: Edit ".$self->GetName()." imin=$imin, imax=$imax\n");

	my $o_newscaff = New Scaffold('name' => $self->GetName(),'maxvalidationlength'=>$self->_FastOneGet('_maxvalidationlength'));
 	$o_newscaff->Push(@a_newpath);
	# je veux aussi supprimer lorsque le scaffold (hormis repeats) est pratiquement englobe dans un autre
	# pour eviter les petits spe-gros pas spe-petits spe
	if ( scalar(@a_newpath) > 0 && ($o_newscaff->GetLength()*$OVL_MARGIN) > $lendup )
	{
		$o_newscaff->DeregisterContigs();
		$self->_Print("TrimRepeats/$tag: Trimming ".$self->GetName()." pre:0,".($imin-1)." post:".($imax+1).",".(scalar(@$ra_path)-1)." lendup=$lendup\n");
		#$self->DeregisterContigs(0,$imin-1);
		#$self->DeregisterContigs($imax+1,scalar(@$ra_path)-1);
		$self->ResetPath();
		$self->Push(@a_newpath);
 		$self->Check();
		$self->InitAnchor(&TRUE);
		return 1;
	}
	else
	{
		$o_newscaff->DeregisterContigs() if ( scalar(@a_newpath) > 0 );
		$self->_Print("TrimRepeats/$tag: Deleting ".$self->GetName()." lendup=$lendup\n");
 		$self->ResetPath(); # pour derenregistrer les contigs
		return -1;
		# Penser a le supprimer si plus de contigs
	}
}


# essaye de trouver des contigs commums aux extremites de deux scaffolds qui sont senses se chevaucher
# si je trouve je merge
# overlap n'est pas une longueur min classique mais c'est la zone dans laquelle je pense avoir le chevauchement
# pour le moment je n'utilise pas l'overhang

sub MergeOverlapping
{
my($self,$o_tgt,$overlap,$gaplen_tolerance,$maxoverhang) = @_;

	$overlap += ( $overlap*$gaplen_tolerance/100 );
	my $ra_srcpath = $self->GetArray();
	my $ra_tgtpath = $o_tgt->GetArray();
	my ($srcmin,$srcmax)=(&MAXINT,-1);
	my ($tgtmin,$tgtmax)=(&MAXINT,-1);
	my $slen = 0;
	my $found = &FALSE;
	$self->_Print("MergeOverlapping: ".$self->GetName()." $#$ra_srcpath => ".$o_tgt->GetName()." $#$ra_srcpath $overlap,$gaplen_tolerance\n");

	my %h_src = ();
	for(my $s=scalar(@$ra_srcpath)-1; $s >= 0 && $slen < $overlap ; $s-=2)
	{
		my($name,$strand,$len)=$$ra_srcpath[$s]->GetArray();
		$h_src{$name}=$s if ( ! defined($h_src{$name}) );
		$slen += $len;
	}

	my $tlen = 0;
	my %h_tgt = ();
	for(my $t=0;$t<scalar(@$ra_tgtpath) && $tlen < $overlap ; $t+=2 )
	{
		my($name,$strand,$len)=$$ra_tgtpath[$t]->GetArray();
		if ( defined($h_src{$name}) )
		{
			$found++;
			$srcmin = &Min($h_src{$name},$srcmin);
			$srcmax = &Max($h_src{$name},$srcmax);
			$tgtmin = &Min($t,$tgtmin);
			$tgtmax = $t;
		}
		$tlen += $len;
	}

	$self->_Print("MergeOverlapping: found=$found ".$self->GetName()." ($srcmin,$srcmax)   ".$o_tgt->GetName()." ($tgtmin,$tgtmax) found=$found\n");

	return &FALSE if ( ! $found );

	# je garde la zone de chevauchement sur le suivant

	# 1. trimmer  l'aval
	for (my $s=scalar(@$ra_srcpath)-1; $s>=$srcmin ;  $s-=2 ) 
	{
		$self->Pop(); # enleve le gap aussi
	}

#	$o_tgt->DereferenceContigs();
	#2. trimmer le suivant
	for (my $t=0;$t<$tgtmin;$t+=2)
	{
		$o_tgt->Shift();
	}

	# 2. merge le suivant
	$ra_tgtpath = $o_tgt->GetArray();
	for (my $i=0;$i<scalar(@$ra_tgtpath);$i++)
	{
		$self->Push($$ra_tgtpath[$i]);
	}

	$self->CheckFormat();



	return &TRUE;
}

#==============================================================================================================
#==============================================================================================================
#==============================================================================================================


#
#	Fasta dump
#
# il y a des overlap inferieurs a la longueur de kmer utilisee pour contiguer,
# je rajoute une heuristique pour fermer ces trous
# pour le moment je ne mets pas de protection sur la taille des gaps car je ne sais pas encore bien la calculer
# et des mots exacts de 30nt ca ne doit pas etre si frequent

sub GetSequence
{
my($self,$o_linkds,$rh_o_seqds,$ra_seqlist,$minmingaplen,$f_gff3,$gaplen_tolerance) =@_;


	$minmingaplen = 0 if (! defined($minmingaplen) );

	# comme j'ai pris la taille max d'insert, il faudrait que je diminue la taille des gaps
	# si je fais ca je perds le recalcul des gaps lorsque j'ai pu le faire $self->CheckGapsOnePath($o_linkds,&FALSE);

	# DEBUGJG
	#$f_gff3=&GetStreamOut("t.gff3");

	my $ra_path = $self->GetArray();

	my %h_ctg_pos=();
	if ( ref($$ra_path[0]) ne 'Contig' )
	{
		print STDERR Dumper($self);
		Carp::confess("GetSequence\n");
	}

	my $fast_mode = $self->_FastOneGet('_fast_getsequence'); 

	$h_ctg_pos{$$ra_path[0]->GetName()}= "1\t".$$ra_path[0]->GetLength();
	my ($previd,$strand)=($$ra_path[0]->GetName(),$$ra_path[0]->GetStrand());
	my $prev_seq    = &LookupSeq($rh_o_seqds,$ra_seqlist,$previd,$strand);

	my $seq = $prev_seq;

	my %h_val=();
	my $prev_N_length=0;
	for(my $i=1;$i<scalar(@$ra_path);$i++)
	{
		if ( $i %  2  == 0 )
		{
			my ($mingaplen,$maxgaplen) = split(',',$$ra_path[$i-1]);
			my ($id,$strand)=($$ra_path[$i]->GetName(),$$ra_path[$i]->GetStrand());
			my $curr_seq = &LookupSeq($rh_o_seqds,$ra_seqlist,$id,$strand);
			my ($id_prev,$strand_prev)=($$ra_path[$i-2]->GetName(),$$ra_path[$i-2]->GetStrand());
			my $found=&FALSE;
			my $index = 0;


			my $precomputed_ovl = &FALSE;
			if ( $o_linkds->GetKV2OrientedContigs($id_prev,$strand_prev,$id,$strand,\%h_val) )
			{
				$precomputed_ovl = $h_val{ovllen} || &FALSE;
			}

			#print STDERR "$id_prev => $id mingaplen=$mingaplen, fast_mode=$fast_mode $$ra_path[$i-1], precomputed_ovl=$precomputed_ovl $mingaplen == $maxgaplen\n";

			if ( $mingaplen <= 1 || ( $precomputed_ovl && $mingaplen < 3000) ) 
			{
				if ($fast_mode || ! $precomputed_ovl) # sur les gros problemes c'est bcp trop lourd de lancer aaln5 a chaque trou, si pas de suspicion d'ovl je fais en rapide
				{
					#my $end_prev_seq = substr($prev_seq,-100);
					my $end_prev_seq = substr($seq,-1 * (100+$prev_N_length)); # 20121226 je change mais pour -100 ca ne doit rien changer, SI CA CHANGE A CAUSE DES GAPS
					#print STDERR "($id_prev,$strand_prev) -> ($id,$strand) FAST $end_prev_seq\n";
					# faut-il mettre une protection sur la taille du gap ?
					my $len = 0;
					for (my $j=0;$j<20 && length($end_prev_seq);$j++)
					{
						my $downstream = substr($curr_seq,$j,30);
						$index = rindex($end_prev_seq,$downstream); # 30 nt strictement identiques dans les 100 derniers nucleotides de la precedente
						if ( $index != -1 )
						{
							my $gaplen = 0;
							while( $seq=~/N$/g )
							{
								chop($seq);
								$gaplen++;
							}
							$index = rindex($seq,$downstream);
							substr($seq,$index)=substr($curr_seq,$j);
							$found = &TRUE;
							$$ra_path[$i-1]="0,0";
							my $len = length($seq);
							$index++;
							$h_ctg_pos{$$ra_path[$i]->GetName()} ="$index\t$len";
							last;
						}
					}
				} 
				else
				{
					# je ne vais pas tenter des chev trop importants, c'est trop long et aaln peut planter
					my $extractlen = &Min(&Min(length($curr_seq)+$prev_N_length,length($seq)),3000); # je prends 3000 au cas ou il y des gap puis 1000 dans GetOverlap
					my $end_prev_seq = substr($seq,-1 * $extractlen); # je ne prends plus prev_seq pour mieux gerer les chevauchements (si prev_seq est trop courte)
					#print STDERR "($id_prev,$strand_prev) -> ($id,$strand) SLOW $end_prev_seq\n";
					$end_prev_seq=~s/N+$//; # j'enleve les N qui correspondent au gap precedent, de toute facon si overlap j'enleve le gap
					#print STDERR "extractlen=$extractlen // $end_prev_seq\n";
					#print STDERR "PREV $end_prev_seq ".length($end_prev_seq)."\n";
					#print STDERR "CURR $curr_seq ".length($curr_seq)."\n";

					my %h_hitsaaln=();
					my $ovl = $self->GetOverlap(\$end_prev_seq,\$curr_seq,\%h_hitsaaln,undef);
					print STDERR "$id_prev => $id $mingaplen == $maxgaplen $i-2 ".$$ra_path[$i-2]->GetName()." $i ".$$ra_path[$i]->GetName()." ovl=$ovl ".join(' ',%h_hitsaaln)."\n";

					if ( $ovl )
					{
						my $gaplen = 0;
						while( $seq=~/N$/g )
						{
							chop($seq);
							$gaplen++;
						}
						my $downstream = substr($end_prev_seq,$h_hitsaaln{qstart}-1,$h_hitsaaln{qend}-$h_hitsaaln{qstart}+1); # je cherche la sequence de la query car avec cette methode query et target ne sont plus 100% identique
		#				print STDERR "downstream $h_hitsaaln{qstart}-1,$h_hitsaaln{qend}-$h_hitsaaln{qstart}+1 $downstream\n";
						my $prev_seqlen = length($seq);
						my $prev_fullseq = $seq;
						$index = rindex($seq,$downstream);
						substr($seq,$index)=substr($curr_seq,$h_hitsaaln{tstart}-1);
						my $curr_seqlen = length($seq);
						$found = &TRUE;
						$$ra_path[$i-1]="0,0";
						my $len = length($seq);
						$index++;
						$h_ctg_pos{$$ra_path[$i]->GetName()} ="$index\t$len";
						if ( $curr_seqlen < $prev_seqlen )
						{
							# apres avoir garder les pos dans h_ctg_pos
							print STDERR "GetSequence Warning: INCLUSION ( $curr_seqlen < $prev_seqlen ), rollback to previous sequence\n";
							$seq = $prev_fullseq;
						}
					}
					print STDERR "OVERLAP ($ovl): prev=".($i-2).", name=".$$ra_path[$i-2]->GetName()."\n";
					print STDERR "OVERLAP ($ovl): curr=$i, name=".$$ra_path[$i]->GetName()."\n";
					print STDERR "OVERLAP ($ovl): ".join(' ',%h_hitsaaln)."\n";
				}
			}
			#else
			#{
			#		print STDERR "($id_prev,$strand_prev) -> ($id,$strand) NO OVL DETECTION\n";
			#}

			if ( ! $found )
			{
				$index = length($seq);
				$seq .= $curr_seq;
				my $len = length($seq);
				$h_ctg_pos{$$ra_path[$i]->GetName()}="$index\t$len";
			}
		}
		else
		{
			$previd = $$ra_path[$i-1]->GetName();
			my ($mingaplen,$maxgaplen) = split(',',$$ra_path[$i]);
			#$mingaplen=&Max($mingaplen,1);	# au cas ou ce soit negatif
			#$maxgaplen=&Max($maxgaplen,1);
			#my $gaplen = 0;
			#$gaplen = $mingaplen+($maxgaplen+$mingaplen)/4;

			my $gaplen=0;
			if  ( $maxgaplen > 0 && $mingaplen < 0)
			{
				$gaplen = $maxgaplen/2;
			}
			elsif ( $maxgaplen < 0 && $mingaplen < 0)
			{
				$gaplen = 0;
			}
			else
			{
				$gaplen = $mingaplen+($maxgaplen+$mingaplen)/4;
			}

			if ( $mingaplen == $maxgaplen && $mingaplen > 0 ) # lorsque les deux valeurs sont identiques c'est que c'est vraissemblablement ok (ou que l'on veut le garder tel quel)
			{
				$gaplen = $mingaplen;
			}
			else
			{
				if ( $gaplen <= 0 && $fast_mode ) 
				{
					$self->_Print("GetSequence: scaffold ".$self->GetName()." Negative overlap between $previd and ".$$ra_path[$i+1]->GetName()." gaplen=$gaplen ==> $minmingaplen\n");
				}
				$gaplen = &Max($minmingaplen,$gaplen);
			}
			$prev_N_length=$gaplen;
			$seq .= 'N' x $gaplen;
		}
	}
	if ( defined($f_gff3) )
	{
		$self->DumpGFF3($o_linkds,$f_gff3,\%h_ctg_pos,length($seq),$gaplen_tolerance);
	}
	return (\$seq,\%h_ctg_pos);
}

# je ne peux pas paralleliser car la sequence prev n'est pas forcement celle du contig precedent (e.g lorsque edite a cause d'un chev avec le n-2)

sub GetOverlap
{
my($self,$r_prev,$r_next,$rh_hitsaaln,$rh_param)=@_;

	my($subprev)=$$r_prev;
	if ( length($$r_prev) > &ALIGN_MAXLEN )
	{
		$subprev = substr($$r_prev,-1 *&ALIGN_MAXLEN);
	}
	my($subcurr)=$$r_next;
	if ( length($$r_next) > &ALIGN_MAXLEN )
	{
		$subcurr = substr($$r_next,0,&ALIGN_MAXLEN);
	}

	my ($minpci,$minlen,$maxoverhang)= $self->GetOverlapParams();
	$minpci = $$rh_param{minovlpci} if ( defined($$rh_param{minovlpci}) );
	$minlen = $$rh_param{minovllen} if ( defined($$rh_param{minovllen}) );
	$maxoverhang = $$rh_param{maxovlovh} if ( defined($$rh_param{maxovlovh}) );

	print STDERR "GetOverlap($minpci,$minlen,$maxoverhang)\n";
	#print STDERR length($$r_prev)."\n";

	my $rh_dejavu = $self->_FastOneGet('_rh_cache_overlaps');
	my $digest = md5_hex("$subprev#$subcurr");
	if ( ! defined($$rh_dejavu{$digest}) )
	{
		my $scaffid = $self->GetName();
		#my $in = ">PREV\n$$r_prev\n>CURR\n$$r_next\n";
		#print STDERR "$in\n";
		#open(FH,"echo '"."$in"."' | aaln5 - --gi 30 --gx 5|");
		#my $aalnhit=<FH>;
		#close(FH);
		$subprev=~s/[^ATGC]/N/ig;
		$subcurr=~s/[^ATGC]/N/ig;
		my $aalnhit = &Alnpack::Aaln5(30,5,"PREV",$subprev,"CURR",$subcurr);
		if ( !defined($aalnhit) )
		{
			print STDERR "GetOverlap Warning: unexpected aaln5 result (maybe a crash, see dmesg)\n";
		}
		else
		{
			chomp($aalnhit);
			$$rh_dejavu{$digest}=$aalnhit;
		}
	}
	my $aalnhit = $$rh_dejavu{$digest};
	&HSPSet::AALN::ParseAALNLine(\$aalnhit,1,$rh_hitsaaln);
	return &FALSE if ( $$rh_hitsaaln{pci} < $minpci );
	return &FALSE if ( $$rh_hitsaaln{len} < $minlen );
	# dans l'alignment je fais en sorte de reorienter pour exposer les extremites => le chev est sur fin de query et debut de target
	return &FALSE if ( $$rh_hitsaaln{qend} < length($subprev)-$maxoverhang ); # && $h_hitsaaln{qstart} > $maxoverhang
	return &FALSE if ( $$rh_hitsaaln{tstart} > $maxoverhang ); # && $h_hitsaaln{tend} < $h_val{tgtlen}-$maxoverhang )
	return &TRUE;
}

sub DumpGFF3
{
my($self,$o_linkds,$f_gff3,$rh_ctg_pos,$scafflen,$gaplen_tolerance)=@_;

	my $vallen = $self->_FastOneGet('_maxvalidationlength');
	$vallen += ($vallen*$gaplen_tolerance/100);
	my $scaffid = $self->GetName();
	my $ra_ctg = $self->GetArrayContigs();
	my %h_val=();
	print $f_gff3 "$scaffid\tknown\tchromosome\t1\t$scafflen\t.\t.\t.\tID=$scaffid;Name=$scaffid\n";
	for(my $i=0;$i<scalar(@$ra_ctg);$i++)
	{
		my ($id,$strand,$ctglen)=$$ra_ctg[$i]->GetArray();
		my $brin = ($strand eq 'p' ) ? '+' : '-';
		my($start,$end)=split(' ',$$rh_ctg_pos{$id});
		printf $f_gff3 "$scaffid\tlynx_contig\tmatch\t$start\t$end\t%d\t$brin\t.\tID=%s;Name=$id;length=$ctglen\n",$end-$start+1,"$scaffid.$id.$i";
		for(my $j=$i+1;$j<scalar(@$ra_ctg);$j++)
		{
			my ($startj,$endj)=split(' ',$$rh_ctg_pos{$$ra_ctg[$j]->GetName()});
			last if ( ($startj-$end) > $vallen );

			my ($src,$srcext,$tgt,$tgtext) = $$ra_ctg[$i]->ConvertPathLink2GraphLink($$ra_ctg[$j],&FALSE);
			if ( $o_linkds->GetKV2Oriented($src,$srcext,$tgt,$tgtext,\%h_val) )
			{
				my ($idj)=$$ra_ctg[$j]->GetName();
				my $parent = "$scaffid.$id.$idj.$i.$j";
				my $str = "src=$src;tgt=$tgt;";
				foreach my $k ('count','srcstart','srcend','srcext','tgtstart','tgtend','tgtext','insert_size','gaplen_min','gaplen_max','hardlink','weak','ovlpci','ovllen','nlib','ambiguous_orient')
				{
					next if ( ! defined($h_val{$k}) );
					$str .= "$k=$h_val{$k};";
				}
				$str =~ s/;$//;
				printf $f_gff3 "$scaffid\tlynx_link\tmatch\t$start\t$endj\t%d\t+\t.\tID=$parent;Name=$id-$idj;$str\n",$start,$endj-$start+1,$parent;
				printf $f_gff3 "$scaffid\tlynx_link\tmatch_part\t$start\t$end\t%d\t+\t.\tID=%s;Name=$id;Parent=$parent\n",$end-$start+1,"part$j.$scaffid.$id.$i";
				printf $f_gff3 "$scaffid\tlynx_link\tmatch_part\t$startj\t$endj\t%d\t+\t.\tID=%s;Name=$idj;Parent=$parent\n",$endj-$startj+1,"part$i.$scaffid.$idj.$j";
			}
		}
	}
}


sub LookupSeq
{
my($rh_o_seqds,$ra_seqlist,$id,$strand)=@_;

	foreach my $seqds (@$ra_seqlist)
	{
		my $seq = $$rh_o_seqds{$seqds}->GetSeq($id);
		next if ( ! defined($seq) || $seq eq '' );
		if ( $strand eq 'm' || $strand eq '-' )
		{
			return &RevCompSeq(\$seq);
		}
		return $seq;
	}
	print STDERR "WARNING: unable to find >$id<\n";
	return 'N' x 111;
}


sub AssertParams
{
my ($self,$rh_param,$ra_expected) = @_;
	
	foreach my $param (@{$ra_expected})
	{
		if ( !exists($$rh_param{$param}) )
		{
			print Dumper(%$rh_param);
			Carp::confess("Fatal error in constructor: $param must be provided\n") 
		}
	}
}

sub _Print
{
my($self,$msg)=@_;

	print STDERR "$msg";
}


1;
